/* gsmstat.c
 * gsmstat   2002 Neil Piercy
 *
 * $Id: epan/dissectors/gsmstat.c 1.3 2008/06/06 05:11:23BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* This module provides GSM statistics common to ethereal and tethereal.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <glib.h>

#include "epan/packet_info.h"
#include <epan/packet.h>
#include <epan/tap.h>
#include "epan/dissectors/packet-gsm.h"
#include "register.h"

/* The following functions are called within the packet disectors to ease the job
 * of this function and avoid re-parsing the packets. They rely on the fact that the
 * disectors are called in a single thread, so each stores some state data, which the
 * packet function then uses to update the stats. This is not really how ethereal taps
 * are intended to work, but I dont want to re-parse and rewrite a disector!
 */

/* start with a VERY simple set of packet description options... */

typedef struct _gsmstat_msg_t {
    int proto;
    int pd;
    int msg_type;
    int chan_type;
    int chan_id;
    int cause;
    int cic;
} gsmstat_msg_t;

#define GSMSTAT_MSG_MAX    100    /* how many messages, including multiple embedding and multiple messages, in one packet ? */

typedef struct _gsmstat_pkt_t {
    int           secs;
    int           intface;
    unsigned      pnum;
    int           msg_count;
    int           intface_count;
    gsmstat_msg_t msg[GSMSTAT_MSG_MAX];
} gsmstat_pkt_t;

static struct _gsmstat_pkt_t gsmstat_pkt;

typedef struct _gsmstat_a_conn_t {
	int		dlr;
	int		slr;
	int		cic;
} gsmstat_a_conn_t;

#define GSMSTAT_A_CONNS_MAX		10000	/* total SCCP connactions to support - CICs + signalling-only connections */
typedef struct _gsmstat_a_t {
	int					conns;
	gsmstat_a_conn_t	conn[GSMSTAT_A_CONNS_MAX];
} gsmstat_a_t;

static struct _gsmstat_a_t gsmstat_a;

int gsmstat_common_message(gsmstat_common_t *gsc, gsmstat_msg_t *gm, packet_info *pinfo);
void update_usage_stats(usage_stats_t *use, int delta, packet_info *pinfo);

/* Value Strings */
static const value_string gsmstat_vals[] = {
    { GSMSTAT_CHAN_REQS,                    "Chan Reqs" },
    { GSMSTAT_PAGE_REQS,                    "Page Reqs" },
    { GSMSTAT_PAGE_RESPONSE,                "Page Resp" },
    { GSMSTAT_IMM_ASS_ATTEMPTS,             "Imm Ass Att" },
    { GSMSTAT_IMM_ASS_SUCCESSES,            "Imm Ass Succ" },
    { GSMSTAT_HANDOVER_COMMAND,             "Handover Command" },
    { GSMSTAT_HANDOVER_COMPLETE,            "Handover Complete" },
    { GSMSTAT_HANDOVER_FAILURE,             "Handover Fail" },
    { GSMSTAT_HANDOVER_DETECT,              "Handover Detect" },
    { GSMSTAT_LU_REQS,                      "LU Reqs" },
    { GSMSTAT_LU_ACCEPT,                    "LU Accept" },
    { GSMSTAT_LU_REJECTS,                   "LU Rejects" },
    { GSMSTAT_SETUPS,                       "Setups" },
    { GSMSTAT_CONNECTS,                     "Connects" },
    { GSMSTAT_DISC_NORMAL,                  "Disc(norm)" },
    { GSMSTAT_DISC_ABNORMAL,                "Disc(abnorm)" },
    { GSMSTAT_ASS_REQS,                     "Assignmt Reqs" },
    { GSMSTAT_ASS_FAILS,                    "Assignmt Fail" },
    { GSMSTAT_CLEAR_REQS,                   "Clear Reqs" },
    { GSMSTAT_CLEAR_CMDS,                   "Clear Cmds" },
    { GSMSTAT_ERROR_REPS,                   "Error Reports" },
    { GSMSTAT_ERROR_INDS,                   "Error Indications" },
    { GSMSTAT_PREPROC_MEAS_RESULTS,         "PreProc Meas" },
    { GSMSTAT_CONN_FAIL_SDCCH,              "Conn Fails (SDCCH)" },
    { GSMSTAT_CONN_FAIL_TCH,                "Conn Fails (TCH)" },
    { GSMSTAT_ERRORS,                       "Stats Errors" },
    { GSMSTAT_MSGS,                         "Total Msgs" },
    { GSMSTAT_IPPKTS,                       "Total IP Pkts" },
    { GSMSTAT_IPLEN,                        "Total IP Octs" },
    { GSMSTAT_HANDOVER_REQUEST,             "Handover Request" },
    { GSMSTAT_HANDOVER_REQUIRED,            "Handover Required" },
    { GSMSTAT_HANDOVER_REQUIRED_REJECT,     "Handover Reqrd Reject" },
    { GSMSTAT_HANDOVER_ALLOC,               "Handover Allocated" },
    { GSMSTAT_ASS_CMD,                      "Assignment Command" },
    { GSMSTAT_ASS_EST,                      "Assign/Hand Est" },
    { GSMSTAT_ASS_CPTE,                     "Assignment Complete" },
    { GSMSTAT_MODE_MOD,                     "Mode Modify" },

    { 0, NULL }
};

static const value_string gsmstat_usage_vals[] = {
    { GSMSTAT_CHAN_TYPE_TCH,    "TCH" },
    { GSMSTAT_CHAN_TYPE_SDCCH,  "SDCCH" },
    { GSMSTAT_CHAN_TYPE_CIC,    "CIC" },

    { 0, NULL }
};

char *gsmstat_name(int tag)
{
    return rd_match_strval(tag, gsmstat_vals);
}

char *gsmstat_usage_name(int tag)
{
    return rd_match_strval(tag, gsmstat_usage_vals);
}


static const value_string intface_vals[] = {
    { GSMSTAT_INTERFACE_A,      "A Interface" },
    { GSMSTAT_INTERFACE_A_BIS,  "A-bis Interface" },
    { 0, NULL }
};

char *gsmstat_intface_name(int tag)
{
    return rd_match_strval(tag, intface_vals);
}


static int a_counts[] = {
                            GSMSTAT_PAGE_REQS,
                            GSMSTAT_PAGE_RESPONSE,
                            GSMSTAT_ASS_REQS,
                            GSMSTAT_ASS_FAILS,
                            GSMSTAT_CLEAR_REQS,
                            GSMSTAT_CLEAR_CMDS,
                            GSMSTAT_HANDOVER_REQUEST,
                            GSMSTAT_HANDOVER_REQUIRED,
                            GSMSTAT_HANDOVER_ALLOC,
                            GSMSTAT_HANDOVER_REQUIRED_REJECT,
                            GSMSTAT_HANDOVER_COMMAND,
                            GSMSTAT_HANDOVER_DETECT,
                            GSMSTAT_HANDOVER_COMPLETE,
                            GSMSTAT_HANDOVER_FAILURE,
                            GSMSTAT_LU_REQS,
                            GSMSTAT_LU_ACCEPT,
                            GSMSTAT_LU_REJECTS,
                            GSMSTAT_SETUPS,
                            GSMSTAT_CONNECTS,
                            GSMSTAT_DISC_NORMAL,
                            GSMSTAT_DISC_ABNORMAL,
                            GSMSTAT_MSGS,
                            GSMSTAT_IPPKTS,
                            GSMSTAT_IPLEN,
                        };

static int a_bis_counts[] = {
                                GSMSTAT_CHAN_REQS,
                                GSMSTAT_PAGE_REQS,
	                            GSMSTAT_PAGE_RESPONSE,
                                GSMSTAT_IMM_ASS_ATTEMPTS,
                                GSMSTAT_IMM_ASS_SUCCESSES,
                                GSMSTAT_ASS_CMD,
                                GSMSTAT_ASS_EST,
                                GSMSTAT_MODE_MOD,
                                GSMSTAT_ASS_CPTE,
                                GSMSTAT_ERROR_REPS,
                                GSMSTAT_ERROR_INDS,
                                GSMSTAT_PREPROC_MEAS_RESULTS,
                                GSMSTAT_CONN_FAIL_SDCCH,
                                GSMSTAT_CONN_FAIL_TCH,
                                GSMSTAT_HANDOVER_COMMAND,
                                GSMSTAT_HANDOVER_COMPLETE,
	                            GSMSTAT_HANDOVER_DETECT,
                                GSMSTAT_HANDOVER_FAILURE,
                                GSMSTAT_LU_REQS,
                                GSMSTAT_LU_ACCEPT,
                                GSMSTAT_LU_REJECTS,
                                GSMSTAT_SETUPS,
                                GSMSTAT_CONNECTS,
                                GSMSTAT_DISC_NORMAL,
                                GSMSTAT_DISC_ABNORMAL,
                                GSMSTAT_MSGS,
                                GSMSTAT_IPPKTS,
                                GSMSTAT_IPLEN,
                             };

static int a_bis_usage[] = {    GSMSTAT_CHAN_TYPE_TCH,
                                GSMSTAT_CHAN_TYPE_SDCCH,
                            };
static int a_usage[] = {    GSMSTAT_CHAN_TYPE_CIC,
                            };

int gsmstat_counts_max( int intface )
{
    switch (intface) {
    case GSMSTAT_INTERFACE_A:
        return (sizeof(a_counts)/sizeof(int));
        break;

    case GSMSTAT_INTERFACE_A_BIS:
        return (sizeof(a_bis_counts)/sizeof(int));
        break;

    default:
        return 0;    /* safe to have none */
        break;
    }
}

int gsmstat_usage_max( int intface )
{
    switch (intface) {
    case GSMSTAT_INTERFACE_A_BIS:
        return (sizeof(a_bis_usage)/sizeof(int));
        break;

    case GSMSTAT_INTERFACE_A:
        return (sizeof(a_usage)/sizeof(int));
        break;

    default:
        return 0;    /* safe to have none */
        break;
    }
}

int gsmstat_counts( int intface, int index )
{
    switch (intface) {
    case GSMSTAT_INTERFACE_A:
        return (a_counts[index]);
        break;

    case GSMSTAT_INTERFACE_A_BIS:
        return (a_bis_counts[index]);
        break;

    default:
        return 0;    /* safe to return first count */
        break;
    }
}

int gsmstat_usage( int intface, int index )
{
    switch (intface) {
    case GSMSTAT_INTERFACE_A_BIS:
        return (a_bis_usage[index]);
        break;

    case GSMSTAT_INTERFACE_A:
        return (a_usage[index]);
        break;

    default:
        return 0;    /* safe to return first count */
        break;
    }
}

void
gsmstat_reset_a(void)
{
	gsmstat_a.conns = 0;
}

int
gsmstat_add_a_conn(int ref)
{
	int i = gsmstat_a.conns;

	gsmstat_a.conn[i].cic = -1;
	gsmstat_a.conn[i].slr = -1;
	gsmstat_a.conn[i].dlr = ref;
	gsmstat_a.conns++;

	return i;
}

int
gsmstat_find_a_conn(int ref)
{
	int i, ret=-1;

	for (i=0; i<gsmstat_a.conns; i++) {
		if (ref == gsmstat_a.conn[i].dlr) {
			ret = i;
			break;
		}
	}

	return ret;
}

void
gsmstat_del_a_conn(int ref)
{
	int i;

	for (i=0; i<gsmstat_a.conns; i++) {
		if (ref == gsmstat_a.conn[i].dlr) {
			/* remove found connection and substitute with the last one if present */
			gsmstat_a.conn[i]=gsmstat_a.conn[--gsmstat_a.conns];
			break;
		}
	}

	return;
}

/* This function resets the collection bins - ensures done is FALSE!
 * Note: we can't rely
 * on the packet function clearing the statics as if we are called with a filter we may get the
 * packet function only called for some of the GSM packets, not all. */
void
gsmstat_set_reset(void)
{
    gsmstat_pkt.secs        = 0;                        /* start at zero time ? */
    gsmstat_pkt.pnum        = 0;                        /* await the first packet */
    gsmstat_pkt.intface        = GSMSTAT_INTERFACE_NULL;    /* mark the collection as empty */
    gsmstat_pkt.msg_count    = -1;

    return;
}

/* This function sets the interface type (A, A-bis, Gb etc) or NULL to clear the setting.
 * Note: we can't rely
 * on the packet function clearing the statics as if we are called with a filter we may get the
 * packet function only called for some of the GSM packets, not all. */
void
gsmstat_set_interface(int intface, packet_info *pinfo)
{
    if (pinfo->fd->num != gsmstat_pkt.pnum) { /* new packet - set the interface type, packet number & clear the rest */
        gsmstat_pkt.secs        = pinfo->fd->abs_ts.secs;
        gsmstat_pkt.intface        = intface;
        gsmstat_pkt.msg_count    = -1;
        gsmstat_pkt.pnum        = pinfo->fd->num;
        gsmstat_pkt.intface_count    = 1;
    }
    else
        gsmstat_pkt.intface_count++;    /* just note it is another top-level call */
    /* all other cases are more info from the same packet - nothing to do */
    return;
}

/* This function sets the protocol (gsm0408 etc.) - it must be called before any other call except interface (above). */
void
gsmstat_set_proto(int proto)
{
    int i;

    i = ++(gsmstat_pkt.msg_count);    /* move to the next message store */
                                    /* reset the next message store - the rest are reset as they are used */
    gsmstat_pkt.msg[i].proto        = proto;    /* store the proto */
    gsmstat_pkt.msg[i].pd            = GSMSTAT_PD_NULL;
    gsmstat_pkt.msg[i].msg_type        = GSMSTAT_MSG_TYPE_NULL;
    gsmstat_pkt.msg[i].cause        = GSMSTAT_CAUSE_NULL;
    gsmstat_pkt.msg[i].chan_type    = GSMSTAT_CHAN_TYPE_NULL;

    return;
}

/* This function sets the protocol discriminator (only really for gsm0408  MM, CC, etc.). */
void
gsmstat_set_pd(int pd)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].pd = pd;

    return;
}

/* This function sets the msg_type (SETUP etc.) - NOTE: only unique within a protocol! */
void
gsmstat_set_msg_type(int msg_type)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].msg_type = msg_type;

    return;
}

/* This function sets the cause (NORMAL etc.) - NOTE: only unique within a protocol and message type! */
void
gsmstat_set_cause(int cause)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].cause = cause;

    return;
}

/* This function sets the Channel Type (TCH, SDCCH etc.) */
void
gsmstat_set_chan_type(int chan_type)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].chan_type = chan_type;

    return;
}

/* This function sets the Channel Id (For A interface: DLR, for A-bis: TS+subchan) */
void
gsmstat_set_chan_id(int chan_id)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].chan_id = chan_id;

    return;
}

/* This function sets the A Interface CIC */
void
gsmstat_set_cic(int cic)
{
    gsmstat_pkt.msg[gsmstat_pkt.msg_count].cic    = cic;

    return;
}

/* This callback is never used by tethereal but it is here for completeness.
 * When registering below, we could just have left this function as NULL.
 *
 * When used by ethereal, this function will be called whenever we would need
 * to reset all state. Such as when ethereal opens a new file, when it
 * starts a new capture, when it rescans the packetlist after some prefs have
 * changed etc.
 * So if your aplication has some state it needs to clean up in those
 * situations, here is a good place to put that code.
 */
void
gsmstat_common_reset(gsmstat_common_t *gsc)
{
    int i;
    /* reset stats */
	gsmstat_reset_a();
    gsc->secs = 0;

    for (i=0; i<GSMSTAT_MAX; i++)
        gsc->count[i] = 0;

    for (i=0; i<GSMSTAT_CHAN_TYPE_MAX; i++) {
        gsc->usage[i].current = 0;
        gsc->usage[i].chan_secs = 0;
        gsc->usage[i].chan_usecs = 0;
        gsc->usage[i].last_secs = 0;
        gsc->usage[i].last_usecs = 0;
        gsc->usage[i].peak = 0;
        gsc->usage[i].peak_count = 0;
        gsc->usage[i].peak_secs = 0;
        gsc->usage[i].peak_usecs = 0;
    }

    return;
}


/* This callback is invoked whenever the tap system has seen a packet
 * we might be interested in.
 * The function is to be used to only update internal state information
 * in the *tapdata structure, and if there were state changes which requires
 * the window to be redrawn, return 1 and (*draw) will be called sometime
 * later.
 *
 * This function should be as lightweight as possible since it executes together
 * with the normal ethereal dissectors. Try to push as much processing as
 * possible into (*draw) instead since that function executes asynchronously
 * and does not affect the main threads performance.
 *
 * If it is possible, try to do all "filtering" explicitely as we do below in
 * this example since you will get MUCH better performance than applying
 * a similar display-filter in the register call.
 *
 * The third parameter is tap dependant. Since we register this one to the "gsm"
 * tap the third parameters type is gsm_call_info_value.
 *
 *
 * The filtering we do is just to check the gsm_call_info_value struct that
 * we were called for the proper program and version. We didnt apply a filter
 * when we registered so we will be called for ALL gsm packets and not just
 * the ones we are collecting stats for.
 *
 *
 * function returns :
 *  0: no updates, no need to call (*draw) later
 * !0: state has changed, call (*draw) sometime later
 */
int
gsmstat_common_packet(gsmstat_common_t *gsc, packet_info *pinfo, const void *pri _U_)
{
    gsmstat_pkt_t *gp=&gsmstat_pkt;

    int rc=0, i;

    gsc->secs = pinfo->fd->rel_ts.secs;

    if (gsc->intface != gp->intface)    /* only collect stats relevant to the chosen interface */
        return 0;

    gsc->count[GSMSTAT_MSGS] +=    gsmstat_pkt.intface_count;    /* number of top-level messages */
    gsc->count[GSMSTAT_IPPKTS]++;                            /* number of IP packets */
    gsc->count[GSMSTAT_IPLEN] += pinfo->iplen;                /* number of IP packet octets */

    for (i=0; i<=gp->msg_count; i++)
        rc |=  gsmstat_common_message(gsc, &(gp->msg[i]), pinfo);

    gp->msg_count = -1;    /* finished processing this packet - if the tap has been called twice, this prevents us counting twice! */

    return rc;
}

int
gsmstat_common_message(gsmstat_common_t *gsc, gsmstat_msg_t *gm, packet_info *pinfo)
{
    int rc=1;            /* default to requesting redraw - should be safe */

    switch (gm->proto) {
    case GSMSTAT_PROTO_GSM0818:
        rc = 0;            /* and no redraw required */
        break;

    case GSMSTAT_PROTO_GSM0858:
        switch (gm->msg_type) {
            case GSM0858_MSG_TYPE_CHANNEL_REQUIRED:
                gsc->count[GSMSTAT_CHAN_REQS]++;
                break;
            case GSM0858_MSG_TYPE_PAGING_COMMAND:
                gsc->count[GSMSTAT_PAGE_REQS]++;
                break;
            case GSM0858_MSG_TYPE_IMMEDIATE_ASSIGN:
                gsc->count[GSMSTAT_IMM_ASS_ATTEMPTS]++;
                break;
            case GSM0858_MSG_TYPE_ESTABLISH_IND:
                if (gm->cause==GSMSTAT_GSM0858_L3INFO)    /* inclusion of L3Info in Est Ind means it was the imediate assign - I think! What about handover ? */
                    gsc->count[GSMSTAT_IMM_ASS_SUCCESSES]++;
                else
                    gsc->count[GSMSTAT_ASS_EST]++;		/* from either Assignment Command or Handover Command */
                break;
            case GSM0858_MSG_TYPE_ERROR_REPORT:
                gsc->count[GSMSTAT_ERROR_REPS]++;
                break;
            case GSM0858_MSG_TYPE_ERROR_INDICATION:
                gsc->count[GSMSTAT_ERROR_INDS]++;
                break;
            case GSM0858_MSG_TYPE_PREPROC_MEAS_RESULTS:
                gsc->count[GSMSTAT_PREPROC_MEAS_RESULTS]++;
                break;
            case GSM0858_MSG_TYPE_CONNECTION_FAIL_IND:
				if (gm->chan_type == GSMSTAT_CHAN_TYPE_SDCCH)
					gsc->count[GSMSTAT_CONN_FAIL_SDCCH]++;
				else if (gm->chan_type == GSMSTAT_CHAN_TYPE_TCH)
					gsc->count[GSMSTAT_CONN_FAIL_TCH]++;
				else
			        gsc->count[GSMSTAT_ERRORS]++;
                break;
            case GSM0858_MSG_TYPE_CHAN_ACTIVATE_ACK:
                update_usage_stats(&gsc->usage[gm->chan_type], 1, pinfo);
                break;
            case GSM0858_MSG_TYPE_RF_CHAN_REL:
                update_usage_stats(&gsc->usage[gm->chan_type], -1, pinfo);
                break;
            case GSM0858_MSG_TYPE_HANDOVER_DETECT:
                gsc->count[GSMSTAT_HANDOVER_DETECT]++;
               break;

            default:        /* do nothing on other messages */
                rc = 0;            /* and no redraw required */
                break;
        }
        break;

    case GSMSTAT_PROTO_GSM0408:
        /* Note: at present if we monitor both A and A-bis interfaces,
         * 04.08 messages except RR will be counted twice! */

        switch (gm->pd) {
        case GSM0408_PD_RR:
            switch (gm->msg_type) {
                case GSM0408_MSG_TYPE_PAGING_RESPONSE:
                    gsc->count[GSMSTAT_PAGE_RESPONSE]++;
                    break;
                case GSM0408_MSG_TYPE_ASSIGNMENT_COMMAND:
                    gsc->count[GSMSTAT_ASS_CMD]++;
                    break;
                case GSM0408_MSG_TYPE_ASSIGNMENT_COMPLETE:
                    gsc->count[GSMSTAT_ASS_CPTE]++;
                    break;
                case GSM0408_MSG_TYPE_CHANNEL_MODE_MODIFY:
                    gsc->count[GSMSTAT_MODE_MOD]++;
                    break;
                case GSM0408_MSG_TYPE_HANDOVER_COMMAND:
                    gsc->count[GSMSTAT_HANDOVER_COMMAND]++;
                    break;
                case GSM0408_MSG_TYPE_HANDOVER_COMPLETE:
                    gsc->count[GSMSTAT_HANDOVER_COMPLETE]++;
                    break;
                case GSM0408_MSG_TYPE_HANDOVER_FAILURE:
                    gsc->count[GSMSTAT_HANDOVER_FAILURE]++;
                    break;
                default:        /* do nothing on other message types */
                    rc = 0;        /* and no redraw required */
                    break;
            }
            break;

        case GSM0408_PD_MM:
            switch (gm->msg_type) {
                case GSM0408_MSG_TYPE_LOCATION_UPDATE_REQUEST:
                    gsc->count[GSMSTAT_LU_REQS]++;
                    break;
                case GSM0408_MSG_TYPE_LOCATION_UPDATE_ACCEPT:
                    gsc->count[GSMSTAT_LU_ACCEPT]++;
                    break;
                case GSM0408_MSG_TYPE_LOCATION_UPDATE_REJECT:
                    gsc->count[GSMSTAT_LU_REJECTS]++;
                    break;
                default:        /* do nothing on other message types */
                    rc = 0;        /* and no redraw required */
                    break;
            }
            break;

        case GSM0408_PD_CC:
            switch (gm->msg_type) {
                case GSM0408_MSG_TYPE_SETUP:
                    gsc->count[GSMSTAT_SETUPS]++;
                    break;
                case GSM0408_MSG_TYPE_CONNECT:
                    gsc->count[GSMSTAT_CONNECTS]++;
                    break;
                case GSM0408_MSG_TYPE_DISCONNECT:
                    switch (gm->cause) {
                    case GSM0408_CAUSE_NORMAL_CALL_CLEARING:
					case GSM0408_CAUSE_USER_BUSY:
					case GSM0408_CAUSE_NO_USER_RESPONDING:
					case GSM0408_CAUSE_USER_ALERTING_NO_ANSWER:
					case GSM0408_CAUSE_CALL_REJECTED:
					case GSM0408_CAUSE_NORMAL_UNSPECIFIED:
                        gsc->count[GSMSTAT_DISC_NORMAL]++;
                        break;
					case GSM0408_CAUSE_UNASSIGNED_NUMBER:
					case GSM0408_CAUSE_NO_ROUTE_TO_DESTINATION:
					case GSM0408_CAUSE_OPERATOR_DETERMINED_BARRING:
                    default:        /* all other causes abnormal */
                        gsc->count[GSMSTAT_DISC_ABNORMAL]++;
                        break;
                    }
                    break;
                default:        /* do nothing on other message types */
                    rc = 0;        /* and no redraw required */
                    break;
            }
            break;

        case GSMSTAT_PD_NULL:    /* PD should have been set for all GSM0408 */
            gsc->count[GSMSTAT_ERRORS]++;
            break;

        default:        /* do nothing on other PDs */
            rc = 0;        /* and no redraw required */
            break;
        }
        break;

    case GSMSTAT_PROTO_GSM0808:
        switch (gm->msg_type) {
            case GSM0808_MSG_TYPE_ASSIGNMENT_REQUEST:
                gsc->count[GSMSTAT_ASS_REQS]++;
                break;
			case GSM0808_MSG_TYPE_ASSIGNMENT_COMPLETE:
				gsmstat_add_a_conn(gm->chan_id);
                update_usage_stats(&gsc->usage[GSMSTAT_CHAN_TYPE_CIC], 1, pinfo);
                break;
            case GSM0808_MSG_TYPE_ASSIGNMENT_FAILURE:
                gsc->count[GSMSTAT_ASS_FAILS]++;
                break;
            case GSM0808_MSG_TYPE_CLEAR_REQUEST:
                gsc->count[GSMSTAT_CLEAR_REQS]++;
                break;
            case GSM0808_MSG_TYPE_CLEAR_COMMAND:
                gsc->count[GSMSTAT_CLEAR_CMDS]++;
                break;
            case GSM0808_MSG_TYPE_CLEAR_COMPLETE:
				if (gsmstat_find_a_conn(gm->chan_id) != -1) {	/* existance will do for now - dont need cic or state */
					update_usage_stats(&gsc->usage[GSMSTAT_CHAN_TYPE_CIC], -1, pinfo);
					gsmstat_del_a_conn(gm->chan_id);
				}
                break;
            case GSM0808_MSG_TYPE_PAGING:
                gsc->count[GSMSTAT_PAGE_REQS]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_REQUEST:
                gsc->count[GSMSTAT_HANDOVER_REQUEST]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_REQUIRED:
                gsc->count[GSMSTAT_HANDOVER_REQUIRED]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_REQUIRED_REJECT:
                gsc->count[GSMSTAT_HANDOVER_REQUIRED_REJECT]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_FAILURE:
                gsc->count[GSMSTAT_HANDOVER_FAILURE]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_COMPLETE:
                gsc->count[GSMSTAT_HANDOVER_COMPLETE]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_DETECT:
                gsc->count[GSMSTAT_HANDOVER_DETECT]++;
                break;
            case GSM0808_MSG_TYPE_HANDOVER_REQUEST_ACK:
				gsmstat_add_a_conn(gm->chan_id);
                update_usage_stats(&gsc->usage[GSMSTAT_CHAN_TYPE_CIC], 1, pinfo);
                gsc->count[GSMSTAT_HANDOVER_ALLOC]++;
                break;
            default:        /* do nothing on other messages */
                rc = 0;            /* and no redraw required */
                break;
        }
        break;

    case GSMSTAT_PROTO_NULL:
        gsc->count[GSMSTAT_ERRORS]++;
        break;

    default:    /* shouldn't get any! */
        gsc->count[GSMSTAT_ERRORS]++;
        break;
    }

    /* update the time-dependant stats with the latest time */
    if (pinfo->fd->rel_ts.secs>0) {
        gsc->secs = pinfo->fd->rel_ts.secs;
        gsc->usecs = pinfo->fd->rel_ts.nsecs/1000;
        update_usage_stats(&gsc->usage[GSMSTAT_CHAN_TYPE_TCH], 0, pinfo );
        update_usage_stats(&gsc->usage[GSMSTAT_CHAN_TYPE_SDCCH], 0, pinfo );
    }

    return rc;
}

void update_usage_stats(usage_stats_t *use, int delta, packet_info *pinfo)
{
    int secs;

    /* update usage based on current channels and time since last update */
    use->chan_secs += use->current * (pinfo->fd->rel_ts.secs - use->last_secs);
    use->chan_usecs += use->current * (pinfo->fd->rel_ts.nsecs/1000 - use->last_usecs);
    if ((use->chan_usecs > 1000000) || (use->chan_usecs < 0)) {
        if (use->chan_usecs < 0)
            secs = (use->chan_usecs / 1000000) - 1;
        else
            secs = use->chan_usecs / 1000000;
        use->chan_secs += secs;
        use->chan_usecs -= secs * 1000000;
    }

    /* update peak stats */
    if (use->current == use->peak) {
        use->peak_secs += (pinfo->fd->rel_ts.secs - use->last_secs);
        use->peak_usecs += (pinfo->fd->rel_ts.nsecs/1000 - use->last_usecs);
        if ((use->peak_usecs > 1000000) || (use->peak_usecs < 0)) {
            if (use->peak_usecs < 0)
                secs = (use->peak_usecs / 1000000) - 1;
            else
                secs = use->peak_usecs / 1000000;
            use->peak_secs += secs;
            use->peak_usecs -= secs * 1000000;
        }
    }
    /* and update update time */
    use->last_secs = pinfo->fd->rel_ts.secs;
    use->last_usecs = pinfo->fd->rel_ts.nsecs/1000;

    /* update current channels */
    use->current += delta;
    if (use->current > use->peak) {
        use->peak = use->current;
        use->peak_count = 0;
        use->peak_secs = 0;
        use->peak_usecs = 0;
    }
	if ((use->current == use->peak) && (delta == 1)) {
		use->peak_count++;
	}

    /* take care with decrease - if channels had been assigned before trace began,
                                current could go negative! Adjust in this case */
    if (use->current < 0) {    /* adjustment needed - from start of trace */
        use->peak -= use->current;
        use->chan_secs -= use->current * pinfo->fd->rel_ts.secs;
        use->chan_usecs -= use->current * (pinfo->fd->rel_ts.nsecs/1000);
        use->current = 0;
        if ((use->chan_usecs > 1000000) || (use->chan_usecs < 0)) {
            if (use->chan_usecs < 0)
                secs = (use->chan_usecs / 1000000) - 1;
            else
                secs = use->chan_usecs / 1000000;
            use->chan_secs += secs;
            use->chan_usecs -= secs * 1000000;
        }
    }

    return;
}

int gsmstat_tap = -1;
void gsmstat_register_tap(void)    /* a function to hide multiple protocols behind a single tap */
{
    if (gsmstat_tap == -1)    /* only register once!!! */
        gsmstat_tap = register_tap("gsm");

    return;
}

