/* packet-gsm.h
 * Headers for GSM message dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <epan/packet.h>

#define GSM0408_HEADER_LEN   2
#define GSM0808_HEADER_LEN   1
#define IPA_HEADER_LEN       3
#define IPA_UDP_HEADER_LEN   4

#define MAX_DATA_STR      5

#define rd_match_strval(a,b)	((char *)val_to_str((a), (const value_string *)(b), "Unknown"))

/* maximum numbers for signalling over IP conversation tracking */
#define BTS_MAX				200
#define CONFIG_MGR_LINK_MAX	10
#define BTS_LINK_MAX		(2*BTS_MAX)			/* OML & RSL */
#define IPA_LINK_MAX		(1+CONFIG_MGR_LINK_MAX+BTS_LINK_MAX)	/* BTS links + config manager + SCCP over IP */

/* defines for the Signalling over IP streams */
#define OML_STREAM			255
#define CTRL_STREAM			254
#define SCCP_STREAM			253
#define SPIRIT_STREAM		253
#define BSC_MGMT_STREAM		251
#define BSC_SUBAGENT_STREAM 250
#define SIGTRAN_STREAM		249
#define FRIP_STREAM			248
#define GNAT_STREAM			247
#define IML_STREAM			129
#define RSL_STREAM_MAX		 32

/* defines for the fixed TCP/UDP port numbers */
#define TCP_PORT_IPA_OML     3002
#define TCP_PORT_IPA_RSL     3003
#define TCP_PORT_IPA_2NDOML  3006
#define TCP_PORT_IPA_HW_MON  3007
#define TCP_PORT_IPA_FRIP	 3008
#define TCP_PORT_IPA_SPIRIT	 3008
#define TCP_PORT_IPA_GNAT    3009
#define TCP_PORT_IPA_BSCMGMT 3010
#define TCP_PORT_IPA_A_IF	 3011
#define TCP_PORT_IPA_SITE_IML 3013
#define TCP_PORT_IPA_BTS_IML 3014
#define UDP_PORT_IPA         3006

/* defines for the GSMSTAT interface */
#define GSMSTAT_INTERFACE_NULL    -1
#define GSMSTAT_INTERFACE_A        0
#define GSMSTAT_INTERFACE_A_BIS    1
#define GSMSTAT_INTERFACE_GB       2
#define GSMSTAT_INTERFACE_MAX      2 /* until we add GB! */

#define GSMSTAT_PROTO_NULL        -1
#define GSMSTAT_PROTO_GSM0408      1
#define GSMSTAT_PROTO_GSM0808      2
#define GSMSTAT_PROTO_GSM0858      3
#define GSMSTAT_PROTO_GSM0818      4
#define GSMSTAT_PROTO_GSM1221      5

#define GSMSTAT_PD_NULL            -1

/* 04.08 Protocol Discriminator values */
#define GSM0408_PD_VGCS     0x0
#define GSM0408_PD_VBS      0x1
#define GSM0408_PD_PDSS1    0x2
#define GSM0408_PD_CC       0x3
#define GSM0408_PD_PDSS2    0x4
#define GSM0408_PD_MM       0x5
#define GSM0408_PD_RR       0x6
#define GSM0408_PD_GMM      0x8
#define GSM0408_PD_SMS      0x9
#define GSM0408_PD_SM       0xa
#define GSM0408_PD_SS       0xb
#define GSM0408_PD_LS       0xc

/* RTP Paylod types in use for A-bis */
#define RTP_PAYLOAD_GSM_FR	0
#define RTP_PAYLOAD_GSM_EFR	1
#define RTP_PAYLOAD_GSM_AMR	2
#define RTP_PAYLOAD_GSM_CSD	3
#define RTP_PAYLOAD_GSM_NANO_GSM_MUX	4

#define GSMSTAT_MSG_TYPE_NULL    -1
/* 08.58 Message Type values */
#define GSM0858_MSG_TYPE_ESTABLISH_IND               0x06
#define GSM0858_MSG_TYPE_ERROR_INDICATION            0x03
#define GSM0858_MSG_TYPE_CHANNEL_REQUIRED            0x13
#define GSM0858_MSG_TYPE_PAGING_COMMAND              0x15
#define GSM0858_MSG_TYPE_IMMEDIATE_ASSIGN            0x16
#define GSM0858_MSG_TYPE_ERROR_REPORT                0x1c
#define GSM0858_MSG_TYPE_CHAN_ACTIVATE               0x21
#define GSM0858_MSG_TYPE_CHAN_ACTIVATE_ACK           0x22
#define GSM0858_MSG_TYPE_CONNECTION_FAIL_IND         0x24
#define GSM0858_MSG_TYPE_HANDOVER_DETECT             0x27
#define GSM0858_MSG_TYPE_RF_CHAN_REL                 0x2e
#define GSM0858_MSG_TYPE_RF_CHAN_REL_ACK             0x33
#define GSM0858_MSG_TYPE_PREPROC_MEAS_RESULTS        0x32
#define GSM0858_MSG_TYPE_CREATE_CONNECTION		     0x70
#define GSM0858_MSG_TYPE_CREATE_CONNECTION_ACK	     0x71
#define GSM0858_MSG_TYPE_MODIFY_CONNECTION		     0x73
#define GSM0858_MSG_TYPE_MODIFY_CONNECTION_ACK	     0x74
#define GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION 0x53
/* 04.08 RR Message Type values */
#define GSM0408_MSG_TYPE_PAGING_RESPONSE             0x27
#define GSM0408_MSG_TYPE_HANDOVER_FAILURE            0x28
#define GSM0408_MSG_TYPE_HANDOVER_COMMAND            0x2b
#define GSM0408_MSG_TYPE_HANDOVER_COMPLETE           0x2c
#define GSM0408_MSG_TYPE_ASSIGNMENT_COMMAND          0x2e
#define GSM0408_MSG_TYPE_ASSIGNMENT_COMPLETE         0x29
#define GSM0408_MSG_TYPE_ASSIGNMENT_FAILURE          0x2f
#define GSM0408_MSG_TYPE_CHANNEL_MODE_MODIFY         0x10
#define GSM0408_MSG_TYPE_CHANNEL_MODE_MODIFY_ACK     0x17
/* 04.08 MM Message Type values */
#define GSM0408_MSG_TYPE_LOCATION_UPDATE_ACCEPT      0x02
#define GSM0408_MSG_TYPE_LOCATION_UPDATE_REJECT      0x04
#define GSM0408_MSG_TYPE_LOCATION_UPDATE_REQUEST     0x08
/* 04.08 CC Message Type values */
#define GSM0408_MSG_TYPE_SETUP                       0x05
#define GSM0408_MSG_TYPE_CONNECT                     0x07
#define GSM0408_MSG_TYPE_DISCONNECT                  0x25
/* GSMSTAT Cause values */
#define GSMSTAT_CAUSE_NULL                            -1
#define GSMSTAT_GSM0858_L3INFO                       0x01
/* GSM 04.08 Cause values */
#define GSM0408_CAUSE_UNASSIGNED_NUMBER              0x01
#define GSM0408_CAUSE_NO_ROUTE_TO_DESTINATION        0x03
#define GSM0408_CAUSE_OPERATOR_DETERMINED_BARRING    0x08
#define GSM0408_CAUSE_NORMAL_CALL_CLEARING           0x10
#define GSM0408_CAUSE_USER_BUSY                      0x11
#define GSM0408_CAUSE_NO_USER_RESPONDING             0x12
#define GSM0408_CAUSE_USER_ALERTING_NO_ANSWER        0x13
#define GSM0408_CAUSE_CALL_REJECTED                  0x15
#define GSM0408_CAUSE_NORMAL_UNSPECIFIED             0x1f
/* GSMSTAT Chan_type values */
#define GSMSTAT_CHAN_TYPE_NULL                        -1
#define GSMSTAT_CHAN_TYPE_TCH                        0x00
#define GSMSTAT_CHAN_TYPE_SDCCH                      0x01
#define GSMSTAT_CHAN_TYPE_CIC                        0x02
#define GSMSTAT_CHAN_TYPE_MAX                        3
/* 08.08 Message Type values */
#define GSM0808_MSG_TYPE_ASSIGNMENT_REQUEST          0x01
#define GSM0808_MSG_TYPE_ASSIGNMENT_COMPLETE         0x02
#define GSM0808_MSG_TYPE_ASSIGNMENT_FAILURE          0x03
#define GSM0808_MSG_TYPE_HANDOVER_REQUEST            0x10
#define GSM0808_MSG_TYPE_HANDOVER_REQUIRED           0x11
#define GSM0808_MSG_TYPE_HANDOVER_REQUEST_ACK        0x12
#define GSM0808_MSG_TYPE_HANDOVER_COMPLETE			 0x14
#define GSM0808_MSG_TYPE_HANDOVER_FAILURE			 0x16
#define GSM0808_MSG_TYPE_HANDOVER_REQUIRED_REJECT    0x1A
#define GSM0808_MSG_TYPE_HANDOVER_DETECT             0x1B
#define GSM0808_MSG_TYPE_CLEAR_COMMAND               0x20
#define GSM0808_MSG_TYPE_CLEAR_COMPLETE              0x21
#define GSM0808_MSG_TYPE_CLEAR_REQUEST               0x22
#define GSM0808_MSG_TYPE_PAGING                      0x52

typedef void (*ie_dissector_t)(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);

/* Note: the value parameter is to keep the dissector easy for type 1 IEs which can be in either half octet of the TVB */
typedef void (*gsm0408_ie_dissector_t)(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);

typedef struct ie_def_t {
    guint8    ie_ei;
    char    *ie_name;
    gint8    ie_len;    /* 0..N=fixed length; -1=1 octet length field, -2=2 octet length field etc. */
    ie_dissector_t    ie_disector;
    void    *info;    /* extra IE-dependent info */
} ie_def_t;

/* IE types defined in 04.07:
    I        1/2 octet V with optional T
    II        T only
    III        V or TV
    IV        LV or TLV
*/

typedef struct ie_t {
    int                tag;        /* tag to match the tag in the ie_id_t */
    guint8            type;        /* as defined in 04.07 - value 1..4 */
    guint8            len;        /* only used for type 3 fixed length */
    gsm0408_ie_dissector_t    dissector;    /* function to dissect the IE */
    void            *info;        /* info to be used by the dissector specific to this IE - e.g cause vale strings */
} ie_t;

typedef struct ie_id_t {
    int                tag;        /* really a tag to match the tag in the ie_t */
    char            *name;        /* display string for the IE name */
    guint8            iei;        /* standard messages have message dependent IEI values */
} ie_id_t;

#define MSG_IE_MAX    40            /* maximum number of IEs in a message -
                                    compiler will throw an error in the initialisers
                                    if this is not big enough ! */

typedef struct msg_t {
    guint8            format;        /* format 1 is normal, format 2 for RACH, format 3 for SACCH */
    guint8            type;        /* msg type value */
    char            *name;        /* name string */
    ie_id_t            ie_id[MSG_IE_MAX];    /* array of mandatory and optional IE identifiers */
} msg_t;

/* Struct for passing info to the GSM stats tap function */
#define ARG_MAX        10

/* struct to hold the data from the message parsing to the tap function */
typedef struct tap_gsmstat_t {
    guint8            proto;            /* GSM protocol */
    guint8            type;            /* msg type value */
} tap_gsmstat_t;

/* definitions of the counters */
#define GSMSTAT_CHAN_REQS                    0
#define GSMSTAT_PAGE_REQS                    1
#define GSMSTAT_IMM_ASS_ATTEMPTS             2
#define GSMSTAT_IMM_ASS_SUCCESSES            3
#define GSMSTAT_HANDOVER_COMMAND             4
#define GSMSTAT_HANDOVER_COMPLETE            5
#define GSMSTAT_HANDOVER_FAILURE             6
#define GSMSTAT_LU_REQS                      7
#define GSMSTAT_LU_ACCEPT                    8
#define GSMSTAT_LU_REJECTS                   9
#define GSMSTAT_SETUPS                      10
#define GSMSTAT_CONNECTS                    11
#define GSMSTAT_DISC_NORMAL                 12
#define GSMSTAT_DISC_ABNORMAL               13
#define GSMSTAT_ASS_REQS                    14
#define GSMSTAT_ASS_FAILS                   15
#define GSMSTAT_CLEAR_REQS                  16
#define GSMSTAT_ERRORS                      17
#define GSMSTAT_PREPROC_MEAS_RESULTS        18
#define GSMSTAT_CONN_FAIL_SDCCH             19
#define GSMSTAT_ERROR_REPS                  20
#define GSMSTAT_MSGS                        21
#define GSMSTAT_OCTETS                      22
#define GSMSTAT_IPPKTS                      23
#define GSMSTAT_IPLEN                       24
#define GSMSTAT_HANDOVER_REQUEST            25
#define GSMSTAT_HANDOVER_REQUIRED           26
#define GSMSTAT_HANDOVER_REQUIRED_REJECT    27
#define GSMSTAT_PAGE_RESPONSE			    28
#define GSMSTAT_HANDOVER_DETECT             29
#define GSMSTAT_CONN_FAIL_TCH               30
#define GSMSTAT_HANDOVER_ALLOC              31
#define GSMSTAT_CLEAR_CMDS                  32
#define GSMSTAT_ASS_CMD                     33
#define GSMSTAT_ASS_EST                     34
#define GSMSTAT_ASS_CPTE                    35
#define GSMSTAT_MODE_MOD                    36
#define GSMSTAT_ERROR_INDS                  37
#define GSMSTAT_MAX                         38

/* used to keep track of erlangs etc */
typedef struct {
    int current;
    int peak;
    int peak_count;
    int peak_secs;
    int peak_usecs;
    int last_secs;
    int last_usecs;
    int chan_usecs;
    int chan_secs;
} usage_stats_t;

/* used to keep track of the statistics for an entire tap instance - only one for GSM, but... */
typedef struct _gsmstat_common_t {
    int  secs;
    int  usecs;
    int  intface;
    int  count[GSMSTAT_MAX];
    usage_stats_t usage[GSMSTAT_CHAN_TYPE_MAX];
} gsmstat_common_t;

void dissect_gsm_generic_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree);

/* gsmstat interface routines */
void gsmstat_set_reset(void);
void gsmstat_set_interface(int intface, packet_info *pinfo);
void gsmstat_set_proto(int proto);
void gsmstat_set_pd(int pd);
void gsmstat_set_msg_type(int msg_type);
void gsmstat_set_cause(int cause);
void gsmstat_set_chan_type(int chan_type);
void gsmstat_set_chan_id(int chan_id);
void gsmstat_set_cic(int cic);
void gsmstat_common_reset(gsmstat_common_t *gsc);
int gsmstat_common_packet(gsmstat_common_t *gsc, packet_info *pinfo, const void *pri);
gchar *gsmstat_name(int tag);
gchar *gsmstat_intface_name(int tag);
int gsmstat_counts_max(int intface);
int gsmstat_counts(int intface, int index);
gchar *gsmstat_usage_name(int tag);
int gsmstat_usage_max(int intface);
int gsmstat_usage(int intface, int index);
/* global data and a helper function to share a single tap across multiple protocols - and thus potentially multiple regsitrations etc */
extern int gsmstat_tap;
void gsmstat_register_tap(void);

/* Some helper routines */
/* gchar *rd_match_strval(guint32 val, const value_string *vs); */
gchar *data_str( tvbuff_t *tvb, int offset, int data_len, gchar *buf, int buf_len );
guint16 reduced_frame_number(guint16 fn);

/* a few nasty cross-protocol routines for analysis */
int analyze_immediate_assignment(guint8 ra, guint16 fn, packet_info *pinfo);
