/* packet-gsm0408.c
 * Routines for GSM 04.08 message dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0408.c 1.5 2009/02/24 07:03:36GMT Mike (mm3) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include <epan/prefs.h>
#include "packet-gsm.h"

static void dissect_gsm0408_ie( const msg_t *p_msg_defs, guint8 msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
static void dissect_default_gsm0408_ie(guint8 ie_type, guint8 value, tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
static void dissect_gsm0408_cause_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_rr_cause_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_request_reference_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_channel_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_ie_mobile_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_ie_unt8_vals(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_ie_unt8_vals_type_1(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_cp_user_data_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_neighbour_cell_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_gsm0408_cell_channel_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value);
static void dissect_rest_octets(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
void proto_reg_handoff_gsm0408(void);

/* useful statics */
static char hex_digit[]="0123456789abcdef";

static dissector_handle_t gsm0408_handle;
static dissector_handle_t gsm0408_l2_handle;
static dissector_handle_t gsm0408_l3_contents_handle;
static dissector_handle_t gsm_a_dtap_handle;
static dissector_handle_t rp_handle;
static dissector_handle_t data_handle;

/* Initialize the protocol and registered fields */
static int proto_gsm0408 = -1;
static int hf_gsm0408_rr_msg_type = -1;
static int hf_gsm0408_mm_msg_type = -1;
static int hf_gsm0408_cc_msg_type = -1;
static int hf_gsm0408_gmm_msg_type = -1;
static int hf_gsm0408_sm_msg_type = -1;
static int hf_gsm0408_sms_msg_type = -1;
static int hf_gsm0408_ss_msg_type = -1;
static int hf_gsm0408_pd = -1;
static int hf_gsm0408_cause = -1;
static int hf_gsm0408_rr_cause = -1;
static int hf_gsm0408_ra = -1;
static int hf_gsm0408_fn = -1;
static int hf_gsm0408_assignment_delay = -1;
static int hf_gsm0408_ts = -1;
static int hf_gsm0408_tsc = -1;
static int hf_gsm0408_chan = -1;
static int hf_gsm0408_subchan = -1;
static int hf_gsm0408_arfcn = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0408 = -1;
static gint ett_gsm0408_ie = -1;

/* Global preferences */
static gboolean global_gsm0408_use_ipa = FALSE;

/* Value Strings */

static const value_string gsm0408_pd_vals[] = {
	{ 0x0, "VGCS" },
	{ 0x1, "VBS" },
	{ 0x2, "PDSS1" },
	{ 0x3, "CC" },
	{ 0x4, "PDSS2" },
	{ 0x5, "MM" },
	{ 0x6, "RR" },
	{ 0x8, "GMM" },
	{ 0X9, "SMS" },
	{ 0xa, "SM" },
	{ 0xb, "SS" },
	{ 0xc, "LS" },
	{ 0, NULL, }
};

static const value_string chan_vals[] = {
	{ 0x01,  "Bm" },
	{ 0x02,  "Lm/0" },
	{ 0x03,  "Lm/1" },
	{ 0x04,  "SDCCH4/0" },
	{ 0x05,  "SDCCH4/1" },
	{ 0x06,  "SDCCH4/2" },
	{ 0x07,  "SDCCH4/3" },
	{ 0x08,  "SDCCH8/0" },
	{ 0x09,  "SDCCH8/1" },
	{ 0x0a,  "SDCCH8/2" },
	{ 0x0b,  "SDCCH8/3" },
	{ 0x0c,  "SDCCH8/4" },
	{ 0x0d,  "SDCCH8/5" },
	{ 0x0e,  "SDCCH8/6" },
	{ 0x0f,  "SDCCH8/7" },
	{ 0x10,  "BCCH" },
	{ 0x11,  "RACH" },
	{ 0x12,  "CCCH" },
	{ 0, NULL }
};

static const value_string cm_service_type_vals[] = {
	{ 0x1, "MO Call or Packet mode" },
	{ 0x2, "Emergency Call" },
	{ 0x4, "SMS" },
	{ 0x8, "SS Activation" },
	{ 0x9, "Voice Group Call" },
	{ 0xa, "Voice Broadcast" },
	{ 0xb, "Location Services" },
	{ 0, NULL, }
};


static const value_string cp_cause_vals[] = {
	{ 0x11, "Network Failure" },
	{ 0x16, "Congestion" },
	{ 0x51, "Invalid Transaction Identifier Value" },
	{ 0x5f, "Semantically incorrect message" },
	{ 0x60, "Invalid mandatory information" },
	{ 0x61, "Message type non-existent or not implemented" },
	{ 0x62, "Message not compatible with the short message protocol state" },
	{ 0x63, "Information element non-existent or not implemented" },
	{ 0x6f, "Protocol error, unspecified" },
	{ 0, NULL }
};

static const value_string rr_cause_vals[] = {
	{ 0x00, "Normal event" },
	{ 0x01, "Abnormal release, unspecified" },
	{ 0x02, "Abnormal release, channel unacceptable" },
	{ 0x03, "Abnormal release, timer expired" },
	{ 0x04, "Abnormal release, no activity on the radio path" },
	{ 0x05, "Preemptive release" },
	{ 0x06, "UTRAN configuration unknown" },
	{ 0x08, "Handover impossible, timing advance out of range" },
	{ 0x09, "Channel mode unacceptable" },
	{ 0x0a, "Frequency not implemented" },
	{ 0x0c, "Lower layer failure" },
	{ 0x21, "Call already cleared" },
	{ 0x5f, "Semantically incorrect message" },
	{ 0x60, "Invalid mandatory information" },
	{ 0x61, "Message type non-existent or not implemented" },
	{ 0x62, "Message type not compatible with protocol state" },
	{ 0x64, "Conditional IE error" },
	{ 0x65, "No cell allocation available" },
	{ 0x6f, "Protocol error unspecified" },
	{ 0, NULL }
};

static const value_string channel_mode_vals[] = {
	{ 0x00, "signalling only" },
	{ 0x01, "speech full rate or half rate version 1" },
	{ 0x21, "speech full rate or half rate version 2" },
	{ 0x41, "speech full rate or half rate version 3" },
	{ 0x61, "data, 43.5 kbit/s (downlink)+14.5 kbps (uplink)" },
	{ 0x62, "data, 29.0 kbit/s (downlink)+14.5 kbps (uplink)" },
	{ 0x64, "data, 43.5 kbit/s (downlink)+29.0 kbps (uplink)" },
	{ 0x67, "data, 14.5 kbit/s (downlink)+43.5 kbps (uplink)" },
	{ 0x65, "data, 14.5 kbit/s (downlink)+29.0 kbps (uplink)" },
	{ 0x66, "data, 29.0 kbit/s (downlink)+43.5 kbps (uplink)" },
	{ 0x47, "data, 43.5 kbit/s radio interface rate" },
	{ 0x63, "data, 32.0 kbit/s radio interface rate" },
	{ 0x43, "data, 29.0 kbit/s radio interface rate" },
	{ 0x0f, "data, 14.5 kbit/s radio interface rate" },
	{ 0x03, "data, 12.0 kbit/s radio interface rate" },
	{ 0x0b, "data, 6.0 kbit/s radio interface rate" },
	{ 0x13, "data, 3.6 kbit/s radio interface rate" },
	{ 0, NULL }
};

static const value_string channel_mode_2_vals[] = {
	{ 0x00, "signalling only" },
	{ 0x05, "speech half rate version 1" },
	{ 0x25, "speech half rate version 2" },
	{ 0x45, "speech half rate version 3" },
	{ 0x0f, "data, 6.0 kbit/s radio interface rate" },
	{ 0x17, "data, 3.6 kbit/s radio interface rate" },
	{ 0, NULL }
};

static const value_string cipher_response_vals[] = {
	{ 0x00, "IMEISV shall not be included" },
	{ 0x01, "IMEISV shall be included" },
	{ 0, NULL }
};

static const value_string gprs_resumption_vals[] = {
	{ 0x00, "resumption of GPRS services not successfully acknowledged" },
	{ 0x01, "resumption of GPRS services successfully acknowledged" },
	{ 0, NULL }
};

static const value_string nc_mode_vals[] = {
	{ 0x00, "NC 0" },
	{ 0x01, "NC 1" },
	{ 0x02, "NC 2" },
	{ 0x03, "information on NC mode to be gathered from the target cell" },
	{ 0, NULL }
};

static const value_string page_mode_vals[] = {
	{ 0x00, "Normal paging" },
	{ 0x01, "Extended paging" },
	{ 0x02, "Paging reorganization" },
	{ 0x03, "Same as before" },
	{ 0, NULL }
};

static const value_string suspension_cause_vals[] = {
	{ 0x00, "Emergency call, mobile originating call or call re-establishment" },
	{ 0x01, "Location Area Update" },
	{ 0x02, "MO Short message service" },
	{ 0x03, "Other procedure which can be completed with an SDCCH" },
	{ 0x04, "MO Voice broadcast or group call" },
	{ 0x05, "Mobile terminating CS connection" },
	{ 0x06, "DTM not supported in the cell" },
	{ 0, NULL }
};

static const value_string ciphering_key_sequence_number_vals[] = {
	{ 0x00, "Key 0" },
	{ 0x01, "Key 1" },
	{ 0x02, "Key 2" },
	{ 0x03, "Key 3" },
	{ 0x04, "Key 4" },
	{ 0x05, "Key 5" },
	{ 0x06, "Key 6" },
	{ 0x07, "No key available" },
	{ 0, NULL }
};

static const value_string priority_level_vals[] = {
	{ 0x00, "no priority applied" },
	{ 0x01, "call priority level 4" },
	{ 0x02, "call priority level 3" },
	{ 0x03, "call priority level 2" },
	{ 0x04, "call priority level 1" },
	{ 0x05, "call priority level 0" },
	{ 0x06, "call priority level B" },
	{ 0x07, "call priority level A" },
	{ 0, NULL }
};

static const value_string identity_type_vals[] = {
	{ 0x00, "No Identity" },
	{ 0x01, "IMSI" },
	{ 0x02, "IMEI" },
	{ 0x03, "IMEISV" },
	{ 0x04, "TMSI/P-TMSI" },
	{ 0, NULL }
};
#define MOBILE_IDENTITY_TYPE_TMSI	0x04

static const value_string location_updating_type_vals[] = {
	{ 0x00, "Normal location updating, no follow-on request pending" },
	{ 0x01, "Periodic updating, no follow-on request pending" },
	{ 0x02, "IMSI attach, no follow-on request pending" },
	{ 0x04, "Normal location updating, follow-on request pending" },
	{ 0x05, "Periodic updating, follow-on request pending" },
	{ 0x06, "IMSI attach, follow-on request pending" },
	{ 0, NULL }
};

static const value_string reject_cause_vals[] = {
	{ 0x02, "IMSI unknown in HLR" },
	{ 0x03, "Illegal MS" },
	{ 0x04, "IMSI unknown in VLR" },
	{ 0x05, "IMEI not accepted" },
	{ 0x06, "Illegal ME" },
	{ 0x0b, "PLMN not allowed" },
	{ 0x0c, "Location Area not allowed" },
	{ 0x0d, "Roaming not allowed in this location area" },
	{ 0x0f, "No Suitable Cells In Location Area" },
	{ 0x11, "Network failure" },
	{ 0x14, "MAC failure" },
	{ 0x15, "Synch failure" },
	{ 0x16, "Congestion" },
	{ 0x17, "GSM authentication unacceptable" },
	{ 0x20, "Service option not supported" },
	{ 0x21, "Requested service option not subscribed" },
	{ 0x22, "Service option temporarily out of order" },
	{ 0x26, "Call cannot be identified" },
	{ 0x30, "retry upon entry into a new cell" },
	{ 0x31, "retry upon entry into a new cell" },
	{ 0x32, "retry upon entry into a new cell" },
	{ 0x33, "retry upon entry into a new cell" },
	{ 0x34, "retry upon entry into a new cell" },
	{ 0x35, "retry upon entry into a new cell" },
	{ 0x36, "retry upon entry into a new cell" },
	{ 0x37, "retry upon entry into a new cell" },
	{ 0x38, "retry upon entry into a new cell" },
	{ 0x39, "retry upon entry into a new cell" },
	{ 0x3a, "retry upon entry into a new cell" },
	{ 0x3b, "retry upon entry into a new cell" },
	{ 0x3c, "retry upon entry into a new cell" },
	{ 0x3d, "retry upon entry into a new cell" },
	{ 0x3e, "retry upon entry into a new cell" },
	{ 0x3f, "retry upon entry into a new cell" },
	{ 0x5f, "Semantically incorrect message" },
	{ 0x60, "Invalid mandatory information" },
	{ 0x61, "Message type non-existent or not implemented" },
	{ 0x62, "Message type not compatible with the protocol state" },
	{ 0x63, "Information element non-existent or not implemented" },
	{ 0x64, "Conditional IE error" },
	{ 0x65, "Message not compatible with the protocol state" },
	{ 0x6f, "Protocol error, unspecified" },
	{ 0, NULL }
};

static const value_string cause_vals[] = {
	{ 1,  "Unassigned (unallocated) number" },
	{ 3,  "No route to destination" },
	{ 6,  "Channel unacceptable" },
	{ 8,  "Operator determined barring" },
	{ 16, "Normal call clearing" },
	{ 17, "User busy" },
	{ 18, "No user responding" },
	{ 19, "User alerting, no answer" },
	{ 21, "Call rejected - user supplied diagnostic" },
	{ 22, "Number changed New destination" },
	{ 25, "Pre-emption" },
	{ 26, "Non selected user clearing" },
	{ 27, "Destination out of order" },
	{ 28, "Invalid number format (incomplete number)" },
	{ 29, "Facility rejected" },
	{ 30, "Response to STATUS ENQUIRY" },
	{ 31, "Normal, unspecified" },
	{ 34, "No circuit/channel available" },
	{ 38, "Network out of order" },
	{ 41, "Temporary failure" },
	{ 42, "Switching equipment congestion" },
	{ 43, "Access information discarded Discarded information element identifiers" },
	{ 44, "Rquested circuit/channel not available" },
	{ 47, "Resources unavailable, unspecified" },
	{ 49, "Quality of service unavailable" },
	{ 50, "Requested facility not subscribed" },
	{ 55, "Incoming calls barred within the CUG" },
	{ 57, "Bearer capability not authorized" },
	{ 58, "Bearer capability not presently available" },
	{ 63, "Service or option not available, unspecified" },
	{ 65, "Bearer service not implemented" },
	{ 68, "ACM equal to or greater than ACMmax" },
	{ 69, "Requested facility not implemented" },
	{ 70, "Only restricted digital information bearer capability is available" },
	{ 79, "Service or option not implemented, unspecified" },
	{ 81, "Invalid transaction identifier value" },
	{ 87, "User not member of CUG" },
	{ 88, "Incompatible destination Incompatible parameter" },
	{ 91, "Invalid transit network selection" },
	{ 95, "Semantically incorrect message" },
	{ 96, "Invalid mandatory information Information element identifier(s)" },
	{ 97, "Message type non-existent or not implemented Message type" },
	{ 98, "Message type not compatible with protocol state Message type" },
	{ 99, "Information element non-existent or not implemented Information element identifier(s)" },
	{ 100, "Conditional IE error Information element identifier(s)" },
	{ 101, "Message not compatible with protocol state" },
	{ 102, "Recovery on timer expiry Timer number" },
	{ 111, "Protocol error, unspecified" },
	{ 127, "Interworking, unspecified" },
	{ 0, NULL }
};

static const value_string daylight_saving_time_vals[] = {
	{ 0x00, "No adjustment for Daylight Saving Time" },
	{ 0x01, "+1 hour adjustment for Daylight Saving Time" },
	{ 0x02, "+2 hours adjustment for Daylight Saving Time" },
	{ 0, NULL }
};

static const value_string congestion_level_vals[] = {
	{ 0x00, "receiver ready" },
	{ 0xff, "receiver not ready" },
	{ 0, NULL }
};

static const value_string notification_indicator_vals[] = {
	{ 0x00, "User suspended" },
	{ 0x01, "User resumed" },
	{ 0x02, "Bearer change" },
	{ 0, NULL }
};

static const value_string repeat_indicator_vals[] = {
	{ 0x01, "Circular for successive selection - mode 1 alternate mode 2" },
	{ 0x02, "Support of fallback - mode 1 preferred, mode 2 selected if setup of mode 1 fails" },
	{ 0x03, "reserved: was allocated in earlier phases of the protocol" },
	{ 0, NULL }
};

static const value_string signal_vals[] = {
	{ 0x00, "dial tone on" },
	{ 0x01, "ring back tone on" },
	{ 0x02, "intercept tone on" },
	{ 0x03, "network congestion tone on" },
	{ 0x04, "busy tone on" },
	{ 0x05, "confirm tone on" },
	{ 0x06, "answer tone on" },
	{ 0x07, "call waiting tone on" },
	{ 0x08, "off-hook warning tone on" },
	{ 0x3f, "tones off" },
	{ 0x4f, "alerting off" },
	{ 0, NULL }
};

static const value_string cause_of_no_cli_vals[] = {
	{ 0x00, "Unavailable" },
	{ 0x01, "Reject by user" },
	{ 0x02, "Interaction with other service" },
	{ 0x03, "Coin line/payphone" },
	{ 0, NULL }
};

static const value_string attach_result_vals[] = {
	{ 0x01, "GPRS only attached" },
	{ 0x03, "Combined GPRS/IMSI attached" },
	{ 0, NULL }
};

static const value_string attach_type_vals[] = {
	{ 0x1, "GPRS attach, no follow-on request pending" },
	{ 0x2, "GPRS attach while IMSI attached, no follow-on request pending" },
	{ 0x3, "Combined GPRS/IMSI attach, no follow-on request pending" },
	{ 0x5, "GPRS attach, no follow-on request pending" },
	{ 0x6, "GPRS attach while IMSI attached, no follow-on request pending" },
	{ 0x7, "Combined GPRS/IMSI attach, no follow-on request pending" },
	{ 0, NULL }
};

static const value_string ciphering_algorithm_vals[] = {
	{ 0x00, "ciphering not used" },
	{ 0x01, "GPRS Encryption Algorithm GEA/1" },
	{ 0x02, "GPRS Encryption Algorithm GEA/2" },
	{ 0x03, "GPRS Encryption Algorithm GEA/3" },
	{ 0x04, "GPRS Encryption Algorithm GEA/4" },
	{ 0x05, "GPRS Encryption Algorithm GEA/5" },
	{ 0x06, "GPRS Encryption Algorithm GEA/6" },
	{ 0x07, "GPRS Encryption Algorithm GEA/7" },
	{ 0, NULL }
};

static const value_string tmsi_status_vals[] = {
	{ 0x00, "no valid TMSI available" },
	{ 0x01, "valid TMSI available" },
	{ 0, NULL }
};

#define CHANNEL_LIST_BIT_MAP_0			0
#define CHANNEL_LIST_1024_RANGE			1
#define CHANNEL_LIST_512_RANGE			2
#define CHANNEL_LIST_256_RANGE			3
#define CHANNEL_LIST_128_RANGE			4
#define CHANNEL_LIST_VARIABLE_BIT_MAP	5
#define CHANNEL_LIST_UNKNOWN			6
#define ARFCN_MAX					 1023

static const value_string channel_list_format_vals[] = {
	{ CHANNEL_LIST_BIT_MAP_0, "Bit Map 0" },
	{ CHANNEL_LIST_1024_RANGE, "1024 Range" },
	{ CHANNEL_LIST_512_RANGE, "512 Range" },
	{ CHANNEL_LIST_256_RANGE, "256 Range" },
	{ CHANNEL_LIST_128_RANGE, "128 Range" },
	{ CHANNEL_LIST_VARIABLE_BIT_MAP, "Variable Bit Map" },
	{ CHANNEL_LIST_UNKNOWN, "Unknown" },
	{ 0, NULL }
};

enum ie_names {
	START_OF_IE_NAMES=0,
	ALERTING_PATTERN,
	ALLOWED_ACTIONS,
	AUTH_PARAMETER_RAND,
	AUTH_PARAMETER_SRES,
	AUXILIARY_STATES,
	BA_LIST_PREF,
	BA_RANGE,
	BEARER_CAPABILITY,
	CALL_CONTROL_CAPABILITIES,
	CALL_STATE,
	CALLING_PARTY_BCD_NUM,
	CALLING_PARTY_SUBADDR,
	CALLED_PARTY_BCD_NUM,
	CALLED_PARTY_SUBADDR,
	CAUSE,
	CELL_CHANNEL_DESCRIPTION,
	CELL_DESCRIPTION,
	CELL_IDENTITY,
	CELL_OPTIONS_BCCH,
	CELL_OPTIONS_SACCH,
	CELL_SELECTION_PARAMETERS,
	CHANNEL_MODE,
	CHANNEL_MODE_2,
	CHANNEL_DESCRIPTION,
	CHANNEL_DESCRIPTION_2,
	CIPHER_MODE_SETTING,
	CIPHER_RESPONSE,
	CIPHERING_KEY_SEQUENCE_NUMBER,
	CLIR_SUPPRESSION,
	CLIR_INVOCATION,
	CM_SERVICE_TYPE,
	CONGESTION_LEVEL,
	CONNECTED_NUMBER,
	CONNECTED_SUBADDR,
	CONTAINER,
	CONTROL_CHANNEL_DESCRIPTION,
	CP_USER_DATA,
	CP_CAUSE,
	CTS_PERMISSION,
	DEDICATED_MODE_OR_TBF,
	EXTENDED_MEASUREMENT_FREQUENCY_LIST,
	EXTENDED_MEASUREMENT_RESULTS,
	FACILITY,
	FOLLOW_ON_PROCEED,
	FREQUENCY_CHANNEL_SEQUENCE,
	FREQUENCY_LIST,
	FREQUENCY_SHORT_LIST,
	GPRS_RESUMPTION,
	GROUP_CHANNEL_DESCRIPTION,
	GROUP_CIPHER_KEY_NUMBER,
	HANDOVER_REFERENCE,
	HIGHER_LAYER_COMP,
	IDENTITY_TYPE,
	KEYPAD_FACILITY,
	LOCATION_AREA_IDENTIFICATION,
	LOCATION_UPDATING_TYPE,
	LOWER_LAYER_COMP,
	LSA_IDENTIFIER,
	MEASUREMENT_RESULTS,
	MOBILE_ALLOCATION,
	MOBILE_IDENTITY,
	MOBILE_STATION_CLASSMARK_1,
	MOBILE_STATION_CLASSMARK_2,
	MOBILE_STATION_CLASSMARK_3,
	MOBILE_TIME_DIFFERENCE,
	MORE_DATA,
	MULTIRATE_CONFIGURATION,
	MULTISLOT_ALLOCATION,
	NCC_PERMITTED,
	NEIGHBOUR_CELL_DESCRIPTION,
	NEIGHBOUR_CELL_DESCRIPTION_2,
	NETWORK_NAME,
	NOTIFICATION_INDICATOR,
	PAGE_MODE,
	POWER_COMMAND,
	POWER_COMMAND_AND_ACCESS_TYPE,
	PRIORITY_LEVEL,
	PROGRESS_INDICATOR,
	RACH_CONTROL_PARAMETERS,
	REAL_TIME_DIFFERENCE,
	RECALL_TYPE,
	REDIRECTING_PARTY_BCD_NUM,
	REDIRECTING_PARTY_SUBADDR,
	REPEAT_INDICATOR,
	REQUEST_REFERENCE,
	ROUTING_AREA_IDENTIFICATION,
	REJECT_CAUSE,
	REVERSE_CALL_SETUP_DIRECTION,
	RR_CAUSE,
	SIGNAL,
	SPARE_HALF_OCTET,
	SS_VERSION,
	STARTING_TIME,
	SUSPENSION_CAUSE,
	SYNCHRONISATION_INDICATION,
	TIME_ZONE,
	TIME_ZONE_AND_TIME,
	TIMING_ADVANCE,
	TLLI,
	USER_USER,
	VGCS_TARGET_MODE_INDICATION,
	WAIT_INDICATION,
	END_OF_IE_NAMES
};

static const ie_t ie_defs[] = {
	{ ALERTING_PATTERN, 4, 0, NULL, NULL },
	{ ALLOWED_ACTIONS, 4, 0, NULL, NULL },
	{ AUTH_PARAMETER_RAND, 3, 17, NULL, NULL},
	{ AUTH_PARAMETER_SRES, 3, 5, NULL, NULL},
	{ AUXILIARY_STATES, 4, 0, NULL, NULL },
	{ BA_LIST_PREF, 4, 0, NULL, NULL },
	{ BA_RANGE, 4, 0, NULL, NULL },
	{ BEARER_CAPABILITY, 4, 0, NULL, NULL },
	{ CALL_CONTROL_CAPABILITIES, 4, 0, NULL, NULL },
	{ CALL_STATE, 3, 2, NULL, NULL},
	{ CALLING_PARTY_BCD_NUM, 4, 0, NULL, NULL },
	{ CALLING_PARTY_SUBADDR, 4, 0, NULL, NULL },
	{ CALLED_PARTY_BCD_NUM, 4, 0, NULL, NULL },
	{ CALLED_PARTY_SUBADDR, 4, 0, NULL, NULL },
	{ CAUSE, 4, 0, dissect_gsm0408_cause_ie, NULL },
	{ CELL_CHANNEL_DESCRIPTION, 3, 17, dissect_gsm0408_cell_channel_description_ie, NULL },
	{ CELL_DESCRIPTION, 3, 3, NULL, NULL },
	{ CELL_IDENTITY, 3, 3, NULL, NULL },
	{ CELL_OPTIONS_BCCH, 3, 2, NULL, NULL },
	{ CELL_OPTIONS_SACCH, 3, 2, NULL, NULL },
	{ CELL_SELECTION_PARAMETERS, 3, 3, NULL, NULL },
	{ CHANNEL_MODE, 3, 2, NULL, (void *)channel_mode_vals },
	{ CHANNEL_MODE_2, 3, 2, NULL, (void *)channel_mode_2_vals },
	{ CHANNEL_DESCRIPTION, 3, 4, dissect_gsm0408_channel_description_ie, NULL },
	{ CHANNEL_DESCRIPTION_2, 3, 4, dissect_gsm0408_channel_description_ie, NULL },
	{ CIPHER_MODE_SETTING, 1, 0, NULL, NULL },
	{ CIPHER_RESPONSE, 1, 0, NULL, (void *)cipher_response_vals },
	{ CIPHERING_KEY_SEQUENCE_NUMBER, 1, 0, NULL, (void *)ciphering_key_sequence_number_vals },
	{ CM_SERVICE_TYPE, 1, 0, NULL, (void *)cm_service_type_vals },
	{ CLIR_SUPPRESSION, 2, 0, NULL, NULL },
	{ CLIR_INVOCATION, 2, 0, NULL, NULL },
	{ CONGESTION_LEVEL, 1, 0, NULL, (void *)congestion_level_vals },
	{ CONTAINER, 4, 0, NULL, NULL },
	{ CONTROL_CHANNEL_DESCRIPTION, 3, 4, NULL, NULL },
	{ CONNECTED_NUMBER, 4, 0, NULL, NULL },
	{ CONNECTED_SUBADDR, 4, 0, NULL, NULL },
	{ CP_USER_DATA, 4, 0, dissect_gsm0408_cp_user_data_ie, NULL },
	{ CP_CAUSE, 3, 1, NULL, (void *)cp_cause_vals },
	{ CTS_PERMISSION, 2, 0, NULL, NULL },
	{ DEDICATED_MODE_OR_TBF, 1, 0, NULL, NULL },
	{ EXTENDED_MEASUREMENT_FREQUENCY_LIST, 3, 17, NULL, NULL},
	{ EXTENDED_MEASUREMENT_RESULTS, 3, 17, NULL, NULL},
	{ FACILITY, 4, 0, NULL, NULL },
	{ FOLLOW_ON_PROCEED, 2, 0, NULL, NULL },
	{ FREQUENCY_CHANNEL_SEQUENCE, 3, 10, NULL, NULL },
	{ FREQUENCY_LIST, 4, 0, NULL, NULL },
	{ FREQUENCY_SHORT_LIST, 3, 10, NULL, NULL},
	{ GPRS_RESUMPTION, 1, 0, NULL, (void *)gprs_resumption_vals },
	{ GROUP_CHANNEL_DESCRIPTION, 4, 0, NULL, NULL },
	{ GROUP_CIPHER_KEY_NUMBER, 1, 0, NULL, NULL },
	{ HANDOVER_REFERENCE, 3, 2, NULL, NULL},
	{ HIGHER_LAYER_COMP, 4, 0, NULL, NULL },
	{ IDENTITY_TYPE, 1, 0, NULL, (void *)identity_type_vals },
	{ KEYPAD_FACILITY, 3, 2, NULL, NULL },
	{ LOCATION_AREA_IDENTIFICATION, 3, 6, NULL, NULL },
	{ LOCATION_UPDATING_TYPE, 1, 0, NULL, (void *)location_updating_type_vals },
	{ LOWER_LAYER_COMP, 3, 2, NULL, NULL },
	{ LSA_IDENTIFIER, 4, 0, NULL, NULL },
	{ MEASUREMENT_RESULTS, 3, 17, NULL, NULL },
	{ MULTIRATE_CONFIGURATION, 4, 0, NULL, NULL },
	{ MULTISLOT_ALLOCATION, 4, 0, NULL, NULL },
	{ MOBILE_ALLOCATION, 4, 0, NULL, NULL },
	{ MOBILE_STATION_CLASSMARK_1, 3, 2, NULL, NULL },
	{ MOBILE_STATION_CLASSMARK_2, 4, 0, NULL, NULL },
	{ MOBILE_STATION_CLASSMARK_3, 4, 0, NULL, NULL },
	{ MOBILE_TIME_DIFFERENCE, 4, 0, NULL, NULL },
	{ MOBILE_IDENTITY, 4, 0, dissect_gsm0408_ie_mobile_identity, NULL },
	{ MORE_DATA, 2, 0, NULL, NULL },
	{ NCC_PERMITTED, 3, 2, NULL, NULL },
	{ NEIGHBOUR_CELL_DESCRIPTION, 3, 17, dissect_gsm0408_neighbour_cell_description_ie, NULL },
	{ NEIGHBOUR_CELL_DESCRIPTION_2, 3, 17, NULL, NULL },
	{ NETWORK_NAME, 4, 0, NULL, NULL },
	{ NOTIFICATION_INDICATOR, 3, 2, NULL, (void *)notification_indicator_vals },
	{ PAGE_MODE, 1, 0, NULL, (void *)page_mode_vals },
	{ POWER_COMMAND, 3, 2, NULL, NULL },
	{ POWER_COMMAND_AND_ACCESS_TYPE, 3, 2, NULL, NULL},
	{ PRIORITY_LEVEL, 1, 0, NULL, (void *)priority_level_vals },
	{ PROGRESS_INDICATOR, 4, 0, NULL, NULL },
	{ RACH_CONTROL_PARAMETERS, 3, 4, NULL, NULL },
	{ REAL_TIME_DIFFERENCE, 4, 0, NULL, NULL},
	{ RECALL_TYPE, 3, 2, NULL, NULL},
	{ REDIRECTING_PARTY_BCD_NUM, 4, 0, NULL, NULL },
	{ REDIRECTING_PARTY_SUBADDR, 4, 0, NULL, NULL },
	{ REPEAT_INDICATOR, 1, 0, NULL, (void *)repeat_indicator_vals },
	{ REQUEST_REFERENCE, 3, 4, dissect_gsm0408_request_reference_ie, NULL },
	{ REVERSE_CALL_SETUP_DIRECTION, 2, 0, NULL, NULL },
	{ ROUTING_AREA_IDENTIFICATION, 3, 7, NULL, NULL },
	{ REJECT_CAUSE, 3, 2, NULL, (void *)reject_cause_vals },
	{ RR_CAUSE, 3, 2, dissect_gsm0408_rr_cause_ie, NULL },
	{ SIGNAL, 3, 2, NULL, (void *)signal_vals },
	{ SPARE_HALF_OCTET, 1, 0, NULL, NULL },
	{ SS_VERSION, 4, 0, NULL, NULL },
	{ STARTING_TIME, 3, 3, NULL, NULL },
	{ SUSPENSION_CAUSE, 3, 2, NULL, (void *)suspension_cause_vals },
	{ SYNCHRONISATION_INDICATION, 1, 0, NULL, NULL},
	{ TIME_ZONE, 3, 2, NULL, NULL },
	{ TIME_ZONE_AND_TIME, 3, 8, NULL, NULL },
	{ TIMING_ADVANCE, 3, 2, NULL, NULL },
	{ TLLI, 3, 5, NULL, NULL },
	{ USER_USER, 4, 0, NULL, NULL },
	{ VGCS_TARGET_MODE_INDICATION, 3, 2, NULL, NULL },
	{ WAIT_INDICATION, 3, 2, NULL, NULL},
	{ NULL, 0, 0, NULL, NULL }
};

static const msg_t sms_msg_defs[] = {
/* SMS Messages */
	{ 1, 0x01, "CP Data", {
		{CP_USER_DATA, "CP User Data", 0},
		{0,"",0}
	}},
	{ 1, 0x04, "CP Ack", {
		{0,"",0}
	}},
	{ 1, 0x10, "CP Error", {
		{CP_CAUSE, "CP Cause", 0},
		{0,"",0}
	}},
	{ 0, 0x00, NULL, {0,0}}
};

static const msg_t rr_msg_defs[] = {
/* Radio Resource Messages */
	{ 1, 0x00, "System Information Type 13", {	/* Only Rest Octets */
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x02, "System Information Type 2bis", {
		{NEIGHBOUR_CELL_DESCRIPTION,"Extended BCCH Frequency List",0},
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x03, "System Information Type 2ter", {
		{NEIGHBOUR_CELL_DESCRIPTION_2,"Extended BCCH Frequency List",0},
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x04, "System Information Type 9", {
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x05, "System Information Type 5bis", {
		{NEIGHBOUR_CELL_DESCRIPTION,"Extension of BCCH Frequency List",0},
		{0,"",0}
	}},
	{ 1, 0x06, "System Information Type 5ter", {
		{NEIGHBOUR_CELL_DESCRIPTION_2,"Extended BCCH Frequency List",0},
		{0,"",0}
	}},
	{ 1, 0x07, "System Information Type 2quater", {	/* Only Rest Octets */
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x0d, "Channel Release", {
		{RR_CAUSE, "RR Cause", 0},
		{BA_RANGE, "BA Range", 0x73},
		{GROUP_CHANNEL_DESCRIPTION, "Group Channel Description", 0x74},
		{GROUP_CIPHER_KEY_NUMBER, "Group Cipher Key Number", 0x8},
		{GPRS_RESUMPTION, "GPRS Resumption", 0xc},
		{BA_LIST_PREF, "BA List Pref", 0x75},
		{0,"",0}
	}},
	{ 1, 0x10, "Channel Mode Modify", {
		{CHANNEL_DESCRIPTION_2, "Channel Description", 0},
		{CHANNEL_MODE, "Channel Mode", 0},
		{VGCS_TARGET_MODE_INDICATION, "VGCS Target Mode Indication", 0x01},
		{MULTIRATE_CONFIGURATION, "Multi-Rate Configuration", 0x03},
		{0,"",0}
	}},
	{ 1, 0x12, "RR Status", {
		{RR_CAUSE, "RR Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x15, "Measuement Report", {
		{MEASUREMENT_RESULTS, "Measurement Results", 0},
		{0,"",0}
	}},
	{ 1, 0x16, "Classmark Change", {
		{MOBILE_STATION_CLASSMARK_2, "Mobile Station Classmark 2", 0},
		{MOBILE_STATION_CLASSMARK_3, "Mobile Station Classmark 3", 0x20},
		{0,"",0}
	}},
	{ 1, 0x17, "Channel Mode Modify Acknowledge", {
		{CHANNEL_DESCRIPTION_2, "Channel Description", 0},
		{CHANNEL_MODE, "Channel Mode", 0},
		{0,"",0}
	}},
	{ 1, 0x18, "System Information Type 8", {	/* Only Rest Octets */
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x19, "System Information Type 1", {
		{CELL_CHANNEL_DESCRIPTION, "Cell Channel Description", 0},
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x1a, "System Information Type 2", {
		{NEIGHBOUR_CELL_DESCRIPTION,"BCCH Frequency List",0},
		{NCC_PERMITTED, "NCC Permitted", 0},
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{0,"",0}
	}},
	{ 1, 0x1b, "System Information Type 3", {
		{CELL_IDENTITY,"Cell Identity",0},
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{CONTROL_CHANNEL_DESCRIPTION, "Control Channel Description", 0},
		{CELL_OPTIONS_BCCH, "Cell Options", 0},
		{CELL_SELECTION_PARAMETERS, "Cell Selection Parameters", 0},
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{0,"",0}
	}},
	{ 1, 0x1c, "System Information Type 4", {
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{CELL_SELECTION_PARAMETERS, "Cell Selection Parameters", 0},
		{RACH_CONTROL_PARAMETERS, "RACH Control Parameters", 0},
		{CHANNEL_DESCRIPTION, "CBCH Channel Description", 0x64},
		{MOBILE_ALLOCATION, "CBCH Mobile Allocation", 0x72},
		{0,"",0}
	}},
	{ 1, 0x1d, "System Information Type 5", {
		{NEIGHBOUR_CELL_DESCRIPTION,"BCCH Frequency List",0},
		{0,"",0}
	}},
	{ 1, 0x1e, "System Information Type 6", {
		{CELL_IDENTITY,"Cell Identity",0},
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{CELL_OPTIONS_SACCH, "Cell Options", 0},
		{NCC_PERMITTED, "NCC Permitted", 0},
		{0,"",0}
	}},
	{ 1, 0x1f, "System Information Type 7", {	/* Only Rest Octets */
		{0,"",0}		/* + Rest Octets */
	}},
	{ 1, 0x27, "Paging Response", {
		{CIPHERING_KEY_SEQUENCE_NUMBER, "Ciphering Key Sequence Number", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{MOBILE_STATION_CLASSMARK_2, "Mobile Station Classmark 2", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{0,"",0}
	}},
	{ 1, 0x2b, "Handover Command", {
		{CELL_DESCRIPTION, "Cell Description", 0},
		{CHANNEL_DESCRIPTION_2, "Description of First Channel after Time", 0},
		{HANDOVER_REFERENCE, "Handover Reference", 0},
		{POWER_COMMAND_AND_ACCESS_TYPE, "Power Comand and Access Type", 0},
		{SYNCHRONISATION_INDICATION, "Synchronisation Indication", 0xd},
		{FREQUENCY_SHORT_LIST, "Frequency Short List after Time", 0x02},
		{FREQUENCY_LIST, "Frequency List after Time", 0x05},
		{CELL_CHANNEL_DESCRIPTION, "Cell Channel Description", 0x62},
		{MULTISLOT_ALLOCATION, "Description of the Multislot Configuration", 0x10},
		{CHANNEL_MODE, "Mode of Channel Set 1", 0x63},
		{CHANNEL_MODE, "Mode of Channel Set 2", 0x11},
		{CHANNEL_MODE, "Mode of Channel Set 3", 0x13},
		{CHANNEL_MODE, "Mode of Channel Set 4", 0x14},
		{CHANNEL_MODE, "Mode of Channel Set 5", 0x15},
		{CHANNEL_MODE, "Mode of Channel Set 6", 0x16},
		{CHANNEL_MODE, "Mode of Channel Set 7", 0x17},
		{CHANNEL_MODE, "Mode of Channel Set 8", 0x18},
		{CHANNEL_DESCRIPTION, "Description of Second Channel after Time", 0x64},
		{CHANNEL_MODE_2, "Mode of Second Channel", 0x66},
		{FREQUENCY_CHANNEL_SEQUENCE, "Frequency Channel Sequence after Time", 0x69},
		{MOBILE_ALLOCATION, "Mobile Allocation after Time", 0x72},
		{STARTING_TIME, "Starting Time", 0x7c},
		{REAL_TIME_DIFFERENCE, "Real Time Difference", 0x7b},
		{TIMING_ADVANCE, "Timing Advance", 0x7d},
		{FREQUENCY_SHORT_LIST, "Frequency Short List before Time", 0x12},
		{FREQUENCY_LIST, "Frequency List before Time", 0x19},
		{CHANNEL_DESCRIPTION_2, "Description of First Channel before Time", 0x1c},
		{CHANNEL_DESCRIPTION, "Description of Second Channel before Time", 0x1d},
		{FREQUENCY_CHANNEL_SEQUENCE, "Fequency Channel Sequence before Time", 0x1e},
		{MOBILE_ALLOCATION, "Mobile Allocation before Time", 0x21},
		{CIPHER_MODE_SETTING, "Cipher Mode Setting", 0x9},
		{VGCS_TARGET_MODE_INDICATION, "VGCS Target Mode Indication", 0x01},
		{MULTIRATE_CONFIGURATION, "Multi-Rate Configuration", 0x03},
		{0,"",0}
	}},
	{ 1, 0x2c, "Handover Complete", {
		{RR_CAUSE, "RR Cause", 0},
		{MOBILE_TIME_DIFFERENCE, "Mobile Observed Time Difference", 0x77},
		{0,"",0}
	}},
	{ 1, 0x28, "Handover Failure", {
		{RR_CAUSE, "RR Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x2d, "Physical Information", {
		{TIMING_ADVANCE, "Timing Advance", 0},
		{0,"",0}
	}},
	{ 1, 0x2e, "Assignment Command", {
		{CHANNEL_DESCRIPTION_2, "Description of First Channel after Time", 0},
		{POWER_COMMAND, "Power Command", 0},
		{FREQUENCY_LIST, "Frequency List after Time", 0x05},
		{CELL_CHANNEL_DESCRIPTION, "Cell Channel Description", 0x62},
		{MULTISLOT_ALLOCATION, "Description of the Multislot Configuration", 0x10},
		{CHANNEL_MODE, "Mode of Channel Set 1", 0x63},
		{CHANNEL_MODE, "Mode of Channel Set 2", 0x11},
		{CHANNEL_MODE, "Mode of Channel Set 3", 0x13},
		{CHANNEL_MODE, "Mode of Channel Set 4", 0x14},
		{CHANNEL_MODE, "Mode of Channel Set 5", 0x15},
		{CHANNEL_MODE, "Mode of Channel Set 6", 0x16},
		{CHANNEL_MODE, "Mode of Channel Set 7", 0x17},
		{CHANNEL_MODE, "Mode of Channel Set 8", 0x18},
		{CHANNEL_DESCRIPTION, "Description of Second Channel after Time", 0x64},
		{CHANNEL_MODE_2, "Mode of Second Channel", 0x66},
		{MOBILE_ALLOCATION, "Mobile Allocation after Time", 0x72},
		{STARTING_TIME, "Starting Time", 0x7c},
		{FREQUENCY_LIST, "Frequency List before Time", 0x19},
		{CHANNEL_DESCRIPTION_2, "Description of First Channel before Time", 0x1c},
		{CHANNEL_DESCRIPTION, "Description of Second Channel before Time", 0x1d},
		{FREQUENCY_CHANNEL_SEQUENCE, "Fequency Channel Sequence before Time", 0x1e},
		{MOBILE_ALLOCATION, "Mobile Allocation before Time", 0x21},
		{CIPHER_MODE_SETTING, "Cipher Mode Setting", 0x9},
		{VGCS_TARGET_MODE_INDICATION, "VGCS Target Mode Indication", 0x01},
		{MULTIRATE_CONFIGURATION, "Multi-Rate Configuration", 0x03},
		{0,"",0}
	}},
	{ 1, 0x29, "Assignment Complete", {
		{RR_CAUSE, "RR Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x32, "Ciphering Mode Complete", {
		{MOBILE_IDENTITY, "Mobile Equipment Identity", 0x17},
		{0,"",0}
	}},
	{ 1, 0x34, "GPRS Suspension Request", {
		{TLLI, "Temporary Logical Link Identifier", 0},
		{ROUTING_AREA_IDENTIFICATION, "Routing Area Identification", 0},
		{SUSPENSION_CAUSE, "Suspension Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x35, "Ciphering Mode Command", {
		{CIPHER_MODE_SETTING, "Cipher Mode Setting", 0},
		{CIPHER_RESPONSE, "Cipher Response", 0},
		{0,"",0}
	}},
	{ 1, 0x36, "Extended Measuement Report", {
		{EXTENDED_MEASUREMENT_RESULTS, "Extended Measurement Results", 0},
		{0,"",0}
	}},
	{ 1, 0x37, "Extended Measuement Order", {
		{EXTENDED_MEASUREMENT_FREQUENCY_LIST, "Extended Measurement Frequency List", 0},
		{0,"",0}
	}},
	{ 1, 0x3a, "Immediate Assignment Reject", {
		{PAGE_MODE, "Page Mode", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{REQUEST_REFERENCE, "Request Reference 1", 0},
		{WAIT_INDICATION, "Wait Indication 1", 0},
		{REQUEST_REFERENCE, "Request Reference 2", 0},
		{WAIT_INDICATION, "Wait Indication 2", 0},
		{REQUEST_REFERENCE, "Request Reference 3", 0},
		{WAIT_INDICATION, "Wait Indication 3", 0},
		{REQUEST_REFERENCE, "Request Reference 4", 0},
		{WAIT_INDICATION, "Wait Indication 4", 0},
		{0,"",0}
	}},
	{ 1, 0x3f, "Immediate Assignment", {
		{PAGE_MODE, "Page Mode", 0},
		{DEDICATED_MODE_OR_TBF, "Dedicated Mode or TBF", 0},
		{CHANNEL_DESCRIPTION, "Channel Description", 0},
		{REQUEST_REFERENCE, "Request Reference", 0},
		{TIMING_ADVANCE, "Timing Advance", 0},
		{MOBILE_ALLOCATION, "Mobile Allocation", 0},
		{STARTING_TIME, "Starting Time", 0x7c},
		{0,"",0}
	}},
	{ 0, 0x00, NULL, {0,0}}
};

static const msg_t mm_msg_defs[] = {
/* Mobility Management */
	/* Registration messages */
	{ 1, 0x01, "IMSI Detach Indication", {
		{MOBILE_STATION_CLASSMARK_1, "Mobile Station Classmark 1", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{0,"",0}
	}},
	{ 1, 0x02, "Location Updating Accept", {
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0x17},
		{FOLLOW_ON_PROCEED, "Follow On Proceed", 0xa1},
		{CTS_PERMISSION, "CTS Permission", 0xa2},
		{0,"",0}
	}},
	{ 1, 0x04, "Location Updating Reject", {
		{REJECT_CAUSE, "Reject Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x08, "Location Updating Request", {
		{LOCATION_UPDATING_TYPE, "Location Updating Type", 0},
		{CIPHERING_KEY_SEQUENCE_NUMBER, "Ciphering Key Sequence Number", 0},
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{MOBILE_STATION_CLASSMARK_1, "Mobile Station Classmark 1", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{0,"",0}
	}},
	{ 1, 0x11, "Authentication Reject", {
		{0,"",0}
	}},
	{ 1, 0x12, "Authentication Request", {
		{CIPHERING_KEY_SEQUENCE_NUMBER, "Ciphering Key Sequence Number", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{AUTH_PARAMETER_RAND, "Authentication Parameter RAND", 0},
		{0,"",0}
	}},
	{ 1, 0x14, "Authentication Response", {
		{AUTH_PARAMETER_SRES, "Authentication Parameter SRES", 0},
		{0,"",0}
	}},
	{ 1, 0x18, "Identity Request", {
		{IDENTITY_TYPE, "Identity Type", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{0,"",0}
	}},
	{ 1, 0x19, "Identity Response", {
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{0,"",0}
	}},
	{ 1, 0x1a, "TMSI Reallocation Command", {
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{0,"",0}
	}},
	{ 1, 0x1b, "TMSI Reallocation Complete", {
		{0,"",0}
	}},
	{ 1, 0x21, "CM Service Accept", {
		{0,"",0}
	}},
	{ 1, 0x22, "CM Service Reject", {
		{REJECT_CAUSE, "Reject Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x23, "CM Service Abort", {
		{0,"",0}
	}},
	{ 1, 0x24, "CM Service Request", {
		{CM_SERVICE_TYPE, "CM Service Type", 0},
		{CIPHERING_KEY_SEQUENCE_NUMBER, "Ciphering Key Sequence Number", 0},
		{MOBILE_STATION_CLASSMARK_2, "Mobile Station Classmark 2", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{PRIORITY_LEVEL, "Priority Level", 0x8},
		{0,"",0}
	}},
	{ 1, 0x28, "CM Re-establishment Request", {
		{CIPHERING_KEY_SEQUENCE_NUMBER, "Ciphering Key Sequence Number", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{MOBILE_STATION_CLASSMARK_2, "Mobile Station Classmark", 0},
		{MOBILE_IDENTITY, "Mobile Identity", 0},
		{LOCATION_AREA_IDENTIFICATION, "Location Area Identification", 0x13},
		{0,"",0}
	}},
	{ 1, 0x29, "Abort", {
		{REJECT_CAUSE, "Reject Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x30, "MM Null", {
		{0,"",0}
	}},
	{ 1, 0x31, "MM Status", {
		{REJECT_CAUSE, "Reject Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x32, "MM Information", {
		{NETWORK_NAME, "Full Name for Network", 0x43},
		{NETWORK_NAME, "Short Name for Network", 0x45},
		{TIME_ZONE, "Network Time Zone", 0x46},
		{TIME_ZONE_AND_TIME, "Network Time Zone and Time", 0x47},
		{LSA_IDENTIFIER, "LSA Identity", 0x48},
		{0,"",0}
	}},
	{ 0, 0x00, NULL, {0,0}}
};

static const msg_t cc_msg_defs[] = {
/* Call Control Messages */
	{ 1, 0x01, "Alerting", {
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{USER_USER, "User-User", 0x7e},
		{0,"",0}
	}},
	{ 1, 0x02, "Call Proceeding", {
		{REPEAT_INDICATOR, "Repeat Indicator", 0xd},
		{BEARER_CAPABILITY, "Bearer Capability 1", 0x04},
		{BEARER_CAPABILITY, "Bearer Capability 2", 0x04},
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{PRIORITY_LEVEL, "Priority Granted", 0x8},
		{0,"",0}
	}},
	{ 1, 0x03, "Progress", {
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{USER_USER, "User-User", 0x7e},
		{0,"",0}
	}},
	{ 1, 0x04, "CC Establishment", {
		{CONTAINER, "Setup Container", 0},
		{0,"",0}
	}},
	{ 1, 0x05, "Setup", {
		{REPEAT_INDICATOR, "BC Repeat Indicator", 0xd},
		{BEARER_CAPABILITY, "Bearer Capability 1", 0x04},
		{BEARER_CAPABILITY, "Bearer Capability 2", 0x04},
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{SIGNAL, "Signal", 0x34},
		{CALLING_PARTY_BCD_NUM, "Calling Party BCD Num", 0x5c},
		{CALLING_PARTY_SUBADDR, "Calling Party Subaddr",0x5d},
		{CALLED_PARTY_BCD_NUM, "Called Party BCD Num", 0x5e},
		{CALLED_PARTY_SUBADDR, "Called Party Subaddr", 0x6d},
		{REDIRECTING_PARTY_BCD_NUM, "Redirecting Party BCD Num", 0x74},
		{REDIRECTING_PARTY_SUBADDR, "Redirecting Party Subaddr", 0x75},
		{REPEAT_INDICATOR, "LLC Repeat Indicator", 0xd},
		{LOWER_LAYER_COMP, "Lower Layer Comp 1", 0x7c},
		{LOWER_LAYER_COMP, "Lower Layer Comp 2", 0x7c},
		{REPEAT_INDICATOR, "HLC Repeat Indicator", 0xd},
		{HIGHER_LAYER_COMP, "Higher Layer Comp 1", 0x7d},
		{HIGHER_LAYER_COMP, "Higher Layer Comp 2", 0x7d},
		{USER_USER, "User-User", 0x7e},
		{PRIORITY_LEVEL, "Priority", 0x8},
		{ALERTING_PATTERN, "Alert", 0x19},
		{SS_VERSION, "SS Version", 0x7f},
		{CLIR_SUPPRESSION, "CLIR Suppression", 0xa1},
		{CLIR_INVOCATION, "CLIR Invocation", 0xa2},
		{CALL_CONTROL_CAPABILITIES, "CC Capabilities", 0x15},
		{FACILITY, "Facility (advanced recall alignment)", 0x1d},
		{FACILITY, "Facility (recall alignment not essential)", 0x1b},
		{0,"",0}
	}},
	{ 1, 0x06, "CC Establishment Confirmed", {
		{REPEAT_INDICATOR, "Repeat Indicator", 0xd},
		{BEARER_CAPABILITY, "Bearer Capability 1", 0x04},
		{BEARER_CAPABILITY, "Bearer Capability 2", 0x04},
		{CAUSE, "Cause", 0x08},
		{0,"",0}
	}},
	{ 1, 0x07, "Connect", {
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{CONNECTED_NUMBER, "Connected Number", 0x4c},
		{CONNECTED_SUBADDR, "Connected Subaddr", 0x4d},
		{USER_USER, "User-User", 0x7e},
		{0,"",0}
	}},
	{ 1, 0x08, "Call Confirmed", {
		{REPEAT_INDICATOR, "Repeat Indicator", 0xd},
		{BEARER_CAPABILITY, "Bearer Capability 1", 0x04},
		{BEARER_CAPABILITY, "Bearer Capability 2", 0x04},
		{CAUSE, "Cause", 0x08},
		{CALL_CONTROL_CAPABILITIES, "CC Capabilities", 0x15},
		{0,"",0}
	}},
	{ 1, 0x09, "Start CC", {
		{CALL_CONTROL_CAPABILITIES, "CC Capabilities", 0x15},
		{0,"",0}
	}},
	{ 1, 0x0b, "Recall", {
		{RECALL_TYPE, "Recall Type", 0},
		{FACILITY, "Facility", 0},
		{0,"",0}
	}},
	{ 1, 0x0e, "Emergency Setup", {
		{REPEAT_INDICATOR, "BC Repeat Indicator", 0xd},
		{BEARER_CAPABILITY, "Bearer Capability 1", 0x04},
		{BEARER_CAPABILITY, "Bearer Capability 2", 0x04},
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{SIGNAL, "Signal", 0x34},
		{CALLING_PARTY_BCD_NUM, "Calling Party BCD Num", 0x5c},
		{CALLING_PARTY_SUBADDR, "Calling Party Subaddr",0x5d},
		{CALLED_PARTY_BCD_NUM, "Called Party BCD Num", 0x5e},
		{CALLED_PARTY_SUBADDR, "Called Party Subaddr", 0x6d},
		{REDIRECTING_PARTY_BCD_NUM, "Redirecting Party BCD Num", 0x74},
		{REDIRECTING_PARTY_SUBADDR, "Redirecting Party Subaddr", 0x75},
		{REPEAT_INDICATOR, "LLC Repeat Indicator", 0xd},
		{LOWER_LAYER_COMP, "Lower Layer Comp 1", 0x7c},
		{LOWER_LAYER_COMP, "Lower Layer Comp 2", 0x7c},
		{REPEAT_INDICATOR, "HLC Repeat Indicator", 0xd},
		{HIGHER_LAYER_COMP, "Higher Layer Comp 1", 0x7d},
		{HIGHER_LAYER_COMP, "Higher Layer Comp 2", 0x7d},
		{USER_USER, "User-User", 0x7e},
		{PRIORITY_LEVEL, "Priority", 0x8},
		{ALERTING_PATTERN, "Alert", 0x19},
		{0,"",0}
	}},
	{ 1, 0x0f, "Connect Acknowledge", {
		{0,"",0}
	}},
	{ 1, 0x17, "Modify", {
		{BEARER_CAPABILITY, "Bearer Capability", 0},
		{LOWER_LAYER_COMP, "Lower Layer Comp", 0x7c},
		{HIGHER_LAYER_COMP, "Higher Layer Comp", 0x7d},
		{REVERSE_CALL_SETUP_DIRECTION, "Reverse Call Setup Direction", 0xa3},
		{0,"",0}
	}},
	{ 1, 0x1f, "Modify Complete", {
		{BEARER_CAPABILITY, "Bearer Capability", 0},
		{LOWER_LAYER_COMP, "Lower Layer Comp", 0x7c},
		{HIGHER_LAYER_COMP, "Higher Layer Comp", 0x7d},
		{REVERSE_CALL_SETUP_DIRECTION, "Reverse Call Setup Direction", 0xa3},
		{0,"",0}
	}},
	{ 1, 0x13, "Modify Reject", {
		{BEARER_CAPABILITY, "Bearer Capability", 0},
		{CAUSE, "Cause", 0},
		{LOWER_LAYER_COMP, "Lower Layer Comp", 0x7c},
		{HIGHER_LAYER_COMP, "Higher Layer Comp", 0x7d},
		{0,"",0}
	}},
	{ 1, 0x10, "User Information", {
		{USER_USER, "User-User", 0},
		{MORE_DATA, "More Data", 0xa0},
		{0,"",0}
	}},
	{ 1, 0x18, "Hold", {
		{0,"",0}
	}},
	{ 1, 0x19, "Hold Acknowledge", {
		{0,"",0}
	}},
	{ 1, 0x1a, "Hold Reject", {
		{CAUSE, "Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x1c, "Retrieve", {
		{0,"",0}
	}},
	{ 1, 0x1d, "Retrieve Acknowledge", {
		{0,"",0}
	}},
	{ 1, 0x1e, "Retrieve Reject", {
		{CAUSE, "Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x25, "Disconnect", {
		{CAUSE, "Cause", 0},
		{FACILITY, "Facility", 0x1c},
		{PROGRESS_INDICATOR, "Progress Indicator", 0x1e},
		{USER_USER, "User-User", 0x7e},
		{ALLOWED_ACTIONS, "Allowed Actions", 0x7b},
		{0,"",0}
	}},
	{ 1, 0x2a, "Release Complete", {
		{CAUSE, "Cause", 0x08},
		{FACILITY, "Facility", 0x1c},
		{0,"",0}
	}},
	{ 1, 0x2d, "Release", {
		{CAUSE, "Cause", 0x08},
		{CAUSE, "Second Cause", 0x08},
		{FACILITY, "Facility", 0x1c},
		{USER_USER, "User-User", 0x7e},
		{0,"",0}
	}},
	{ 1, 0x39, "Congestion Control", {
		{CONGESTION_LEVEL, "Congestion Level", 0},
		{SPARE_HALF_OCTET, "Spare Half Octet", 0},
		{CAUSE, "Cause", 0x08},
		{0,"",0}
	}},
	{ 1, 0x3e, "Notify", {
		{NOTIFICATION_INDICATOR, "Notification Indicator", 0},
		{0,"",0}
	}},
	{ 1, 0x3d, "Status", {
		{CAUSE, "Cause", 0},
		{CALL_STATE, "Call State", 0},
		{AUXILIARY_STATES, "Auxiliary States", 0x24},
		{0,"",0}
	}},
	{ 1, 0x34, "Status Enquiry", {
		{0,"",0}
	}},
	{ 1, 0x35, "Start DTMF", {
		{KEYPAD_FACILITY, "Keypad Facility", 0x2c},
		{0,"",0}
	}},
	{ 1, 0x31, "Stop DTMF", {
		{0,"",0}
	}},
	{ 1, 0x36, "Start DTMF Acknowledge", {
		{KEYPAD_FACILITY, "Keypad Facility", 0x2c},
		{0,"",0}
	}},
	{ 1, 0x32, "Stop DTMF Acknowledge", {
		{0,"",0}
	}},
	{ 1, 0x37, "Start DTMF Reject", {
		{CAUSE, "Cause", 0},
		{0,"",0}
	}},
	{ 1, 0x3a, "Facility", {
		{FACILITY, "Facility", 0},
		{0,"",0}
	}},
	{ 1, 0x3b, "Register", {
		{FACILITY, "Facility", 0x1c},
		{SS_VERSION, "SS Version", 0x7f},
		{0,"",0}
	}},
	{ 0, 0x00, NULL, {0,0}}
};


static const value_string gsm0408_rr_msg_type_vals[] = {
/* Radio Resource Messages */
	/* Channel Establishment Messages */
	{ 0x3c, "RR Initialisation Request" },
	{ 0x3b, "Additional Assignment" },
	{ 0x3f, "Immediate Assignment" },
	{ 0x39, "Immediate Assignment Extended" },
	{ 0x3a, "Immediate Assignment Reject" },
	/* Cipher Messages */
	{ 0x35, "Cipher Mode Command" },
	{ 0x32, "Cipher Mode Complete" },
	/* Configuration Change Messages */
	{ 0x30, "Configuration Change Command" },
	{ 0x31, "Configuration Change Ack" },
	{ 0x33, "Configuration Change Reject" },
	/* Handover Messages */
	{ 0x2e, "Assignment Command" },
	{ 0x29, "Assignment Complete" },
	{ 0x2f, "Assignment Failure" },
	{ 0x2b, "Handover Command" },
	{ 0x2c, "Handover Complete" },
	{ 0x28, "Handover Failure" },
	{ 0x2d, "Physical Information" },
	{ 0x08, "RR Cell Change Order" },
	{ 0x23, "PDCH Assignment Command" },
	/* Channel Release Messages */
	{ 0x0d, "Channel Release" },
	{ 0x0a, "Partial Release" },
	{ 0x0f, "Partial Release Complete" },
	/* Paging and Notification Messages */
	{ 0x21, "Paging Request Type 1" },
	{ 0x22, "Paging Request Type 2" },
	{ 0x24, "Paging Request Type 3" },
	{ 0x27, "Paging Response" },
	{ 0x20, "Notification/NCH" },
	{ 0x25, "Notification/FACCH" },
	/* System Information Messages */
	{ 0x18, "System Information Type 8" },
	{ 0x19, "System Information Type 1" },
	{ 0x1a, "System Information Type 2" },
	{ 0x1b, "System Information Type 3" },
	{ 0x1c, "System Information Type 4" },
	{ 0x1d, "System Information Type 5" },
	{ 0x1e, "System Information Type 6" },
	{ 0x1f, "System Information Type 7" },
	{ 0x02, "System Information Type 2bis" },
	{ 0x03, "System Information Type 2ter" },
	{ 0x05, "System Information Type 5bis" },
	{ 0x06, "System Information Type 5ter" },
	{ 0x07, "System Information Type 2qtr" },
	{ 0x04, "System Information Type 9" },
	{ 0x00, "System Information Type 13" },
	{ 0x01, "System Information Type 14" },
	{ 0x3d, "System Information Type 16" },
	{ 0x3e, "System Information Type 17" },
	/* Miscellaneous Messages */
	{ 0x10, "Channel Mode Modify" },
	{ 0x12, "RR Status" },
	{ 0x17, "Channel Mode Modify Acknowledge" },
	{ 0x14, "Frequency Redefinition" },
	{ 0x15, "Measurement Report" },
	{ 0x16, "Classmark Change" },
	{ 0x13, "Classmark Enquiry" },
	{ 0x36, "Extended Measurement Report" },
	{ 0x37, "Extended Measurement Order" },
	{ 0x34, "GPRS Suspension Request" },
	/* VGCS Uplink Control Messages */
	{ 0x09, "VGCS Uplink Grant" },
	{ 0x0e, "Uplink Release" },
	{ 0x0c, "Uplink Free" },
	{ 0x2a, "Uplink Busy" },
	{ 0x11, "Talker Indication" },
	{ 0, NULL }
};
static const value_string gsm0408_mm_msg_type_vals[] = {
/* Mobility Management */
	/* Registration messages */
	{ 0x01, "IMSI Detach Indication" },
	{ 0x02, "Location Updating Accept" },
	{ 0x04, "Location Updating Reject" },
	{ 0x08, "Location Updating Request" },
	/* Security Messages */
	{ 0x11, "Authentication Reject" },
	{ 0x12, "Authentication Request" },
	{ 0x14, "Authentication Response" },
	{ 0x18, "Identity Request" },
	{ 0x19, "Identity Response" },
	{ 0x1a, "TMSI Reallocation Command" },
	{ 0x1b, "TMSI Reallocation Complete" },
	/* Connection Management Messages */
	{ 0x21, "CM Service Accept" },
	{ 0x22, "CM Service Reject" },
	{ 0x23, "CM Service Abort" },
	{ 0x24, "CM Service Request" },
	{ 0x25, "CM Service Prompt" },
	{ 0x26, "Notification Response" },
	{ 0x28, "CM Re-establishment Request" },
	{ 0x29, "Abort" },
	/* Miscellaneous Messages */
	{ 0x30, "MM Null" },
	{ 0x31, "MM Status" },
	{ 0x32, "MM Information" },
	{ 0, NULL }
};
static const value_string gsm0408_cc_msg_type_vals[] = {
/* Call Control & SS */
	/* Call Establishment */
	{ 0x01, "Alerting" },
	{ 0x08, "Call Confirmed" },
	{ 0x02, "Call Proceeding" },
	{ 0x07, "Connect" },
	{ 0x0f, "Connect Acknowledge" },
	{ 0x0e, "Emergency Setup" },
	{ 0x03, "Progress" },
	{ 0x04, "CC-Establishment" },
	{ 0x06, "CC-Establishment Confirmed" },
	{ 0x0b, "Recall" },
	{ 0x09, "Start CC" },
	{ 0x05, "Setup" },
	/* Call information phase messages */
	{ 0x17, "Modify" },
	{ 0x1f, "Modify Complete" },
	{ 0x13, "Modify Reject" },
	{ 0x10, "User Information" },
	{ 0x18, "Hold" },
	{ 0x19, "Hold Acknowledge" },
	{ 0x1a, "Hold Reject" },
	{ 0x1c, "Retrieve" },
	{ 0x1d, "Retrieve Acknowledge" },
	{ 0x1e, "Retrieve Reject" },
	/* Call Clearing Messages */
	{ 0x25, "Disconnect" },
	{ 0x2d, "Release" },
	{ 0x2a, "Release Complete" },
	/* Miscellaneous Messages */
	{ 0x39, "Congestion Control" },
	{ 0x3e, "Notify" },
	{ 0x3d, "Status" },
	{ 0x34, "Status Enquiry" },
	{ 0x35, "Start DTMF" },
	{ 0x31, "Stop DTMF" },
	{ 0x32, "Stop DTMF Acknowledge" },
	{ 0x36, "Start DTMF Acknowledge" },
	{ 0x37, "Start DTMF Reject" },
	{ 0x3a, "Facility" },
	/* Supplementary Services */
	{ 0x3b, "Register" },
	{ 0, NULL }
};

static const value_string gsm0408_gmm_msg_type_vals[] = {
	{ 0, NULL }
};

static const value_string gsm0408_sm_msg_type_vals[] = {
	{ 0, NULL }
};


static const value_string gsm0408_sms_msg_type_vals[] = {
	{ 0x01, "CP-DATA" },
	{ 0x04, "CP-ACK" },
	{ 0x10, "CP-ERROR" },
	{ 0, NULL }
};

typedef enum {
    GSM0408_FORMAT_L2,        /* including L2 Pseudo Length and following Rest Octets */
    GSM0408_FORMAT_L3,            /* normal L3 format */
    GSM0408_FORMAT_L3_CONTENTS,   /* L3 format, but omitting the message type IE (see 08.08 L3 Message Contents IE) */
} gsm0408_format_t;


/* Code to dissect the 04.08 message */
static void
dissect_gsm0408_format(gsm0408_format_t format, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len, l2_pseudo_len;
	guint8 msg_type, pd;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL, *next_tree=NULL;
	const msg_t *p_msg_defs=0;

	/* first take a sneak preview of the PD and if it isn't RR, maybe handoff to the GSM DTAP dissector */
	if (format==GSM0408_FORMAT_L3) {
		pd = (tvb_get_guint8(tvb, offset) & 0x0f);

		/* The GSM A DTAP dissector is generally better for all non-RR messages */
		if ((pd != GSM0408_PD_RR) && (!global_gsm0408_use_ipa)) {
			call_dissector(gsm_a_dtap_handle, tvb, pinfo, main_tree);
			return;
		}
	}

	/* load the display labels 	*/
	pinfo->current_proto = "gsm0408";
	gsmstat_set_proto(GSMSTAT_PROTO_GSM0408);

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM");

	msg_len = tvb_length_remaining(tvb,offset);

	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm0408, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_gsm0408);
	}

	if (format==GSM0408_FORMAT_L2) {
		l2_pseudo_len = (tvb_get_guint8(tvb, offset)-1) >> 2;
		offset+=1;
	}

	if ((msg_len>=GSM0408_HEADER_LEN) || (format==GSM0408_FORMAT_L3_CONTENTS)) {
		if (format==GSM0408_FORMAT_L3_CONTENTS) {	/* this is only used by 08.08 to embed RR contents */
			p_msg_defs = rr_msg_defs;
			msg_type = 0x32;						/* in fact, only ever used to embed Cipher Mode Complete contents !!! */
			if (tree) {
				proto_item_set_text(ti, "   04.08:RR:Message Contents");
			}
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:RR:Message Contents");
		}
		else {
			pd = (tvb_get_guint8(tvb, offset) & 0x0f);
			if (tree)
				proto_tree_add_uint(tree, hf_gsm0408_pd, tvb, offset, 1, pd);
			offset+=1;
			gsmstat_set_pd(pd);

			msg_type = (tvb_get_guint8(tvb, offset) & 0x3f);
			gsmstat_set_msg_type(msg_type);

			if (pd == GSM0408_PD_RR) {
				p_msg_defs = rr_msg_defs;
				if (tree) {
					proto_item_set_text(ti, "   04.08:RR:%s",rd_match_strval(msg_type&0x3f,gsm0408_rr_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_rr_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:RR:%s",rd_match_strval(msg_type&0x3f,gsm0408_rr_msg_type_vals));
			}
			else if (pd == GSM0408_PD_MM) {
				p_msg_defs = mm_msg_defs;
				if (tree) {
					proto_item_set_text(ti, "   04.08:MM:%s",rd_match_strval(msg_type&0x3f,gsm0408_mm_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_mm_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:MM:%s",rd_match_strval(msg_type&0x3f,gsm0408_mm_msg_type_vals));
			}
			else if (pd == GSM0408_PD_CC) {
				p_msg_defs = cc_msg_defs;
				if (tree) {
					proto_item_set_text(ti, "   04.08:CC:%s",rd_match_strval(msg_type&0x3f,gsm0408_cc_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_cc_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:CC:%s",rd_match_strval(msg_type&0x3f,gsm0408_cc_msg_type_vals));
			}
			else if (pd == GSM0408_PD_SS) {
				p_msg_defs = cc_msg_defs;
				if (tree) {
					proto_item_set_text(ti, "   04.08:SS:%s",rd_match_strval(msg_type&0x3f,gsm0408_cc_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_ss_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:SS:%s",rd_match_strval(msg_type&0x3f,gsm0408_cc_msg_type_vals));
			}
			else if (pd == GSM0408_PD_GMM) {
				if (tree) {
					proto_item_set_text(ti, "   04.08:GMM:%s",rd_match_strval(msg_type&0x3f,gsm0408_gmm_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_gmm_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:GMM:%s",rd_match_strval(msg_type&0x3f,gsm0408_gmm_msg_type_vals));
			}
			else if (pd == GSM0408_PD_SM) {
				if (tree) {
					proto_item_set_text(ti, "   04.08:SM:%s",rd_match_strval(msg_type&0x3f,gsm0408_sm_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_sm_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:SM:%s",rd_match_strval(msg_type&0x3f,gsm0408_sm_msg_type_vals));
			}
			else if (pd == GSM0408_PD_SMS) {
				p_msg_defs = sms_msg_defs;
				if (tree) {
					proto_item_set_text(ti, "   04.08:SMS:%s",rd_match_strval(msg_type,gsm0408_sms_msg_type_vals));
					proto_tree_add_uint(tree, hf_gsm0408_sms_msg_type, tvb, offset, 1, msg_type);
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:SMS:%s",rd_match_strval(msg_type,gsm0408_sms_msg_type_vals));
			}
			else {
				if (tree) {
					proto_item_set_text(ti, "   04.08:Unknown PD=%d, Msg Type %d", tvb_get_guint8(tvb, 0), tvb_get_guint8(tvb, 1));
				}
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_fstr(pinfo->cinfo, COL_INFO, "04.08:Unknown PD=%d, Msg Type %d", tvb_get_guint8(tvb, 0), tvb_get_guint8(tvb, 1));
			}
			offset += 1;
		}

		if (p_msg_defs) {
			/* create a new subset TVB for disecting the payload, and call the correct payload disector */
			if (format==GSM0408_FORMAT_L2) {
				if (l2_pseudo_len == 0)
					next_tvb = 0;
				else
					next_tvb = tvb_new_subset(tvb, offset, l2_pseudo_len-offset+1, -1);
			}
			else
				next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);

			for ( ;p_msg_defs->format>0; p_msg_defs++) {	/* search for message def to match this message type */
				if (p_msg_defs->type == msg_type)
					break;
			}

			if (p_msg_defs->type == msg_type)	/* if msg has a definition, decode using it */
			{
				if (next_tvb)
					dissect_gsm0408_ie( p_msg_defs, msg_type, next_tvb, pinfo, tree);	/* dissect main formatted IEs */

				if ((format==GSM0408_FORMAT_L2) && (msg_len-l2_pseudo_len-1>0)) {	/* dissect Rest Octets */
					if (l2_pseudo_len) {
						next_tvb = tvb_new_subset(tvb, l2_pseudo_len+1, -1, -1);
						ti = proto_tree_add_text(tree, next_tvb, 0, msg_len-l2_pseudo_len-1, "Rest Octets");
					}
					else {														/* cope with new mesages which have L2 Pseudo length zero - but have 3 octets before rest octets! */
						next_tvb = tvb_new_subset(tvb, 3, -1, -1);
						ti = proto_tree_add_text(tree, next_tvb, 0, msg_len-3, "Rest Octets");
					}
					next_tree = proto_item_add_subtree(ti, ett_gsm0408_ie);
					dissect_rest_octets( next_tvb, pinfo, next_tree);
				}
			}
			else								/* default to decoding as raw data if no matching message def found */
				call_dissector(data_handle, tvb, pinfo, tree);
		}
		else								/* default to decoding as raw data if no matching message defs for this PD */
			call_dissector(data_handle, tvb, pinfo, tree);
	}
	else { /* Duff 04.08 or ... FIXME - partial message !!! */
		if (tree)
			proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"Short 04.08 Message!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Short 04.08 Message!!!");
	}

	return;
}

/* entry points for the variations on the theme */
static void
dissect_gsm0408(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	dissect_gsm0408_format(GSM0408_FORMAT_L3, tvb, pinfo, main_tree);
}

static void
dissect_gsm0408_l2(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	dissect_gsm0408_format(GSM0408_FORMAT_L2, tvb, pinfo, main_tree);
}

static void
dissect_gsm0408_l3_contents(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	dissect_gsm0408_format(GSM0408_FORMAT_L3_CONTENTS, tvb, pinfo, main_tree);
}

/* Code to dissect the 04.08 IEs */
static void
dissect_gsm0408_ie( const msg_t *p_msg_defs, guint8 msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len, iei_offset, val_len, tot_len, i;
	guint8	octet, iei;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];
	const ie_id_t *p_ie_id;
	const ie_t	*p_ie;
	int		half_done=0;

	msg_len = tvb_length_remaining(tvb,offset);

	for (p_ie_id = p_msg_defs->ie_id; ((msg_len>offset) && (p_ie_id->tag)); p_ie_id++) {	/* decode each IE at a time */

		for (p_ie=ie_defs; p_ie->tag; p_ie++) {	/* find the IE def for the next IE listed */
			if (p_ie_id->tag == p_ie->tag)
				break;
		}
		if (p_ie->tag == 0) {
			/* unknown IE - help! */
			new_tvb = tvb_new_subset(tvb, offset, -1, -1);
			ti = proto_tree_add_text(main_tree, new_tvb, 0, msg_len-offset, "Unrecognised Data");
			tree = proto_item_add_subtree(ti, ett_gsm0408_ie);
			call_dissector(data_handle, new_tvb, pinfo, tree);
		}

		iei_offset = offset;		/* remember for later */
		val_len = 0;				/* default to 0 length for the value part */
		tot_len = 0;				/* default to 0 length for the total */
		octet = 0;					/* initialise for safety for non-type 1 IEs */

		if (p_ie_id->iei != 0) {	/* optional or conditional IE */
			tot_len = 1;			/* default to 1 length for the total */
			octet = (tvb_get_guint8(tvb, offset));
			if (p_ie->type==1) {	/* treat half octet IEIs specially */
				iei = (octet >> 4) & 0x0f;
				octet = octet & 0x0f;
			}
			else {
				iei = octet;		/* all other IEs have whole octet IEIs */
			}
			if (p_ie_id->iei != iei)	/* if this optional IE not included, skip to next IE */
				continue;

			offset++;				/* IEI octet finished */
		}

		if ((p_ie->type == 1)&&(p_ie_id->iei == 0)) {	/* mandatory half octet - which half ? */
			if (!half_done) {	/* first half octet */
				octet = tvb_get_guint8(tvb, offset) & 0x0f;
			}
			else {				/* second half octet */
				octet = (tvb_get_guint8(tvb, offset) >> 4) & 0x0f;
				offset++;		/* finished with this octet now */
			}
			tot_len = 1;			/* 1 length for the total */
		}
		else if (p_ie->type == 3) {
			val_len = p_ie->len-1;
			tot_len += val_len;
		}
		else if (p_ie->type == 4) {
			val_len = tvb_get_guint8(tvb, offset);
			tot_len += (val_len + 1);
			offset++;		/* finished with length octet too */
		}
		else {
				/* unknown IE type - help! */
		}

		if (p_ie->type == 1) {					/* single octet IE */
			if (!half_done) {					/* first half octet */
				if (p_ie_id->iei == 0) {		/* Only optional type 1 IEs have a second to follow */
					sprintf(str, "%x.", octet);	/* mandatory type 1 IE - value in first half */
					half_done=TRUE;
				}
				else {
					sprintf(str, ".%x", octet);	/* optional type 1 IE has value in second half */
				}
			}
			else {								/* second half octet - must be optional to have got to here! */
				sprintf(str, ".%x", octet);
				half_done=FALSE;
			}
		}
		else
			data_str( tvb, offset, val_len, str, sizeof(str) );

		if (main_tree) {
			if (p_ie->dissector == NULL && p_ie->info && (p_ie->type == 1 || val_len == 1)) {
				if (p_ie->type == 1)
					ti = proto_tree_add_text(main_tree, tvb, iei_offset, tot_len, "%s: %s", p_ie_id->name, rd_match_strval(octet,(value_string *)p_ie->info));
				else if (val_len == 1)
					ti = proto_tree_add_text(main_tree, tvb, iei_offset, tot_len, "%s: %s", p_ie_id->name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)p_ie->info));
				else {
					ti = proto_tree_add_text(main_tree, tvb, iei_offset, tot_len, "%s: %s", p_ie_id->name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)(p_ie->info)));
					for (i=offset+1; i<tot_len; i++)
						proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie->info)));
				}
			}
			else {
				if (p_ie->type == 2)
					ti = proto_tree_add_text(main_tree, tvb, iei_offset, tot_len, "%s", p_ie_id->name);
				else
					ti = proto_tree_add_text(main_tree, tvb, iei_offset, tot_len, "%s: %s", p_ie_id->name, str);
				tree = proto_item_add_subtree(ti, ett_gsm0408_ie);
			}
		}

		if ((p_ie->type == 1)&&(p_ie_id->iei != 0))	/* optional type 1 IE - IEI & value in same octet */
			new_tvb = tvb_new_subset(tvb, iei_offset, val_len, -1);
		else
			new_tvb = tvb_new_subset(tvb, offset, val_len, -1);
		if (p_ie->dissector)
			(p_ie->dissector)(msg_type, new_tvb, pinfo, tree, p_ie->info, octet);
		else if (val_len > MAX_DATA_STR)
			dissect_default_gsm0408_ie(p_ie->type, octet, new_tvb, 0, pinfo, tree);
		offset += val_len;
	}

	if (msg_len > offset) {
		new_tvb = tvb_new_subset(tvb, offset, -1, -1);
		data_str( new_tvb, 0, msg_len-offset, str, sizeof(str) );
		ti = proto_tree_add_text(main_tree, new_tvb, 0, msg_len-offset, "Undefined IE data: %s", str);
		tree = proto_item_add_subtree(ti, ett_gsm0408_ie);
		call_dissector(data_handle, new_tvb, pinfo, tree);
	}
	return;
}

static void dissect_default_gsm0408_ie(guint8 ie_type, guint8 value, tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree)
{
	tvbuff_t *new_tvb;

	if (ie_type == 1) {
		if (tree)
			proto_tree_add_text(tree, tvb, 0, 1, "Data: 0x%x", value);
	}
	else {
		new_tvb = tvb_new_subset(tvb, offset, -1, -1);
		call_dissector(data_handle, new_tvb, pinfo, tree);
	}

	return;
}

#define REST_OCTETS_MAX	23

static void dissect_rest_octets(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset=0, msg_len, i, j, last;
	guint8 octet[REST_OCTETS_MAX], *c, charbuf[9*REST_OCTETS_MAX+1];

	msg_len = tvb_length_remaining(tvb,offset);

	last = -1;
	for (i=0; i<msg_len; i++) {
		octet[i] = tvb_get_guint8(tvb, i);
		if (octet[i] ^ 0x2b)	/* look for the last octet with "useful" information */
			last=i;
	}
	last++;

	charbuf[0]='\0';
	for (i=0; i<last; i++) {
		sprintf(&charbuf[3*i], "%2.2x ", octet[i]);
	}
	if (tree)
		proto_tree_add_text(tree, tvb, 0, msg_len, "Octets: %s", charbuf);

	c = charbuf;
	for (i=0; i<last; i++) {
		for (j=0; j<8; j++) {
			*c++=((octet[i]&(0x80>>j))?'1':'0');
		}
		*c++=' ';
	}
	*c++='\0';
	if (tree)
		proto_tree_add_text(tree, tvb, 0, msg_len, "Binary: %s", charbuf);

	return;
}

static void dissect_gsm0408_cause_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0;
	char octet2, octet3;

	do {		/* skip over octet 2 - including extensions 2a etc */
		octet2 = tvb_get_guint8(tvb, offset++);
	} while ((octet2 & 0x80)==0);

	octet3 = tvb_get_guint8(tvb, offset) & 0x7f;	/* mask off the ext bit */

	if (tree)
		proto_tree_add_uint(tree, hf_gsm0408_cause, tvb, offset, 1, octet3);
	offset++;

	gsmstat_set_cause(octet3);

	return;
}

static void dissect_gsm0408_channel_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0;
	guint16 octets3and4;
	guint8	chan_no, chan, subchan_no=8; /* sub-channel 8 is too big to be a real sub-channel: flag meaning no sub-channel */

	chan_no = tvb_get_guint8(tvb, offset);
	chan = chan_no>>3;

	if (chan>>1== 0x01) /* TCH/H */
		subchan_no = chan & 0x01;
	else if (chan>>2 == 0x01) /* SDCCH/4 */
		subchan_no = chan & 0x03;
	else if (chan>>3 == 0x01) /* SDCCH/8 */
		subchan_no = chan & 0x07;

	if (tree) {
		proto_tree_add_uint(tree, hf_gsm0408_ts, tvb, offset, 1, chan_no);
		if (subchan_no<8)
			proto_tree_add_uint_hidden(tree, hf_gsm0408_subchan, tvb, offset, 1, chan_no);
		proto_tree_add_uint(tree, hf_gsm0408_chan, tvb, offset, 1, chan_no);
	}
	offset++;

	octets3and4 = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_uint(tree, hf_gsm0408_tsc, tvb, offset, 2, octets3and4);
		proto_tree_add_uint(tree, hf_gsm0408_arfcn, tvb, offset, 2, octets3and4);
	}
	offset += 2;

	return;
}

static void dissect_gsm0408_rr_cause_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0;
	char octet2;

	octet2 = tvb_get_guint8(tvb, offset);

	if (tree)
		proto_tree_add_uint(tree, hf_gsm0408_rr_cause, tvb, offset, 1, octet2);
	offset++;

	gsmstat_set_cause(octet2);

	return;
}

static void dissect_gsm0408_request_reference_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0, msg_len;
	guint8	ra;
	guint16	fn;
	guint32 delay;

	msg_len = tvb_length_remaining(tvb,offset);

	ra = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0408_ra, tvb, offset, 1, ra);
	offset++;

	fn = reduced_frame_number(tvb_get_ntohs(tvb, offset));
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0408_fn, tvb, offset, 2, fn);
	offset+=2;

	delay = analyze_immediate_assignment(ra, fn, pinfo);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0408_assignment_delay, tvb, offset, 0, delay);

	return;
}

static void display_channel_list(guint8 *list, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int arfcn;
	proto_item *ti=NULL;

	ti = proto_tree_add_text(tree, tvb, 0, 16, "Channel List: ");
	for (arfcn=0; arfcn<=ARFCN_MAX; arfcn++) {
		if (list[arfcn])
			proto_item_append_text(ti, "%d ", arfcn);
	}

	return;
}

static void dissect_gsm0408_channel_list_bit_map_0(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 octet2)
{
	int offset, b;
	guint8 octet, list[1024];

	memset((void*)list,0,sizeof(list));

	for (offset=0; offset<16; offset++) {
		octet = tvb_get_guint8(tvb, offset);
		if (offset==0)	/* first octet has format id as first 4 bits */
			octet &= 0x0f;

		for (b=0; b<8; b++) {
			if (octet&1)
				list[8*(15-offset)+b+1]=1;
			octet >>= 1;
		}
	}

	display_channel_list(list, tvb, pinfo, tree);

	return;
}

static void dissect_gsm0408_channel_list_variable_bit_map(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 octet2)
{
	int i, offset=0, b, orig_arfcn, arfcn;
	guint8 octet, list[1024];

	memset((void*)list,0,sizeof(list));

	orig_arfcn = (tvb_get_guint8(tvb, offset++) & 0x01); /* MSB is LSB of first octet */
	orig_arfcn = (orig_arfcn<<8) + tvb_get_guint8(tvb, offset++); /* then whole of next octet, then LSB is first bit of next octet */
	orig_arfcn = (orig_arfcn<<1)  + (tvb_get_guint8(tvb, offset)>>7); /* then LSB is MSB of next octet */

	list[orig_arfcn] = 1;	/* origin ARFCN is always included */

	for (i=0; i<14; i++,offset++) {
		octet = tvb_get_guint8(tvb, offset);
		if (i==0)	/* first octet has LSB of orig_arfcn as MSB */
			octet &= 0x7f;

		for (b=0; b<8; b++) {
			if (octet&1) {
				arfcn = (orig_arfcn + 8*i + 7-b) % 1024;	/* ARFCN bit map is modulus 1024 */
				list[arfcn]=1;
			}
			octet >>= 1;
		}
	}

	display_channel_list(list, tvb, pinfo, tree);

	return;
}

static int f_k(int k, int *w, int range)
{
	int index=k, j=1, n;

	/* J := GREATEST_POWER_OF_2_LESSER_OR_EQUAL_TO(INDEX); */
	if (index>1) {
		do {
			j<<=1;
		} while (j<=index);
		j >>= 1;
	}

   n = w[index];

   while (index>1) {
	   if (2*index < 3*j) {				/* left child */
			index -= j>>1;
			n = (n + w[index] - range/j - 1)%((2*range/j) - 1) + 1;
	   }
	   else	{							/* right child */
			index -= j;
			n = (n + w[index] - 1)%((2*range)/j - 1) + 1;
		}
		j >>= 1;
	}

	return n%1024;
}

static void dissect_gsm0408_channel_list_n_range(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int range, guint8 octet2)
{
	int offset=1, f0, arfcn_orig, bits, w[64], wsize, i, wi;
	int octet, nwi=1, jwi=0, wbits, imax, iused, arfcn;
	guint8 list[1024];

	memset((void*)list,0,sizeof(list));

	octet = octet2;
	if (range==1024) {
		f0 = (octet>>2)&1;
		if (f0)
			list[0] = 1;
		bits = 2;
		arfcn_orig = 0;
		wsize = 10;
		imax = 16;
	}
	else {
		arfcn_orig = (octet&1);
		arfcn_orig = (arfcn_orig << 8) + tvb_get_guint8(tvb, offset++);
		octet = tvb_get_guint8(tvb, offset++);
		arfcn_orig = (arfcn_orig << 1) + (octet>>7);
		list[arfcn_orig] = 1;
		bits = 7;
		switch (range) {
		case 512:
			wsize=9;
			imax = 17;
			break;
		case 256:
			wsize=8;
			imax = 21;
			break;
		case 128:
			wsize=7;
			imax = 29;
			break;
		}
	}
	iused = imax;	/* in case the list is actually full */

	/* extract the variable size w[] elements */
	for (i=1; i<=imax; i++) {
		wi = octet & ~(0xff<<bits);		/* mask "bits" low bits to start wi from existing octet */
		wbits = bits;
		if (wsize>wbits) {				/* need to extract more bits from the next octet */
			octet = tvb_get_guint8(tvb, offset++);
			wi = (wi << 8) + octet;
			bits = 8;
			wbits += 8;
		}

		if (wbits>wsize)	{			/* now we have too many bits - save some */
			bits = wbits - wsize;
			wi >>= bits;
		}
		else							/* just right number of bits */
			bits = 0;

		w[i] = wi;
		if (w[i]==0) {
			iused = i - 1;
			break;		/* all remaining elements must also be zero */
		}

		if (++jwi==nwi)	{		/* check if the number of wi at this wsize has been extracted */
			jwi = 0;			/* reset the count of wi at this size */
			nwi <<= 1;			/* get twice as many of the next size */
			wsize--;			/* make the next size 1 bit smaller */
		}
	}

	for (i=1; i<=iused; i++) {
		arfcn = (f_k(i, w, range) + arfcn_orig)%1024;
		list[arfcn] = 1;
	}

	display_channel_list(list, tvb, pinfo, tree);

	return;
}

static void dissect_gsm0408_channel_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 octet2)
{
	int offset=1, format;

	if ((octet2>>6)==0)
		format=CHANNEL_LIST_BIT_MAP_0;
	else if ((octet2>>6)==2) {
		if (((octet2>>3)&1)==0)
			format=CHANNEL_LIST_1024_RANGE;
		else
			switch ((octet2>>1)&0x03) {
			case 0:
				format=CHANNEL_LIST_512_RANGE;
				break;
			case 1:
				format=CHANNEL_LIST_256_RANGE;
				break;
			case 2:
				format=CHANNEL_LIST_128_RANGE;
				break;
			case 3:
				format=CHANNEL_LIST_VARIABLE_BIT_MAP;
				break;
			default:
				break;
			}
	}
	else
		format=CHANNEL_LIST_UNKNOWN;

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, 1, "Format=%s", rd_match_strval(format,channel_list_format_vals));
	}

	switch (format) {
	case CHANNEL_LIST_BIT_MAP_0:
		dissect_gsm0408_channel_list_bit_map_0(msg_type, tvb, pinfo, tree, extra, octet2);
		break;
	case CHANNEL_LIST_1024_RANGE:
		dissect_gsm0408_channel_list_n_range(msg_type, tvb, pinfo, tree, 1024, octet2);
		break;
	case CHANNEL_LIST_512_RANGE:
		dissect_gsm0408_channel_list_n_range(msg_type, tvb, pinfo, tree, 512, octet2);
		break;
	case CHANNEL_LIST_256_RANGE:
		dissect_gsm0408_channel_list_n_range(msg_type, tvb, pinfo, tree, 256, octet2);
		break;
	case CHANNEL_LIST_128_RANGE:
		dissect_gsm0408_channel_list_n_range(msg_type, tvb, pinfo, tree, 128, octet2);
		break;
	case CHANNEL_LIST_VARIABLE_BIT_MAP:
		dissect_gsm0408_channel_list_variable_bit_map(msg_type, tvb, pinfo, tree, extra, octet2);
		break;
	default:
		break;
	}


	return;
}

static void dissect_gsm0408_neighbour_cell_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0;
	guint8 octet2;

	octet2 = tvb_get_guint8(tvb, offset++);

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, 1, "EXT-IND=%d;  BA-IND=%d", (octet2>>5)&1, (octet2>>4)&1);
	}

	dissect_gsm0408_channel_list(msg_type, tvb, pinfo, tree, extra, octet2);

	return;
}

static void dissect_gsm0408_cell_channel_description_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0;
	guint8 octet2;

	octet2 = tvb_get_guint8(tvb, offset++);

	dissect_gsm0408_channel_list(msg_type, tvb, pinfo, tree, extra, octet2);

	return;
}

static void dissect_gsm0408_cp_user_data_ie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	call_dissector(rp_handle, tvb, pinfo, tree);

	return;
}

static void dissect_gsm0408_ie_mobile_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0, msg_len;
	guint8 type, odd;
	char *id, id_string[17];	/* 16 digits is max size + NULL*/

	id = id_string;

	msg_len = tvb_length_remaining(tvb,offset);

	type = tvb_get_guint8(tvb, offset) & 0x07;
	odd = (tvb_get_guint8(tvb, offset) >> 3) & 0x01;
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Identity Type: %s", rd_match_strval(type,identity_type_vals));
	}

	if (type == MOBILE_IDENTITY_TYPE_TMSI) {
		offset++;	/* TMSIs have pad in rest of octet 3, followed by 32 bit value */
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 4, "Identity: 0x%8.8x", tvb_get_ntohl(tvb,offset));
		}
	}
	else {
		for (; offset<msg_len; offset++) {
			if (offset>0) {	/* first octet only has upper nibble used for digits */
				*id++ = hex_digit[tvb_get_guint8(tvb, offset) & 0x0f];
			}
			if (odd || (offset<msg_len-1)) { /* last octet has last nibble spare if even number of digits */
				*id++ = hex_digit[(tvb_get_guint8(tvb, offset) >> 4) & 0x0f];
			}
		}
		*id = 0;	/* NULL terminator */

		if (tree) {
			proto_tree_add_text(tree, tvb, 0, msg_len, "Identity: %s", id_string);
		}
	}

	return;
}

/* Code to dissect enum IEs with val_strings */
static void dissect_ie_unt8_vals_type_1(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0, msg_len;
    const value_string *vals = (const value_string *)extra;

	msg_len = 1;

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, msg_len, "%s", rd_match_strval(value,vals));
	}
	offset += msg_len;

	return;
}

/* Code to dissect enum IEs with val_strings */
static void dissect_ie_unt8_vals(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *extra, guint8 value)
{
	int offset=0, msg_len, val;
    const value_string *vals = (const value_string *)extra;

	msg_len = tvb_length_remaining(tvb,offset);
	val = tvb_get_guint8(tvb, offset);

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, msg_len, "%s", rd_match_strval(val,vals));
	}
	offset += msg_len;

	return;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0408(void)
{
    module_t *gsm0408_module;

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_gsm0408_pd,
			{ "PD", "gsm0408.pd",
			FT_UINT8, BASE_HEX, VALS(gsm0408_pd_vals), 0x0f,
			"04.08 Protocol Descriminator", HFILL }
		},
		{ &hf_gsm0408_rr_msg_type,
			{ "RR Msg Type", "gsm0408.rr.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_rr_msg_type_vals), 0x3f,
			"04.08 RR Message Type" , HFILL}
		},
		{ &hf_gsm0408_mm_msg_type,
			{ "MM Msg Type", "gsm0408.mm.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_mm_msg_type_vals), 0x3f,
			"04.08 MM Message Type", HFILL }
		},
		{ &hf_gsm0408_cc_msg_type,
			{ "CC Msg Type", "gsm0408.cc.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_cc_msg_type_vals), 0x3f,
			"04.08 CC Message Type" , HFILL}
		},
		{ &hf_gsm0408_gmm_msg_type,
			{ "GMM Msg Type", "gsm0408.gmm.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_gmm_msg_type_vals), 0x3f,
			"04.08 GMM Message Type", HFILL }
		},
		{ &hf_gsm0408_sm_msg_type,
			{ "SM Msg Type", "gsm0408.sm.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_sm_msg_type_vals), 0x3f,
			"04.08 SM Message Type", HFILL }
		},
		{ &hf_gsm0408_sms_msg_type,
			{ "SMS Msg Type", "gsm0408.sms.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_sms_msg_type_vals), 0xff,
			"04.08 SMS Message Type", HFILL }
		},
		{ &hf_gsm0408_ss_msg_type,
			{ "SS Msg Type", "gsm0408.ss.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0408_cc_msg_type_vals), 0x3f,
			"04.08 SS Message Type", HFILL }
		},
		{ &hf_gsm0408_cause,
			{ "Cause", "gsm0408.cause",
			FT_UINT8, BASE_HEX, VALS(cause_vals), 0x7f,
			"04.08 Cause", HFILL }
		},
		{ &hf_gsm0408_rr_cause,
			{ "RR Cause", "gsm0408.rr.cause",
			FT_UINT8, BASE_HEX, VALS(rr_cause_vals), 0x0,
			"04.08 RR Cause", HFILL }
		},
		{ &hf_gsm0408_fn,
			{ "FN", "gsm0408.fn",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Reduced Frame Number", HFILL }
		},
		{ &hf_gsm0408_ra,
			{ "RA", "gsm0408.ra",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"Random Access Information", HFILL }
		},
		{ &hf_gsm0408_assignment_delay,
			{ "[Assignment Delay (ms)]", "gsm0408.assign.delay",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Assignment Delay (ms)", HFILL }
		},
		{ &hf_gsm0408_ts,
			{ "TS", "gsm0408.ts",
			FT_UINT8, BASE_DEC, NULL, 0x07,
			"Timeslot", HFILL }
		},
		{ &hf_gsm0408_tsc,
			{ "TSC", "gsm0408.tsc",
			FT_UINT16, BASE_DEC, NULL, 0xE000,
			"Timeslot Code", HFILL }
		},
		{ &hf_gsm0408_arfcn,
			{ "ARFCN", "gsm0408.arfcn",
			FT_UINT16, BASE_DEC, NULL, 0x03FF,
			"ARFCN", HFILL }
		},
		{ &hf_gsm0408_chan,
			{ "Chan", "gsm0408.chan",
			FT_UINT8, BASE_DEC, VALS(chan_vals), 0xF8,
			"Channel", HFILL }
		},
		{ &hf_gsm0408_subchan,
			{ "Sub-Channel", "gsm0408.subchan",
			FT_UINT8, BASE_DEC, NULL, 0x38,
			"Sub-Channel", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0408,
		&ett_gsm0408_ie,
	};

	proto_gsm0408 = proto_register_protocol("GSM 04.08 Air Interface Layer 3", "GSM 04.08", "gsm0408");

	register_dissector("gsm0408", dissect_gsm0408, proto_gsm0408);
	register_dissector("gsm0408_l2", dissect_gsm0408_l2, proto_gsm0408);
	register_dissector("gsm0408_l3_contents", dissect_gsm0408_l3_contents, proto_gsm0408);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0408, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

    /* Register preferences */
    gsm0408_module = prefs_register_protocol(proto_gsm0408, proto_reg_handoff_gsm0408);

    /* Register preferences */
    prefs_register_bool_preference(gsm0408_module, "use_ipa",
         "Use ip.access GSM 04.08 dissector for all",
         "Allows the use of the ip.access or the standard GSM 04.08 dissector for the non-RR messages.",
         &global_gsm0408_use_ipa);

}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0408(void)
{
    static int initialized = FALSE;

    if (!initialized) {
        data_handle = find_dissector("data");
        rp_handle = find_dissector("gsm_a_rp");
        gsm_a_dtap_handle = find_dissector("gsm_a_dtap");
        gsm0408_handle = find_dissector("gsm0408");
        gsm0408_l2_handle = find_dissector("gsm0408_l2");
        gsm0408_l3_contents_handle = find_dissector("gsm0408_l3_contents");

        initialized = TRUE;
	}
}

gchar *data_str( tvbuff_t *tvb, int offset, int data_len, gchar *buf, int buf_len )
{
	int i;
	int len=(buf_len/3);
	gchar *s=buf;

	*s = '\0';
	if (data_len > len) {
		for (i=0; i<len-1; i++) {
			sprintf(s, "%2.2x ", tvb_get_guint8(tvb, offset+i));
			s+=3;
		}
		sprintf(s,"...");
		s+=3;
	}
	else {
		for (i=0; i<data_len; i++) {
			sprintf(s, "%2.2x ", tvb_get_guint8(tvb, offset+i));
			s+=3;
		}
	}

	return buf;
}
