/* packet-gsm0464.c
 * Routines for GSM 04.64
 * Copyright 2001, ip.access Ltd <Calvin.Monk@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0464.c 1.2 2006/08/09 18:41:52BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"

/* dissectors for the data portion of this protocol
 */
static dissector_handle_t gsm24008gmm_handle;
static dissector_handle_t sndcp_handle;
static dissector_handle_t data_handle;

/* Initialize the protocol and registered fields */
static int proto_gsm0464         = -1;
static int hf_gsm0464_sapi       = -1; /* SAPI */
static int hf_gsm0464_cr_bit     = -1; /* Command/Response Bit */
static int hf_gsm0464_ctrl_format_i= -1; /* Control Field Format */
static int hf_gsm0464_ctrl_format_s= -1; /* Control Field Format */
static int hf_gsm0464_ctrl_format_ui= -1; /* Control Field Format */
static int hf_gsm0464_ctrl_format_u= -1; /* Control Field Format */
static int hf_gsm0464_a_bit_i    = -1; /* Acknowledgement Request Bit in I-Frame */
static int hf_gsm0464_a_bit_s    = -1; /* Acknowledgement Request Bit in S-Frame */
static int hf_gsm0464_e_bit      = -1; /* Encryption Function Bit */
static int hf_gsm0464_m_bits     = -1; /* Unnumbered Function Bits (M4,M3,M2,M1) */
static int hf_gsm0464_nr         = -1; /* Transmitter Receive Sequence Number */
static int hf_gsm0464_ns         = -1; /* Transmitter Send Sequence Number */
static int hf_gsm0464_nu         = -1; /* Transmitter Unconfirmed Sequence Number */
static int hf_gsm0464_pm_bit     = -1; /* Protected Mode Bit */
static int hf_gsm0464_pf_bit     = -1; /* Poll/Final Bit */
static int hf_gsm0464_s_bits     = -1; /* Supervisory Function Bits (S1,S2) */

/* Initialize the subtree pointers */
static gint ett_gsm0464 = -1;
static gint ett_gsm0464_sack = -1;
static gint ett_gsm0464_sapi = -1;

/**** ADDRESS FIELD MACROS ****/

/* Values for SAPI */
#define SAPI_LLGMM                  (0x01)
#define SAPI_TOM2                   (0x02)
#define SAPI_LL3                    (0x03)
#define SAPI_LL5                    (0x05)
#define SAPI_LLSMS                  (0x07)
#define SAPI_TOM8                   (0x08)
#define SAPI_LL9                    (0x09)
#define SAPI_LL11                   (0x0b)

/* Mask for SAPI */
#define SAPI_MASK                   (0x0f)

/* Mask for Protocol Discriminator */
#define PD_MASK                     (0x80)

/* Mask for Command/Response bit */
#define CR_MASK                     (0x40)

/**** CONTROL FIELD MACROS ****/

/* Format values & masks */
#define I_FORMAT                    (0x00)
#define I_FORMAT_MASK               (0x80)
#define S_FORMAT                    (0x80)
#define S_FORMAT_MASK               (0xc0)
#define UI_FORMAT                   (0xc0)
#define UI_FORMAT_MASK              (0xe0)
#define U_FORMAT                    (0xe0)
#define U_FORMAT_MASK               (0xe0)

/* S-bits */
#define S_RR                        (0x00)
#define S_ACK                       (0x01)
#define S_RNR                       (0x02)
#define S_SACK                      (0x01)

#define S_BITS_MASK                 (0x03)

/* M-bits */
#define M_NULL                      (0x00)
#define M_DM                        (0x01)
#define M_DISC                      (0x04)
#define M_UA                        (0x06)
#define M_SABM                      (0x07)
#define M_FRMR                      (0x08)
#define M_XID                       (0x0b)

#define M_BITS_MASK                 (0x0f)

/* A-bit */

#define I_FRAME_A_BIT_MASK          (0x40)
#define S_FRAME_A_BIT_MASK          (0x20)

/* E-bit */
#define E_BIT_MASK                  (0x02)

/* PM-bit */
#define PM_BIT_MASK                 (0x01)

/* PF-bit */
#define PF_BIT_MASK                 (0x10)

/* NS */ 
#define NS_MASK                     (0x1ff0)

/* NR */ 
#define NR_MASK                     (0x07fc)

/* NU */ 
#define NU_MASK                     (0x07fc)

#define GET_SEQ_NUM( MSB, LSB) ( \
    (((guint16) (MSB)) << 8) + ((guint16) (LSB)) \
    )

/* SACK data length */
#define SACK_DATA_LEN_MASK          (0x1f)

/* Value Strings */
static const value_string sapi_vals[] = {
  { SAPI_LLGMM, "LLGMM (GPRS Mobility Management)" },
  { SAPI_TOM2,  "TOM2 (Tunnelling of messages 2)" },
  { SAPI_LL3,   "LL3 (User data 3)" },
  { SAPI_LL5,   "LL5 (User data 5)" },
  { SAPI_LLSMS, "LLSMS (SMS)" },
  { SAPI_TOM8,  "TOM8 (Tunnelling of messages 8)" },
  { SAPI_LL9,   "LL9 (User data 9)" },
  { SAPI_LL11,  "LL11 (User data 11)" },
  { 0, NULL }
};

static const value_string sapi_short_vals[] = {
  { SAPI_LLGMM, "LLGMM" },
  { SAPI_TOM2,  "TOM2" },
  { SAPI_LL3,   "LL3" },
  { SAPI_LL5,   "LL5" },
  { SAPI_LLSMS, "LLSMS" },
  { SAPI_TOM8,  "TOM8" },
  { SAPI_LL9,   "LL9" },
  { SAPI_LL11,  "LL11" },
  { 0, NULL }
};

static const value_string ctrl_format_vals[] = {
  { 0x0,   "Information Transfer (I)" },
  { 0x4,   "Supervisory (S)" },
  { 0x6,   "Unconfirmed Information (UI)" },
  { 0x7,   "Unnumbered (U)" },
  { 0, NULL }
};

static const value_string s_bits_vals[] = {
  { S_RR,       "RR" },
  { S_ACK,      "ACK" },
  { S_RNR,      "RNR" },
  { S_SACK,     "SACK" },
  { 0, NULL }
};

static const value_string m_bits_vals[] = {
  { M_NULL,     "NULL" },
  { M_DM,       "DM" },
  { M_DISC,     "DISC" },
  { M_UA,       "UA" },
  { M_SABM,     "SABM" },
  { M_FRMR,     "FRMR" },
  { M_XID,      "XID" },
  { 0, NULL }
};


/* Code to dissect the NS messages */
static void
dissect_gsm0464(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
  int offset=0, msg_len;
  guint8 msg_addr;                          /* Address byte */
  guint8 sapi, pd_bit;                      /* Fields within Address byte */
  guint8 msg_ctrl0, msg_ctrl1, msg_ctrl2;   /* Control Field byte(s) */
  guint8 ctrl_format, s_bits;               /* Field within Control Field byte(s) */
  guint16 nr, ns, nu;                       /* Fields within Control Field byte(s) (cont.)*/
  guint8 sack_length;
  int close_brackets=TRUE;
  tvbuff_t *next_tvb;                       
  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  /* load the display labels   */
  pinfo->current_proto = "gsm0464";
  
  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM 04.64");
  if (check_col(pinfo->cinfo, COL_INFO))
    col_clear(pinfo->cinfo, COL_INFO);

  msg_len = tvb_length_remaining(tvb,offset);

  if (main_tree) {
    /* create display subtree for the protocol */
    ti = proto_tree_add_item(main_tree, proto_gsm0464, tvb, offset, msg_len, FALSE);
    tree = proto_item_add_subtree(ti, ett_gsm0464);
  }

  /*************************/
  /* Process Address field */
  /*************************/
  msg_addr = tvb_get_guint8(tvb, offset);
  pd_bit = (msg_addr & PD_MASK) ? 1 : 0;
  sapi = msg_addr & SAPI_MASK;

  if (tree) 
  {
    /* create display subtree for the protocol */
    if(pd_bit)
    {
      proto_item_set_text(ti, "GSM LLC: Service Access Point (*** BAD PD ***): %s",rd_match_strval(sapi,sapi_vals));
    }
    else
    {
      proto_item_set_text(ti, "GSM LLC: Service Access Point: %s",rd_match_strval(sapi,sapi_vals));
    }
    proto_tree_add_uint(tree, hf_gsm0464_cr_bit, tvb, offset, 1, msg_addr);
    proto_tree_add_uint(tree, hf_gsm0464_sapi, tvb, offset, 1, msg_addr);
  }
  if (check_col(pinfo->cinfo, COL_INFO))
  {
    if(pd_bit)
    {
      col_append_fstr(pinfo->cinfo, COL_INFO, "++%s (*** BAD PD ***)(",rd_match_strval(sapi,sapi_short_vals));
    }
    else
    {
      col_append_fstr(pinfo->cinfo, COL_INFO, "++%s(",rd_match_strval(sapi,sapi_short_vals));
    } }
  offset += 1;

  /*************************/
  /* Process Control field */
  /*************************/

  msg_ctrl0 = tvb_get_guint8(tvb, offset);

  if( (msg_ctrl0 & I_FORMAT_MASK) == I_FORMAT)
  {
    /* Get I-frame Control Field stuff */
    ctrl_format = I_FORMAT;
    msg_ctrl1 = tvb_get_guint8(tvb, offset + 1);
    msg_ctrl2 = tvb_get_guint8(tvb, offset + 2);
    ns = GET_SEQ_NUM( msg_ctrl0, msg_ctrl1);
    nr = GET_SEQ_NUM( msg_ctrl1, msg_ctrl2);
    s_bits = msg_ctrl2 & S_BITS_MASK;

    /* Display I-frame Control Field stuff */
    if(tree)
    {
      proto_tree_add_uint(tree, hf_gsm0464_ctrl_format_i, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_a_bit_i, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_ns, tvb, offset, 2, ns);
      proto_tree_add_uint(tree, hf_gsm0464_nr, tvb, offset + 1, 2, nr);
      proto_tree_add_uint(tree, hf_gsm0464_s_bits, tvb, offset + 2, 1, msg_ctrl2);
    }
    offset += 3;

    /* Display SACK if present */
    if(s_bits == S_SACK)
    {
      sack_length = tvb_get_guint8(tvb, offset);
      sack_length &= SACK_DATA_LEN_MASK;
      offset++;
      if(tree)
      {
        ti = proto_tree_add_text(tree, tvb, offset, sack_length, "SACK Bitmap");
        tree = proto_item_add_subtree(ti, ett_gsm0464_sack);
        /* create a new subset TVB for disecting the unknown payload */
        next_tvb = tvb_new_subset(tvb, offset, sack_length, -1);
        call_dissector(data_handle, next_tvb, pinfo, tree);
      }
      offset += sack_length;
    }
  }
  else if( (msg_ctrl0 & S_FORMAT_MASK) == S_FORMAT)
  {
    /* Get S-frame Control Field stuff */
    ctrl_format = S_FORMAT;
    msg_ctrl1 = tvb_get_guint8(tvb, offset + 1);
    nr = GET_SEQ_NUM( msg_ctrl0, msg_ctrl1);
    s_bits = msg_ctrl1 & S_BITS_MASK;

    /* Display S-frame Control Field stuff */
    if(tree)
    {
      proto_tree_add_uint(tree, hf_gsm0464_ctrl_format_s, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_a_bit_s, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_nr, tvb, offset, 2, nr);
      proto_tree_add_uint(tree, hf_gsm0464_s_bits, tvb, offset + 1, 1, msg_ctrl1);
    }
    offset += 2;

    /* Display SACK if present */
    if(s_bits == S_SACK)
    {
      if(tree)
      {
        ti = proto_tree_add_text(tree, tvb, offset, -1, "SACK Bitmap");
        tree = proto_item_add_subtree(ti, ett_gsm0464_sack);
        /* create a new subset TVB for disecting the unknown payload */
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(data_handle, next_tvb, pinfo, tree);
      }
    }
  }
  else if( (msg_ctrl0 & UI_FORMAT_MASK) == UI_FORMAT)
  {
    /* Get UI-frame Control Field stuff */
    ctrl_format = UI_FORMAT;
    msg_ctrl1 = tvb_get_guint8(tvb, offset + 1);
    nu = GET_SEQ_NUM( msg_ctrl0, msg_ctrl1);

    /* Display UI-frame Control Field stuff */
    if(tree)
    {
      proto_tree_add_uint(tree, hf_gsm0464_ctrl_format_ui, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_nu, tvb, offset, 2, nu);
      proto_tree_add_uint(tree, hf_gsm0464_e_bit, tvb, offset + 1, 1, msg_ctrl1);
      proto_tree_add_uint(tree, hf_gsm0464_pm_bit, tvb, offset + 1, 1, msg_ctrl1);
    }
    offset += 2;
  }
  else
  {
    /* Get U-frame Control Field stuff */
    ctrl_format = U_FORMAT;

    /* Display UI-frame Control Field stuff */
    if(tree)
    {
      proto_tree_add_uint(tree, hf_gsm0464_ctrl_format_u, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_pf_bit, tvb, offset, 1, msg_ctrl0);
      proto_tree_add_uint(tree, hf_gsm0464_m_bits, tvb, offset, 1, msg_ctrl0);
    }
    offset++;
  }

  /***********************/
  /* Process Information */
  /***********************/

  if( (ctrl_format == I_FORMAT) || (ctrl_format == UI_FORMAT) )
  {
    switch(sapi)
    {
      case SAPI_LLGMM:
      {
        /* use gmm decoder */
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(gsm24008gmm_handle, next_tvb, pinfo, main_tree);
        close_brackets=FALSE;
        break;
      }
      case SAPI_LL3:
      case SAPI_LL5:
      case SAPI_LL9:
      case SAPI_LL11:
      {
        /* use sndcp decoder for the user data */
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(sndcp_handle, next_tvb, pinfo, main_tree);
        close_brackets=FALSE;
        break;
      }
      default:
      {
        /* Dump data */
        ti = proto_tree_add_text(tree, tvb, offset, -1, "SAPI %s:", rd_match_strval(sapi,sapi_vals));
        tree = proto_item_add_subtree(ti, ett_gsm0464_sapi);
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(data_handle,next_tvb, pinfo, tree);
        break;
      }
    }
  }

  if ( (close_brackets) && (check_col(pinfo->cinfo, COL_INFO)))
    col_append_fstr(pinfo->cinfo, COL_INFO, ")");

  return;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0464(void)
{     

/* Setup list of header fields  See Section 1.6.1 for details*/
  static hf_register_info hf[] = {
    { &hf_gsm0464_sapi,
      { "SAPI", "gsm0464.sapi",
      FT_UINT8, BASE_HEX, VALS(sapi_vals), SAPI_MASK,
      "Service Access Point Indicator", HFILL }
    },
    { &hf_gsm0464_cr_bit,
      { "CR", "gsm0464.cr_bit",
      FT_UINT8, BASE_DEC, NULL, CR_MASK,
      "Command/Response Bit", HFILL }
    },
    { &hf_gsm0464_ctrl_format_i,
      { "Format", "gsm0464.ctrl_format",
      FT_UINT8, BASE_HEX, VALS(ctrl_format_vals), I_FORMAT_MASK,
      "Control Field Format", HFILL }
    },
    { &hf_gsm0464_ctrl_format_s,
      { "Format", "gsm0464.ctrl_format",
      FT_UINT8, BASE_HEX, VALS(ctrl_format_vals), S_FORMAT_MASK,
      "Control Field Format", HFILL }
    },
    { &hf_gsm0464_ctrl_format_ui,
      { "Format", "gsm0464.ctrl_format",
      FT_UINT8, BASE_HEX, VALS(ctrl_format_vals), UI_FORMAT_MASK,
      "Control Field Format", HFILL }
    },
    { &hf_gsm0464_ctrl_format_u,
      { "Format", "gsm0464.ctrl_format",
      FT_UINT8, BASE_HEX, VALS(ctrl_format_vals), U_FORMAT_MASK,
      "Control Field Format", HFILL }
    },
    { &hf_gsm0464_a_bit_i,
      { "A", "gsm0464.a_bit",
      FT_UINT8, BASE_DEC, NULL, I_FRAME_A_BIT_MASK,
      "Acknowledgement Request Bit", HFILL }
    },
    { &hf_gsm0464_a_bit_s,
      { "A", "gsm0464.a_bit",
      FT_UINT8, BASE_DEC, NULL, S_FRAME_A_BIT_MASK,
      "Acknowledgement Request Bit", HFILL }
    },
    { &hf_gsm0464_e_bit,
      { "E", "gsm0464.e_bit",
      FT_UINT8, BASE_DEC, NULL, E_BIT_MASK,
      "Encryption Function Bit", HFILL }
    },
    { &hf_gsm0464_m_bits,
      { "M4M3M2M1", "gsm0464.m_bits",
      FT_UINT8, BASE_DEC, NULL, M_BITS_MASK,
      "Unnumbered Function Bits (M4,M3,M2,M1)", HFILL }
    },
    { &hf_gsm0464_nr,
      { "N(R)", "gsm0464.nr",
      FT_UINT16, BASE_DEC, NULL, NR_MASK,    
      "Transmitter Receive Sequence Number", HFILL }
    },
    { &hf_gsm0464_ns,
      { "N(S)", "gsm0464.ns",
      FT_UINT16, BASE_DEC, NULL, NS_MASK,    
      "Transmitter Send Sequence Number", HFILL }
    },
    { &hf_gsm0464_nu,
      { "N(U)", "gsm0464.nu",
      FT_UINT16, BASE_DEC, NULL, NU_MASK,    
      "Transmitter Unconfirmed Sequence Number", HFILL }
    },
    { &hf_gsm0464_pm_bit,
      { "PM", "gsm0464.pm_bit",
      FT_UINT8, BASE_DEC, NULL, PM_BIT_MASK,
      "Protected Mode Bit", HFILL }
    },
    { &hf_gsm0464_pf_bit,
      { "PF", "gsm0464.pf_bit",
      FT_UINT8, BASE_DEC, NULL, 0x00,    
      "Poll/Final Bit" , HFILL}
    },
    { &hf_gsm0464_s_bits,
      { "S1S2", "gsm0464.s_bits",
      FT_UINT8, BASE_DEC, NULL, S_BITS_MASK,
      "Supervisory Function Bits (S1,S2)", HFILL }
    },
  };


/* Setup protocol subtree array */
  static gint *ett[] = {
    &ett_gsm0464,
    &ett_gsm0464_sack,
    &ett_gsm0464_sapi
  };

  proto_gsm0464 = proto_register_protocol("GSM 04.64 Subnetwork Dependent Convergence Protocol (SNDCP)", "GSM 04.64", "gsm0464");

  register_dissector("gsm0464", dissect_gsm0464, proto_gsm0464);

/* Required function calls to register the header fields and subtrees used */
  proto_register_field_array(proto_gsm0464, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0464(void)
{
  dissector_handle_t gsm0464_handle;

  gsm0464_handle = find_dissector("gsm0464");
  gsm24008gmm_handle   = find_dissector("gsm24008gmm"); 
  sndcp_handle   = find_dissector("sndcp"); 
  data_handle    = find_dissector("data");

#ifdef UDP_GB_PORT
  dissector_add("udp.port", UDP_GB_PORT, gsm0464_handle);
#endif
}
