/* packet-gsm0806.c
 * Routines for GSM 08.06
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0806.c 1.2 2007/03/09 15:24:16GMT Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "prefs.h"
#include "packet-gsm.h"
#include "tap.h"

#define SCCP_SSN_BSSAP          254     /* the ITU-T   SCCP SSN value for GSM BSS Application Part */
#define SCCP_SSN_BSSAP_EIA_634  252     /* the EIA-634 SCCP SSN value for GSM BSS Application Part */
#define SCCP_SSN_BSSAP_222      222     /* the SCCP SSN value used by the T-Mobile Ericsson MSC */

/* function prototypes for this file */
void proto_reg_handoff_gsm0806(void);
gchar *data_str( tvbuff_t *tvb, int offset, int data_len, gchar *buf, int buf_len );

/* Initialize the protocol and registered fields */
static int proto_gsm0806 = -1;

static int hf_gsm0806_desc = -1;
static int hf_gsm0806_dlci = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0806 = -1;

static dissector_handle_t gsm0806_handle;
static dissector_handle_t gsm0808_handle;
static dissector_handle_t gsm0408_handle;

/* Initialise the Preferences */
static int global_bssap_sccp_ssn = SCCP_SSN_BSSAP;

/* Value Strings */
static const value_string gsm0806_desc_vals[] = {
	{ 0x00, "BSSMAP" },
	{ 0x01, "DTAP" },
	{ 0, NULL }
};

/* Code to dissect the BSAP messages */
static int
dissect_gsm0806(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 desc, len, dlci;
	tvbuff_t *next_tvb;

	/* load the display labels 	*/
	pinfo->current_proto = "GSM 08.06";

	/* following call moved up to packet-sccp.c */
	/* gsmstat_set_interface(GSMSTAT_INTERFACE_A,pinfo); */

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>1) {

		desc = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm0806_desc, tvb, offset, 1, desc);
		offset += 1;

		if (desc == 0) {	/* BSSMAP */
			len = tvb_get_guint8(tvb, offset);
			offset += 1;

			/* create a new subset TVB for disecting the payload */
			next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, len);
			offset += call_dissector(gsm0808_handle, next_tvb, pinfo, main_tree);
		}
		else if (desc == 1) {	/* DTAP */

			dlci = tvb_get_guint8(tvb, offset);
			if (main_tree)
				proto_tree_add_uint_hidden(main_tree, hf_gsm0806_dlci, tvb, offset, 1, dlci);
			offset += 1;

			len = tvb_get_guint8(tvb, offset);
			offset += 1;

			/* create a new subset TVB for disecting the payload */
			next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, len);
			offset += call_dissector(gsm0408_handle, next_tvb, pinfo, main_tree);

		}
		else {	/* wrong! */
			if (main_tree)
				proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"Illegal Descriminator value!!!");
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, " Illegal Descriminator value!!!");
		}
	}

	if (offset)
        tap_queue_packet(gsmstat_tap,pinfo,NULL);

	return offset;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0806(void)
{
    module_t *gsm0806_module;

    /* Setup list of header fields */
	static hf_register_info hf[] = {
		{ &hf_gsm0806_desc,
			{ "Descriminator", "gsm0806.desc",
			FT_UINT8, BASE_HEX, VALS(gsm0806_desc_vals), 0x0,
			"Message Descriminator", HFILL }
		},
		{ &hf_gsm0806_dlci,
			{ "DLCI", "gsm0806.dlci",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Data Link Connection Identifier", HFILL }
		},
	};

    /* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0806,
	};

	proto_gsm0806 = proto_register_protocol("GSM 08.06 A Interface BSAP", "GSM 08.06", "gsm0806");
    gsm0806_handle = new_create_dissector_handle(dissect_gsm0806, proto_gsm0806);

    /* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0806, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

    /* Register preferences - basically SCCP SSN */
    gsm0806_module = prefs_register_protocol(proto_gsm0806, proto_reg_handoff_gsm0806);

    prefs_register_uint_preference(gsm0806_module, "bssap_sccp_ssn", "SCCP SSN for BSSAP",
				 "The SCCP SubStation Number for BSSAP",
				 10, (gint *)&global_bssap_sccp_ssn);

	/* common gsmstat tap registration
	 - do NOT use the register tap directly in each protocol - it confuses ethereal taps! */
	gsmstat_register_tap();

	return;
}

/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0806(void)
{
    static int initialized = FALSE;
    static int local_bssap_sccp_ssn;

    if (!initialized) {
        gsm0408_handle = find_dissector("gsm0408");
        gsm0808_handle = find_dissector("gsm0808");

        initialized = TRUE;
	}
    else {
        dissector_delete("sccp.ssn", global_bssap_sccp_ssn, gsm0806_handle);
	}

    /* Set our SCCP SSN for linking to this dissector */
    dissector_add("sccp.ssn", global_bssap_sccp_ssn, gsm0806_handle);
    local_bssap_sccp_ssn = global_bssap_sccp_ssn;

    return;
}
