/* packet-gsm0816.c
 * Routines for GSM 08.06
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0816.c 1.2 2007/03/09 15:29:00GMT Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include <epan/prefs.h>

#include "packet-gsm.h"

/* function prototypes for this file */
static int dissect_gsm0816_ie(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
void proto_reg_handoff_gsm0816(void);

/* Initialize the protocol and registered fields */
static int proto_gsm0816 = -1;
static int hf_gsm0816_msg_type = -1;
static int hf_gsm0816_iei = -1;
static int hf_gsm0816_ie_len = -1;
static int hf_gsm0816_bvci = -1;
static int hf_gsm0816_ns_vci = -1;
static int hf_gsm0816_nsei = -1;
static int hf_gsm0816_cause = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0816 = -1;
static gint ett_gsm0816_ie = -1;

static dissector_handle_t gprs_ns_handle;
static dissector_handle_t gsm0816_handle;
static dissector_handle_t gsm0818_handle;
static dissector_handle_t data_handle;

/* Value Strings */
static const value_string msg_type_vals[] = {
	{ 0x00, "NS-UNITDATA" },
	{ 0x02, "NS-RESET" },
	{ 0x03, "NS-RESET-ACK" },
	{ 0x04, "NS-BLOCK" },
	{ 0x05, "NS-BLOCK-ACK" },
	{ 0x06, "NS-UNBLOCK" },
	{ 0x07, "NS-UNBLOCK-ACK" },
	{ 0x08, "NS-STATUS" },
	{ 0x0a, "NS-ALIVE" },
	{ 0x0b, "NS-ALIVE-ACK" },
	{ 0, NULL }
};
#define NS_UNITDATA		0x00

static const value_string iei_vals[] = {
	{ 0x00, "Cause" },
	{ 0x01, "NS-VCI" },
	{ 0x02, "NS-PDU" },
	{ 0x03, "BVCI" },
	{ 0x04, "NSEI" },
	{ 0, NULL }
};

#define IEI_CAUSE	0x00
#define IEI_NS_VCI	0x01
#define IEI_NS_PDU	0x02
#define IEI_BVCI	0x03
#define IEI_NSEI	0x04

static const value_string cause_vals[] = {
	{ 0x00, "Transit network failure" },
	{ 0x01, "O&M intervention" },
	{ 0x02, "Equipment failure" },
	{ 0x03, "NS-VC blocked" },
	{ 0x04, "NS-VC unknown" },
	{ 0x05, "BVCI unknown on that NSE" },
	{ 0x08, "Semantically incorrect PDU" },
	{ 0x0a, "PDU not compatible with the protocol state" },
	{ 0x0b, "Protocol error - unspecified" },
	{ 0x0c, "Invalid essential IE" },
	{ 0x0d, "Missing essential IE" },
	{ 0, NULL }
};


/* Code to dissect the NS messages */
static int
dissect_gsm0816(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type;
	guint16 bvci;
	tvbuff_t *next_tvb;
  int close_brackets = TRUE;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "gsm0816";
	
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM 08:16");
	if (check_col(pinfo->cinfo, COL_INFO))
		col_clear(pinfo->cinfo, COL_INFO);

	msg_len = tvb_length_remaining(tvb,offset);

	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm0816, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_gsm0816);
	}

	msg_type = tvb_get_guint8(tvb, offset);
	if (tree) {
		/* create display subtree for the protocol */
		proto_item_set_text(ti, "GSM Gb Network Service:%s",rd_match_strval(msg_type,msg_type_vals));
		proto_tree_add_uint(tree, hf_gsm0816_msg_type, tvb, offset, 1, msg_type);
	}
	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s(",rd_match_strval(msg_type,msg_type_vals));
	offset += 1;

	if (msg_type == NS_UNITDATA) {
		if (msg_len > 4) {	/* should be 5+ for UnitData - treat specially: uses V format IEs only! */
			offset += 1;	/* skip spare octet - should be zero anyway */

			bvci = tvb_get_ntohs(tvb, offset);
			if (main_tree)
				proto_tree_add_uint(tree, hf_gsm0816_bvci, tvb, offset, 2, bvci);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "BVCI=%d,", bvci);
			offset += 2;

			/* create a new subset TVB for disecting the payload */
			next_tvb = tvb_new_subset(tvb, offset, -1, -1);
            offset += call_dissector(gsm0818_handle, next_tvb, pinfo, main_tree);
      close_brackets = FALSE;
		}
	}
	else {	/* All other messages use TLV IEs throughout */

		while (offset < msg_len) {	/* simply parse messages blind to extract any IEs */
			next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
			offset += dissect_gsm0816_ie(next_tvb, pinfo, tree);
		}
	}

	if((close_brackets) &&  (check_col(pinfo->cinfo, COL_INFO)))
		col_append_fstr(pinfo->cinfo, COL_INFO, ")");

    return offset;
}

/* Code to dissect the IEs */
static int
dissect_gsm0816_ie(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8	iei, cause;
	guint16 ie_len, bvci, ns_vci, nsei;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		iei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm0816_iei, tvb, offset, 1, iei);
		offset += 1;

		ie_len = tvb_get_guint8(tvb, offset);
		offset += 1;
		if ((ie_len & 0x80))	{	/* extension bit indicates NOT extended length field */
			ie_len = ie_len & 0x7f;
		}
		else	{	/* extended length field */
			ie_len = (ie_len<<8) + tvb_get_guint8(tvb, offset);
			offset += 1;
		}

		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm0816_ie_len, tvb, offset, 2, ie_len);
			
		switch (iei) {
		case IEI_BVCI:
			bvci = tvb_get_ntohs(tvb, offset);
			if (main_tree)
				proto_tree_add_uint(main_tree, hf_gsm0816_bvci, tvb, offset, 2, bvci);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "BVCI=%d,", bvci);
			break;

		case IEI_NS_VCI:
			ns_vci = tvb_get_ntohs(tvb, offset);
			if (main_tree)
				proto_tree_add_uint(main_tree, hf_gsm0816_ns_vci, tvb, offset, 2, ns_vci);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "NS-VCI=%d,", ns_vci);
			break;

		case IEI_NSEI:
			nsei = tvb_get_ntohs(tvb, offset);
			if (main_tree)
				proto_tree_add_uint(main_tree, hf_gsm0816_nsei, tvb, offset, 2, nsei);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "NSEI=%d,", nsei);
			break;

		case IEI_CAUSE:
			cause = tvb_get_guint8(tvb, offset);
			if (main_tree)
				proto_tree_add_uint(main_tree, hf_gsm0816_cause, tvb, offset, 1, cause);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "Cause=%s,", rd_match_strval(cause,cause_vals));
			break;

		case IEI_NS_PDU:
			ti = proto_tree_add_text(main_tree, tvb, 0, offset, "NS PDU");
			tree = proto_item_add_subtree(ti, ett_gsm0816_ie);
			/* create a new subset TVB for disecting the PDU payload */
			next_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
			call_dissector(gsm0816_handle,next_tvb, pinfo, tree);
			break;

		default:
			ti = proto_tree_add_text(main_tree, tvb, 0, offset, "Unknown IEI %d", iei);
			tree = proto_item_add_subtree(ti, ett_gsm0816_ie);
			/* create a new subset TVB for disecting the unknown payload */
			next_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
			call_dissector(data_handle,next_tvb, pinfo, tree);
			break;
		}
		offset += ie_len;
	}

	return offset;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0816(void)
{		 
    /* Setup list of header fields */
	static hf_register_info hf[] = {
		{ &hf_gsm0816_msg_type,
			{ "Msg Type", "gsm0816.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,	  
			"Message Type", HFILL }
		},
		{ &hf_gsm0816_iei,
			{ "IEI", "gsm0816.iei",
			FT_UINT8, BASE_HEX, VALS(iei_vals), 0x0,	  
			"Information Element Identifier" , HFILL}
		},
		{ &hf_gsm0816_ie_len,
			{ "IE Len", "gsm0816.ie_len",
			FT_UINT8, BASE_HEX, NULL, 0x0,	  
			"Information Element Length", HFILL }
		},
		{ &hf_gsm0816_bvci,
			{ "BVCI", "gsm0816.bvci",
			FT_UINT16, BASE_DEC, NULL, 0x0,	  
			"BVCI" , HFILL}
		},
		{ &hf_gsm0816_ns_vci,
			{ "NS-VCI", "gsm0816.ns_vci",
			FT_UINT16, BASE_DEC, NULL, 0x0,	  
			"NS-VCI", HFILL }
		},
		{ &hf_gsm0816_nsei,
			{ "NSEI", "gsm0816.nsei",
			FT_UINT16, BASE_DEC, NULL, 0x0,	  
			"NSEI", HFILL }
		},
		{ &hf_gsm0816_cause,
			{ "Cause", "gsm0816.cause",
			FT_UINT8, BASE_HEX, VALS(cause_vals), 0x0,	  
			"Cause", HFILL }
		},
	};

    /* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0816,
		&ett_gsm0816_ie,
	};

	proto_gsm0816 = proto_register_protocol("GSM 08.16 Gb Interface Network Service", "GSM 08.16", "gsm0816");
	register_dissector("gsm0816", dissect_gsm0816, proto_gsm0816);

    /* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0816, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

}

/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0816(void)
{
    static gboolean initialised = FALSE;
    static guint16 local_start_port;
    static guint16 local_end_port;
    static gboolean local_use_ipa_dissector;

    if (!initialised) {
    gsm0816_handle = find_dissector("gsm0816");
        gprs_ns_handle = find_dissector("gprs_ns");
    gsm0818_handle = find_dissector("gsm0818");
    data_handle = find_dissector("data");

        initialised = TRUE;
    }
}
