/* packet-gsm0818.c
 * Routines for GSM 08.06
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0818.c 1.2 2007/02/09 17:47:31GMT Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"

/* Local data */
static int close_brackets = TRUE;

/* function prototypes for this file */
int dissect_gsm0818_ie(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ie_tree, proto_tree *main_tree);

/* Initialize the protocol and registered fields */
static int proto_gsm0818 = -1;
static int hf_gsm0818_msg_type = -1;
static int hf_gsm0818_iei = -1;
static int hf_gsm0818_ie_len = -1;
static int hf_gsm0818_bvci = -1;
static int hf_gsm0818_cause = -1;
static int hf_gsm0818_tlli = -1;
static int hf_gsm0818_qos_profile = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0818 = -1;
static gint ett_gsm0818_ie = -1;

static dissector_handle_t llcgprs_handle;
static dissector_handle_t gsm0818_handle;
static dissector_handle_t data_handle;

/* Value Strings */
static const value_string msg_type_vals[] = {
	/* PDUs between RL and BSSGP SAPs */
	{ 0x00, "DL-UNITDATA" },
	{ 0x01, "UL-UNITDATA" },
	{ 0x02, "RA-CAPABILITY" },
	{ 0x03, "PTM-UNITDATA" },
	/* PDUs between GMM SAPs */
	{ 0x06, "PAGING PS" },
	{ 0x07, "PAGING CS" },
	{ 0x08, "RA-CAPABILITY-UPDATE" },
	{ 0x09, "RA-CAPABILITY-UPDATE-ACK" },
	{ 0x0a, "RADIO-STATUS" },
	{ 0x0b, "SUSPEND" },
	{ 0x0c, "SUSPEND-ACK" },
	{ 0x0d, "SUSPEND-NACK" },
	{ 0x0e, "RESUME" },
	{ 0x0f, "RESUME-ACK" },
	{ 0x10, "RESUME-NACK" },
	/* PDUs between NM SAPs */
	{ 0x20, "BVC-BLOCK" },
	{ 0x21, "BVC-BLOCK-ACK" },
	{ 0x22, "BVC-RESET" },
	{ 0x23, "BVC-RESET-ACK" },
	{ 0x24, "BVC-UNBLOCK" },
	{ 0x25, "BVC-UNBLOCK-ACK" },
	{ 0x26, "FLOW-CONTROL-BVC" },
	{ 0x27, "FLOW-CONTROL-BVC-ACK" },
	{ 0x28, "FLOW-CONTROL-MS" },
	{ 0x29, "FLOW-CONTROL-MS-ACK" },
	{ 0x2a, "FLUSH-LL" },
	{ 0x2b, "FLUSH-LL-ACK" },
	{ 0x2c, "LLC-DISCARDED" },
	/* */
	{ 0x40, "SGSN-INVOKE-TRACE" },
	{ 0x41, "STATUS" },
	/* */
	{ 0x50, "DOWNLOAD-BSS-PFC" },
	{ 0x51, "CREATE-BSS-PFC" },
	{ 0x52, "CREATE-BSS-PFC-ACK" },
	{ 0x53, "CREATE-BSS-PFC-NACK" },
	{ 0x54, "MODIFY-BSS-PFC" },
	{ 0x55, "MODIFY-BSS-PFC-ACK" },
	{ 0x56, "DELETE-BSS-PFC" },
	{ 0x57, "DELETE-BSS-PFC-ACK" },
	{ 0, NULL }
};
#define DL_UNITDATA		0x00
#define UL_UNITDATA		0x01

static const value_string iei_vals[] = {
	{ 0x00, "Alignment Octets" },
	{ 0x01, "Bmax default MS" },
	{ 0x02, "BSS Area Indication" },
	{ 0x03, "Bucket Leak Rate" },
	{ 0x04, "BVCI" },
	{ 0x05, "BVC Bucket Size" },
	{ 0x06, "BVC Measurement" },
	{ 0x07, "Cause" },
	{ 0x08, "Cell Identifier" },
	{ 0x09, "Channel needed" },
	{ 0x0a, "DRX Parameters" },
	{ 0x0b, "eMLPP Priority" },
	{ 0x0c, "Flush Action" },
	{ 0x0d, "IMSI" },
	{ 0x0e, "LLC-PDU" },
	{ 0x0f, "LLC Frames Discarded" },
	{ 0x10, "Location Area" },
	{ 0x11, "Mobile Id" },
	{ 0x12, "MS Bucket Size" },
	{ 0x13, "MS Radio Access Capability" },
	{ 0x14, "OMC Id" },
	{ 0x15, "PDU In Error" },
	{ 0x16, "PDU Lifetime" },
	{ 0x17, "Priority" },
	{ 0x18, "QoS Profile" },
	{ 0x19, "Radio Cause" },
	{ 0x1a, "RA-Cap-Upd-Cause" },
	{ 0x1b, "Routeing Area" },
	{ 0x1c, "R_default_MS" },
	{ 0x1d, "Suspend Reference Number" },
	{ 0x1e, "Tag" },
	{ 0x1f, "TLLI" },
	{ 0x20, "TMSI" },
	{ 0x21, "Trace Reference" },
	{ 0x22, "Trace Type" },
	{ 0x23, "Transaction Id" },
	{ 0x24, "Trigger Id" },
	{ 0x25, "Number of octets affected" },
	{ 0x26, "LSA Identifier List" },
	{ 0x27, "LSA Information" },
	{ 0x28, "Packet Flow Identifier" },
	{ 0x29, "Packet Flow Timer" },
	{ 0x3a, "Aggregate BSS QoS Profile (ABQP)" },
	{ 0x3b, "Feature Bitmap" },
	{ 0x3c, "Bucket_Full Ratio" },
	{ 0x3d, "Service UTRAN CCO (Cell Change Order)" },
	{ 0x51, "PFC transfer result" },
	{ 0, NULL }
};

#define IEI_BVCI		0x04
#define IEI_CAUSE		0x07
#define IEI_QOS_PROFILE	0x18
#define IEI_TLLI		0x1f
#define IEI_LLC_PDU		0x0e

static const value_string cause_vals[] = {
	{ 0x00, "Processor overload" },
	{ 0x01, "Equipment failure" },
	{ 0x02, "Transit network service failure" },
	{ 0x03, "Network service trasmission capacity modified from 0kbps to >0kbps" },
	{ 0x04, "Unknown MS" },
	{ 0x05, "BVCI unknown" },
	{ 0x06, "cell traffic congestion" },
	{ 0x07, "SGSN congestion" },
	{ 0x08, "O&M intervention" },
	{ 0x09, "BVCI blocked" },
	{ 0x20, "Semantically incorrect PDU" },
	{ 0x21, "Invalid mandatory information" },
	{ 0x22, "Missing mandatory IE" },
	{ 0x23, "Missing conditional IE" },
	{ 0x24, "Unexpected conditional IE" },
	{ 0x25, "Conditional IE error" },
	{ 0x26, "PDU not compatible with the protocol state" },
	{ 0x27, "Protocol error - reserved for future use" },
	{ 0, NULL }
};


/* Code to dissect the NS messages */
void
dissect_gsm0818(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type;
	guint32 tlli, qos_profile;
	tvbuff_t *next_tvb;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

  close_brackets = TRUE;

	/* load the display labels 	*/
	pinfo->current_proto = "gsm0818";

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM 08:18");
  if (check_col(pinfo->cinfo, COL_INFO))
    col_clear(pinfo->cinfo, COL_INFO);

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm0818, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_gsm0818);
	}

	msg_type = tvb_get_guint8(tvb, offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		proto_item_set_text(ti, "GSM Gb BSS GPRS Protocol:%s",rd_match_strval(msg_type,msg_type_vals));
		proto_tree_add_uint(tree, hf_gsm0818_msg_type, tvb, offset, 1, msg_type);
	}
	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_fstr(pinfo->cinfo, COL_INFO, "+%s(",rd_match_strval(msg_type,msg_type_vals));
	offset += 1;

	if ((msg_type == DL_UNITDATA) || (msg_type == UL_UNITDATA)) {
		if (msg_len > 8) {	/* should be 9+ for UnitData - treat specially: uses V format IEs only! */

			tlli = tvb_get_ntohl(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_gsm0818_tlli, tvb, offset, 4, tlli);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "TLLI=0x%lx,", tlli);
			offset += 4;

			qos_profile = tvb_get_ntoh24(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_gsm0818_qos_profile, tvb, offset, 3, qos_profile);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "QoS=0x%lx", qos_profile);
			offset += 3;
		}
	}
	/* Rest of ?L_UNITDATA and all other messages use TLV IEs throughout */

	while (offset < msg_len) {	/* simply parse messages blind to extract any IEs */
		next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
		offset += dissect_gsm0818_ie(next_tvb, pinfo, tree, main_tree);
	}

	if( (close_brackets) && (check_col(pinfo->cinfo, COL_INFO)) )
		col_append_fstr(pinfo->cinfo, COL_INFO, ")");

	return;
}

/* Code to dissect the IEs */
int
dissect_gsm0818_ie(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ie_tree, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8	iei, cause;
	guint16 ie_len, bvci;
	guint32 tlli, qos_profile;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		iei = tvb_get_guint8(tvb, offset);
		if (ie_tree)
			proto_tree_add_uint_hidden(ie_tree, hf_gsm0818_iei, tvb, offset, 1, iei);
		offset += 1;

		ie_len = tvb_get_guint8(tvb, offset);
		offset += 1;
		if ((ie_len & 0x80))	{	/* extension bit indicates NOT extended length field */
			ie_len = ie_len & 0x7f;
		}
		else	{	/* extended length field */
			ie_len = (ie_len<<8) + tvb_get_guint8(tvb, offset);
			offset += 1;
		}

		if (ie_tree)
			proto_tree_add_uint_hidden(ie_tree, hf_gsm0818_ie_len, tvb, offset, 2, ie_len);

		switch (iei) {
		case IEI_BVCI:
			bvci = tvb_get_ntohs(tvb, offset);
			if (ie_tree)
				proto_tree_add_uint(ie_tree, hf_gsm0818_bvci, tvb, offset, 2, bvci);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "BVCI=%d,", bvci);
			break;

		case IEI_CAUSE:
			cause = tvb_get_guint8(tvb, offset);
			if (ie_tree)
				proto_tree_add_uint(ie_tree, hf_gsm0818_cause, tvb, offset, 1, cause);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "Cause=%s,", rd_match_strval(cause,cause_vals));
			break;

		case IEI_QOS_PROFILE:
			qos_profile = tvb_get_ntoh24(tvb, offset);
			if (ie_tree)
				proto_tree_add_uint(ie_tree, hf_gsm0818_qos_profile, tvb, offset, 3, qos_profile);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "QoS=0x%lx,", qos_profile);
			break;

		case IEI_TLLI:
			tlli = tvb_get_ntohl(tvb, offset);
			if (ie_tree)
				proto_tree_add_uint(ie_tree, hf_gsm0818_tlli, tvb, offset, 2, tlli);
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "TLLI=0x%lx,", tlli);
			break;

		case IEI_LLC_PDU:
			if (ie_tree)
			  ti = proto_tree_add_text(ie_tree, tvb, 0, offset+ie_len, "LLC-PDU");
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, ", LLC_PDU");
      /* create a new subset TVB for disecting the payload */
			next_tvb = tvb_new_subset(tvb, offset, -1, -1);
			call_dissector(llcgprs_handle, next_tvb, pinfo, main_tree);
      close_brackets = FALSE;
			break;

		default:
			if (ie_tree)
			  ti = proto_tree_add_text(ie_tree, tvb, 0, offset+ie_len, "%s", rd_match_strval(iei,iei_vals), data_str( tvb, offset, ie_len, str, sizeof(str) ));
			tree = proto_item_add_subtree(ti, ett_gsm0818_ie);
			/* create a new subset TVB for disecting the unknown payload */
			next_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
			call_dissector(data_handle, next_tvb, pinfo, tree);
			break;
		}
		offset += ie_len;
	}

	return offset;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0818(void)
{

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_gsm0818_msg_type,
			{ "Msg Type", "gsm0818.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,
			"Message Type", HFILL }
		},
		{ &hf_gsm0818_iei,
			{ "IEI", "gsm0818.iei",
			FT_UINT8, BASE_HEX, VALS(iei_vals), 0x0,
			"Information Element Identifier", HFILL }
		},
		{ &hf_gsm0818_ie_len,
			{ "IE Len", "gsm0818.ie_len",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Information Element Length", HFILL }
		},
		{ &hf_gsm0818_bvci,
			{ "BVCI", "gsm0818.bvci",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"BVCI" , HFILL}
		},
		{ &hf_gsm0818_cause,
			{ "Cause", "gsm0818.cause",
			FT_UINT8, BASE_HEX, VALS(cause_vals), 0x0,
			"Cause", HFILL }
		},
		{ &hf_gsm0818_qos_profile,
			{ "QoS Profile", "gsm0818.qos_profile",
			FT_UINT24, BASE_HEX, NULL, 0x0,
			"QoS Profile", HFILL }
		},
		{ &hf_gsm0818_tlli,
			{ "TLLI", "gsm0818.tlli",
			FT_UINT32, BASE_HEX, NULL, 0x0,
			"TLLI", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0818,
		&ett_gsm0818_ie,
	};

	proto_gsm0818 = proto_register_protocol("GSM 08.18 Gb Interface BSS GPRS Protocol", "GSM 08.18", "gsm0818");

	register_dissector("gsm0818", dissect_gsm0818, proto_gsm0818);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0818, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0818(void)
{
    llcgprs_handle = find_dissector("llcgprs");
    gsm0818_handle = find_dissector("gsm0818");
    data_handle = find_dissector("data");
}
