/* packet-gsm0858.c
 * Routines for ip.access generic A-bis dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0858.c 1.9 2008/09/06 22:07:47BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/emem.h>
#include <epan/packet.h>
#include <epan/conversation.h>
#include <epan/prefs.h>
#include "packet-gsm.h"
#include "tap.h"
#include "packet-rtp.h"
#include <epan/rtp_pt.h>
#include "packet-rtcp.h"

static dissector_handle_t gsm0858_handle=NULL;
static dissector_handle_t data_handle=NULL;
static dissector_handle_t rtp_handle=NULL;
static dissector_handle_t rtcp_handle=NULL;
static dissector_handle_t gsm_a_dtap_handle;
static dissector_handle_t gsm_a_ccch_handle;
static dissector_handle_t gsm0408_handle;
static dissector_handle_t gsm0408_l2_handle;
static dissector_handle_t l2_handle;
static dissector_handle_t l3_handle;

static void dissect_gsm0858_ie(guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree);
static void dissect_gsm0858_ie_l3info(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_sacch_info(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_channel_number(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_link_identifier(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_resource_information(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_activation_type(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_bs_power(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_uplink_measurements(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_timing_offset(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_ms_power(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_channel_mode(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_encryption_information(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_cause(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_paging_group(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_handover_reference(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_cb_command_type(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_smscb_message(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_sys_info_type(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_access_delay(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_timing_advance(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_rlm_cause(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_ip_address(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_ip_port(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_ip_speech_mode(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_request_reference(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_frame_number(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_connection_statistics(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_rtp_csd_format(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_rtp_payload_type(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_rtp_jitter_buffer_control(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_rtp_compression_control(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_preprocessing_parameters(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_multiplex_control(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_srtp_configuration(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_ie_multiplex_connection_id(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_measurement_averaging_configure(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_power_control_thresholds(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_handover_thresholds(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_handover_candidates(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_ncell_defaults(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_ncell_list(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_ho_threshold_comparators(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie_pc_threshold_comparators(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0858_eie(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static char *channel_number_string( guint8 chan_no, char *chan_no_str );
void proto_reg_handoff_gsm0858(void);

/* Initialize the protocol and registered fields */
static int proto_gsm0858 = -1;
static int hf_gsm0858_msg_desc = -1;
static int hf_gsm0858_msg_type = -1;
static int hf_gsm0858_chan = -1;
static int hf_gsm0858_chan_no = -1;
static int hf_gsm0858_subchan_no = -1;
static int hf_gsm0858_conn_start_frame = -1;
static int hf_gsm0858_conn_msg_count = -1;
static int hf_gsm0858_conn_state = -1;
static int hf_gsm0858_ie_ei = -1;
static int hf_gsm0858_ie_len8 = -1;
static int hf_gsm0858_ie_len16 = -1;
static int hf_gsm0858_eie_ei = -1;
static int hf_gsm0858_eie_len = -1;
static int hf_gsm0858_sig_chan = -1;
static int hf_gsm0858_sapi = -1;
static int hf_gsm0858_link_id_priority = -1;
static int hf_gsm0858_activation_r = -1;
static int hf_gsm0858_activation_type = -1;
static int hf_gsm0858_bs_power_level = -1;
static int hf_gsm0858_ms_power_level = -1;
static int hf_gsm0858_bs_fpc = -1;
static int hf_gsm0858_ms_fpc = -1;
static int hf_gsm0858_timing_offset = -1;
static int hf_gsm0858_rfu = -1;
static int hf_gsm0858_rxlev_full_up = -1;
static int hf_gsm0858_rxlev_sub_up = -1;
static int hf_gsm0858_rxqual_full_up = -1;
static int hf_gsm0858_rxqual_sub_up = -1;
static int hf_gsm0858_dtxd_meas = -1;
static int hf_gsm0858_dtxd = -1;
static int hf_gsm0858_dtxu = -1;
static int hf_gsm0858_speech_data = -1;
static int hf_gsm0858_rate_type = -1;
static int hf_gsm0858_speech_ver = -1;
static int hf_gsm0858_t_nt = -1;
static int hf_gsm0858_nt_rate = -1;
static int hf_gsm0858_t_rate = -1;
static int hf_gsm0858_encryption_algorithm = -1;
static int hf_gsm0858_encryption_key = -1;
static int hf_gsm0858_rfn = -1;
static int hf_gsm0858_ho_ref = -1;
static int hf_gsm0858_paging_group = -1;
static int hf_gsm0858_paging_buffer_space = -1;
static int hf_gsm0858_access_delay = -1;
static int hf_gsm0858_ra = -1;
static int hf_gsm0858_release_mode = -1;
static int hf_gsm0858_interference_band = -1;
static int hf_gsm0858_rlm_cause = -1;
static int hf_gsm0858_timing_advance = -1;
static int hf_gsm0858_cause = -1;
static int hf_gsm0858_cause_diagnostics = -1;
static int hf_gsm0858_sacch_info_number_of_messages = -1;
static int hf_gsm0858_sys_info_type = -1;
static int hf_gsm0858_sacch_info_length = -1;
static int hf_gsm0858_preprocessed_reporting = -1;
static int hf_gsm0858_smscb_frame = -1;
static int hf_gsm0858_channel_needed = -1;
static int hf_gsm0858_cb_command = -1;
static int hf_gsm0858_cb_default_broadcast = -1;
static int hf_gsm0858_cb_last_block = -1;
static int hf_gsm0858_smscb_message = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0858 = -1;
static gint ett_gsm0858_ie = -1;
static gint ett_gsm0858_eie = -1;
static gint ett_gsm0858_analysis = -1;

/* Global preferences */
static gboolean global_gsm0858_use_ipa = TRUE;

/* Value Strings */
static const value_string rtp_payload_vals[] = {
	{ RTP_PAYLOAD_GSM_FR,   "GSM" },
	{ RTP_PAYLOAD_GSM_EFR,  "GSM-EFR" },
	{ RTP_PAYLOAD_GSM_AMR,  "AMR" },
	{ RTP_PAYLOAD_GSM_CSD,  "GSM-CSD" },
	{ RTP_PAYLOAD_GSM_NANO_GSM_MUX,  "RTP-MUX" },
	{ 0, NULL }
};

#define GSM0858_RTP_MUX_PT	96	/* standard fixed PT for RTP MUX packets */

static const value_string msg_desc_vals[] = {
	{ 0x01,  "RLLM" },
	{ 0x04,  "DCM" },
	{ 0x06,  "CCM" },
	{ 0x08,  "TRXM" },
	{ 0x10,  "LCS" },
	{ 0x3f,  "CMGMT" },
	{ 0x33,  "MCMGMT" },
	{ 0, NULL }
};
#define MSG_DESC_RLLM	1
#define MSG_DESC_DCM	4
#define MSG_DESC_CCM	6
#define MSG_DESC_TRXM	8
#define MSG_DESC_LCS	0x10
#define MSG_DESC_CMGMT	0x3f
#define MSG_DESC_MCMGMT	0x33

static const value_string sig_chan_vals[] = {
	{ 0x00,  "Main (FACCH/SDCCH)" },
	{ 0x01,  "SACCH" },
	{ 0, NULL }
};

static const value_string link_id_priority_vals[] = {
	{ 0x00,  "Normal" },
	{ 0x01,  "High" },
	{ 0x02,  "Low" },
	{ 0, NULL }
};

static const value_string chan_vals[] = {
	{ 0x01,  "Bm" },
	{ 0x02,  "Lm/0" },
	{ 0x03,  "Lm/1" },
	{ 0x04,  "SDCCH4/0" },
	{ 0x05,  "SDCCH4/1" },
	{ 0x06,  "SDCCH4/2" },
	{ 0x07,  "SDCCH4/3" },
	{ 0x08,  "SDCCH8/0" },
	{ 0x09,  "SDCCH8/1" },
	{ 0x0a,  "SDCCH8/2" },
	{ 0x0b,  "SDCCH8/3" },
	{ 0x0c,  "SDCCH8/4" },
	{ 0x0d,  "SDCCH8/5" },
	{ 0x0e,  "SDCCH8/6" },
	{ 0x0f,  "SDCCH8/7" },
	{ 0x10,  "BCCH" },
	{ 0x11,  "RACH" },
	{ 0x12,  "CCCH" },
	{ 0, NULL }
};

#define CHAN_BCCH 0x10
#define CHAN_RACH 0x11
#define CHAN_NULL 0x1f	/* dummy value for Mux connection tracking */

static const value_string msg_type_vals[] = {
	{ 0x01,  "DATA_REQ" },
	{ 0x02,  "DATA_IND" },
	{ 0x03,  "ERROR_IND" },
	{ 0x04,  "EST_REQ" },
	{ 0x05,  "EST_CONF" },
	{ 0x06,  "EST_IND" },
	{ 0x07,  "REL_REQ" },
	{ 0x08,  "REL_CONF" },
	{ 0x09,  "REL_IND" },
	{ 0x0a,  "UNIT_DATA_REQ" },
	{ 0x0b,  "UNIT_DATA_IND" },
	{ 0x11,  "BCCH_INFO" },
	{ 0x12,  "CCCH_LOAD_IND" },
	{ 0x13,  "CHAN_RQD" },
	{ 0x14,  "DELETE_IND" },
	{ 0x15,  "PAGING_CMD" },
	{ 0x16,  "IMM_ASS" },
	{ 0x17,  "SMS_BCAST_REQ" },
	{ 0x19,  "RF_RES_IND" },
	{ 0x1a,  "SACCH_FILLING" },
	{ 0x1b,  "OVERLOAD" },
	{ 0x1c,  "ERROR_REPORT" },
	{ 0x1d,  "SMS_BCAST_CMD" },
	{ 0x1e,  "CBCH_LOAD_IND" },
	{ 0x1f,  "NOTIFIC_CMD" },
	{ 0x21,  "CHAN_ACTIV" },
	{ 0x22,  "CHAN_ACTIV_ACK" },
	{ 0x23,  "CHAN_ACTIV_NACK" },
	{ 0x24,  "CONN_FAIL" },
	{ 0x25,  "DEACT_SACCH" },
	{ 0x26,  "ENCR_CMD" },
	{ 0x27,  "HANDO_DET" },
	{ 0x28,  "MEAS_RES" },
	{ 0x29,  "MODE_MOD_REQ" },
	{ 0x2a,  "MODE_MOD_ACK" },
	{ 0x2b,  "MODE_MOD_NACK" },
	{ 0x2c,  "PHY_CONTEXT_REQ" },
	{ 0x2d,  "PHY_CONTEXT_CONF" },
	{ 0x2e,  "RF_CHAN_REL" },
	{ 0x2f,  "MS_POWER_CONTROL" },
	{ 0x30,  "BS_POWER_CONTROL" },
	{ 0x31,  "PREPROC_CONFIG" },
	{ 0x32,  "PREPROC_MEAS_RES" },
	{ 0x33,  "RF_CHAN_REL_ACK" },
	{ 0x34,  "SACCH_INFO_MODIFY" },
	{ 0x35,  "TALKER_DET" },
	{ 0x36,  "LISTENER_DET" },
	{ 0x37,  "REMOTE_CODEC_CONF_REP" },
	{ 0x38,  "ROUND_TRIP_DELAY_REP" },
	{ 0x39,  "PRE_HANDOVER_NOTIF" },
	{ 0x3A,  "MULTIRATE_CODEC_MOD_REQ" },
	{ 0x3B,  "MULTIRATE_CODEC_MOD_ACK" },
	{ 0x3C,  "MULTIRATE_CODEC_MOD_NACK" },
	{ 0x3D,  "MULTIRATE_CODEC_MOD_PERF" },
	{ 0x3E,  "TFO_REPORT" },
	{ 0x3F,  "TFO_MOD_REQ" },
	{ 0x40,  "DIRECTED_RETRY_ENQ" },
	{ 0x41,  "LOCATION_INFO" },
	{ 0x48,  "PDCH_ACTIVATION" },
	{ 0x49,  "PDCH_ACTIVATION_ACK" },
	{ 0x4a,  "PDCH_ACTIVATION_NACK" },
	{ 0x4b,  "PDCH_DEACTIVATION" },
	{ 0x4c,  "PDCH_DEACTIVATION_ACK" },
	{ 0x4d,  "PDCH_DEACTIVATION_NACK" },
	{ 0x50,  "MD_MUX_CX" },
	{ 0x51,  "MD_MUX_CX_ACK" },
	{ 0x52,  "MD_MUX_CX_NACK" },
	{ 0x53,  "CR_MUX_CX" },
	{ 0x54,  "CR_MUX_CX_ACK" },
	{ 0x55,  "CR_MUX_CX_NACK" },
	{ 0x56,  "DL_MUX_CX" },
	{ 0x57,  "DL_MUX_CX_ACK" },
	{ 0x58,  "DL_MUX_CX_NACK" },
	{ 0x60,  "MEAS_PREPROC_DFLT" },
	{ 0x61,  "HANDO_CAND_ENQ" },
	{ 0x62,  "HANDO_CAND_RESP" },
	{ 0x70,  "CRCX" },
	{ 0x71,  "CRCX_ACK" },
	{ 0x72,  "CRCX_NACK" },
	{ 0x73,  "MDCX" },
	{ 0x74,  "MDCX_ACK" },
	{ 0x75,  "MDCX_NACK" },
	{ 0x76,  "DLCX_IND" },
	{ 0x77,  "DLCX" },
	{ 0x78,  "DLCX_ACK" },
	{ 0x79,  "DLCX_NACK" },
	{ 0, NULL }
};
/* for special L2 Pseudo Length formats */
#define BCCH_INFO			0x11
#define SACCH_FILLING		0x1a
#define IMMEDIATE_ASSIGN	0x16
#define SACCH_INFO_MODIFY	0x34

static const value_string ie_ei_vals[] = {
	{ 0x01,  "Channel Number" },
	{ 0x02,  "Link Identifier" },
	{ 0x03,  "Activation Type" },
	{ 0x04,  "BS Power" },
	{ 0x05,  "Channel Identification" },
	{ 0x06,  "Channel Mode" },
	{ 0x07,  "Encryption Information" },
	{ 0x08,  "Frame Number" },
	{ 0x09,  "Handover Reference" },
	{ 0x0a,  "L1 Information" },
	{ 0x0b,  "L3 Information" },
	{ 0x0c,  "MS Identity" },
	{ 0x0d,  "MS Power" },
	{ 0x0e,  "Paging Group" },
	{ 0x0f,  "Paging Load" },
	{ 0x10,  "Physical Context" },
	{ 0x11,  "Access Delay" },
	{ 0x12,  "RACH Load" },
	{ 0x13,  "Request Reference" },
	{ 0x14,  "Release Mode" },
	{ 0x15,  "Resource Information" },
	{ 0x16,  "RLM Cause" },
	{ 0x17,  "Starting Time" },
	{ 0x18,  "Timing Advance" },
	{ 0x19,  "Uplink Measurements" },
	{ 0x1a,  "Cause" },
	{ 0x1b,  "Measurement Result Number" },
	{ 0x1c,  "Message Identifier" },
	{ 0x1d,  "Message Indicator (Ph 1)" },
	{ 0x1e,  "System Info Type" },
	{ 0x1f,  "MS Power Parameters" },
	{ 0x20,  "BS Power Parameters" },
	{ 0x21,  "Pre-processing Parameters" },
	{ 0x22,  "Pre-processed Measurements" },
	{ 0x23,  "Immediate Assign Info (Ph 1)" },
	{ 0x24,  "SMSCB Information" },
	{ 0x25,  "MS Timing Offset" },
	{ 0x26,  "Erroneous Message" },
	{ 0x27,  "Full BCCH Information" },
	{ 0x28,  "Channel Needed" },
	{ 0x29,  "CB Command Type" },
	{ 0x2a,  "SMSCB Message" },
	{ 0x2b,  "Full Immediate Assign Info" },
	{ 0x2c,  "SACCH Information" },
	{ 0x2d,  "CBCH Load Indication" },
	{ 0x2e,  "SMSCB Channel Indicator" },
	{ 0x2f,  "Group Call Reference" },
	{ 0x30,  "Channel Description" },
	{ 0x31,  "NCH DRX Information" },
	{ 0x32,  "Command Indicator" },
	{ 0x33,  "eMLPP Priority" },
	{ 0x34,  "UIC" },
	{ 0x35,  "Main Channel Reference" },
	{ 0x36,  "MultiRate Configuration" },
	{ 0x37,  "MultiRate Control" },
	{ 0x38,  "Supported Codec Types" },
	{ 0x39,  "Codec Configuration" },
	{ 0x3A,  "Round Trip Delay" },
	{ 0x3B,  "TFO Status" },
	{ 0x3C,  "LLP APDU" },
	{ 0xe0,  "SRTP Configuration" },
	{ 0xe1,  "BSC Proxy UDP Port" },
	{ 0xe2,  "BSC Multiplexor Keep Alive Timeout" },
	{ 0xf0,  "Destination IP Address" },
	{ 0xf1,  "Destination IP Port" },
	{ 0xf2,  "RTP Payload Type" },
	{ 0xf3,  "Source IP Port" },
	{ 0xf4,  "IP Speech Mode" },
	{ 0xf5,  "Source IP Address" },
	{ 0xf6,  "Connection Statistics" },
	{ 0xf7,  "Handover Candidate Parameters" },
	{ 0xf8,  "Connection Id" },
	{ 0xf9,  "RTP CSD Format" },
	{ 0xfa,  "RTP Jitter Buffer Control" },
	{ 0xfb,  "RTP Compression Control" },
	{ 0xfc,  "RTP Payload Type 2" },
	{ 0xfd,  "Multiplex Control" },
	{ 0xfe,  "Multiplex Connection Id" },
	{ 0, NULL }
};

static const value_string release_mode_vals[] = {
	{ 0x00,  "Normal" },
	{ 0x01,  "Local End" },
	{ 0, NULL }
};

static const value_string activation_type_vals[] = {
	{ 0x00,  "Related to intra-cell channel change - immediate assignment" },
	{ 0x01,  "Related to intra-cell channel change - normal assignment" },
	{ 0x02,  "Related to inter-cell channel change (asynchronous handover)" },
	{ 0x03,  "Related to inter-cell channel change (synchronous handover)" },
	{ 0x04,  "Related to secondary channels - additional assignment" },
	{ 0x05,  "Related to secondary channels - multislot configuration" },
	{ 0, NULL }
};

static const value_string activation_r_vals[] = {
	{ 0x00,  "Initial Activation" },
	{ 0x01,  "Reactivation" },
	{ 0, NULL }
};

static const value_string bs_power_vals[] = {
	{ 0x00,  "Pn" },
	{ 0x01,  "Pn - 2dB" },
	{ 0x02,  "Pn - 4dB" },
	{ 0x03,  "Pn - 6dB" },
	{ 0x04,  "Pn - 8dB" },
	{ 0x05,  "Pn - 10dB" },
	{ 0x06,  "Pn - 12dB" },
	{ 0x07,  "Pn - 14dB" },
	{ 0x08,  "Pn - 16dB" },
	{ 0x09,  "Pn - 18dB" },
	{ 0x0a,  "Pn - 20dB" },
	{ 0x0b,  "Pn - 22dB" },
	{ 0x0c,  "Pn - 24dB" },
	{ 0x0d,  "Pn - 26dB" },
	{ 0x0e,  "Pn - 28dB" },
	{ 0x0f,  "Pn - 30dB" },
	{ 0, NULL }
};

static const value_string ms_power_vals[] = {
	{ 0x00,  "30dBm (18/1900MHz) / 39dBm (900MHz)" },
	{ 0x01,  "28dBm (18/1900MHz) / 39dBm (900MHz)" },
	{ 0x02,  "26dBm (18/1900MHz) / 39dBm (900MHz)" },
	{ 0x03,  "24dBm (18/1900MHz) / 37dBm (900MHz)" },
	{ 0x04,  "22dBm (18/1900MHz) / 35dBm (900MHz)" },
	{ 0x05,  "20dBm (18/1900MHz) / 33dBm (900MHz)" },
	{ 0x06,  "18dBm (18/1900MHz) / 31dBm (900MHz)" },
	{ 0x07,  "16dBm (18/1900MHz) / 29dBm (900MHz)" },
	{ 0x08,  "14dBm (18/1900MHz) / 27dBm (900MHz)" },
	{ 0x09,  "12dBm (18/1900MHz) / 25dBm (900MHz)" },
	{ 0x0a,  "10dBm (18/1900MHz) / 23dBm (900MHz)" },
	{ 0x0b,  "8dBm (18/1900MHz) / 21dBm (900MHz)" },
	{ 0x0c,  "6dBm (18/1900MHz) / 19dBm (900MHz)" },
	{ 0x0d,  "4dBm (18/1900MHz) / 17dBm (900MHz)" },
	{ 0x0e,  "2dBm (18/1900MHz) / 15dBm (900MHz)" },
	{ 0x0f,  "0dBm (18/1900MHz) / 13dBm (900MHz)" },
	{ 0x10,  "0dBm (1800MHz) / 11dBm (900MHz)" },
	{ 0x11,  "0dBm (1800MHz) / 9dBm (900MHz)" },
	{ 0x12,  "0dBm (1800MHz) / 7dBm (900MHz)" },
	{ 0x13,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x14,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x15,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x16,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x17,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x18,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x19,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x1a,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x1b,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x1c,  "0dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x1d,  "36dBm (1800MHz) / 5dBm (900MHz)" },
	{ 0x1e,  "34dBm (1800MHz) / 33dBm (1900MHz) / 5dBm (900MHz)" },
	{ 0x1f,  "32dBm (18/1900MHz) / 5dBm (900MHz)" },
	{ 0, NULL }
};

static const value_string fpc_vals[] = {
	{ 0x00,  "Fast Power Control not in use" },
	{ 0x01,  "Fast Power Control in use" },
	{ 0, NULL }
};

static const value_string smscb_channel_indicator_vals[] = {
	{ 0x00,  "Basic CBCH" },
	{ 0x01,  "Extended CBCH" },
	{ 0, NULL }
};

static const value_string speech_data_vals[] = {
	{ 0x01,  "Speech" },
	{ 0x02,  "Data" },
	{ 0x03,  "Signalling" },
	{ 0, NULL }
};
#define CHANNEL_MODE_SPEECH			0x01
#define CHANNEL_MODE_DATA			0x02
#define CHANNEL_MODE_SIGNALLING		0x03

static const value_string rate_type_vals[] = {
	{ 0x01,  "SDCCH" },
	{ 0x08,  "Full rate TCH channel Bm" },
	{ 0x09,  "Half rate TCH channel Lm" },
	{ 0x0a,  "Full rate TCH channel bi-directional Bm, multislot configuration" },
	{ 0x1a,  "Full rate TCH channel uni-directional downlink Bm, multislot configuration" },
	{ 0x18,  "Full rate TCH channel Bm Group call channel" },
	{ 0x19,  "Half rate TCH channel Lm Group call channel" },
	{ 0x28,  "Full rate TCH channel Bm Broadcast call channel" },
	{ 0x29,  "Half rate TCH channel Lm Broadcast call channel" },
	{ 0, NULL }
};

static const value_string speech_ver_vals[] = {
	{ 0x00,  "No resources required" },
	{ 0x01,  "GSM speech coding algorithm version 1" },
	{ 0x11,  "GSM speech coding algorithm version 2" },
	{ 0x21,  "GSM speech coding algorithm version 3" },
	{ 0, NULL }
};

static const value_string t_nt_vals[] = {
	{ 0x00,  "Transparent" },
	{ 0x01,  "Non-Transparent" },
	{ 0, NULL }
};

static const value_string nt_rate_vals[] = {
	{ 0x21,  "asymmetric 43.5 kbit/s (downlink) + 14.5 kbit/s (uplink)" },
	{ 0x22,  "asymmetric 29.0 kbit/s (downlink) + 14.5 kbit/s (uplink)" },
	{ 0x23,  "asymmetric 43.5 kbit/s (downlink) + 29.0 kbit/s (uplink)" },
	{ 0x29,  "asymmetric 14.5 kbit/s (downlink) + 43.5 kbit/s (uplink)" },
	{ 0x2a,  "asymmetric 14.5 kbit/s (downlink) + 29.0 kbit/s (uplink)" },
	{ 0x2b,  "asymmetric 29.0 kbit/s (downlink) + 43.5 kbit/s (uplink)" },
	{ 0x34,  "43.5 kbit/s" },
	{ 0x31,  "28.8 kbit/s" },
	{ 0x18,  "14.5 kbit/s" },
	{ 0x10,  "12 kbit/s" },
	{ 0x11,  "6 kbit/s" },
	{ 0, NULL }
};

static const value_string t_rate_vals[] = {
	{ 0x38,  "32 kbit/s" },
	{ 0x39,  "29 kbit/s" },
	{ 0x18,  "14.4 kbit/s" },
	{ 0x10,  "9.6 kbit/s" },
	{ 0x11,  "4.8 kbit/s" },
	{ 0x12,  "2.4 kbit/s" },
	{ 0x13,  "1.2 kbit/s" },
	{ 0x14,  "600 bit/s" },
	{ 0x15,  "1200/75 bit/s" },
	{ 0, NULL }
};

static const value_string command_indicator_vals[] = {
	{ 0x00,  "Start" },
	{ 0x01,  "Stop" },
	{ 0, NULL }
};

static const value_string ip_speech_mode_s_vals[] = {
	{ 0x00,  "GSM FR codec (GSM type 1, FS)" },
	{ 0x01,  "GSM EFR codec (GSM type 2, FS)" },
	{ 0x02,  "GSM AMR/FR codec (GSM type 3, FS)" },
	{ 0x03,  "GSM HR codec (GSM type 1, HS)" },
	{ 0x05,  "GSM AMR/HR codec (GSM type 3,HS)" },
	{ 0x0f,  "As specified by RTP Payload Type IE" },
	{ 0, NULL }
};

static const value_string ip_speech_mode_m_vals[] = {
	{ 0x00,  "Send and Receive" },
	{ 0x01,  "Receive Only" },
	{ 0x02,  "Send Only" },
	{ 0, NULL }
};

static const value_string rtp_csd_format_d_vals[] = {
	{ 0x00,  "External TRAU format" },
	{ 0x01,  "Non-TRAU Packed format" },
	{ 0x02,  "TRAU within the BTS" },
	{ 0x03,  "IWF-Free BTS-BTS data" },
	{ 0, NULL }
};

static const value_string rtp_csd_format_ir_vals[] = {
	{ 0x00,  "8kb/s" },
	{ 0x01,  "16kb/s" },
	{ 0x02,  "32kb/s" },
	{ 0x03,  "64kb/s" },
	{ 0, NULL }
};

static const value_string rtp_payload_type_vals[] = {
	{ 0x03,  "GSM FR" },
	{ 0x54,  "Dynamic" },
	{ 0, NULL }
};

static const value_string measurement_averaging_configure_param_id_vals[] = {
	{ 0x00,  "RxLev" },
	{ 0x01,  "RxQual" },
	{ 0x02,  "MS-BTS Distance" },
	{ 0, NULL }
};

static const value_string measurement_averaging_configure_method_vals[] = {
	{ 0x00,  "Unweighted Average" },
	{ 0x01,  "Weighted Average" },
	{ 0x02,  "Modified Median" },
	{ 0, NULL }
};

static const value_string rtpa_vals[] = {
	{ 0x00,  "RTP Authentication Disabled" },
	{ 0x01,  "RTP Authentication Disabled" },
	{ 0, NULL }
};

static const value_string channel_needed_vals[] = {
	{ 0x00,  "Any Channel" },
	{ 0x01,  "SDCCH" },
	{ 0x02,  "TCH/F (Full Rate)" },
	{ 0x03,  "TCH/F or TCH/H (Dual rate)" },
	{ 0, NULL }
};

static const value_string encryption_algorithm_vals[] = {
	{ 0x01,  "No encryption shall be used" },
	{ 0x02,  "GSM A5/1" },
	{ 0x03,  "GSM A5/2" },
	{ 0x04,  "GSM A5/3" },
	{ 0x05,  "GSM A5/4" },
	{ 0x06,  "GSM A5/5" },
	{ 0x07,  "GSM A5/6" },
	{ 0x08,  "GSM A5/7" },
	{ 0, NULL }
};

static const value_string rlm_cause_vals[] = {
	{ 0x01,  "T200 expired (N200+1) times" },
	{ 0x02,  "Re-establishment request" },
	{ 0x03,  "Unsolicited UA response" },
	{ 0x04,  "Unsolicited DM response" },
	{ 0x05,  "Unsolicated DM response, multiple frame established state" },
	{ 0x06,  "Unsolicited supervisory response" },
	{ 0x07,  "Sequence error" },
	{ 0x08,  "U-frame with incorrect parameters" },
	{ 0x09,  "S-frame with incorrect parameters" },
	{ 0x0a,  "I-frame with incorrect use of M bit" },
	{ 0x0b,  "I-frame with incorrect length" },
	{ 0x0c,  "Frame not implemented" },
	{ 0x0d,  "SABM command, multiple frame established state" },
	{ 0x0e,  "SABM frame with information not allowed in this state" },
	{ 0, NULL }
};

static const value_string cause_vals[] = {
	{ 0x00,  "radio interface failure" },
	{ 0x01,  "radio link failure" },
	{ 0x02,  "handover access failure" },
	{ 0x03,  "talker access failure" },
	{ 0x07,  "O&M intervention" },
	{ 0x0f,  "normal event, unspecified" },
	{ 0x20,  "equipment failure" },
	{ 0x21,  "radio resource not available" },
	{ 0x22,  "terrestrial channel failure" },
	{ 0x23,  "CCCH overload" },
	{ 0x24,  "ACCH overload" },
	{ 0x25,  "processor overload" },
	{ 0x27,  "BTS not equipped" },
	{ 0x28,  "remote transcoder failure" },
	{ 0x29,  "notification overflow" },
	{ 0x2f,  "resource not available, unspecified" },
	{ 0x30,  "requested transcoding/rate adaption not available" },
	{ 0x3f,  "service or option not available, unspecified" },
	{ 0x40,  "encryption algorithm not implemented" },
	{ 0x4f,  "service or option not implemented, unspecified" },
	{ 0x50,  "radio channel already activated/allocated" },
	{ 0x51,  "Radio Channel not activated/allocated" },
	{ 0x52,  "Connection invalid" },
	{ 0x5f,  "invalid message, unspecified" },
	{ 0x60,  "message discriminator error" },
	{ 0x61,  "message type error" },
	{ 0x62,  "message sequence error" },
	{ 0x63,  "general information element error" },
	{ 0x64,  "mandatory information element error" },
	{ 0x65,  "optional information element error" },
	{ 0x66,  "information element non-existent" },
	{ 0x67,  "information element length error" },
	{ 0x68,  "invalid information element contents" },
	{ 0x6f,  "protocol error, unspecified" },
	{ 0x7f,  "interworking, unspecified" },
	{ 0, NULL }
};

static const value_string sys_info_type_vals[] = {
	{ 0x00,  "System Information 8" },
	{ 0x01,  "System Information 1" },
	{ 0x02,  "System Information 2" },
	{ 0x03,  "System Information 3" },
	{ 0x04,  "System Information 4" },
	{ 0x05,  "System Information 5" },
	{ 0x06,  "System Information 6" },
	{ 0x07,  "System Information 7" },
	{ 0x08,  "System Information 16" },
	{ 0x09,  "System Information 17" },
	{ 0x0a,  "System Information 2bis" },
	{ 0x0b,  "System Information 2ter" },
	{ 0x0d,  "System Information 5bis" },
	{ 0x0e,  "System Information 5ter" },
	{ 0x0f,  "System Information 10" },
	{ 0x47,  "Extended Measurement Order " },
	{ 0x48,  "Measurement Information " },
	{ 0x28,  "System Information 13" },
	{ 0x29,  "System Information 2quater" },
	{ 0x2a,  "System Information 9" },
	{ 0x2b,  "System Information 18" },
	{ 0x2c,  "System Information 19" },
	{ 0x2d,  "System Information 20" },
	{ 0, NULL }
};

static const ie_def_t ie_defs[] = {
	{ 0x01,  "Channel Number", 1, &dissect_gsm0858_ie_channel_number, NULL },
	{ 0x02,  "Link Identifier", 1, &dissect_gsm0858_ie_link_identifier, NULL },
	{ 0x03,  "Activation Type", 1, &dissect_gsm0858_ie_activation_type, NULL },
	{ 0x04,  "BS Power", 1, &dissect_gsm0858_ie_bs_power, NULL },
	{ 0x05,  "Channel Identification", -1, NULL, NULL },
	{ 0x06,  "Channel Mode", -1, &dissect_gsm0858_ie_channel_mode, NULL },
	{ 0x07,  "Encryption Information", -1, &dissect_gsm0858_ie_encryption_information, NULL },
	{ 0x08,  "Frame Number", 2, &dissect_gsm0858_ie_frame_number, NULL },
	{ 0x09,  "Handover Reference", 1, &dissect_gsm0858_ie_handover_reference, NULL },
	{ 0x0a,  "L1 Information", 2, NULL, NULL },
	{ 0x0b,  "L3 Information", -2, &dissect_gsm0858_ie_l3info, NULL },
	{ 0x0c,  "MS Identity", -1, NULL, NULL },
	{ 0x0d,  "MS Power", 1, &dissect_gsm0858_ie_ms_power, NULL },
	{ 0x0e,  "Paging Group", 1, &dissect_gsm0858_ie_paging_group, NULL },
	{ 0x0f,  "Paging Load", 2, NULL, NULL },
	{ 0x10,  "Physical Context", -1, &dissect_gsm0858_eie, NULL },
	{ 0x11,  "Access Delay", 1, &dissect_gsm0858_ie_access_delay, NULL },
	{ 0x12,  "RACH Load", -1, NULL, NULL },
	{ 0x13,  "Request Reference", 3, &dissect_gsm0858_ie_request_reference, NULL },
	{ 0x14,  "Release Mode", 1, NULL, (void *)release_mode_vals },
	{ 0x15,  "Resource Information", -1, &dissect_gsm0858_ie_resource_information, NULL },
	{ 0x16,  "RLM Cause", -1, &dissect_gsm0858_ie_cause, NULL },
	{ 0x17,  "Starting Time", 2, &dissect_gsm0858_ie_frame_number, NULL },
	{ 0x18,  "Timing Advance", 1, &dissect_gsm0858_ie_timing_advance, NULL },
	{ 0x19,  "Uplink Measurements", -1, &dissect_gsm0858_ie_uplink_measurements, NULL },
	{ 0x1a,  "Cause", -1, &dissect_gsm0858_ie_cause, NULL },
	{ 0x1b,  "Measurement Result Number", 1, NULL, NULL },
	{ 0x1c,  "Message Identifier", 1, NULL, NULL },
	{ 0x1d,  "Message Indicator (Ph 1)", -3, NULL, NULL },	/* Phase 1 IE of zero length - rest of message contains a complete 08.58 message!  */
	{ 0x1e,  "System Info Type", 1, &dissect_gsm0858_ie_sys_info_type, NULL },
	{ 0x1f,  "MS Power Parameters", -1, &dissect_gsm0858_eie, NULL },
	{ 0x20,  "BS Power Parameters", -1, &dissect_gsm0858_eie, NULL },
	{ 0x21,  "Pre-processing Parameters", -1, &dissect_gsm0858_ie_preprocessing_parameters, NULL },
	{ 0x22,  "Pre-processed Measurements", -1, &dissect_gsm0858_eie, NULL },
	{ 0x23,  "Immediate Assign Info (Ph 1)", -1, NULL, NULL },
	{ 0x24,  "SMSCB Information", 23, NULL, NULL },
	{ 0x25,  "MS Timing Offset", 1, &dissect_gsm0858_ie_timing_offset, NULL },
	{ 0x26,  "Erroneous Message", -1, NULL, NULL },
	{ 0x27,  "Full BCCH Information", -1, &dissect_gsm0858_ie_l3info, NULL },
	{ 0x28,  "Channel Needed", 1, NULL, (void *)channel_needed_vals },
	{ 0x29,  "CB Command Type", 1, &dissect_gsm0858_ie_cb_command_type, NULL },
	{ 0x2a,  "SMSCB Message", -1, &dissect_gsm0858_ie_smscb_message, NULL },
	{ 0x2b,  "Full Immediate Assign Info", -1, &dissect_gsm0858_ie_l3info, NULL },
	{ 0x2c,  "SACCH Information", -1, &dissect_gsm0858_ie_sacch_info, NULL },
	{ 0x2d,  "CBCH Load Indication", 1, NULL, NULL },
	{ 0x2e,  "SMSCB Channel Indicator", 1, NULL, (void *)smscb_channel_indicator_vals },
	{ 0x2f,  "Group Call Reference", -1, NULL, NULL },
	{ 0x30,  "Channel Description", -1, NULL, NULL },
	{ 0x31,  "NCH DRX Information", -1, NULL, NULL },
	{ 0x32,  "Command Indicator", -1, NULL, (void *)command_indicator_vals },
	{ 0x33,  "eMLPP Priority", 1, NULL, NULL },
	{ 0x34,  "UIC", -1, NULL, NULL },
	{ 0x35,  "Main Channel Reference", 1, NULL, NULL },
	{ 0x36,  "MultiRate Configuration", -1, NULL, NULL },
	{ 0x37,  "MultiRate Control", 1, NULL, NULL },
	{ 0x38,  "Supported Codec Types", -1, NULL, NULL },
	{ 0x39,  "Codec Configuration", -1, NULL, NULL },
	{ 0x3A,  "Round Trip Delay", 1, NULL, NULL },
	{ 0x3B,  "TFO Status", 1, NULL, NULL },
	{ 0x3C,  "LLP APDU", -1, NULL, NULL },
	{ 0xe0,  "SRTP Configuration", -1, &dissect_gsm0858_ie_srtp_configuration, NULL },
	{ 0xe1,  "BSC Proxy UDP Port", 2, &dissect_gsm0858_ie_ip_port, NULL },
	{ 0xe2,  "BSC Multiplexor Keep Alive Timeout", 1, NULL, NULL },
	{ 0xf0,  "Destination IP Address", 4, &dissect_gsm0858_ie_ip_address, NULL },
	{ 0xf1,  "Destination IP Port", 2, &dissect_gsm0858_ie_ip_port, NULL },
	{ 0xf2,  "RTP Payload Type", 1, &dissect_gsm0858_ie_rtp_payload_type, NULL },
	{ 0xf3,  "Source IP Port", 2, &dissect_gsm0858_ie_ip_port, NULL },
	{ 0xf4,  "IP Speech Mode", 1, &dissect_gsm0858_ie_ip_speech_mode, NULL },
	{ 0xf5,  "Source IP Address", 4, &dissect_gsm0858_ie_ip_address, NULL },
	{ 0xf6,  "Connection Statistics", -1, &dissect_gsm0858_ie_connection_statistics, NULL },
	{ 0xf7,  "Handover Candidate Parameters", -1, NULL, NULL },
	{ 0xf8,  "Connection Id", 2, NULL, NULL },
	{ 0xf9,  "RTP CSD Format", 1, &dissect_gsm0858_ie_rtp_csd_format, NULL },
	{ 0xfa,  "RTP Jitter Buffer Control", 2, &dissect_gsm0858_ie_rtp_jitter_buffer_control, NULL },
	{ 0xfb,  "RTP Compression Control", 1, dissect_gsm0858_ie_rtp_compression_control, NULL },
	{ 0xfc,  "RTP Payload Type 2", 1, &dissect_gsm0858_ie_rtp_payload_type, NULL },
	{ 0xfd,  "Multiplex Control", 8, &dissect_gsm0858_ie_multiplex_control, NULL },
	{ 0xfe,  "Multiplex Connection Id", 1, &dissect_gsm0858_ie_multiplex_connection_id, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

static const value_string handover_required_cause_vals[] = {
	{ 0x01,  "Lower RXLEV Uplink Handover" },
	{ 0x02,  "Lower RXLEV Downlink Handover" },
	{ 0x03,  "Lower RXQUAL Uplink Handover" },
	{ 0x04,  "Lower RXQUAL Downlink Handover" },
	{ 0x05,  "RXLEV Uplink Intra-cell Handover" },
	{ 0x06,  "RXLEV Downlink Intra-cell Handover" },
	{ 0x07,  "Max MS Range" },
	{ 0x08,  "Power Budget" },
	{ 0x09,  "Enquiry" },
	{ 0x0a,  "Enquiry Failed" },
	{ 0, NULL }
};

static const value_string cb_command_vals[] = {
	{ 0x00,  "Normal message" },
	{ 0x08,  "Schedule message" },
	{ 0x0e,  "Default message" },
	{ 0x0f,  "Null message" },
	{ 0, NULL }
};

static const value_string cb_default_broadcast_vals[] = {
	{ 0x00,  "Normal message" },
	{ 0x01,  "Null message" },
	{ 0, NULL }
};

static const value_string cb_last_block_vals[] = {
	{ 0x00,  "Block 4" },
	{ 0x01,  "Block 1" },
	{ 0x02,  "Block 2" },
	{ 0x03,  "Block 3" },
	{ 0, NULL }
};

static const value_string eie_ei_vals[] = {
	{ 0x00,  "RXLEV" },
	{ 0x01,  "RXQUAL" },
	{ 0x02,  "Frequency Error" },
	{ 0x03,  "Frame Timing Error" },
	{ 0x04,  "Measurement Averaging Configure" },
	{ 0x05,  "BS Power Control Threshold" },
	{ 0x06,  "MS Power Control Threshold" },
	{ 0x07,  "Handover Threshold" },
	{ 0x08,  "NCell Limits Defaults" },
	{ 0x09,  "NCell List" },
	{ 0x0a,  "PC Threshold Comparators" },
	{ 0x0b,  "HO Threshold Comparators" },
	{ 0x0c,  "Handover Cause" },
	{ 0x0d,  "Handover Candidates" },
	{ 0x0e,  "NCell BA-list Change List" },
	{ 0x10,  "Number of MSs" },
	{ 0x11,  "Handover Candidates Extension" },
	{ 0x12,  "NCell Limits Defaults Extension" },
	{ 0x13,  "NCell List Extension" },
	{ 0x14,  "Master Key" },
	{ 0x15,  "Master Salt" },
	{ 0, NULL }
};

static const ie_def_t eie_defs[] = {
	{ 0x00,  "RXLEV", 1, NULL, NULL },
	{ 0x01,  "RXQUAL", 1, NULL, NULL },
	{ 0x02,  "Frequency Error", 2, NULL, NULL },
	{ 0x03,  "Frame Timing Error", 1, NULL, NULL },
	{ 0x04,  "Measurement Averaging Configure", -1, &dissect_gsm0858_eie_measurement_averaging_configure, NULL },
	{ 0x05,  "BS Power Control Threshold", 3, &dissect_gsm0858_eie_power_control_thresholds, NULL },
	{ 0x06,  "MS Power Control Threshold", 3, &dissect_gsm0858_eie_power_control_thresholds, NULL },
	{ 0x07,  "Handover Threshold", 6, &dissect_gsm0858_eie_handover_thresholds, NULL },
	{ 0x08,  "NCell Limits Defaults", 3, &dissect_gsm0858_eie_ncell_defaults, NULL },
	{ 0x09,  "NCell List", -1, &dissect_gsm0858_eie_ncell_list, NULL },
	{ 0x0a,  "PC Threshold Comparators", 10, &dissect_gsm0858_eie_pc_threshold_comparators, NULL },
	{ 0x0b,  "HO Threshold Comparators", 10, &dissect_gsm0858_eie_ho_threshold_comparators, NULL },
	{ 0x0c,  "Handover Cause", -1, NULL, (void *)handover_required_cause_vals },
	{ 0x0d,  "Handover Candidates", -1, &dissect_gsm0858_eie_handover_candidates, NULL },
	{ 0x0e,  "NCell BA-list Change List", -1, NULL, NULL },
	{ 0x10,  "Number of MSs", 1, NULL, NULL },
	{ 0x11,  "Handover Candidates Extension", -1, NULL, NULL },
	{ 0x12,  "NCell Limits Defaults Extension", -1, NULL, NULL },
	{ 0x13,  "NCell List Extension", -1, NULL, NULL },
	{ 0x14,  "Master Key", -1, NULL, NULL },
	{ 0x15,  "Master Salt", -1, NULL, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

static GMemChunk *chan_analysis_chunk = NULL;
static GMemChunk *mux_analysis_chunk = NULL;
static GMemChunk *rach_keys_chunk = NULL;
static GMemChunk *rach_values_chunk = NULL;
static GHashTable *rach_hash_table=NULL;

#define ACTIVATIONS_MAX 10000			/* one for each Channel activation conversation - lots! */
#define RACH_MAX 1000000					/* one for each RACH conversation - lots! */
#define MUX_MAX 1000					/* one for each RSL conversation - lots! */

typedef struct chan_analysis {
	guint32	start_frame;
	guint32 end_frame;
	int msg_count;
	int state;
	GHashTable *rtp_dyn_payload;
	address rtp_src_addr;
	address rtp_dst_addr;
	guint8  rtp_src_addr_data[4];
	guint8  rtp_dst_addr_data[4];
	guint16 rtp_src_port;
	guint16 rtp_dst_port;
	guint16 rtp_payload;
	guint16 rtp_pt;
	guint8  rtp_mux_used;
} chan_analysis;

typedef struct mux_analysis {
	guint32	start_frame;
	guint32 end_frame;
	GHashTable *rtp_dyn_payload;
	address rtp_addr;
	guint8  rtp_addr_data[4];
	guint16 rtp_port;
	guint16 rtp_pt;
	struct  srtp_info *srtp_info;
} mux_analysis;

typedef struct rach_value {
	guint32	rach_frame;
	nstime_t rach_ts;
	gint32 delay_ms;
} rach_value;

typedef struct rach_key {
	gint32 secs;
	guint32 conv;
	guint16 fn;
	guint8	ra;
} rach_key;

static chan_analysis *ca;
static mux_analysis *ma;

#define FN_WRAP_TIME		((42432*4615)/1000000)
#define HALF_FN_WRAP_TIME	(FN_WRAP_TIME/2)

/* Hash Functions */
static gint rach_equal(gconstpointer v, gconstpointer w)
{
  const struct rach_key *v1 = (const struct rach_key *)v;
  const struct rach_key *v2 = (const struct rach_key *)w;

  if ((v1->fn == v2->fn)
	  && (v1->ra == v2->ra)
	  && (v1->conv == v2->conv)
	  && (abs(v1->secs-v2->secs)<HALF_FN_WRAP_TIME))
	return 1;

  return 0;
}

static guint rach_hash(gconstpointer v)
{
  const struct rach_key *key = (const struct rach_key *)v;
  guint val;

  val = (key->fn)<<8 | (key->ra);

  return val;
}

static void analyze_init(void)
{
	ca = NULL;

	/*
	 * Now destroy the chunk from which the conversation table
	 * structures were allocated.
	 */
	if (chan_analysis_chunk) {
		g_mem_chunk_destroy(chan_analysis_chunk);
		chan_analysis_chunk = NULL;
	}
	if (mux_analysis_chunk) {
		g_mem_chunk_destroy(mux_analysis_chunk);
		mux_analysis_chunk = NULL;
	}
	if (rach_keys_chunk) {
		g_mem_chunk_destroy(rach_keys_chunk);
		rach_keys_chunk = NULL;
	}
	if (rach_values_chunk) {
		g_mem_chunk_destroy(rach_values_chunk);
		rach_values_chunk = NULL;
	}
	if (rach_hash_table) {
		g_hash_table_destroy(rach_hash_table);
		rach_hash_table = NULL;
	}

	/* and create new chunks */
	chan_analysis_chunk = g_mem_chunk_new("chan_analysis_chunk",
			sizeof(struct chan_analysis),
			ACTIVATIONS_MAX * sizeof(struct chan_analysis),
			G_ALLOC_AND_FREE);
	mux_analysis_chunk = g_mem_chunk_new("mux_analysis_chunk",
			sizeof(struct mux_analysis),
			MUX_MAX * sizeof(struct mux_analysis),
			G_ALLOC_AND_FREE);
	rach_keys_chunk = g_mem_chunk_new("rach_keys_chunk",
			sizeof(struct rach_key),
			RACH_MAX * sizeof(struct rach_key),
			G_ALLOC_AND_FREE);
	rach_values_chunk = g_mem_chunk_new("rach_values_chunk",
			sizeof(struct rach_value),
			RACH_MAX * sizeof(struct rach_value),
			G_ALLOC_AND_FREE);
	rach_hash_table = g_hash_table_new(rach_hash, rach_equal);
}

static conversation_t *find_chan_conversation(guint8 chan_no, int msg_type, packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct chan_analysis *ca=NULL;
	guint16 srcport, destport;
	guint8 chan;

	chan = chan_no >> 3; /* ignore the timeslot number bit */
	if (chan==CHAN_BCCH || (chan>CHAN_RACH && chan!=CHAN_NULL))	/* channel numbers with the top bit set are for the common channels - no activations tracked except rach */
		return 0;

	/* use the fact that BSC port is always constant to override it for the
	   conversation id to allow a conversation per chan_no */
	if (pinfo->srcport == TCP_PORT_IPA_RSL) {
		srcport = chan_no;
		destport = pinfo->destport;
	}
	else if (pinfo->destport == TCP_PORT_IPA_RSL) {
		destport = chan_no;
		srcport = pinfo->srcport;
	}
	else	/* something wrong: dont try to track the conversation */
		return 0;

	conv = find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, srcport, destport, 0);

	if (((chan==CHAN_RACH || chan==CHAN_NULL) && !conv) || msg_type == GSM0858_MSG_TYPE_CHAN_ACTIVATE) {	/* start a new conversation */
		conv=conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, srcport, destport, 0);
	}

	return conv;
}

static struct rach_value *analyze_rach(guint8 ra, guint16 fn, packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct rach_value *rv=NULL;
	struct rach_key key, *rk=NULL;

	conv = find_chan_conversation(CHAN_RACH<<3, GSM0858_MSG_TYPE_CHANNEL_REQUIRED, pinfo);

	if (!conv)	/* something wrong - abort */
		return 0;

	key.conv = conv->index;
	key.fn = fn;
	key.ra = ra;
	key.secs = pinfo->fd->rel_ts.secs;

	rv = (struct rach_value *)g_hash_table_lookup(rach_hash_table, &key);
	if (!rv) {	/* new table entry required if there is no matching entry */
		rk = g_mem_chunk_alloc(rach_keys_chunk);
		memcpy(rk,&key,sizeof(key));
		rv = g_mem_chunk_alloc(rach_values_chunk);
		rv->rach_ts = pinfo->fd->rel_ts;
		rv->rach_frame = pinfo->fd->num;
		g_hash_table_insert(rach_hash_table, rk, rv);
	}

	return rv;
}

int analyze_immediate_assignment(guint8 ra, guint16 fn, packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct rach_value *rv=NULL;
	struct rach_key key, *rk=NULL;

	conv = find_chan_conversation(CHAN_RACH<<3, GSM0858_MSG_TYPE_CHANNEL_REQUIRED, pinfo);

	if (!conv)	/* something wrong - abort */
		return 0;

	key.conv = conv->index;
	key.fn = fn;
	key.ra = ra;
	key.secs = pinfo->fd->rel_ts.secs;

	rv = (struct rach_value *)g_hash_table_lookup(rach_hash_table, &key);
	if (!rv)
		return 0;

	rv->delay_ms = 1000*(pinfo->fd->rel_ts.secs - rv->rach_ts.secs) + (pinfo->fd->rel_ts.nsecs - rv->rach_ts.nsecs)/1000000;

	return rv->delay_ms;
}

static struct chan_analysis *analyze_chan(guint8 chan_no, int msg_type, packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct chan_analysis *c_a=NULL;
	guint8 chan;

	chan = chan_no >> 3; /* ignore the timeslot number bit */
	if (chan>=CHAN_BCCH)	/* channel numbers with the top bit set are for the common channels - no activations tracked except rach */
		return 0;

	conv=find_chan_conversation(chan_no, msg_type, pinfo);

	if (!conv)
		return 0;

	c_a=conversation_get_proto_data(conv, proto_gsm0858);
	if (c_a && c_a->end_frame>pinfo->fd->num)	/* if this packet is already in the conversation, return the result */
		return c_a;

	/* to get this far we are either starting a new conversation, or are already in one, with conv and ca set */

	if (msg_type == GSM0858_MSG_TYPE_CHAN_ACTIVATE) {	/* Start of a new conversation */
		/* This is a new conversation */
		c_a=g_mem_chunk_alloc(chan_analysis_chunk);
		if (c_a) {
			c_a->start_frame = pinfo->fd->num;
			c_a->end_frame = pinfo->fd->num;
			c_a->msg_count = 1;
			c_a->state = 1;
			c_a->rtp_mux_used = 0;
			c_a->rtp_dyn_payload = g_hash_table_new( g_int_hash, g_int_equal);
			c_a->rtp_src_addr.type=0;
			c_a->rtp_src_addr.len=0;
			c_a->rtp_src_addr.data=NULL;
			c_a->rtp_dst_addr.type=0;
			c_a->rtp_dst_addr.len=0;
			c_a->rtp_dst_addr.data=NULL;
			conversation_add_proto_data(conv, proto_gsm0858, (void *)c_a);
		}
	}
	else if (c_a) {	/* Another message in the middle of this connection - if there is one */
		if (c_a->state == 1)	{ /* keep incrementing the message count until the conversation is over */
			c_a->msg_count++;
			c_a->end_frame = pinfo->fd->num;
		}
		if (msg_type == GSM0858_MSG_TYPE_RF_CHAN_REL_ACK)	/* the conversation is now over */
			c_a->state = 2;
	}

	return c_a;
}

static struct mux_analysis *analyze_mux(int msg_type, packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct mux_analysis *m_a=NULL;

	conv=find_chan_conversation(CHAN_NULL<<3, msg_type, pinfo);

	if (!conv)
		return 0;

	m_a=conversation_get_proto_data(conv, proto_gsm0858);
	if (m_a && m_a->end_frame>pinfo->fd->num)	/* if this packet is already in the conversation, return the result */
		return m_a;

	/* to get this far we are either starting a new conversation, or are already in one, with conv and m_a set */

	if (msg_type == GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION) {	/* Start of a new conversation */
		/* This is a new conversation */
		m_a=g_mem_chunk_alloc(mux_analysis_chunk);
		if (m_a) {
			m_a->start_frame = pinfo->fd->num;
			m_a->end_frame = pinfo->fd->num;
			m_a->rtp_dyn_payload = g_hash_table_new( g_int_hash, g_int_equal);
			m_a->srtp_info = NULL;
			conversation_add_proto_data(conv, proto_gsm0858, (void *)m_a);
		}
	}

	return m_a;
}

/* Code to dissect the GSM 08.58 packets */
static void
dissect_gsm0858(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_desc, msg_type, chan_no;
	tvbuff_t *next_tvb;
	guint8	pt;
	gint* key;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_item *item=NULL;
	proto_tree *tree=NULL;
	proto_item *analysis_ti=NULL;
	proto_tree *analysis_tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "gsm0858";

	gsmstat_set_interface(GSMSTAT_INTERFACE_A_BIS,pinfo);

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM");

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm0858, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_gsm0858);
	}

	gsmstat_set_proto(GSMSTAT_PROTO_GSM0858);

	msg_desc = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0858_msg_desc, tvb, offset, 1, msg_desc);
	offset += 1;

	msg_type = tvb_get_guint8(tvb, offset);
	gsmstat_set_msg_type(msg_type);
	if (tree) {
		proto_tree_add_uint(tree, hf_gsm0858_msg_type, tvb, offset, 1, msg_type);
		proto_item_set_text(ti, " 08.58: %s", val_to_str(msg_type, msg_type_vals, "Unknown"));
	}
	offset += 1;

	if ((msg_desc>>1==MSG_DESC_RLLM)||(msg_desc>>1==MSG_DESC_DCM)||(msg_desc>>1==MSG_DESC_CCM)||(msg_desc>>1==MSG_DESC_CMGMT)) {
		chan_no = tvb_get_guint8(tvb, offset+1);/* skip over the Channel Number IE EI */

		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, "08.58:%s(TS=%d,%s,",val_to_str(msg_type,msg_type_vals,"Unknown"),chan_no&0x07,val_to_str(chan_no>>3,chan_vals,"Unknown"));

		ca = analyze_chan(chan_no, msg_type, pinfo);
		if (tree && ca && pinfo->fd->num>=ca->start_frame) {
			analysis_ti = proto_tree_add_text(tree, tvb, 0, 0, "[Analysis]");
			analysis_tree = proto_item_add_subtree(analysis_ti, ett_gsm0858_analysis);

			if (analysis_tree) {
				item=proto_tree_add_uint(analysis_tree, hf_gsm0858_conn_start_frame, tvb, 0, 0, ca->start_frame);
				PROTO_ITEM_SET_GENERATED(item);
				if (msg_type == GSM0858_MSG_TYPE_RF_CHAN_REL_ACK) { /* display analysis on release */
					item=proto_tree_add_uint(analysis_tree, hf_gsm0858_conn_msg_count, tvb, 0, 0, ca->msg_count);
					PROTO_ITEM_SET_GENERATED(item);
					item=proto_tree_add_uint(analysis_tree, hf_gsm0858_conn_state, tvb, 0, 0, ca->state);
					PROTO_ITEM_SET_GENERATED(item);
				}
			}
		}
	}
	else {
		if (msg_desc>>1==MSG_DESC_MCMGMT) {
			ma = analyze_mux(msg_type, pinfo);
		}
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, "08.58:%s(",val_to_str(msg_type,msg_type_vals,"Unknown"));
	}

	next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);

	dissect_gsm0858_ie(msg_type, ie_defs, next_tvb, pinfo, tree, main_tree);

	if (ca && ca->rtp_src_port && ca->rtp_src_addr.len && ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION)||(msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION))) {
		if (rtp_handle) {
			rtp_add_address(pinfo, &(ca->rtp_src_addr), ca->rtp_src_port, 0, "GSM0858", pinfo->fd->num, ca->rtp_dyn_payload);
		}
		if (rtcp_handle) {
			rtcp_add_address(pinfo, &(ca->rtp_src_addr), ca->rtp_src_port+1, 0, "GSM0858", pinfo->fd->num);
		}
	}
	else if (ca && ca->rtp_dst_port && ca->rtp_dst_addr.len && !ca->rtp_mux_used && ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION_ACK)||(msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION_ACK))) {
		if (rtp_handle) {
			rtp_add_address(pinfo, &(ca->rtp_dst_addr), ca->rtp_dst_port, 0, "GSM0858", pinfo->fd->num, ca->rtp_dyn_payload);
		}
		if (rtcp_handle) {
			rtcp_add_address(pinfo, &(ca->rtp_dst_addr), ca->rtp_dst_port+1, 0, "GSM0858", pinfo->fd->num);
		}
	}
	else if (ma && ma->rtp_port && (msg_type == GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION)) {
		pt = GSM0858_RTP_MUX_PT;
		key = g_malloc(sizeof(gint));
		*key = pt;
		ma->rtp_pt = pt;
		g_hash_table_insert(ma->rtp_dyn_payload, key, (gchar *)val_to_str(RTP_PAYLOAD_GSM_NANO_GSM_MUX,rtp_payload_vals,"Unknown"));

		if (rtp_handle) {
			if (ma->srtp_info)
				srtp_add_address(pinfo, &(ma->rtp_addr), ma->rtp_port, 0, "GSM0858", pinfo->fd->num, ma->rtp_dyn_payload, ma->srtp_info);
			else
				rtp_add_address(pinfo, &(ma->rtp_addr), ma->rtp_port, 0, "GSM0858", pinfo->fd->num, ma->rtp_dyn_payload);
		}
		if (rtcp_handle) {
			if (ma->srtp_info)
				srtcp_add_address(pinfo, &(ma->rtp_addr), ma->rtp_port+1, 0, "GSM0858", pinfo->fd->num, ma->srtp_info);
			else
				rtcp_add_address(pinfo, &(ma->rtp_addr), ma->rtp_port+1, 0, "GSM0858", pinfo->fd->num);
		}
	}

	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_fstr(pinfo->cinfo, COL_INFO, ")");

	tap_queue_packet(gsmstat_tap,pinfo,NULL);

	/* null the nasty statics for safety */
	if (ca)
		ca = NULL;
	if (ma)
		ma = NULL;

	return;
}

/* Code to dissect the IEs */
static void
dissect_gsm0858_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree)
{
	int offset=0, msg_len, i;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		ie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm0858_ie_ei, tvb, offset, 1, ie_ei);
		offset += 1;

		for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==ie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				ie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				ie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, val_to_str(tvb_get_guint8(tvb, offset),(value_string *)(p_ie_def->info),"Unknown"));
						for (i=2; i<ie_len+1; i++)
							proto_item_append_text(ti, ", %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm0858_ie);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				ie_len8 = tvb_get_guint8(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_gsm0858_ie_len8, tvb, offset, 1, ie_len8);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s:", p_ie_def->ie_name);
						for (i=2; i<ie_len8+2; i++) {
							if (i==2)
								proto_item_append_text(ti, " %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
							else
								proto_item_append_text(ti, ", %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm0858_ie);
					}
				}
				offset += 1;
				ie_len = ie_len8;
			}
			else if (p_ie_def->ie_len == -2) {
				ie_len16 = tvb_get_ntohs(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_gsm0858_ie_len16, tvb, offset, 2, ie_len16);
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s:", p_ie_def->ie_name);
						for (i=3; i<ie_len16+3; i++) {
							if (i==3)
								proto_item_append_text(ti, " %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
							else
								proto_item_append_text(ti, ", %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm0858_ie);
					}
				}
				offset += 2;
				ie_len = ie_len16;
			}
			else if (p_ie_def->ie_len == -3) {	/* special flag: last IE, whole 08.58 message follows! */
				if (main_tree) {
					ti = proto_tree_add_text(main_tree, tvb, 0, msg_len, "%s", p_ie_def->ie_name);
					tree = proto_item_add_subtree(ti, ett_gsm0858_ie);
				}
				ie_len=-1;	/* flag for special handling */
			}
			else {			/* should never happen */
				ie_len=0; /* safety! */
			}

			if (ie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle,new_tvb, pinfo, tree);
				offset+=ie_len;
			}
			else if (ie_len==-1) {	/* special Phase 1 IE: rest of IE is a complete 08.58 message - recurse! */
				new_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
				dissect_gsm0858(new_tvb, pinfo, tree);
				offset = msg_len;
			}

				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);

				dissect_gsm0858_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree, top_tree);
			}
		}
	}
	return;
}

/* Code to dissect the Embedded IEs */
static void
dissect_gsm0858_eie( int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;
	guint8	eie_ei, eie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		eie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm0858_eie_ei, tvb, offset, 1, eie_ei);
		offset += 1;

		for (p_ie_def=eie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==eie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				eie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				eie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s:", p_ie_def->ie_name);
						for (i=1; i<eie_len+1; i++)
							if (i==1)
								proto_item_append_text(ti, " %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
							else
								proto_item_append_text(ti, ", %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm0858_eie);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				eie_len = tvb_get_guint8(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_gsm0858_eie_len, tvb, offset, 1, eie_len);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+2, "%s:", p_ie_def->ie_name);
						for (i=2; i<eie_len+2; i++)
							if (i==2)
								proto_item_append_text(ti, " %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
							else
								proto_item_append_text(ti, ", %s", val_to_str(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info),"Unknown"));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm0858_eie);
					}
				}
				offset += 1;
			}

			if (eie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, eie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle,new_tvb, pinfo, tree);
				offset+=eie_len;
			}

				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);

				dissect_gsm0858_eie(msg_type, new_tvb, pinfo, main_tree, top_tree, extra);
			}
		}
	}
	return;
}

static void dissect_gsm0858_ie_l3info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	gsmstat_set_cause(GSMSTAT_GSM0858_L3INFO);

	switch (msg_type)
	{
	case SACCH_FILLING:
	case SACCH_INFO_MODIFY:
		if (tvb_length_remaining(tvb,0) >= 19)	/* check whether this is the phase 1 or phase 2 definition */
			call_dissector(l2_handle, tvb, pinfo, top_tree);	/* phase 2 */
		else
			call_dissector(l3_handle, tvb, pinfo, top_tree);	/* phase 1 */
		break;

	case BCCH_INFO:
	case IMMEDIATE_ASSIGN:
		if (tvb_length_remaining(tvb,0) >= 23)	/* check whether this is the phase 1 or phase 2 definition */
			call_dissector(l2_handle, tvb, pinfo, top_tree);	/* phase 2 */
		else
			call_dissector(l3_handle, tvb, pinfo, top_tree);	/* phase 1 */
		break;

	default:
		col_set_fence(pinfo->cinfo, COL_INFO);
		call_dissector(l3_handle, tvb, pinfo, top_tree);
	}
}

static void dissect_gsm0858_ie_sacch_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8 number_of_messages, message, sacch_info_length;
	tvbuff_t *new_tvb;

	number_of_messages = tvb_get_guint8(tvb,offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0858_sacch_info_number_of_messages, tvb, offset, 1, number_of_messages);
	offset++;

	for (message=0; message<number_of_messages; message++) {
		if (tree)
			proto_tree_add_item(tree, hf_gsm0858_sys_info_type, tvb, offset, 1, FALSE);
		offset++;
		sacch_info_length = tvb_get_guint8(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_gsm0858_sacch_info_length, tvb, offset, 1, sacch_info_length);
		offset++;

		new_tvb = tvb_new_subset(tvb, offset, sacch_info_length, sacch_info_length);
		call_dissector(gsm_a_ccch_handle, new_tvb, pinfo, top_tree);
		offset += sacch_info_length;
	}

	return;
}

static void dissect_gsm0858_ie_channel_number(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	chan_no, chan, subchan_no=8; /* sub-channel 8 is too big to be a real sub-channel: flag meaning no sub-channel */

	chan_no = tvb_get_guint8(tvb, offset);
	chan = chan_no>>3;

	if (chan == 0x01) {	/* TCH/F */
		gsmstat_set_chan_type(GSMSTAT_CHAN_TYPE_TCH);
	}
	else if (chan>>1== 0x01) { /* TCH/H */
		subchan_no = chan & 0x01;
		gsmstat_set_chan_type(GSMSTAT_CHAN_TYPE_TCH);
	}
	else if (chan>>2 == 0x01) { /* SDCCH/4 */
		subchan_no = chan & 0x03;
		gsmstat_set_chan_type(GSMSTAT_CHAN_TYPE_SDCCH);
	}
	else if (chan>>3 == 0x01) { /* SDCCH/8 */
		subchan_no = chan & 0x07;
		gsmstat_set_chan_type(GSMSTAT_CHAN_TYPE_SDCCH);
	}

	if (tree) {
		proto_tree_add_uint(tree, hf_gsm0858_chan_no, tvb, offset, 1, chan_no);
		if (subchan_no<8)
			proto_tree_add_uint_hidden(tree, hf_gsm0858_subchan_no, tvb, offset, 1, chan_no);
		proto_tree_add_uint(tree, hf_gsm0858_chan, tvb, offset, 1, chan_no);
	}

	return;
}

static void dissect_gsm0858_ie_link_identifier(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	link_id;

	if (tree) {
		link_id = tvb_get_guint8(tvb, offset);
		proto_tree_add_item(tree, hf_gsm0858_sapi, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_sig_chan, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_link_id_priority, tvb, offset, 1, FALSE);
	}

	return;
}

static void dissect_gsm0858_ie_activation_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	act;

	if (tree) {
		act = tvb_get_guint8(tvb, offset);
		proto_tree_add_item(tree, hf_gsm0858_activation_type, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_activation_r, tvb, offset, 1, FALSE);
	}

	return;
}

static void dissect_gsm0858_ie_bs_power(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	power;

	if (tree) {
		power = tvb_get_guint8(tvb, offset);
		proto_tree_add_item(tree, hf_gsm0858_bs_power_level, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_bs_fpc, tvb, offset, 1, FALSE);
	}

	return;
}

static void dissect_gsm0858_ie_uplink_measurements(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_rfu, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_dtxd_meas, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_rxlev_full_up, tvb, offset, 1, FALSE);
		offset++;
		proto_tree_add_item(tree, hf_gsm0858_rxlev_sub_up, tvb, offset, 1, FALSE);
		offset++;
		proto_tree_add_item(tree, hf_gsm0858_rxqual_full_up, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_rxqual_sub_up, tvb, offset, 1, FALSE);
		offset++;
	}

	return;
}

static void dissect_gsm0858_ie_timing_offset(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	int timing_offset;

	timing_offset = tvb_get_guint8(tvb, offset) - 63;

	if (tree) {
		proto_tree_add_int(tree, hf_gsm0858_timing_offset, tvb, offset, 1, timing_offset);
		offset++;
	}

	return;
}

static void dissect_gsm0858_ie_ms_power(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	power, db, fpc;

	if (tree) {
		power = tvb_get_guint8(tvb, offset);
		fpc = ((power>>4)&0x01);
		power &= 0x1f;
		if (power<16)
			db = 30 - 2*power;
		else if (power>28)
			db = 36-2*(power-29);
		else
			db = 0;
		proto_tree_add_uint_format(tree, hf_gsm0858_ms_power_level, tvb, offset, 1, db, "MS Power %ddB", db);
		proto_tree_add_item(tree, hf_gsm0858_ms_fpc, tvb, offset, 1, FALSE);
		offset++;
	}

	return;
}

static void dissect_gsm0858_ie_channel_mode(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	dtx, speech_data, rate_type, code_rate, nt;

	dtx = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_dtxd, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_dtxu, tvb, offset, 1, FALSE);
	}
	offset++;

	speech_data = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_speech_data, tvb, offset, 1, FALSE);
	}
	offset++;

	rate_type = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_rate_type, tvb, offset, 1, FALSE);
	}
	offset++;

	code_rate = tvb_get_guint8(tvb, offset);
	if (speech_data == CHANNEL_MODE_DATA) {
		nt = (code_rate>>6)&0x01;
		proto_tree_add_item(tree, hf_gsm0858_t_nt, tvb, offset, 1, FALSE);
		if (ca)
			ca->rtp_payload = RTP_PAYLOAD_GSM_CSD;
	}
	else if (speech_data == CHANNEL_MODE_SPEECH) {
		if (ca)
			ca->rtp_payload = (code_rate>>4);
	}

	if (tree) {
		if (speech_data == CHANNEL_MODE_DATA) {
			if (nt)
				proto_tree_add_item(tree, hf_gsm0858_nt_rate, tvb, offset, 1, FALSE);
			else
				proto_tree_add_item(tree, hf_gsm0858_t_rate, tvb, offset, 1, FALSE);
		}
		else {
	   		proto_tree_add_item(tree, hf_gsm0858_speech_ver, tvb, offset, 1, FALSE);
		}
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_encryption_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	enc;

	if (tree) {
		enc = tvb_get_guint8(tvb, offset);
		proto_tree_add_item(tree, hf_gsm0858_encryption_algorithm, tvb, offset, 1, FALSE);
		offset++;

		if (enc && tvb_length_remaining(tvb,offset))
		proto_tree_add_item(tree, hf_gsm0858_encryption_key, tvb, offset, tvb_length_remaining(tvb,offset), FALSE);
	}

	return;
}

static void dissect_gsm0858_ie_resource_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8	chan_no, interference_level;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		for (offset=0; offset<msg_len; ) {
			chan_no = tvb_get_guint8(tvb, offset);
			proto_tree_add_item(tree, hf_gsm0858_chan_no, tvb, offset, 1, FALSE);
			offset++;

			interference_level = tvb_get_guint8(tvb, offset);
			proto_tree_add_item(tree, hf_gsm0858_interference_band, tvb, offset, 1, FALSE);
			offset++;
		}
	}

	return;
}

static void dissect_gsm0858_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_cause, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_paging_group(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_paging_group, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_handover_reference(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_ho_ref, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_sys_info_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_sys_info_type, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_cb_command_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_cb_command, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_cb_default_broadcast, tvb, offset, 1, FALSE);
		proto_tree_add_item(tree, hf_gsm0858_cb_last_block, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_smscb_message(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_smscb_message, tvb, offset, tvb_length_remaining(tvb,offset), FALSE);
	}
	offset += tvb_length_remaining(tvb,offset);

	return;
}

static void dissect_gsm0858_ie_access_delay(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_access_delay, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_timing_advance(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_timing_advance, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_rlm_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tree) {
		proto_tree_add_item(tree, hf_gsm0858_rlm_cause, tvb, offset, 1, FALSE);
	}
	offset++;

	return;
}

static void dissect_gsm0858_ie_preprocessing_parameters(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	tvbuff_t *new_tvb;

	if (tree)
		proto_tree_add_item(tree, hf_gsm0858_preprocessed_reporting, tvb, offset, 1, FALSE);
		offset++;

		/* only recurse to this disector if this ie was decoded OK */
		if (tvb_length_remaining(tvb,offset)>0) {
			/* use a new subset TVB for the recursion */
			new_tvb = tvb_new_subset(tvb, offset, -1, -1);
			dissect_gsm0858_eie(msg_type, new_tvb, pinfo, tree, top_tree, extra);
		}

	return;
}

static void dissect_gsm0858_ie_rtp_compression_control(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	rcc;

	if (tree) {
		rcc = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, offset-1, 1, "UDP Checksums: %s", rcc?"Disabled":"Enabled");
	}

	return;
}

static void dissect_gsm0858_ie_rtp_payload_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	pt;
	gint* key;

	pt = tvb_get_guint8(tvb, offset++) & 0x7f;	/* PT is only 7 bits - MSB is reserved */
	if (tree) {
		proto_tree_add_text(tree, tvb, offset-1, 1, "RTP Payload Type: %d", pt);
	}

	if (ca) {
		key = se_alloc(sizeof(gint));
		*key = pt;
		ca->rtp_pt = pt;
		if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION)
		 || (msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION_ACK) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION_ACK)) {
			g_hash_table_insert(ca->rtp_dyn_payload, key, (gchar *)val_to_str(ca->rtp_payload,rtp_payload_vals,"Unknown"));
		}
	}
	if (ma) {
		key = se_alloc(sizeof(gint));
		*key = pt;
		ma->rtp_pt = pt;
		if (msg_type == GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION) {
			g_hash_table_insert(ma->rtp_dyn_payload, key, (gchar *)val_to_str(RTP_PAYLOAD_GSM_NANO_GSM_MUX,rtp_payload_vals,"Unknown"));
		}
	}

	return;
}

static void dissect_gsm0858_ie_rtp_jitter_buffer_control(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	ija, bar, ar;

	if (tree) {
		ija = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, offset-1, 1, "Initial Jitter Allowance: %dms", 5*ija);
		bar = tvb_get_guint8(tvb, offset++);
		ar = bar & 0x7f;
		if (ar==0)
			proto_tree_add_text(tree, tvb, offset-1, 1, "%s Buffer; Adaptation Period: Immediate", (bar&0x80)?"Grow":"Grow & Shrink");
		else if (ar==127)
			proto_tree_add_text(tree, tvb, offset-1, 1, "%s Buffer; Adaptation Period: Fixed buffer (no adaptation)", (bar&0x80)?"Grow":"Grow & Shrink");
		else if (ar<=100)
			proto_tree_add_text(tree, tvb, offset-1, 1, "%s Buffer; Adaptation Period: %ds", (bar&0x80)?"Grow":"Grow & Shrink", ar);
		else
			proto_tree_add_text(tree, tvb, offset-1, 1, "%s Buffer; Adaptation Period: unknown", (bar&0x80)?"Grow":"Grow & Shrink");
	}

	return;
}

static void dissect_gsm0858_ie_ip_speech_mode(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	ipsm;

	if (tree) {
		ipsm = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, offset-1, 1, "Speech Codec: %s", val_to_str((ipsm>>0)&0x0f,ip_speech_mode_s_vals,"Unknown"));
		proto_tree_add_text(tree, tvb, offset-1, 1, "Speech Mode: %s", val_to_str((ipsm>>4)&0x03,ip_speech_mode_m_vals,"Unknown"));
	}

	return;
}

static void dissect_gsm0858_ie_rtp_csd_format(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	rcf;

	if (tree) {
		rcf = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, offset-1, 1, "Payload Format: %s", val_to_str((rcf>>0)&0x0f,rtp_csd_format_d_vals,"Unknown"));
		proto_tree_add_text(tree, tvb, offset-1, 1, "Intermediate Rate: %s", val_to_str((rcf>>4)&0x03,rtp_csd_format_ir_vals,"Unknown"));
	}

	return;
}

static void dissect_gsm0858_ie_connection_statistics(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint32 pkts_sent, octets_sent;
	guint32 pkts_recv, octets_recv;
	guint32 pkts_lost, jitter, delay;

	if (tree) {
		pkts_sent = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Packets Sent: %d", pkts_sent);
		offset += 4;
		octets_sent = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Octets Sent: %d", octets_sent);
		offset += 4;
		pkts_recv = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Packets Received: %d", pkts_recv);
		offset += 4;
		octets_recv = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Octets Received: %d", octets_recv);
		offset += 4;
		pkts_lost = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Packets Lost: %d", pkts_lost);
		offset += 4;
		jitter = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Inter-arrival Jitter: %d", jitter);
		offset += 4;
		delay = tvb_get_ntohl(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 4, "Average Transmission Delay: %d", delay);
		offset += 4;
	}

	return;
}

static void dissect_gsm0858_ie_ip_address(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	a[4];

	a[0] = tvb_get_guint8(tvb, offset++);
	a[1] = tvb_get_guint8(tvb, offset++);
	a[2] = tvb_get_guint8(tvb, offset++);
	a[3] = tvb_get_guint8(tvb, offset++);

	if (tree) {
		proto_tree_add_text(tree, tvb, offset-4, 4, "IP Address: %d.%d.%d.%d", a[0],a[1],a[2],a[3]);
	}
	if (ca) {
		if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION)) {
			ca->rtp_src_addr.type=AT_IPv4;
			ca->rtp_src_addr.len=4;
			ca->rtp_src_addr.data=ca->rtp_src_addr_data;
			memcpy(ca->rtp_src_addr_data,a,4);
		}
		else if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION_ACK) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION_ACK)) {
			ca->rtp_dst_addr.type=AT_IPv4;
			ca->rtp_dst_addr.len=4;
			ca->rtp_dst_addr.data=ca->rtp_dst_addr_data;
			memcpy(ca->rtp_dst_addr_data,a,4);
		}
	}
	if (ma) {
		if (msg_type == GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION) {
			ma->rtp_addr.type=AT_IPv4;
			ma->rtp_addr.len=4;
			ma->rtp_addr.data=ma->rtp_addr_data;
			memcpy(ma->rtp_addr_data,a,4);
		}
	}

	return;
}

static void dissect_gsm0858_ie_ip_port(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint16	port;

	msg_len = tvb_length_remaining(tvb,offset);

	port = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, msg_len, "Port: %d", port);
	}
	if (ca) {
		if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION))
			ca->rtp_src_port = port;
		else if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION_ACK) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION_ACK))
			ca->rtp_dst_port = port;
	}
	if (ma) {
		if (msg_type == GSM0858_MSG_TYPE_CREATE_MULTIPLEX_CONNECTION)
			ma->rtp_port = port;
	}

	return;
}

guint16 reduced_frame_number(guint16 fn)
{
	/* great care needed with signed/unsigned - -1 in unsigned is 0xffff, which mod(26) is not what you think !!! */
	gint16	t2, t3, t;
	guint16	frame, t1;

	t1 = (fn >> 11) & 0x1f;
	t2 = (fn >> 0) & 0x1f;
	t3 = (fn >> 5) & 0x3f;

	t = (t3-t2)%26;
	if (t<0)
		t += 26;

	frame = 51*(unsigned)t+(unsigned)t3+51*26*t1;

	return frame;
}

static void dissect_gsm0858_ie_frame_number(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint16	fn;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		fn = tvb_get_ntohs(tvb, offset);
		proto_tree_add_uint(tree, hf_gsm0858_rfn, tvb, offset, msg_len, reduced_frame_number(fn));
	}

	return;
}

static void dissect_gsm0858_ie_srtp_configuration(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	if (ma) {
		ma->srtp_info = se_alloc(sizeof(struct srtp_info));
		ma->srtp_info->encryption_algorithm = SRTP_ENC_ALG_AES_CM;
		ma->srtp_info->encryption_algorithm = SRTP_AUTH_ALG_HMAC_SHA1;
		ma->srtp_info->mki_len = 0;
		ma->srtp_info->auth_tag_len = 10;
	}

	dissect_gsm0858_eie(msg_type, tvb, pinfo, tree, top_tree, extra);

	return;
}

static void dissect_gsm0858_ie_multiplex_control(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint16	dl_pkt_size, ul_pkt_size, dl_interval, ul_interval;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		dl_pkt_size = tvb_get_ntohs(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 2, "Downlink Maximum Multiplex RTP Packet Size: %d bytes", dl_pkt_size);
		offset+=2;

		ul_pkt_size = tvb_get_ntohs(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 2, "Uplink Maximum Multiplex RTP Packet Size: %d bytes", ul_pkt_size);
		offset+=2;

		dl_interval = tvb_get_ntohs(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 2, "Downlink Maximum Multiplex RTP Time Interval: %dms", (guint32)dl_interval*20);
		offset+=2;

		ul_interval = tvb_get_ntohs(tvb, offset);
		proto_tree_add_text(tree, tvb, offset, 2, "Uplink Maximum Multiplex RTP Time Interval: %dms", (guint32)ul_interval*20);
		offset+=2;
	}

	return;
}

static void dissect_gsm0858_ie_multiplex_connection_id(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8 mux_conn_id;

	mux_conn_id = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Multiplex Connection Id: %d", mux_conn_id);
	}
	if (ca) {
		if ((msg_type == GSM0858_MSG_TYPE_CREATE_CONNECTION) || (msg_type == GSM0858_MSG_TYPE_MODIFY_CONNECTION))
			ca->rtp_mux_used = TRUE;
	}

	return;
}

static void dissect_gsm0858_ie_request_reference(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8	ra;
	guint16	fn;
	struct rach_value *rv;

	ra = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_item(tree, hf_gsm0858_ra, tvb, offset, 1, FALSE);
	offset++;

	fn = reduced_frame_number(tvb_get_ntohs(tvb, offset));
	if (tree)
		proto_tree_add_uint(tree, hf_gsm0858_rfn, tvb, offset, 2, fn);
	offset += 2;

	rv = analyze_rach(ra, fn, pinfo);

	return;
}

static void dissect_gsm0858_eie_measurement_averaging_configure(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8	octet3, octet4;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		octet3 = tvb_get_guint8(tvb, offset++);
		octet4 = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, 0, 2, "%s: HReqAve=%d HReqT=%d  Method:%s", val_to_str((octet3>>5)&0x03,measurement_averaging_configure_param_id_vals,"Unknown"), (octet3>>0)&0x1f, (octet4>>0)&0x1f, val_to_str((octet4>>5)&0x07,measurement_averaging_configure_method_vals,"Unknown"));
	}

	return;
}

static void dissect_gsm0858_eie_power_control_thresholds(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8	octet2, octet3, octet4;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		octet2 = tvb_get_guint8(tvb, offset++);
		octet3 = tvb_get_guint8(tvb, offset++);
		octet4 = tvb_get_guint8(tvb, offset++);
		proto_tree_add_text(tree, tvb, 0, 2, "RxLev:  %d - %d", (octet2>>0)&0x3f, (octet3>>0)&0x3f);
		proto_tree_add_text(tree, tvb, 2, 1, "RxQual: %d - %d", (octet4>>4)&0x07, (octet4>>0)&0x07);
	}

	return;
}

static void dissect_gsm0858_eie_handover_thresholds(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, 2, "Handover RxLev:		 uplink %d  downlink %d", tvb_get_guint8(tvb,0)&0x3f, tvb_get_guint8(tvb,1)&0x3f);
		proto_tree_add_text(tree, tvb, 4, 1, "Handover RxQual:		uplink %d  downlink %d", tvb_get_guint8(tvb,4)&0x07, (tvb_get_guint8(tvb,4)>>4)&0x07);
		proto_tree_add_text(tree, tvb, 2, 2, "Interference H/O RxLev: uplink %d  downlink %d", tvb_get_guint8(tvb,2)&0x3f, tvb_get_guint8(tvb,3)&0x3f);
		proto_tree_add_text(tree, tvb, 5, 1, "MS Range Max:		   %d", tvb_get_guint8(tvb,5)&0x3f);
	}

	return;
}

static void dissect_gsm0858_eie_ncell_defaults(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_text(tree, tvb, 0, 1, "RxLev Min:	   %d", tvb_get_guint8(tvb,0)&0x3f);
		proto_tree_add_text(tree, tvb, 1, 1, "Handover Margin: %d", tvb_get_guint8(tvb,1)&0x1f);
		proto_tree_add_text(tree, tvb, 2, 1, "MS Tx Power Max: %d", tvb_get_guint8(tvb,2)&0x1f);
	}

	return;
}

static void dissect_gsm0858_eie_ncell_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int cell=0, i, offset=0, msg_len, octet, len, bsic, ncell_freq_ind, baind, rxlevmin, txpwrmax, homargin;

	msg_len = tvb_length_remaining(tvb,offset);

	while (offset<msg_len) {
		octet = tvb_get_guint8(tvb,offset);
		len = octet >> 6;
		bsic = octet & 0x3f;

		octet = tvb_get_guint8(tvb,offset+1);
		baind = (octet >> 5) & 0x01;
		ncell_freq_ind = octet & 0x1f;

		if (tree)
			proto_tree_add_text(tree, tvb, offset, 2, "Cell %d: BSIC=0x%2.2x, BA-ind=%d, Freq-Ind=%d", cell, bsic, baind, ncell_freq_ind);

		offset += 2;

		for (i=0; i<=len; i++) {
			octet = tvb_get_guint8(tvb,offset);
			switch (octet>>6) {
			case 0:
				baind = (octet >> 5) & 0x01;
				ncell_freq_ind = octet & 0x1f;
				if (tree)
					proto_tree_add_text(tree, tvb, offset, 1, "		BA-ind=%d, Freq-Ind=%d", baind, ncell_freq_ind);
				offset++;
				break;

			case 1:
				rxlevmin = octet & 0x3f;
				if (tree)
					proto_tree_add_text(tree, tvb, offset, 1, "		RxLevMin=%d", rxlevmin);
				offset++;
				break;

			case 2:
				homargin = octet & 0x1f;
				if (tree)
					proto_tree_add_text(tree, tvb, offset, 1, "		HoMargin=%d", homargin);
				offset++;
				break;

			case 3:
				txpwrmax = octet & 0x1f;
				if (tree)
					proto_tree_add_text(tree, tvb, offset, 1, "		TxPwrMax=%d", txpwrmax);
				offset++;
				break;
			}
		}

		cell++;
	}

	return;
}

static void dissect_gsm0858_eie_pc_threshold_comparators(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		for(i=1; i<5; i++) {
			proto_tree_add_text(tree, tvb, offset, 2, "P%d: %d   N%d: %d", i, tvb_get_guint8(tvb,offset)&0x1f, i, tvb_get_guint8(tvb,offset+1)&0x1f);
			offset+=2;
		}
		proto_tree_add_text(tree, tvb, 8, 2, "Control interval %d  Increment step size %d  Decrement stepsize %d", tvb_get_guint8(tvb,8)&0x1f, (tvb_get_guint8(tvb,9)>>4)&0x0f,tvb_get_guint8(tvb,9)&0x0f);
	}

	return;
}

static void dissect_gsm0858_eie_ho_threshold_comparators(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		for(i=5; i<9; i++) {
			proto_tree_add_text(tree, tvb, offset, 2, "P%d: %d   N%d: %d", i, tvb_get_guint8(tvb,offset)&0x1f, i, tvb_get_guint8(tvb,offset+1)&0x1f);
			offset+=2;
		}
		proto_tree_add_text(tree, tvb, 8, 1, "Handover Required interval %ds", tvb_get_guint8(tvb,8)&0x1f);
		proto_tree_add_text(tree, tvb, 9, 1, "Directed Retry Timer %ds", tvb_get_guint8(tvb,9)&0x1f);
	}

	return;
}

static void dissect_gsm0858_eie_handover_candidates(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;
	guint8	octet1, octet2;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		for(i=0; i<msg_len>>1; i++) {
			octet1 = tvb_get_guint8(tvb,offset);
			octet2 = tvb_get_guint8(tvb,offset+1);
			if ((octet2>>6)&0x01)
				proto_tree_add_text(tree, tvb, offset, 2, "Serving Cell");
			else
				proto_tree_add_text(tree, tvb, offset, 2, "BA-Ind=%d  ARFCN Index=%d  BSIC=%d", (octet2>>5)&0x01, (octet2>>0)&0x1f, (octet1>>0)&0x3f);
			offset+=2;
		}
	}

	return;
}

static char *channel_number_string( guint8 chan_no, char *chan_no_str )
{
	sprintf(chan_no_str,"TS=%d,%s", chan_no&0x07, val_to_str(chan_no>>3,chan_vals,"Unknown"));

	return chan_no_str;
}
/* Register the protocol with Ethereal */

void
proto_register_gsm0858(void)
{
	module_t *gsm0858_module;

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_gsm0858_msg_desc,
			{ "Msg Desc", "gsm0858.msg_desc",
			FT_UINT8, BASE_HEX, VALS(msg_desc_vals), 0xfe,
			"Message Descriminator", HFILL }
		},
		{ &hf_gsm0858_msg_type,
			{ "Msg Type", "gsm0858.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,
			"Message Type", HFILL }
		},
		{ &hf_gsm0858_chan_no,
			{ "Timeslot", "gsm0858.chan_no",
			FT_UINT8, BASE_DEC, NULL, 0x07,
			"Timeslot", HFILL }
		},
		{ &hf_gsm0858_chan,
			{ "Channel", "gsm0858.chan",
			FT_UINT8, BASE_DEC, VALS(chan_vals), 0xF8,
			"Channel", HFILL }
		},
		{ &hf_gsm0858_subchan_no,
			{ "Sub-Channel No", "gsm0858.subchan_no",
			FT_UINT8, BASE_DEC, NULL, 0x38,
			"Sub-Channel Number", HFILL }
		},
		{ &hf_gsm0858_conn_start_frame,
			{ "Connection Start Frame Number", "gsm0858.conn.start_frame",
			FT_FRAMENUM, BASE_NONE, NULL, 0x0,
			"Connection Start Frame Number", HFILL }
		},
		{ &hf_gsm0858_conn_msg_count,
			{ "Connection Message Count", "gsm0858.conn.msg_count",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Connection Message Count", HFILL }
		},
		{ &hf_gsm0858_conn_state,
			{ "Connection State", "gsm0858.conn.state",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Connection State", HFILL }
		},
		{ &hf_gsm0858_ie_ei,
			{ "IE", "gsm0858.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ie_ei_vals), 0x0,
			"IE Element Identifier", HFILL }
		},
		{ &hf_gsm0858_ie_len8,
			{ "IE Len", "gsm0858.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_gsm0858_ie_len16,
			{ "IE Len", "gsm0858.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_gsm0858_eie_ei,
			{ "EIE", "gsm0858.eie.ei",
			FT_UINT8, BASE_HEX, VALS(eie_ei_vals), 0x0,
			"Embedded IE Element Identifier", HFILL }
		},
		{ &hf_gsm0858_eie_len,
			{ "EIE Len", "gsm0858.eie.len",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"EIE Length", HFILL }
		},
		{ &hf_gsm0858_sig_chan,
			{ "Signalling Channel", "gsm0858.sig_chan",
			FT_UINT8, BASE_DEC, VALS(sig_chan_vals), 0xc0,
			"Signalling Channel", HFILL }
		},
		{ &hf_gsm0858_sapi,
			{ "SAPI", "gsm0858.sapi",
			FT_UINT8, BASE_DEC, NULL, 0x07,
			"SAPI", HFILL }
		},
		{ &hf_gsm0858_link_id_priority,
			{ "Priority", "gsm0858.priority",
			FT_UINT8, BASE_DEC, VALS(link_id_priority_vals), 0x18,
			"Link Identifier Priority", HFILL }
		},
		{ &hf_gsm0858_activation_r,
			{ "Activation", "gsm0858.activation_r",
			FT_UINT8, BASE_DEC, VALS(activation_r_vals), 0x80,
			"Activation R Bit", HFILL }
		},
		{ &hf_gsm0858_activation_type,
			{ "Activation Type", "gsm0858.activation_type",
			FT_UINT8, BASE_DEC, VALS(activation_type_vals), 0x07,
			"Activation Type", HFILL }
		},
		{ &hf_gsm0858_bs_power_level,
			{ "BS Power", "gsm0858.bs_power",
			FT_UINT8, BASE_DEC, VALS(bs_power_vals), 0x0f,
			"BS Power Level", HFILL }
		},
		{ &hf_gsm0858_ms_power_level,
			{ "MS Power", "gsm0858.ms_power",
			FT_UINT8, BASE_DEC, VALS(ms_power_vals), 0x1f,
			"MS Power Level", HFILL }
		},
		{ &hf_gsm0858_bs_fpc,
			{ "FPC", "gsm0858.fpc",
			FT_UINT8, BASE_DEC, VALS(fpc_vals), 0x10,
			"BS Power Fast Power Control flag", HFILL }
		},
		{ &hf_gsm0858_ms_fpc,
			{ "FPC", "gsm0858.fpc",
			FT_UINT8, BASE_DEC, VALS(fpc_vals), 0x20,
			"MS Power Fast Power Control flag", HFILL }
		},
		{ &hf_gsm0858_dtxd_meas,
			{ "DTXd (measurement)", "gsm0858.dtxd_meas",
			FT_UINT8, BASE_DEC, NULL, 0x40,
			"Downlink DTX (measurement)", HFILL }
		},
		{ &hf_gsm0858_dtxd,
			{ "DTXd", "gsm0858.dtxd",
			FT_UINT8, BASE_DEC, NULL, 0x02,
			"Downlink DTX", HFILL }
		},
		{ &hf_gsm0858_dtxu,
			{ "DTXu", "gsm0858.dtxu",
			FT_UINT8, BASE_DEC, NULL, 0x01,
			"Uplink DTX", HFILL }
		},
		{ &hf_gsm0858_timing_offset,
			{ "Timing Offset", "gsm0858.timing_offset",
			FT_INT8, BASE_DEC, NULL, 0x0,
			"Timing Offset", HFILL }
		},
		{ &hf_gsm0858_rfu,
			{ "RFU", "gsm0858.rfu",
			FT_UINT8, BASE_DEC, NULL, 0x80,
			"RFU", HFILL }
		},
		{ &hf_gsm0858_rxlev_full_up,
			{ "RXLEV FULL up", "gsm0858.rxlev_full_up",
			FT_UINT8, BASE_DEC, NULL, 0x3f,
			"RXLEV FULL up", HFILL }
		},
		{ &hf_gsm0858_rxlev_sub_up,
			{ "RXLEV SUB up", "gsm0858.rxlev_sub_up",
			FT_UINT8, BASE_DEC, NULL, 0x3f,
			"RXLEV SUB up", HFILL }
		},
		{ &hf_gsm0858_rxqual_full_up,
			{ "RXQual FULL up", "gsm0858.rxqual_full_up",
			FT_UINT8, BASE_DEC, NULL, 0x38,
			"RXQual FULL up", HFILL }
		},
		{ &hf_gsm0858_rxqual_sub_up,
			{ "RXQual SUB up", "gsm0858.rxqual_sub_up",
			FT_UINT8, BASE_DEC, NULL, 0x07,
			"RXQual SUB up", HFILL }
		},
		{ &hf_gsm0858_speech_data,
			{ "Speech or Data", "gsm0858.speech_data",
			FT_UINT8, BASE_DEC, VALS(speech_data_vals), 0x0,
			"Speech or Data", HFILL }
		},
		{ &hf_gsm0858_rate_type,
			{ "Rate and Type", "gsm0858.rate_type",
			FT_UINT8, BASE_DEC, VALS(rate_type_vals), 0x0,
			"Channel Rate and Type", HFILL }
		},
		{ &hf_gsm0858_speech_ver,
			{ "Speech Version", "gsm0858.speech_ver",
			FT_UINT8, BASE_DEC, VALS(speech_ver_vals), 0x0,
			"Speech Version", HFILL }
		},
		{ &hf_gsm0858_t_nt,
			{ "T/NT", "gsm0858.t_nt",
			FT_UINT8, BASE_DEC, VALS(t_nt_vals), 0x40,
			"Transparent or Non-Transparent flag", HFILL }
		},
		{ &hf_gsm0858_nt_rate,
			{ "NT Rate", "gsm0858.nt_rate",
			FT_UINT8, BASE_DEC, NULL, 0x3f,
			"Radio Interface Data Rate", HFILL }
		},
		{ &hf_gsm0858_t_rate,
			{ "T Rate", "gsm0858.t_rate",
			FT_UINT8, BASE_DEC, NULL, 0x3f,
			"Data Rate", HFILL }
		},
		{ &hf_gsm0858_encryption_algorithm,
			{ "Encryption Alg", "gsm0858.enc_alg",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Encryption Algorithm", HFILL }
		},
		{ &hf_gsm0858_encryption_key,
			{ "Encryption Key", "gsm0858.enc_key",
			FT_BYTES, BASE_HEX, NULL, 0x0,
			"Encryption Key", HFILL }
		},
		{ &hf_gsm0858_rfn,
			{ "Reduced Frame No", "gsm0858.rfn",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Reduced Frame Number", HFILL }
		},
		{ &hf_gsm0858_ho_ref,
			{ "Handover Ref", "gsm0858.ho_ref",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Handover Reference", HFILL }
		},
		{ &hf_gsm0858_paging_group,
			{ "Paging Group", "gsm0858.paging_group",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Paging Group", HFILL }
		},
		{ &hf_gsm0858_paging_buffer_space,
			{ "Paging Buffer Space", "gsm0858.paging_buffer_space",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Paging Buffer Space", HFILL }
		},
		{ &hf_gsm0858_access_delay,
			{ "Access Delay", "gsm0858.access_delay",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Access Delay", HFILL }
		},
		{ &hf_gsm0858_ra,
			{ "Random Access", "gsm0858.ra",
			FT_UINT8, BASE_HEX_DEC, NULL, 0x0,
			"Request Reference Random Access", HFILL }
		},
		{ &hf_gsm0858_release_mode,
			{ "Release Mode", "gsm0858.release_mode",
			FT_UINT8, BASE_DEC, VALS(release_mode_vals), 0x01,
			"Release Mode", HFILL }
		},
		{ &hf_gsm0858_interference_band,
			{ "Interference band", "gsm0858.interference_band",
			FT_UINT8, BASE_DEC, NULL, 0xe0,
			"Interference Band", HFILL }
		},
		{ &hf_gsm0858_rlm_cause,
			{ "RLM Cause", "gsm0858.rlm_cause",
			FT_UINT8, BASE_DEC, VALS(rlm_cause_vals), 0x0,
			"RLM Cause", HFILL }
		},
		{ &hf_gsm0858_timing_advance,
			{ "TA", "gsm0858.timing_advance",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Timing Advance", HFILL }
		},
		{ &hf_gsm0858_cause,
			{ "Cause", "gsm0858.cause",
			FT_UINT8, BASE_DEC, VALS(cause_vals), 0x0,
			"Cause", HFILL }
		},
		{ &hf_gsm0858_cause_diagnostics,
			{ "Cause diagnostics", "gsm0858.cause_diagnostics",
			FT_BYTES, BASE_HEX, NULL, 0x0,
			"Cause diagnostics", HFILL }
		},
		{ &hf_gsm0858_sacch_info_number_of_messages,
			{ "SACCH Info Number of Messages", "gsm0858.sacch_info_number_of_messages",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"SACCH Info Number of Messages", HFILL }
		},
		{ &hf_gsm0858_sys_info_type,
			{ "Sys Info Type", "gsm0858.sys_info_type",
			FT_UINT8, BASE_DEC, VALS(sys_info_type_vals), 0x0,
			"System Information Type", HFILL }
		},
		{ &hf_gsm0858_sacch_info_length,
			{ "SACCH Info Length", "gsm0858.sacch_info_length",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"SACCH Info Length", HFILL }
		},
		{ &hf_gsm0858_preprocessed_reporting,
			{ "Pre-processed reporting", "gsm0858.preprocessed_reporting",
			FT_BOOLEAN, BASE_DEC, NULL, 0x01,
			"Pre-processed reporting", HFILL }
		},
		{ &hf_gsm0858_smscb_frame,
			{ "SMSCB Frame", "gsm0858.smscb_frame",
			FT_BYTES, BASE_HEX, NULL, 0x0,
			"SMSCB Frame", HFILL }
		},
		{ &hf_gsm0858_channel_needed,
			{ "Channel needed", "gsm0858.channel_needed",
			FT_UINT8, BASE_DEC, VALS(channel_needed_vals), 0x03,
			"Channel needed", HFILL }
		},
		{ &hf_gsm0858_cb_command,
			{ "CB Command", "gsm0858.cb_command",
			FT_UINT8, BASE_HEX, (cb_command_vals), 0xF0,
			"CB Command", HFILL }
		},
		{ &hf_gsm0858_cb_default_broadcast,
			{ "CB Default Broadcast", "gsm0858.cb_default_broadcast",
			FT_UINT8, BASE_DEC, VALS(cb_default_broadcast_vals), 0x08,
			"CB Default Broadcast", HFILL }
		},
		{ &hf_gsm0858_cb_last_block,
			{ "CB Last block", "gsm0858.cb_last_block",
			FT_UINT8, BASE_DEC, VALS(cb_last_block_vals), 0x03,
			"CB Last Block", HFILL }
		},
		{ &hf_gsm0858_smscb_message,
			{ "SMSCB Message", "gsm0858.smscb_message",
			FT_BYTES, BASE_HEX, NULL, 0x0,
			"SMSCB Message", HFILL }
		}
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0858,
		&ett_gsm0858_ie,
		&ett_gsm0858_eie,
		&ett_gsm0858_analysis,
	};

	proto_gsm0858 = proto_register_protocol("GSM 08.58 A-bis Interface Radio Signalling Layer 3", "GSM 08.58", "gsm0858");

	register_dissector("gsm0858", dissect_gsm0858, proto_gsm0858);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0858, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	/* Register preferences */
	gsm0858_module = prefs_register_protocol(proto_gsm0858, proto_reg_handoff_gsm0858);

	/* Register preferences */
	prefs_register_bool_preference(gsm0858_module, "use_ipa",
		 "Use ip.access GSM 04.08 dissector",
		 "Allows the use of the ip.access or the standard GSM 04.08 dissector.",
		 &global_gsm0858_use_ipa);

	/* common gsmstat tap registration
	 - do NOT use the register tap directly in each protocol - it confuses taps! */
	gsmstat_register_tap();

	/* Init rountines - for conversations */
	register_init_routine(analyze_init);
}

void
proto_reg_handoff_gsm0858(void)
{
	static int initialized = FALSE;

	if (!initialized) {
		gsm0858_handle = find_dissector("gsm0858");
		data_handle = find_dissector("data");
		rtp_handle = find_dissector("rtp");
		rtcp_handle = find_dissector("rtcp");
		gsm_a_dtap_handle = find_dissector("gsm_a_dtap");
		gsm_a_ccch_handle = find_dissector("gsm_a_ccch");
		gsm0408_handle = find_dissector("gsm0408");
		gsm0408_l2_handle = find_dissector("gsm0408_l2");

		initialized = TRUE;
	}

	if (global_gsm0858_use_ipa) {
		l2_handle = gsm0408_l2_handle;
		l3_handle = gsm0408_handle;
	}
	else {
		l2_handle = gsm_a_ccch_handle;
		l3_handle = gsm_a_dtap_handle;
	}
}
