/* packet-gsm_csd.c
 * Routines for GSM-CSD dissection
 * Copyright 2005, Neil Piercy <Neil.Piercy[at]ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm_csd.c 1.3 2008/06/06 05:11:22BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <epan/packet.h>
#include <epan/proto.h>
#include <epan/conversation.h>

#include "prefs.h"
#include "packet-gsm.h"

/* Initialize the protocol and registered fields */
static int proto_gsm_csd = -1;
static int hf_gsm_csd_rate = -1;

/* Initialize the subtree pointers */
static int ett_gsm_csd = -1;

/* The dynamic payload type which will be dissected as GSM-CSD */

static guint dynamic_payload_type = 0x68;
static guint temp_dynamic_payload_type = 0x68;

static const value_string gsm_csd_rate_vals[] = {
	{ 0,	NULL }
};

static GMemChunk *csd_analysis_chunk = NULL;
#define CSD_STREAM_MAX 100			/* one for each CSD conversation */
#define CSD_BUF_MAX 512	/* not really sure how big this needs to be yet... */

typedef struct csd_analysis {
	int state;			/* 0=null, 1=syncing, 2=frame-sync, 3=multi-frame sync */
	gboolean nt;		/* non-transparent ? */
	gboolean v110;		/* V110 frames or E-TRAU ? */
	int bit_rate;		/* 300,600, etc */
	int sub_chan_rate;	/* number of active bits: 0=unset, 1=8kb/s, 2=16kb/s etc. */
	int	sub_chan_offset;/* offset required in 64k octets to get to start of sync word */
	unsigned end_frame;		/* keep track of how far we have already analysed */
	unsigned sync_frame;	/* when did we find sync ? */
	int buf_len;		/* how full is the buffer ? */
	gchar buf[CSD_BUF_MAX];	/* buffer of sub-chan data - eventually frame then multi-frame aligned */
} csd_analysis;

static void analyze_init(void)
{
	/*
	 * Now destroy the chunk from which the conversation table
	 * structures were allocated.
	 */
	if (csd_analysis_chunk) {
		g_mem_chunk_destroy(csd_analysis_chunk);
		csd_analysis_chunk = NULL;
	}

	/* and create new chunks */
	csd_analysis_chunk = g_mem_chunk_new("csd_analysis_chunk",
			sizeof(struct csd_analysis),
			CSD_STREAM_MAX * sizeof(struct csd_analysis),
			G_ALLOC_AND_FREE);
}

static struct csd_analysis *analyze_csd(packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct csd_analysis *csd=NULL;

	conv = find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);

	if (!conv) {	/* start a new conversation */
		conv=conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
	}

	csd=conversation_get_proto_data(conv, proto_gsm_csd);
	if (csd && csd->end_frame>pinfo->fd->num)	/* if this packet is already in the conversation, return the result */
		return csd;

	if (csd) {	/* Another block in the middle of this connection */
		csd->end_frame = pinfo->fd->num;
	}
	else {
		/* This is a new analysis */
		csd=g_mem_chunk_alloc(csd_analysis_chunk);
		if (csd) {
			csd->state = 0;
			csd->sub_chan_offset = 0;
			csd->bit_rate = 0;
			csd->nt = FALSE;
			csd->v110 = FALSE;
			csd->sub_chan_rate = 0;
			csd->end_frame = pinfo->fd->num;
			csd->buf_len = 0;

			conversation_add_proto_data(conv, proto_gsm_csd, (void *)csd);
		}
	}

	return csd;
}

/* static unsigned char mask[9]={0, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff}; */
static unsigned char mask[9]={0, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};

static int guess_sub_chan_rate(tvbuff_t *tvb)
{
	int i, sub_chan_rate=0;
	unsigned char b=0xff;

	for (i=0; i<tvb_length_remaining(tvb,0); i++)
		b = b & tvb_get_guint8(tvb, i);	/* unused bits are 1s, data is mix of 0s and 1s - b will acquire 0s in the data bits */

	for (i=0; i<=8; i++) {
		if ((mask[i]|b) == 0xff) {	/* when the mask 1s make up for the data 0s, we have the right rate */
			sub_chan_rate = i;
			break;
		}
	}

	return sub_chan_rate;
}

static int find_sync(int sub_chan_rate, tvbuff_t *tvb, int *z)
{
	int v, i, zeros=0, sync=-1;

	/* does not yet cope with sync word crossing TVB boundary */

	for (i=0; i<tvb_length_remaining(tvb,0); i++) {
		v = tvb_get_guint8(tvb, i) & mask[sub_chan_rate];
		if (v==0)
			zeros+=sub_chan_rate;
		else if (zeros >= 8) {	/* end of a run of zeros - enough ? */
			sync = i - (zeros/sub_chan_rate);/* should really check 8 or 16 zeros here - any other is wrong! */
			*z = zeros;
			break;
		}
		else					/* not yet a sync word - start counting zeros again */
			zeros = 0;
	}

	/* should really check for the following ones at the start of every octet here too */

	return sync;	/* octets of the original 64k data to skip to START of sync word */
}

unsigned char get_byte(int sub_chan_rate, tvbuff_t *tvb, int offset)
{
	int byte=0;

	switch(sub_chan_rate)
	{
		case 8:		/* 64kb/s - just copy */
			byte = tvb_get_guint8(tvb,offset);
			break;

		case 4:		/* 32kb/s - 4 bits per octet */
			byte  = (tvb_get_guint8(tvb,offset++) & 0x0f)<<4;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x0f)<<0;
			break;

		case 2:		/* 16kb/s - 2 bits per octet */
			byte  = (tvb_get_guint8(tvb,offset++) & 0x03)<<6;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x03)<<4;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x03)<<2;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x03)<<0;
			break;

		case 1:		/* 8kb/s - 1 bit per octet */
			byte  = (tvb_get_guint8(tvb,offset++) & 0x01)<<7;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<6;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<5;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<4;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<3;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<2;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<1;
			byte |= (tvb_get_guint8(tvb,offset++) & 0x01)<<0;
			break;

		default:
			break;

	}
	return byte;
}

#if 0
static void copy_sub_chan(int sub_chan_rate, tvbuff_t *tvb, int offset, int len, unsigned char *sub_chan)
{
	int i, bytes_per_octet, sub_chan_octets;

	bytes_per_octet = 8/sub_chan_rate;
	sub_chan_octets = len/bytes_per_octet;

	for (i=0; i<sub_chan_octets; i++) {
		*sub_chan++ = get_byte(sub_chan_rate, tvb, offset);
		offset += bytes_per_octet;
	}
	return;
}
#endif

/* Code to actually dissect the packets */
static void
dissect_gsm_csd(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0, sync, zeros;
	struct csd_analysis *csd=NULL;
/*	guint8 buf[CSD_BUF_MAX]; */

/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti;
	proto_tree *gsm_csd_tree;

/* Make entries in Protocol column and Info column on summary display */
	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "GSM-CSD");
	if (tree) {
		ti = proto_tree_add_item(tree, proto_gsm_csd, tvb, 0, -1, FALSE);
		gsm_csd_tree = proto_item_add_subtree(ti, ett_gsm_csd);

		csd = analyze_csd(pinfo);
		if (csd->sub_chan_rate==0)
			csd->sub_chan_rate = guess_sub_chan_rate(tvb);

		if (csd->sub_chan_rate) {	/* only try to find sync once we have the sub-rate evaluated */
			if (csd->state==0) {
				sync = find_sync(csd->sub_chan_rate, tvb, &zeros);
				if (sync>=0) {
					if (zeros ==8) {
						csd->sub_chan_offset = sync;
						csd->sync_frame = pinfo->fd->num;
						csd->state = 1;
						csd->v110 =	TRUE;
						proto_tree_add_text(gsm_csd_tree, tvb, sync, -1, "%d bit sync found at offset %d", zeros, sync);
					}
					else if (zeros == 16) {
						csd->sub_chan_offset = sync;
						csd->sync_frame = pinfo->fd->num;
						csd->state = 1;
						csd->v110 =	FALSE;
						proto_tree_add_text(gsm_csd_tree, tvb, sync, -1, "%d bit sync found at offset %d", zeros, sync);
					}
					else {
						proto_tree_add_text(gsm_csd_tree, tvb, sync, -1, "%d bit sync found at offset %d - ignored", zeros, sync);
						sync = -1;
						csd->sub_chan_rate = 0;	/* if we didnt find sync, dont trust the sub-chan guess - probably noise! */
					}
				}
			}
			if (pinfo->fd->num>=csd->sync_frame && csd->sub_chan_rate)
				proto_tree_add_text(gsm_csd_tree, tvb, offset, -1, "Sub-channel rate %dkb/s", csd->sub_chan_rate*8);
			if (pinfo->fd->num>=csd->sync_frame && csd->state>0)
				proto_tree_add_text(gsm_csd_tree, tvb, csd->sub_chan_offset, -1, "Offset %d", csd->sub_chan_offset);
		}

		/* now transfer the data to the frame-aligned internal buffer until we have a full frame of data */
		if (csd->state==1) { /* prime state buffer */
			tvb_memcpy(tvb, csd->buf, csd->sub_chan_offset, tvb_length_remaining(tvb,csd->sub_chan_offset));
			csd->buf_len += tvb_length_remaining(tvb,csd->sub_chan_offset);
			csd->state = 2;
		}
		else if (csd->state==2) {	/* buffer already primed - finish the old and start the new */
			tvb_memcpy(tvb, &csd->buf[csd->buf_len], offset, csd->sub_chan_offset);
			/* commented out for now for safety - until we handle the growth! */
			/* csd->buf_len += csd->sub_chan_offset; */
		}

		/* check to see if we have a frame of data, and if so, process it */
		if (csd->state==2) {
		/*	if (csd->v110 && ((csd->buf_len*csd->sub_chan_rate)/8 > V110_FRAME_LEN)) {	*/ /* need to evaluate rate to determine the length needed perhaps ? */
		/*	}
			else {
			} */
		}
	}

	return;
}

/* Register the protocol with Ethereal */
/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm_csd(void)
{
	dissector_handle_t gsm_csd_handle;
	static int gsm_csd_prefs_initialized = FALSE;

	gsm_csd_handle = create_dissector_handle(dissect_gsm_csd, proto_gsm_csd);

	if (!gsm_csd_prefs_initialized) {
		gsm_csd_prefs_initialized = TRUE;
	  }
	else {
		if ( dynamic_payload_type > 95 )
			dissector_delete("rtp.pt", dynamic_payload_type, gsm_csd_handle);
	}
	dynamic_payload_type = temp_dynamic_payload_type;

	if ( dynamic_payload_type > 95 ){
		dissector_add("rtp.pt", dynamic_payload_type, gsm_csd_handle);
	}
	dissector_add_string("rtp_dyn_payload_type","GSM-CSD", gsm_csd_handle);
}

/* This format is required because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm_csd(void)
{
	module_t *gsm_csd_module;

	static hf_register_info hf[] = {
		{ &hf_gsm_csd_rate,
			{ "Rate",           "gsm-csd.rate",
			FT_UINT8, BASE_DEC, VALS(gsm_csd_rate_vals), 0x0,
			"CSD Rate", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm_csd,
	};

	/* Register the protocol name and description */
	proto_gsm_csd = proto_register_protocol("GSM Circuit-Switched Data","GSM-CSD", "gsm-csd");

	/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm_csd, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
	/* Register a configuration option for port */

	gsm_csd_module = prefs_register_protocol(proto_gsm_csd, proto_reg_handoff_gsm_csd);

	prefs_register_uint_preference(gsm_csd_module, "dynamic.payload.type",
								   "GSM-CSD dynamic payload type",
								   "The dynamic payload type which will be interpreted as GSM-CSD",
								   10,
								   &temp_dynamic_payload_type);

	/* Init rountines - for conversations */
	register_init_routine(analyze_init);
}
