/* packet-ipa-bscmgmt.c
 * Routines for ip.access BSC Management dissection
 * Copyright 2003, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-ipa-bscmgmt.c 1.6 2008/10/22 22:45:31BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"

#include "ipa-bscmgmt-mib.h"

#define DN_STRING_MAX			132	/* max length of the DN instance strings */
#define INDEX_STRING_MAX		132	/* max length of the structured attribute id strings */

typedef struct bscmgmt_version {
	int					version;
	const value_string *msg_type_vals;
	const value_string *ie_ei_vals;
	const ie_def_t	   *ie_defs;
	const value_string *attr_value_type_vals;
	int				   *hf_bscmgmt_msg_type;
	int				   *hf_bscmgmt_ie_ei;
	int				   *hf_bscmgmt_attr_val_type;
} bscmgmt_version_t;

#define BSCMGMT_V0		0
#define BSCMGMT_V1		1
#define BSCMGMT_V2		2

int dissect_bscmgmt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, int version);
static int dissect_bscmgmt_ie(guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, bscmgmt_version_t *p_version);
static void dissect_bscmgmt_ie_netobjid(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_objinstid(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_sequence_number(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_timestamp(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_attribute(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_attribute_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_attribute_id_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_attribute_failure_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_object_instance_id_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_attribute_id_failure_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_additional_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_additional_text(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_v1_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_v2_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_alarm_flag(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_alarm_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_text_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_action_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_perceived_severity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_probable_cause_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_probable_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static int dissect_bscmgmt_ie_attribute_element(unsigned char *index_str, int attr_id, tvbuff_t *tvb, int offset, int element_len, packet_info *pinfo, proto_tree *tree, bscmgmt_version_t *p_version);
static guint32 dissect_bscmgmt_dn(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned char *inst_str, unsigned inst_str_len);
static void dissect_bscmgmt_ie_timezone_offset(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_bscmgmt_ie_number_of_notifications(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);

/* Initialize the protocol and registered fields */
static int proto_bscmgmt = -1;
static int hf_bscmgmt_v1_msg_type = -1;
static int hf_bscmgmt_v2_msg_type = -1;
static int hf_bscmgmt_obj_class = -1;
static int hf_bscmgmt_v2_obj_class = -1;
static int hf_bscmgmt_v2_obj_dn = -1;
static int hf_bscmgmt_v2_obj_inst = -1;
static int hf_bscmgmt_obj_bsc = -1;
static int hf_bscmgmt_obj_site = -1;
static int hf_bscmgmt_obj_bts = -1;
static int hf_bscmgmt_obj_ts = -1;
static int hf_bscmgmt_obj_trx = -1;
static int hf_bscmgmt_v1_ie_ei = -1;
static int hf_bscmgmt_v2_ie_ei = -1;
static int hf_bscmgmt_ie_len8 = -1;
static int hf_bscmgmt_ie_len16 = -1;
static int hf_bscmgmt_attr = -1;
static int hf_bscmgmt_attr_id = -1;
static int hf_bscmgmt_v1_attr_val_type = -1;
static int hf_bscmgmt_v2_attr_val_type = -1;
static int hf_bscmgmt_attr_val_double = -1;
static int hf_bscmgmt_attr_val_int32 = -1;
static int hf_bscmgmt_attr_val_uint32 = -1;
static int hf_bscmgmt_attr_val_boolean = -1;
static int hf_bscmgmt_attr_val_uint8 = -1;
static int hf_bscmgmt_attr_val_str = -1;
static int hf_bscmgmt_attr_cause = -1;
static int hf_bscmgmt_add_text = -1;
static int hf_bscmgmt_v1_cause = -1;
static int hf_bscmgmt_v2_cause = -1;
static int hf_bscmgmt_alarm_type = -1;
static int hf_bscmgmt_text_type = -1;
static int hf_bscmgmt_action_type = -1;
static int hf_bscmgmt_alarm_flag = -1;
static int hf_bscmgmt_perceived_severity = -1;
static int hf_bscmgmt_probable_cause_type = -1;
static int hf_bscmgmt_probable_cause = -1;
static int hf_bscmgmt_sequence_number = -1;
static int hf_bscmgmt_timestamp = -1;
static int hf_bscmgmt_timezone_offset = -1;
static int hf_bscmgmt_number_of_notifications = -1;

/* Initialize the subtree pointers */
static gint ett_bscmgmt = -1;
static gint ett_bscmgmt_ie = -1;
static gint ett_bscmgmt_dn = -1;

static dissector_handle_t bscmgmt_v0_handle;
static dissector_handle_t bscmgmt_v1_handle;
static dissector_handle_t bscmgmt_v2_handle;
static dissector_handle_t data_handle;

static const value_string msg_type_vals_v1[] = {
	{ 0x01,  "State Change" },
	{ 0x02,  "Alarm" },
	{ 0x03,  "Object Creation" },
	{ 0x04,  "Object Deletion" },
	{ 0x05,  "Text Event" },
	{ 0x06,  "AVC" },
	{ 0x10,  "Action" },
	{ 0x11,  "Action Ack" },
	{ 0x12,  "Action Nack" },
	{ 0x20,  "Create Object" },
	{ 0x21,  "Create Object Ack" },
	{ 0x22,  "Create Object Nack" },
	{ 0x30,  "Delete Object" },
	{ 0x31,  "Delete Object Ack" },
	{ 0x32,  "Delete Object Nack" },
	{ 0x40,  "Set Attributes" },
	{ 0x41,  "Set Attributes Response" },
	{ 0x42,  "Set Attributes Nack" },
	{ 0x50,  "Get Attributes" },
	{ 0x51,  "Get Attributes Response" },
	{ 0x52,  "Get Attributes Nack" },
	{ 0x60,  "Get Objects" },
	{ 0x61,  "Get Objects Response" },
	{ 0x62,  "Get Objects Nack" },
	{ 0x70,  "Get Events" },
	{ 0x71,  "Get Events Response" },
	{ 0x72,  "Get Events Nack" },
	{ 0x80,  "Delete Events" },
	{ 0x81,  "Delete Events Ack" },
	{ 0x82,  "Delete Events Nack" },
	{ 0, NULL }
};

static const value_string msg_type_vals_v2[] = {
	{ 0x01,  "Alarm" },
	{ 0x02,  "AVC Notification" },
	{ 0x03,  "Object Creation Notification" },
	{ 0x04,  "Object Deletion Notification" },
	{ 0x10,  "Action" },
	{ 0x11,  "Action Ack" },
	{ 0x12,  "Action Nack" },
	{ 0x20,  "Create Object" },
	{ 0x21,  "Create Object Ack" },
	{ 0x22,  "Create Object Nack" },
	{ 0x30,  "Delete Object" },
	{ 0x31,  "Delete Object Ack" },
	{ 0x32,  "Delete Object Nack" },
	{ 0x40,  "Set Attributes" },
	{ 0x41,  "Set Attributes Response" },
	{ 0x42,  "Set Attributes Nack" },
	{ 0x50,  "Get Attributes" },
	{ 0x51,  "Get Attributes Response" },
	{ 0x52,  "Get Attributes Nack" },
	{ 0x60,  "Get Objects" },
	{ 0x61,  "Get Objects Response" },
	{ 0x62,  "Get Objects Nack" },
	{ 0x80,  "Get Measurements" },
	{ 0x81,  "Get Measurements Response" },
	{ 0x82,  "Get Measurements Nack" },
	{ 0x83,  "Start Measurements" },
	{ 0x84,  "Start Measurements Ack" },
	{ 0x85,  "Start Measurements Nack" },
	{ 0x86,  "Stop Measurements" },
	{ 0x87,  "Stop Measurements Ack" },
	{ 0x88,  "Stop Measurements Nack" },
	{ 0x90,  "Configuration Request" },
	{ 0x91,  "Synchronisation Request" },
	{ 0x92,  "OpStart" },
	{ 0x93,  "OpStart Ack" },
	{ 0x94,  "OpStart Nack" },
	{ 0xA0,  "Text Event" },
	{ 0, NULL }
};

/* definitions needed for better decode of netObjId */
#define OBJ_CLS_BSS				0
#define OBJ_CLS_BSC				1
#define OBJ_CLS_SITE			2
#define OBJ_CLS_BTS				3
#define OBJ_CLS_RC				4
#define OBJ_CLS_BBT				5
#define OBJ_CLS_TS				6
#define OBJ_CLS_NSE				7
#define OBJ_CLS_CELL			8
#define OBJ_CLS_NSVC			9

#define TEXT_TYPE_LOG_MESSAGE			0x00
#define TEXT_TYPE_CHANNELS				0x01
#define TEXT_TYPE_RADIO_RESOURCES		0x02
#define TEXT_TYPE_SCCP_CONNECTIONS		0x03
#define TEXT_TYPE_MGCP_CONNECTIONS		0x04
#define TEXT_TYPE_OML_CONNECTIONS		0x05
#define TEXT_TYPE_RSL_CONNECTIONS		0x06
#define TEXT_TYPE_MSC_CONNECTION		0x07
#define TEXT_TYPE_OBJECT_TREE			0x08

static const value_string text_type_vals[] = {
	{ TEXT_TYPE_LOG_MESSAGE, "Log Message" },
	{ TEXT_TYPE_CHANNELS, "Channels" },
	{ TEXT_TYPE_RADIO_RESOURCES, "Radio Resources" },
	{ TEXT_TYPE_SCCP_CONNECTIONS, "SCCP Connections" },
	{ TEXT_TYPE_MGCP_CONNECTIONS, "MGCP Connections" },
	{ TEXT_TYPE_OML_CONNECTIONS, "OML" },
	{ TEXT_TYPE_RSL_CONNECTIONS, "RSL Connection" },
	{ TEXT_TYPE_MSC_CONNECTION, "MSC Connection" },
	{ TEXT_TYPE_OBJECT_TREE, "Object Tree" },
	{ 0, NULL }
};

static const value_string ie_ei_vals_v1[] = {
	{ 0x01, "Network Object Id" },
	{ 0x02, "Action Type" },
	{ 0x03, "Timestamp" },
	{ 0x04, "Sequence Number" },
	{ 0x05, "Additional Text" },
	{ 0x06, "Event Flag" },
	{ 0x07, "Alarm Type" },
	{ 0x08, "Alarm Severity" },
	{ 0x09, "Alarm Probable Cause Type" },
	{ 0x0a, "Alarm Probable Cause" },
	{ 0x0b, "Text Type" },
	{ 0x0c, "Uint32 Parameter" },
	{ 0x10, "Attribute" },
	{ 0x11, "Attribute List" },
	{ 0x12, "Attribute Ids List" },
	{ 0x13, "Attribute Failure List" },
	{ 0x14, "Attribute Failure" },
	{ 0x15, "Object List" },
	{ 0x16, "Object Information" },
	{ 0xff, "Nack Cause" },
	{ 0, NULL }
};

static const ie_def_t ie_defs_v1[] = {
	{ 0x01, "Network Object Id", 8, dissect_bscmgmt_ie_netobjid, NULL },
	{ 0x02, "Action Type", 2, dissect_bscmgmt_ie_action_type, NULL },
	{ 0x03, "Timestamp", 4, dissect_bscmgmt_ie_timestamp, NULL },
	{ 0x04, "Sequence Number", 4, dissect_bscmgmt_ie_sequence_number, NULL },
	{ 0x05, "Additional Text", -2, dissect_bscmgmt_ie_additional_text, NULL },
	{ 0x06, "Event Flag", 1, dissect_bscmgmt_ie_alarm_flag, NULL },
	{ 0x07, "Alarm Type", 1, dissect_bscmgmt_ie_alarm_type, NULL },
	{ 0x08, "Alarm Severity", 1, dissect_bscmgmt_ie_perceived_severity, NULL },
	{ 0x09, "Alarm Probable Cause Type", 1, dissect_bscmgmt_ie_probable_cause_type, NULL },
	{ 0x0a, "Alarm Probable Cause", 2, dissect_bscmgmt_ie_probable_cause, NULL },
	{ 0x0b, "Text Type", 1, dissect_bscmgmt_ie_text_type, NULL },
	{ 0x0c, "Uint32 Parameter", 4, NULL, NULL },
	{ 0x10, "Attribute", -2, dissect_bscmgmt_ie_attribute, NULL },
	{ 0x11, "Attribute List", -2, dissect_bscmgmt_ie_list, NULL },
	{ 0x12, "Attribute Ids List", -2, dissect_bscmgmt_ie_attribute_id_list, NULL },
	{ 0x13, "Attribute Failure List", -2, dissect_bscmgmt_ie_list, NULL },
	{ 0x14, "Attribute Failure", -2, NULL, NULL },
	{ 0x15, "Object List", -2, dissect_bscmgmt_ie_list, NULL },
	{ 0x16, "Object Information", -2, NULL, NULL },
	{ 0xff, "Nack Cause", -2, dissect_bscmgmt_v1_ie_cause, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

static const value_string ie_ei_vals_v2[] = {
	{ 0x01, "Object Instance" },
	{ 0x02, "Timestamp" },
	{ 0x03, "Additional Text" },
	{ 0x04, "Additional Info" },
	{ 0x05, "Alarm Flag" },
	{ 0x06, "Alarm Type" },
	{ 0x07, "Percieved Severity" },
	{ 0x08, "Alarm Probable Cause Type" },
	{ 0x09, "Alarm Probable Cause" },
	{ 0x0a, "Action Type" },
	{ 0x0b, "Attribute List" },
	{ 0x0c, "Attribute Id List" },
	{ 0x0d, "Attribute Failure List" },
	{ 0x0e, "Object Instance List" },
	{ 0x0f, "Parameter List" },
	{ 0x10, "Cause" },
	{ 0x11, "Attribute Id Failure List" },
	{ 0x12, "Text Type" },
	{ 0x13, "Number of Notifications" },
	{ 0x14, "Parameter Failure List" },
	{ 0x15, "Parameter Id Failure List" },
	{ 0x16, "Timezone Offset" },
	{ 0, NULL }
};

static const ie_def_t ie_defs_v2[] = {
	{ 0x01, "Object Instance Id", -2, dissect_bscmgmt_ie_objinstid, NULL },
	{ 0x02, "Timestamp", 4, dissect_bscmgmt_ie_timestamp, NULL },
	{ 0x03, "Additional Text", -2, dissect_bscmgmt_ie_additional_text, NULL },
	{ 0x04, "Additional Info", -2, dissect_bscmgmt_ie_additional_info, NULL },
	{ 0x05, "Alarm Flag", 1, dissect_bscmgmt_ie_alarm_flag, NULL },
	{ 0x06, "Alarm Type", 1, dissect_bscmgmt_ie_alarm_type, NULL },
	{ 0x07, "Perceived Severity", 1, dissect_bscmgmt_ie_perceived_severity, NULL },
	{ 0x08, "Alarm Probable Cause Type", 1, dissect_bscmgmt_ie_probable_cause_type, NULL },
	{ 0x09, "Alarm Probable Cause", 2, dissect_bscmgmt_ie_probable_cause, NULL },
	{ 0x0a, "Action Type", 2, dissect_bscmgmt_ie_action_type, NULL },
	{ 0x0b, "Attribute List", -2, dissect_bscmgmt_ie_attribute_list, NULL },
	{ 0x0c, "Attribute Id List", -2, dissect_bscmgmt_ie_attribute_id_list, NULL },
	{ 0x0d, "Attribute Failure List", -2, dissect_bscmgmt_ie_attribute_failure_list, NULL },
	{ 0x0e, "Object Instance Id List", -2, dissect_bscmgmt_ie_object_instance_id_list, NULL },
	{ 0x0f, "Parameter List", -2, dissect_bscmgmt_ie_attribute_list, NULL },
	{ 0x10, "Cause", 2, dissect_bscmgmt_v2_ie_cause, NULL },
	{ 0x11, "Attribute Id Failure List", -2, dissect_bscmgmt_ie_attribute_id_failure_list, NULL },
	{ 0x12, "Text Type", 1, dissect_bscmgmt_ie_text_type, NULL },
	{ 0x13, "Number of Notifications", 4, dissect_bscmgmt_ie_number_of_notifications, NULL },
	{ 0x14, "Parameter Failure List", -2, dissect_bscmgmt_ie_attribute_failure_list, NULL },
	{ 0x15, "Parameter Id Failure List", -2, dissect_bscmgmt_ie_attribute_id_failure_list, NULL },
	{ 0x16, "Timezone Offset", 4, dissect_bscmgmt_ie_timezone_offset, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

static const value_string alarm_flag_vals[] = {
	{ 0, "Normal" },
	{ 1, "Reply" },
	{ 0, NULL }
};

static const value_string attr_val_boolean_vals[] = {
	{ 0, "FALSE" },
	{ 1, "TRUE" },
	{ 0, NULL }
};

static const value_string alarm_type_vals[] = {
	{ 0, "Communications" },
	{ 1, "QoS" },
	{ 2, "Processing" },
	{ 3, "Equipment" },
	{ 4, "Environmental" },
	{ 0, NULL }
};

static const value_string action_type_vals[] = {
	{ 1, "reinitialise" },
	{ 3, "refresh" },
	{ 4, "lock" },
	{ 5, "unlock" },
	{ 6, "shutdown" },
	{ 7, "changePassword" },
	{ 8, "backup" },
	{ 0, NULL }
};

static const value_string perceived_severity_vals[] = {
	{ 0, "Indeterminate" },
	{ 1, "Critical" },
	{ 2, "Major" },
	{ 3, "Minor" },
	{ 4, "Warning" },
	{ 5, "Cleared" },
	{ 0, NULL }
};

static const value_string probable_cause_type_vals[] = {
	{ 1, "X.721" },
	{ 2, "GSM" },
	{ 3, "BTS" },
	{ 4, "BSC" },
	{ 5, "Gateway" },
	{ 6, "FRIP" },
	{ 7, "GNAT" },
	{ 0, NULL }
};

static const value_string probable_cause_vals[] = {
	{ 0, NULL }
};

/* Names for the actual V1 MIB attributes carried by this protocol */
static const value_string attr_id_v1_vals[] = {
	{ 31, "administrativeState" },
	{ 33, "availabilityStatus" },
	{ 35, "operationalState" },
	{ 10310, "objectSequenceNumber" },
	{ 0, NULL }
};

typedef struct {
	int					attr_id;
	const value_string	*vals;
} enum_values;

static const value_string administrative_state_vals[] = {
	{ 0, "LOCKED" },
	{ 1, "UNLOCKED" },
	{ 2, "SHUTTING DOWN" },
	{ 0, NULL }
};

static const value_string operational_state_vals[] = {
	{ 0, "DISABLED" },
	{ 1, "ENABLED" },
	{ 0, NULL }
};

static const value_string availability_status_vals[] = {
	{ 0, "IN TEST" },
	{ 1, "FAILED" },
	{ 2, "POWER OFF" },
	{ 3, "OFF LINE" },
	{ 4, "OFF DUTY" },
	{ 5, "DEPENDENCY" },
	{ 6, "DEGRADED" },
	{ 7, "NOT INSTALLED" },
	{ 8, "LOG FULL" },
	{ 0, NULL }
};

static const value_string administrative_state_v1_vals[] = {
	{ 1, "LOCKED" },
	{ 2, "UNLOCKED" },
	{ 3, "SHUTTING DOWN" },
	{ 0, NULL }
};

static const value_string operational_state_v1_vals[] = {
	{ 1, "DISABLED" },
	{ 2, "ENABLED" },
	{ 0, NULL }
};

static const enum_values enum_vals[] = {
	{ 31, administrative_state_vals },
	{ 35, operational_state_vals },
	{ 33, availability_status_vals },
	{ 0, NULL }
};

static const enum_values enum_v1_vals[] = {
	{ 31, administrative_state_v1_vals },
	{ 35, operational_state_v1_vals },
	{ 33, availability_status_vals },
	{ 0, NULL }
};

/* Names for the Nack causes */
static const value_string cause_vals[] = {
	{ 0, "Success" },
	{ 1, "Object Class Invalid" },
	{ 2, "Object Class Missing" },
	{ 3, "Object Instance does not exist" },
	{ 4, "Object Instance already exists" },
	{ 5, "Missing Mandatory Object Instance" },
	{ 6, "Invalid Object Instance" },
	{ 7, "Maximum number of Object Instances exceeded" },
	{ 8, "Missing Mandatory Attribute" },
	{ 9, "Invalid Attribute Id" },
	{ 10, "Invalid Attribute Value" },
	{ 11, "Invalid Attributes" },
	{ 12, "Contained Objects Exist" },
	{ 13, "Action Invalid" },
	{ 14, "Access not supported" },
	{ 15, "Command not supported" },
	{ 16, "Invalid Unless Locked" },
	{ 17, "Command Invalid Missing Mandatory IE" },
	{ 18, "Command Timeout" },
	{ 19, "BSS not Available" },
	{ 20, "Unexpected Event" },
	{ 21, "Internal Software Error" },
	{ 22, "Malformed DN" },
	{ 65535, "Success" },
	{ 0, NULL }
};


#define V1_ATTR_TYPE_UINT32				0x00
#define V1_ATTR_TYPE_STRING				0x01
#define V1_ATTR_TYPE_SEQ_OF				0x03
#define V1_ATTR_TYPE_ATTRIBUTE			0x04
#define V1_ATTR_TYPE_ARRAY_OF_UINT32	0x05
#define V1_ATTR_TYPE_ARRAY_OF_BYTES		0x06

static const value_string attr_val_type_vals_v1[] = {
	{ V1_ATTR_TYPE_UINT32, "Uint32" },
	{ V1_ATTR_TYPE_STRING, "String" },
	{ V1_ATTR_TYPE_SEQ_OF, "SequenceOf" },
	{ V1_ATTR_TYPE_ATTRIBUTE, "Attribute" },
	{ V1_ATTR_TYPE_ARRAY_OF_UINT32, "Array of Uint32" },
	{ V1_ATTR_TYPE_ARRAY_OF_BYTES, "Array of Bytes" },
	{ 0, NULL }
};

#define V2_ATTR_TYPE_UINT32				0x00
#define V2_ATTR_TYPE_STRING				0x01
#define V2_ATTR_TYPE_ARRAY_OF_ATTR		0x02
#define V2_ATTR_TYPE_STRUCT_ATTR		0x03
#define V2_ATTR_TYPE_DOUBLE				0x04
#define V2_ATTR_TYPE_DN					0x05
#define V2_ATTR_TYPE_BOOLEAN			0x06

static const value_string attr_val_type_vals_v2[] = {
	{ V2_ATTR_TYPE_UINT32, "Int32" },
	{ V2_ATTR_TYPE_STRING, "String" },
	{ V2_ATTR_TYPE_ARRAY_OF_ATTR, "Array" },
	{ V2_ATTR_TYPE_STRUCT_ATTR, "Sequence" },
	{ V2_ATTR_TYPE_DOUBLE, "Double" },
	{ V2_ATTR_TYPE_DN, "DN" },
	{ V2_ATTR_TYPE_DN, "Boolean" },
	{ 0, NULL }
};

static const value_string proto_version_vals[] = {
	{ BSCMGMT_V0, "BSC Mgmt(V0)" },
	{ BSCMGMT_V1, "BSC Mgmt(V1)" },
	{ BSCMGMT_V2, "BSC Mgmt(V2)" },
	{ 0, NULL }
};

bscmgmt_version_t bscmgmt_version[] = {
	{ BSCMGMT_V0, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
	{ BSCMGMT_V1, msg_type_vals_v1, ie_ei_vals_v1, ie_defs_v1, attr_val_type_vals_v1, &hf_bscmgmt_v1_msg_type, &hf_bscmgmt_v1_ie_ei, &hf_bscmgmt_v1_attr_val_type },
	{ BSCMGMT_V2, msg_type_vals_v2, ie_ei_vals_v2, ie_defs_v2, attr_val_type_vals_v2, &hf_bscmgmt_v2_msg_type, &hf_bscmgmt_v1_ie_ei, &hf_bscmgmt_v2_attr_val_type },
};

int
dissect_bscmgmt_v0(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	return dissect_bscmgmt(tvb, pinfo, main_tree, BSCMGMT_V0);
}

int
dissect_bscmgmt_v1(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	return dissect_bscmgmt(tvb, pinfo, main_tree, BSCMGMT_V1);
}

int
dissect_bscmgmt_v2(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	return dissect_bscmgmt(tvb, pinfo, main_tree, BSCMGMT_V2);
}

int
dissect_bscmgmt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, int version)
{
	int offset=0, msg_len;
	guint8 msg_type;
	tvbuff_t *next_tvb;
	bscmgmt_version_t *p_version=&bscmgmt_version[version];

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "bscmgmt";

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_fstr(pinfo->cinfo, COL_PROTOCOL, "%s", rd_match_strval(version, proto_version_vals));

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_bscmgmt, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_bscmgmt);
	}

	if (msg_len>=1) {
		if (p_version->msg_type_vals != NULL) {
			msg_type = tvb_get_guint8(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, *(p_version->hf_bscmgmt_msg_type), tvb, offset, 1, msg_type);
			offset += 1;

			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s(", rd_match_strval(msg_type,p_version->msg_type_vals));
			if (tree)
				proto_item_set_text(ti, "BSCMGMT: %s", rd_match_strval(msg_type,p_version->msg_type_vals));

			/* create a new subset TVB for disecting the payload, and call the correct payload disector */
			next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
			offset += dissect_bscmgmt_ie(msg_type, p_version->ie_defs, next_tvb, pinfo, tree, p_version);

			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, ")");
		}
	}

	return offset;
}

/* Code to dissect the IEs */
static int
dissect_bscmgmt_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, bscmgmt_version_t *p_version)
{
	int offset=0, msg_len;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		ie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, *(p_version->hf_bscmgmt_ie_ei), tvb, offset, 1, ie_ei);
		offset += 1;

		for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==ie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				ie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				ie_len = p_ie_def->ie_len;
				if (tvb_length_remaining(tvb,offset)<ie_len) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len, tvb_length_remaining(tvb,offset));
					ie_len=tvb_length_remaining(tvb,offset);
				}

				if (main_tree) {
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_bscmgmt_ie);
				}
			}
			else if (p_ie_def->ie_len == -1) {
				ie_len8 = tvb_get_guint8(tvb, offset);
				if (tvb_length_remaining(tvb,offset+1)<ie_len8) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len8, tvb_length_remaining(tvb,offset+1));
					ie_len8=tvb_length_remaining(tvb,offset+1);
				}

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_bscmgmt_ie_len8, tvb, offset, 1, ie_len8);
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_bscmgmt_ie);
				}
				offset += 1;
				ie_len = ie_len8;
			}
			else if (p_ie_def->ie_len == -2) {
				ie_len16 = tvb_get_ntohs(tvb, offset);
				if (tvb_length_remaining(tvb,offset+2)<ie_len16) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len16, tvb_length_remaining(tvb,offset+2));
					ie_len16=tvb_length_remaining(tvb,offset+2);
				}

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_bscmgmt_ie_len16, tvb, offset, 2, ie_len16);
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_bscmgmt_ie);
				}
				offset += 2;
				ie_len = ie_len16;
			}
			else {			/* should never happen */
				ie_len=0; /* safety! */
			}

			if (ie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
				if ((p_ie_def->ie_disector!=NULL) && (p_ie_def->info!=NULL))
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, main_tree, p_ie_def->info);
				else if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, main_tree, (void *)p_version);
				else
					call_dissector(data_handle, new_tvb, pinfo, tree);
				offset+=ie_len;
			}

				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {
				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);
				offset += dissect_bscmgmt_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree, p_version);
			}
		}
	}
	return offset;
}

/* Code to dissect the NetObjId IE */
static void dissect_bscmgmt_ie_netobjid(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	guint8 obj_class;
	guint16 obj_site, obj_bsc;
	guint8 obj_bts, obj_ts, obj_trx;
	proto_item *ti=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len==8) {
		obj_class = tvb_get_guint8(tvb, offset);
		if (tree)
			ti = proto_tree_add_uint(tree, hf_bscmgmt_obj_class, tvb, offset, 1, obj_class);
		offset += 1;

		obj_bsc = tvb_get_ntohs(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_bscmgmt_obj_bsc, tvb, offset, 2, obj_bsc);
		offset += 2;

		obj_site = tvb_get_ntohs(tvb, offset);
		if (obj_class!=OBJ_CLS_BSC)
			if (tree)
				proto_tree_add_uint(tree, hf_bscmgmt_obj_site, tvb, offset, 2, obj_site);
		offset += 2;

		obj_bts = tvb_get_guint8(tvb, offset);
		if ((obj_class!=OBJ_CLS_BSC)&&(obj_class!=OBJ_CLS_SITE))
			if (tree)
				proto_tree_add_uint(tree, hf_bscmgmt_obj_bts, tvb, offset, 1, obj_bts);
		offset += 1;

		obj_trx = tvb_get_guint8(tvb, offset);
		if ((obj_class==OBJ_CLS_BBT)||(obj_class==OBJ_CLS_RC)||(obj_class==OBJ_CLS_TS)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
			if (tree)
				proto_tree_add_uint(tree, hf_bscmgmt_obj_trx, tvb, offset, 1, obj_trx);
		offset += 1;

		obj_ts = tvb_get_guint8(tvb, offset);
		if (obj_class==OBJ_CLS_TS)
			if (tree)
				proto_tree_add_uint(tree, hf_bscmgmt_obj_ts, tvb, offset, 1, obj_ts);
		offset += 1;

		if (check_col(pinfo->cinfo, COL_INFO)) {
			if (obj_class==OBJ_CLS_BSC)
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s", rd_match_strval(obj_class,dn_vals));
			else if ((obj_class==OBJ_CLS_SITE))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s_%d", rd_match_strval(obj_class,dn_vals), obj_site);
			else if ((obj_class==OBJ_CLS_BTS)||(obj_class==OBJ_CLS_NSE))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s_%d_%d", rd_match_strval(obj_class,dn_vals), obj_site, obj_bts);
			else if ((obj_class==OBJ_CLS_BBT)||(obj_class==OBJ_CLS_RC)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s_%d_%d_%d", rd_match_strval(obj_class,dn_vals), obj_site, obj_bts, obj_trx);
			else if (obj_class==OBJ_CLS_TS)
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s_%d_%d_%d_%d", rd_match_strval(obj_class,dn_vals), obj_site, obj_bts, obj_trx, obj_ts);
			else /* should not happen */
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s", rd_match_strval(obj_class,dn_vals));
		}
		if (tree) {
			if (obj_class==OBJ_CLS_BSC)
				proto_item_set_text(ti, "BSCMGMT: %s", rd_match_strval(obj_class,dn_vals));
			else if (obj_class==OBJ_CLS_SITE)
				proto_item_set_text(ti, "BSCMGMT: %s_%d", rd_match_strval(obj_class,dn_vals),obj_site);
			else if ((obj_class==OBJ_CLS_BTS)||(obj_class==OBJ_CLS_NSE))
				proto_item_set_text(ti, "BSCMGMT: %s_%d_%d", rd_match_strval(obj_class,dn_vals),obj_site,obj_bts);
			else if ((obj_class==OBJ_CLS_BBT)||(obj_class==OBJ_CLS_RC)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
				proto_item_set_text(ti, "BSCMGMT: %s_%d_%d_%d", rd_match_strval(obj_class,dn_vals),obj_site,obj_bts,obj_trx);
			else if (obj_class==OBJ_CLS_TS)
				proto_item_set_text(ti, "BSCMGMT: %s_%d_%d_%d_%d", rd_match_strval(obj_class,dn_vals),obj_site,obj_bts,obj_trx,obj_ts);
			else /* should not happen */
				proto_item_set_text(ti, "BSCMGMT: %s", rd_match_strval(obj_class,dn_vals));
		}
	}
	else {
	 /* Duff formatted */
		if (tree)
			proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"NetObjId wrong length!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " NetObjId wrong length!!!");
	}

	return;
}

/* Code to dissect the ObjInstId IE */
static void dissect_bscmgmt_ie_objinstid(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	guint32 obj_dn;
	unsigned char inst_str[DN_STRING_MAX];
	proto_item *ti=NULL;
	proto_tree *dn_tree=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		ti = proto_tree_add_text(tree, tvb, 0, msg_len, "");
	}

	dn_tree = proto_item_add_subtree(ti, ett_bscmgmt_dn);

	obj_dn = dissect_bscmgmt_dn(tvb, pinfo, dn_tree, inst_str, sizeof(inst_str));

	if (check_col(pinfo->cinfo, COL_INFO)) {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s%s", rd_match_strval(obj_dn,dn_vals), inst_str);
	}
	if (tree) {
		proto_item_set_text(ti, "DN=%s%s", rd_match_strval(obj_dn,dn_vals), inst_str);
	}

	return;
}

/* Code to dissect the Distinguised Name element */
static guint32 dissect_bscmgmt_dn(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned char *inst_str, unsigned inst_str_len)
{
	int offset=0, msg_len;
	guint32 obj_dn=NULL, obj_inst;
	proto_item *ti=NULL;

	if (inst_str)
		inst_str[0]=0;	/* start with a NULL terminator */

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len==0) {
		if (inst_str && (inst_str_len > 5)) {
			sprintf(&inst_str[strlen(inst_str)], "NULL");
		}
	}

	while (tvb_length_remaining(tvb,offset)>=8) {
		obj_dn = tvb_get_ntohl(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_bscmgmt_v2_obj_dn, tvb, offset, 4, obj_dn);
		offset += 4;

		obj_inst = tvb_get_ntohl(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_bscmgmt_v2_obj_inst, tvb, offset, 4, obj_inst);
		offset += 4;

		if (inst_str && (inst_str_len > strlen(inst_str)+15)) {
			if (obj_inst == 0xffffffff)
				sprintf(&inst_str[strlen(inst_str)], "_NULL");
			else
				sprintf(&inst_str[strlen(inst_str)], "_%d", obj_inst);
		}
		else if (inst_str && (inst_str_len > strlen(inst_str)+1))
			sprintf(&inst_str[strlen(inst_str)], ".");

	}

	return obj_dn;
}

/* Code to dissect the V2 Object Instance Id List IEs */
static void dissect_bscmgmt_ie_object_instance_id_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, start=0;
	proto_item *ti=NULL;
	guint16 dn_len;
	tvbuff_t *new_tvb;
	guint32 obj_class;
	unsigned char inst_str[DN_STRING_MAX];

	while (tvb_length_remaining(tvb,offset)>=2) {
		start = offset;
		dn_len = tvb_get_ntohs(tvb, offset);
		offset += 2;

		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, dn_len, dn_len);

		/* The Attribute List IEs have a payload which is simply a set of Attributes -
		 - use the Attribute (IE) parser */
		obj_class = dissect_bscmgmt_dn(new_tvb, pinfo, tree, inst_str, sizeof(inst_str));
		offset += dn_len;

		if (tree) {
			proto_tree_add_text(tree, tvb, start, offset-start, "%s_%s", rd_match_strval(obj_class,dn_vals), inst_str);
		}
	}

	return;
}

/* Code find hand-coded enum values */
static const value_string *find_enum_vals(int attr_id, int version)
{
	const enum_values	*e;

	if (version==BSCMGMT_V1)
		e=enum_v1_vals;
	else
		e=enum_vals;

	for (; e->vals; e++) {
		if (e->attr_id == attr_id)
			break;
	}

	return e->vals;
}

/* Code to dissect the V1 list IEs */
static void dissect_bscmgmt_ie_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	bscmgmt_version_t *p_version=(bscmgmt_version_t *)info;

	/* The V1 List IEs have a payload which is simply a set of IEs -
	 - recurse into the general message payload (IE) parser */
	dissect_bscmgmt_ie((guint8)msg_type, p_version->ie_defs, tvb, pinfo, tree, p_version);

	return;
}

/* Code to dissect the V2 Attribute List IEs */
static void dissect_bscmgmt_ie_attribute_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	proto_item *ti=NULL;
	guint16 attr_len;
	tvbuff_t *new_tvb;

		/* only recurse to this disector if this ie was decoded OK */
	while (tvb_length_remaining(tvb,offset)>=7) {
		attr_len = tvb_get_ntohs(tvb, offset);
		offset += 2;

		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, attr_len, attr_len);

		/* The Attribute List IEs have a payload which is simply a set of Attributes -
		 - use the Attribute (IE) parser */
		dissect_bscmgmt_ie_attribute(msg_type, new_tvb, pinfo, tree, main_tree, info);
		offset += attr_len;
	}

	return;
}

/* Code to dissect the V2 Attribute Id List IEs */
static void dissect_bscmgmt_ie_attribute_id_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	proto_item *ti=NULL;
	guint32 attr_id;

		/* only recurse to this disector if this ie was decoded OK */
	while (tvb_length_remaining(tvb,offset)>=4) {
		attr_id = tvb_get_ntohl(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_bscmgmt_attr_id, tvb, offset, 4, attr_id);
		offset += 4;
	}

	return;
}

/* Code to dissect the V2 Attribute Failure List IEs */
static void dissect_bscmgmt_ie_attribute_failure_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, str_len, info_len, attr_len;
	proto_item *ti=NULL;
	guint16 cause;
	tvbuff_t *new_tvb;

		/* only recurse to this disector if this ie was decoded OK */
	while (tvb_length_remaining(tvb,offset)>=8) {
		attr_len = tvb_get_ntohs(tvb, offset);
		offset += 2;

		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, attr_len, attr_len);

		/* The Attribute List IEs have a payload which is simply a set of Attributes -
		 - use the Attribute (IE) parser */
		dissect_bscmgmt_ie_attribute(msg_type, new_tvb, pinfo, tree, main_tree, info);
		offset += attr_len;

		cause = tvb_get_ntohs(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_bscmgmt_v2_cause, tvb, offset, 2, cause);
		offset += 2;

		str_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		if (str_len > 0) {
			new_tvb = tvb_new_subset(tvb, offset, str_len, str_len);
			dissect_bscmgmt_ie_additional_text(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
		offset += str_len;

		info_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		if (info_len > 0) {
			new_tvb = tvb_new_subset(tvb, offset, info_len, info_len);
			dissect_bscmgmt_ie_additional_info(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
		offset += info_len;
	}

	return;
}

/* Code to dissect the Attribute IEs */
static void dissect_bscmgmt_ie_attribute(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	int	attr_id;
	guint8 *attr_val_str=0;
	proto_item *ti=NULL;
	bscmgmt_version_t *p_version=(bscmgmt_version_t *)info;

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>=5) {
		attr_id = tvb_get_ntohl(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_id, tvb, offset, 4, attr_id);
		offset += 4;

		offset = dissect_bscmgmt_ie_attribute_element(0, attr_id, tvb, offset, msg_len-offset, pinfo, tree, p_version);
	}
	else {
	 /* Duff formatted or ... FIXME - partial message !!! */
		if (tree)
			proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"Attribute wrong length!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Attribute wrong length!!!");
	}

	return;
}

/* Code to dissect the Attribute Element */
static int dissect_bscmgmt_ie_attribute_element(unsigned char *index_str, int attr_id, tvbuff_t *tvb, int offset, int element_len, packet_info *pinfo, proto_tree *tree, bscmgmt_version_t *p_version)
{
	int i, msg_len, str_len, start=offset;
	int	attr_val_uint32, attr_val_boolean, no_of_elements, new_element_len, element_offset;
	guint8 *attr_val_str=0, attr_type, attr_val_uint8;
	gdouble attr_val_double;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	guint32 obj_dn;
	unsigned char inst_str[DN_STRING_MAX], index_string[INDEX_STRING_MAX];
	const value_string	*vals;

	if (index_str==0) {				/* initialise the indexing string for describing sequences and struct elements */
		index_str=index_string;
		index_str[0] = 0;			/* null terminate ready for use */
	}

	vals = find_enum_vals(attr_id, p_version->version);

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>=1) {
		attr_type = tvb_get_guint8(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, *(p_version->hf_bscmgmt_attr_val_type), tvb, offset, 1, attr_type);
		offset += 1;
		element_len -= 1;	/* and reduce the remaining element length by the same */

		switch ((p_version->version << 8) | attr_type) {

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_UINT32:
			attr_val_uint32 = tvb_get_ntohl(tvb, offset);
			if (tree) {
				proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_val_uint32, tvb, offset, 4, attr_val_uint32);
				if (vals)
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s%s=%s (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),index_str,rd_match_strval(attr_val_uint32,vals),attr_val_uint32);
				else
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s%s=%d (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),index_str,attr_val_uint32,attr_val_uint32);
			}
			offset += 4;
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_UINT32:
			attr_val_uint32 = tvb_get_ntohl(tvb, offset);
			if (tree) {
				proto_tree_add_int_hidden(tree, hf_bscmgmt_attr_val_int32, tvb, offset, 4, attr_val_uint32);
				if (vals)
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s%s=%s (0x%x)", rd_match_strval(attr_id,attr_id_vals),index_str,rd_match_strval(attr_val_uint32,vals),attr_val_uint32);
				else
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s%s=%d (0x%x)", rd_match_strval(attr_id,attr_id_vals),index_str,attr_val_uint32,attr_val_uint32);
			}
			offset += 4;
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_BOOLEAN:
			attr_val_boolean = tvb_get_ntohl(tvb, offset);
			if (tree) {
				proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_val_boolean, tvb, offset, 4, attr_val_boolean);
				proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s%s=%s", rd_match_strval(attr_id,attr_id_vals),index_str,attr_val_boolean?"TRUE":"FALSE");
			}
			offset += 4;
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_DOUBLE:
			attr_val_double = tvb_get_ntohieee_double(tvb, offset);
			if (tree) {
				proto_tree_add_double_hidden(tree, hf_bscmgmt_attr_val_double, tvb, offset, 8, attr_val_double);
				proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 8, "%s%s=%e", rd_match_strval(attr_id,attr_id_vals),index_str,attr_val_double);
			}
			offset += 8;
			break;

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_STRING:
		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_STRING:
			str_len = element_len;
			attr_val_str = malloc(str_len+1);
			attr_val_str[str_len] = 0;						/* cope with no NULL terminator on the original !!! */
			tvb_memcpy(tvb, attr_val_str, offset, str_len);	/* copy remaining bytes as string value to display */
			if (tree) {
				proto_tree_add_string_hidden(tree, hf_bscmgmt_attr_val_str, tvb, offset, str_len, attr_val_str);
				proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, str_len, "%s%s=\042%s\042", rd_match_strval(attr_id,attr_id_vals),index_str,attr_val_str);
			}
			offset += str_len;
			free(attr_val_str);
			attr_val_str =  0;
			break;

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_SEQ_OF:
			no_of_elements = tvb_get_ntohs(tvb, offset);
			offset += 2;
			str_len=strlen(index_str);
			for (i=0; i<no_of_elements; i++) {
				sprintf(&index_str[str_len],"[%d]",i);
				offset = dissect_bscmgmt_ie_attribute_element(index_str, attr_id, tvb, offset, element_len/no_of_elements, pinfo, tree, p_version);
			}
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_ARRAY_OF_ATTR:
			attr_type = tvb_get_guint8(tvb, offset);	/* the type of the array elements */
			if (tree)
				proto_tree_add_uint_hidden(tree, *(p_version->hf_bscmgmt_attr_val_type), tvb, offset, 1, attr_type);
			offset += 1;
			i=0;
			str_len=strlen(index_str);
			while (tvb_length_remaining(tvb,offset)>1) {
				sprintf(&index_str[str_len],"[%d]",i);
				new_element_len = tvb_get_ntohs(tvb, offset);
				offset += 2;
				offset = dissect_bscmgmt_ie_attribute_element(index_str, attr_id, tvb, offset, new_element_len, pinfo, tree, p_version);
				i++;
			}
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_DN:
			/* use a new subset TVB for the recursion */
			new_tvb = tvb_new_subset(tvb, offset, element_len, element_len);

			obj_dn = dissect_bscmgmt_dn(new_tvb, pinfo, (proto_tree *)NULL, inst_str, sizeof(inst_str));
			if (tree) {
				if (obj_dn==NULL)
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, element_len, "%s%s=""%s""", rd_match_strval(attr_id,attr_id_vals),index_str,"NULL");
				else
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, element_len, "%s%s=""%s%s""", rd_match_strval(attr_id,attr_id_vals),index_str,rd_match_strval(obj_dn,dn_vals),inst_str);
			}
			offset += element_len;
			break;

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_ATTRIBUTE:
			attr_id = tvb_get_ntohl(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_bscmgmt_attr_id, tvb, offset, 4, attr_id);
			offset += 4;

			offset = dissect_bscmgmt_ie_attribute_element(index_str, attr_id, tvb, offset, element_len-5, pinfo, tree, p_version);
			break;

		case (BSCMGMT_V2 << 8) | V2_ATTR_TYPE_STRUCT_ATTR:
			i=0;
			str_len=strlen(index_str);
			element_offset=offset;
			if (element_len-(offset-element_offset)==0) {	/* empty sets still need some display */
				proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 0, "%s%s=[]", rd_match_strval(attr_id,attr_id_vals),index_str);
			}
			else {
				while (element_len-(offset-element_offset)>2) {
					new_element_len = tvb_get_ntohs(tvb, offset);
					offset += 2;
					sprintf(&index_str[str_len],".%d",i);

					offset = dissect_bscmgmt_ie_attribute_element(index_str, attr_id, tvb, offset, new_element_len, pinfo, tree, p_version);
					i++;
				}
			}
			break;

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_ARRAY_OF_UINT32:
			if (tvb_length_remaining(tvb,offset)==0) {
				if (tree)
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset-4, 4, "%s=<>", rd_match_strval(attr_id,attr_id_v1_vals));
			}
			for (i=0; tvb_length_remaining(tvb,offset)>0; offset+=4,i++) {
				attr_val_uint32 = tvb_get_ntohl(tvb, offset);
				if (tree) {
					ti = proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_val_int32, tvb, offset, 4, attr_val_uint32);
					if (vals)
						proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s[%d]=%s (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),i,rd_match_strval(attr_val_uint32,vals),attr_val_uint32);
					else
						proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 4, "%s[%d]=%d (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),i,attr_val_uint32,attr_val_uint32);
				}
			}
			break;

		case (BSCMGMT_V1 << 8) | V1_ATTR_TYPE_ARRAY_OF_BYTES:
			if (tvb_length_remaining(tvb,offset)==0) {
				if (tree)
					proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset-4, 4, "%s=<>", rd_match_strval(attr_id,attr_id_v1_vals));
			}
			for (i=0; tvb_length_remaining(tvb,offset)>0; offset+=1,i++) {
				attr_val_uint8 = tvb_get_guint8(tvb, offset);
				if (tree) {
					ti = proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_val_uint8, tvb, offset, 1, attr_val_uint8);
					if (vals)
						proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 1, "%s[%d]=%s (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),i,rd_match_strval(attr_val_uint8,vals),attr_val_uint8);
					else
						proto_tree_add_none_format(tree, hf_bscmgmt_attr, tvb, offset, 1, "%s[%d]=%d (0x%x)", rd_match_strval(attr_id,attr_id_v1_vals),i,attr_val_uint8,attr_val_uint8);
				}
			}
			break;

		default: /* dont know how to treat this - opaque data */
			if (tvb_length_remaining(tvb,offset)>0) {
				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);
				call_dissector(data_handle, new_tvb, pinfo, tree);
			}
			offset += element_len - 1;
			break;
		}
	}
	else {
	 /* Duff formatted or ... FIXME - partial message !!! */
		if (tree)
			proto_tree_add_text(tree, tvb, offset, tvb_length_remaining(tvb,offset),"Attribute wrong length!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Attribute wrong length!!!");
	}

	return offset;
}

/* Code to dissect the Attribute Id Failure IEs */
static void dissect_bscmgmt_ie_attribute_id_failure_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len, str_len, info_len, start;
	int	attr_id, cause;
	tvbuff_t *new_tvb;

	msg_len = tvb_length_remaining(tvb,offset);

	while (tvb_length_remaining(tvb,offset)>=8) {
		start = offset;

		attr_id = tvb_get_ntohl(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, hf_bscmgmt_attr_id, tvb, offset, 4, attr_id);
		offset += 4;

		cause = tvb_get_ntohs(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, hf_bscmgmt_v2_cause, tvb, offset, 2, cause);
		offset += 2;
		proto_tree_add_text(tree, tvb, start, offset-start,"%s: %s", rd_match_strval(attr_id,attr_id_vals), rd_match_strval(cause,cause_vals));

		str_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		if (str_len > 0) {
			new_tvb = tvb_new_subset(tvb, offset, str_len, str_len);
			dissect_bscmgmt_ie_additional_text(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
		offset += str_len;

		info_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		if (info_len > 0) {
			new_tvb = tvb_new_subset(tvb, offset, info_len, info_len);
			dissect_bscmgmt_ie_additional_info(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
		offset += info_len;
	}

	return;
}

/* Code to dissect the Additional Info IEs */
static void dissect_bscmgmt_ie_additional_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	call_dissector(data_handle, tvb, pinfo, tree);

	return;
}

/* Code to dissect the Additional Text IEs */
static void dissect_bscmgmt_ie_additional_text(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_add_text, tvb, offset, msg_len, FALSE);
	}
	offset += msg_len;

	return;
}

/* Code to dissect the V1 Cause IEs */
static void dissect_bscmgmt_v1_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_v1_cause, tvb, offset, 4, FALSE);
	}
	offset += 4;

	return;
}

/* Code to dissect the V2 Cause IEs */
static void dissect_bscmgmt_v2_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_v2_cause, tvb, offset, 2, FALSE);
	}
	offset += 2;

	return;
}

/* Code to dissect the Alarm Type IEs */
static void dissect_bscmgmt_ie_alarm_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_alarm_type, tvb, offset, 1, FALSE);
	}
	offset += 1;

	return;
}

/* Code to dissect the Text Type IEs */
static void dissect_bscmgmt_ie_text_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_text_type, tvb, offset, 1, FALSE);
	}
	offset += 1;

	return;
}

/* Code to dissect the Action Type IEs */
static void dissect_bscmgmt_ie_action_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_action_type, tvb, offset, 2, FALSE);
	}
	offset += 2;

	return;
}

/* Code to dissect the Alarm Flag IEs */
static void dissect_bscmgmt_ie_alarm_flag(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_alarm_flag, tvb, offset, 1, FALSE);
	}
	offset += 1;

	return;
}

/* Code to dissect the Perceived Severity IEs */
static void dissect_bscmgmt_ie_perceived_severity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_perceived_severity, tvb, offset, 1, FALSE);
	}
	offset += 1;

	return;
}

/* Code to dissect the Probable Cause Type IEs */
static void dissect_bscmgmt_ie_probable_cause_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_probable_cause_type, tvb, offset, 1, FALSE);
	}
	offset += 1;

	return;
}

/* Code to dissect the Probable Cause IEs */
static void dissect_bscmgmt_ie_probable_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_probable_cause, tvb, offset, 2, FALSE);
	}
	offset += 2;

	return;
}

/* Code to dissect the Sequence Number IEs */
static void dissect_bscmgmt_ie_sequence_number(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_bscmgmt_sequence_number, tvb, offset, 4, FALSE);
	}
	offset += 4;

	return;
}

/* Code to dissect the Timestamp IEs */
static void dissect_bscmgmt_ie_timestamp(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	time_t	timestamp;
	struct tm *utc;
	proto_item *ti=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	timestamp = tvb_get_ntohl(tvb, offset);
	utc = gmtime( &timestamp );
	if (tree) {
		proto_tree_add_uint_hidden(tree, hf_bscmgmt_timestamp, tvb, offset, 4, FALSE);
		proto_tree_add_text(tree, tvb, offset, 4,"%s", asctime(utc));
	}
	offset += 4;

	return;
}

/* Code to dissect the Timezone Offset IEs */
static void dissect_bscmgmt_ie_timezone_offset(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	long timezone_offset;
	proto_item *ti=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	timezone_offset = tvb_get_ntohl(tvb, offset);
	if (tree) {
		proto_tree_add_int_hidden(tree, hf_bscmgmt_timezone_offset, tvb, offset, 4, FALSE);
		proto_tree_add_text(tree, tvb, offset, 4,"%dh %dm East of UTC", timezone_offset/60, timezone_offset-60*(timezone_offset/60));
	}
	offset += 4;

	return;
}

/* Code to dissect the Number of Notifications IEs */
static void dissect_bscmgmt_ie_number_of_notifications(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	unsigned long number;
	proto_item *ti=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	number = tvb_get_ntohl(tvb, offset);
	if (tree) {
		proto_tree_add_uint(tree, hf_bscmgmt_number_of_notifications, tvb, offset, 4, FALSE);
	}
	offset += 4;

	return;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_bscmgmt(void)
{

    /* Setup list of header fields */
	static hf_register_info hf[] = {
		{ &hf_bscmgmt_v1_msg_type,
			{ "Msg Type", "bscmgmt.v1.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals_v1), 0x0,
			"Message Type", HFILL }
		},
		{ &hf_bscmgmt_v2_msg_type,
			{ "Msg Type", "bscmgmt.v2.msg_type",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"Message Type", HFILL }
		},
		{ &hf_bscmgmt_obj_class,
			{ "Obj Class", "bscmgmt.v1.obj_class",
			FT_UINT8, BASE_HEX, VALS(dn_vals), 0x0,
			"Object Class", HFILL }
		},
		{ &hf_bscmgmt_v2_obj_dn,
			{ "Obj Dn", "bscmgmt.v2.obj_dn",
			FT_UINT8, BASE_HEX, VALS(dn_vals), 0x0,
			"Object Distinguished Name", HFILL }
		},
		{ &hf_bscmgmt_v2_obj_class,
			{ "Obj Class", "bscmgmt.v2.obj_class",
			FT_UINT32, BASE_HEX, VALS(dn_vals), 0x0,
			"Object Class", HFILL }
		},
		{ &hf_bscmgmt_v2_obj_inst,
			{ "Obj Inst", "bscmgmt.v2.obj_instance",
			FT_UINT32, BASE_HEX, NULL, 0x0,
			"Object Instance", HFILL }
		},
		{ &hf_bscmgmt_obj_bsc,
			{ "BSC", "bscmgmt.v1.obj_bsc",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Object BSC", HFILL }
		},
		{ &hf_bscmgmt_obj_site,
			{ "Site", "bscmgmt.v1.obj_site",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Object Site", HFILL }
		},
		{ &hf_bscmgmt_obj_bts,
			{ "BTS", "bscmgmt.v1.obj_bts",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object BTS", HFILL }
		},
		{ &hf_bscmgmt_obj_trx,
			{ "TRX", "bscmgmt.v1.obj_trx",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object Transceiver", HFILL }
		},
		{ &hf_bscmgmt_obj_ts,
			{ "Chan", "bscmgmt.v1.obj_ts",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object Timeslot", HFILL }
		},
		{ &hf_bscmgmt_v1_ie_ei,
			{ "IE", "bscmgmt.v1.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ie_ei_vals_v1), 0x0,
			"IE Element Identifier", HFILL }
		},
		{ &hf_bscmgmt_v2_ie_ei,
			{ "IE", "bscmgmt.v2.ie.ei",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"IE Element Identifier", HFILL }
		},
		{ &hf_bscmgmt_ie_len8,
			{ "IE Len", "bscmgmt.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_bscmgmt_ie_len16,
			{ "IE Len", "bscmgmt.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_bscmgmt_attr,
			{ "Attr", "bscmgmt.ie.attr",
			FT_NONE, NULL, NULL, 0x0,
			"Attribute", HFILL }
		},
		{ &hf_bscmgmt_attr_id,
			{ "Attr Id", "bscmgmt.ie.attr.id",
			FT_UINT32, BASE_DEC, attr_id_vals, 0x0,
			"Attribute Identifier", HFILL }
		},
		{ &hf_bscmgmt_v1_attr_val_type,
			{ "Attr Val Type", "bscmgmt.v1.ie.attr.val.type",
			FT_UINT8, BASE_DEC, attr_val_type_vals_v1, 0x0,
			"Attribute Value Type", HFILL }
		},
		{ &hf_bscmgmt_v2_attr_val_type,
			{ "Attr Val Type", "bscmgmt.v2.ie.attr.val.type",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Attribute Value Type", HFILL }
		},
		{ &hf_bscmgmt_attr_val_int32,
			{ "Attr Val Int32", "bscmgmt.ie.attr.val.int32",
			FT_INT32, BASE_DEC, NULL, 0x0,
			"Attribute Value Int32", HFILL }
		},
		{ &hf_bscmgmt_attr_val_uint32,
			{ "Attr Val UInt32", "bscmgmt.ie.attr.val.uint32",
			FT_UINT32, BASE_HEX, NULL, 0x0,
			"Attribute Value UInt32", HFILL }
		},
		{ &hf_bscmgmt_attr_val_boolean,
			{ "Attr Val Boolean", "bscmgmt.ie.attr.val.boolean",
			FT_UINT32, BASE_HEX, attr_val_boolean_vals, 0x0,
			"Attribute Value Boolean", HFILL }
		},
		{ &hf_bscmgmt_attr_val_double,
			{ "Attr Val Dble", "bscmgmt.ie.attr.val.double",
			FT_DOUBLE, BASE_HEX, NULL, 0x0,
			"Attribute Value Double", HFILL }
		},
		{ &hf_bscmgmt_attr_val_uint8,
			{ "Attr Val Uint8", "bscmgmt.ie.attr.val.uint8",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"Attribute Value Uint8", HFILL }
		},
		{ &hf_bscmgmt_attr_val_str,
			{ "Attr Val Str", "bscmgmt.ie.attr.val.str",
			FT_STRING, BASE_HEX, NULL, 0x0,
			"Attribute Value String", HFILL }
		},
		{ &hf_bscmgmt_attr_cause,
			{ "Attr Cause", "bscmgmt.ie.attr.cause",
			FT_UINT32, BASE_HEX, cause_vals, 0x0,
			"Attribute Failure Cause", HFILL }
		},
		{ &hf_bscmgmt_add_text,
			{ "Additional Text", "bscmgmt.ie.add_text",
			FT_STRING, BASE_HEX, NULL, 0x0,
			"Additional Text", HFILL }
		},
		{ &hf_bscmgmt_v1_cause,
			{ "Cause", "bscmgmt.v1.ie.cause",
			FT_UINT32, BASE_HEX, cause_vals, 0x0,
			"Cause", HFILL }
		},
		{ &hf_bscmgmt_v2_cause,
			{ "Cause", "bscmgmt.v2.ie.cause",
			FT_UINT16, BASE_HEX, cause_vals, 0x0,
			"Cause", HFILL }
		},
		{ &hf_bscmgmt_alarm_type,
			{ "AlarmType", "bscmgmt.ie.alarm_type",
			FT_UINT8, BASE_DEC, alarm_type_vals, 0x0,
			"Alarm Type", HFILL }
		},
		{ &hf_bscmgmt_text_type,
			{ "TextType", "bscmgmt.ie.text_type",
			FT_UINT8, BASE_DEC, text_type_vals, 0x0,
			"Text Type", HFILL }
		},
		{ &hf_bscmgmt_action_type,
			{ "ActionType", "bscmgmt.ie.action_type",
			FT_UINT32, BASE_DEC, action_type_vals, 0x0,
			"Action Type", HFILL }
		},
		{ &hf_bscmgmt_alarm_flag,
			{ "AlarmFlag", "bscmgmt.ie.alarm_flag",
			FT_UINT8, BASE_DEC, alarm_flag_vals, 0x0,
			"Alarm Flag", HFILL }
		},
		{ &hf_bscmgmt_perceived_severity,
			{ "PerceivedSeverity", "bscmgmt.ie.perceived_severity",
			FT_UINT8, BASE_DEC, perceived_severity_vals, 0x0,
			"Perceived Severity", HFILL }
		},
		{ &hf_bscmgmt_probable_cause_type,
			{ "ProbableCauseType", "bscmgmt.ie.probable_cause_type",
			FT_UINT8, BASE_DEC, probable_cause_type_vals, 0x0,
			"Probable Cause Type", HFILL }
		},
		{ &hf_bscmgmt_probable_cause,
			{ "ProbableCause", "bscmgmt.ie.probable_cause",
			FT_UINT16, BASE_DEC, probable_cause_vals, 0x0,
			"Probable Cause", HFILL }
		},
		{ &hf_bscmgmt_sequence_number,
			{ "SequenceNumber", "bscmgmt.ie.sequence_number",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Sequence Number", HFILL }
		},
		{ &hf_bscmgmt_timestamp,
			{ "Timestamp", "bscmgmt.ie.timestamp",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Timestamp", HFILL }
		},
		{ &hf_bscmgmt_timezone_offset,
			{ "TimezoneOffset", "bscmgmt.ie.timezone_offset",
			FT_INT32, BASE_DEC, NULL, 0x0,
			"Timezone Offset", HFILL }
		},
		{ &hf_bscmgmt_number_of_notifications,
			{ "NoOfNotifications", "bscmgmt.ie.number_of_notifications",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"Number of Notifications", HFILL }
		},
	};

    /* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_bscmgmt,
		&ett_bscmgmt_ie,
		&ett_bscmgmt_dn,
	};

	proto_bscmgmt = proto_register_protocol("ip.access BSC Management Layer 3", "BSC Mgmt", "bscmgmt");
    register_dissector("bscmgmt_v0", dissect_bscmgmt_v0, proto_bscmgmt);
    register_dissector("bscmgmt_v1", dissect_bscmgmt_v1, proto_bscmgmt);
    register_dissector("bscmgmt_v2", dissect_bscmgmt_v2, proto_bscmgmt);

    /* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_bscmgmt, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}

/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.

   We use a Heuristic registration if the TCP port number isn't fixed
   but if there is a defined port number register that instead
*/
void
proto_reg_handoff_bscmgmt(void)
{
    static int initialized = FALSE;

    if (!initialized) {
        bscmgmt_v0_handle = find_dissector("bscmgmt_v0");
        bscmgmt_v1_handle = find_dissector("bscmgmt_v1");
        bscmgmt_v2_handle = find_dissector("bscmgmt_v2");

        data_handle = find_dissector("data");

    	dissector_add("ipa.stream", BSC_MGMT_STREAM, bscmgmt_v1_handle);
	    dissector_add("ipa.stream", BSC_SUBAGENT_STREAM, bscmgmt_v2_handle);

        initialized = TRUE;
	}
}

