/* packet-ipa-iml.c
 * Routines for ip.access BTS Internal Management Link dissection
 * Copyright 2005, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-ipa-iml.c 1.1 2006/07/13 18:50:51BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"
#include "tap.h"

static void dissect_ipa_iml_ie(guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree);

/* Initialize the protocol and registered fields */
static int proto_ipa_iml = -1;
static int hf_ipa_iml_msg_type = -1;
static int hf_ipa_iml_ie_ei = -1;
static int hf_ipa_iml_ie_len8 = -1;
static int hf_ipa_iml_ie_len16 = -1;

/* Initialize the subtree pointers */
static gint ett_ipa_iml = -1;
static gint ett_ipa_iml_ie = -1;

static dissector_handle_t ipa_iml_handle;
static dissector_handle_t data_handle;

/* Value Strings */
static const value_string msg_type_vals[] = {
	{ 0, NULL }
};

static const value_string ie_ei_vals[] = {
	{ 0, NULL }
};


static const ie_def_t ie_defs[] = {
	{ 0, NULL, 0, NULL, NULL }
};


/* Code to dissect the iml packets */
void
dissect_ipa_iml(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type;
	tvbuff_t *next_tvb;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "ipa_iml";
	
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "iml");

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_ipa_iml, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_ipa_iml);
	}

	if (msg_len>=1) {
		msg_type = tvb_get_guint8(tvb, offset);
		if (tree)
			proto_tree_add_uint(tree, hf_ipa_iml_msg_type, tvb, offset, 1, msg_type);
		offset += 1;
		if (tree)
			proto_item_set_text(ti, " iml: %s", rd_match_strval(msg_type, msg_type_vals));

		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, "iml:%s",rd_match_strval(msg_type,msg_type_vals));

		next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);

		dissect_ipa_iml_ie(msg_type, ie_defs, next_tvb, pinfo, tree, main_tree);

	}
	else {
	 /* Duff or ... FIXME - partial message !!! */
		if (main_tree)
			proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"Short Message!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Short Message!!!");
	}
	return;
}

/* Code to dissect the IEs */
void
dissect_ipa_iml_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree)
{
	int offset=0, msg_len, i;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		ie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_ipa_iml_ie_ei, tvb, offset, 1, ie_ei);
		offset += 1;

		for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==ie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				ie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				ie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)(p_ie_def->info)));
						for (i=2; i<ie_len+1; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_iml_ie);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				ie_len8 = tvb_get_guint8(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ipa_iml_ie_len8, tvb, offset, 1, ie_len8);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s:", p_ie_def->ie_name);
						for (i=2; i<ie_len8+2; i++) {
							if (i==2)
								proto_item_append_text(ti, " %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
							else
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_iml_ie);
					}
				}
				offset += 1;
				ie_len = ie_len8;
			}
			else if (p_ie_def->ie_len == -2) {
				ie_len16 = tvb_get_ntohs(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ipa_iml_ie_len16, tvb, offset, 2, ie_len16);
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s:", p_ie_def->ie_name);
						for (i=3; i<ie_len16+3; i++) {
							if (i==3)
								proto_item_append_text(ti, " %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
							else
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_iml_ie);
					}
				}
				offset += 2;
				ie_len = ie_len16;
			}
			else {			/* should never happen */
				ie_len=0; /* safety! */
			}

			if (ie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle, new_tvb, pinfo, tree);
				offset+=ie_len;
			}
		
				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);

				dissect_ipa_iml_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree, top_tree);
			}
		}
	}
	return;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_ipa_iml(void)
{		 

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_ipa_iml_msg_type,
			{ "Msg Type", "ipa_iml.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,	  
			"Message Type", HFILL }
		},
		{ &hf_ipa_iml_ie_ei,
			{ "IE", "ipa_iml.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ie_ei_vals), 0x0,	  
			"IE Element Identifier", HFILL }
		},
		{ &hf_ipa_iml_ie_len8,
			{ "IE Len", "ipa_iml.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,	  
			"IE Length", HFILL }
		},
		{ &hf_ipa_iml_ie_len16,
			{ "IE Len", "ipa_iml.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,	  
			"IE Length", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_ipa_iml,
		&ett_ipa_iml_ie,
	};

	proto_ipa_iml = proto_register_protocol("ip.access Internal Management Link (IML)", "IML", "ipa_iml");

	register_dissector("ipa_iml", dissect_ipa_iml, proto_ipa_iml);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_ipa_iml, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.

   We use a Heuristic registration if the TCP port number isn't fixed
   but if there is a defined port number register that instead
*/
void
proto_reg_handoff_ipa_iml(void)
{
    ipa_iml_handle = find_dissector("ipa_iml");
    data_handle = find_dissector("data");
}
