/* packet-ipa.c
 * Routines for ip.access generic A-bis dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-ipa.c 1.4 2008/06/06 05:11:43BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/emem.h>
#include <epan/packet.h>
#include <epan/conversation.h>
#include <epan/prefs.h>

#include "packet-gsm.h"

/* function prototypes for this file */
void proto_reg_handoff_ipa(void);
static int dissect_ipa(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static int dissect_ipa_transport(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *main_tree, int *desegment);
static void dissect_ipa_ctrl(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void dissect_ipa_ctrl_ie(guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void dissect_ie_identity_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_ssl_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);

/* Constants used to validate message lengths and in the heuristic checker */
#define MSG_LEN_MIN		  3
#define MSG_LEN_MAX	  16384

/* Constants for the saved length field in the conversation data used in the heuristic checker */
#define CONV_IN_PROGRESS	((void *)-1)
#define CONV_AWAIT_LEN		((void *)0)

/* Initialize the protocol and registered fields */
static int proto_ipa = -1;
static int hf_ipa_msg_len = -1;
static int hf_ipa_stream = -1;
static int hf_ipa_ctrl_msg_type = -1;
static int hf_ipa_ctrl_ie_ei = -1;
static int hf_ipa_ctrl_ie_len8 = -1;
static int hf_ipa_ctrl_ie_len16 = -1;
static int hf_ipa_ctrl_heartbeat_rtt = -1;
static int hf_ipa_ctrl_heartbeat_peak_rtt = -1;
static int hf_ipa_ctrl_heartbeat_avrg_rtt = -1;
static int hf_ipa_ctrl_heartbeat_frame = -1;

static dissector_table_t subdissector_table;

static dissector_handle_t ipa_handle;
static dissector_handle_t data_handle;
static dissector_handle_t sccp_handle;
static dissector_handle_t gsm0858_handle;
static dissector_handle_t gprs_ns_handle;
static dissector_handle_t gsm0816_handle;
static dissector_handle_t bscmgmt_v0_handle;
static dissector_handle_t bscmgmt_v1_handle;
static dissector_handle_t bscmgmt_v2_handle;
static dissector_handle_t ipa_spirit_handle;

/* Initialize the subtree pointers */
static gint ett_ipa = -1;
static gint ett_ipa_ctrl = -1;
static gint ett_ipa_ctrl_ie = -1;

/* Global preferences */
static gboolean global_gprs_use_ipa = FALSE;
static gint  global_gprs_start_port = 23000;
static gint  global_gprs_end_port = 23003;
static gint  global_oml_port = TCP_PORT_IPA_OML;
static gint  global_rsl_port = TCP_PORT_IPA_RSL;

/* Value Strings */

static const value_string ctrl_msg_type_vals[] = {
	{ 0x00, "Heartbeat" },
	{ 0x01, "Heartbeat Ack" },
	{ 0x04, "Identity Request" },
	{ 0x05, "Identity" },
	{ 0x06, "Identity Ack" },
	{ 0x07, "Identity Nack" },
	{ 0x08, "Proxy" },
	{ 0x09, "Proxy Ack" },
	{ 0x0a, "Proxy Nack" },
	{ 0x0b, "SSL Info" },
	{ 0, NULL }
};
#define IPA_HEARTBEAT		0x00
#define IPA_HEARTBEAT_ACK	0x01

static const value_string identity_type_vals[] = {
	{ 0x00, "Unit Serial Number" },
	{ 0x01, "Unit Name" },
	{ 0x02, "Unit Location" },
	{ 0x03, "Unit Type" },
	{ 0x04, "Hardware Version" },
	{ 0x05, "Software Version" },
	{ 0x06, "IP Address" },
	{ 0x07, "Ethernet Address" },
	{ 0x08, "Unit Id" },
	{ 0x09, "User Name" },
	{ 0x0a, "Password" },
	{ 0x0b, "Access Class" },
	{ 0x0c, "Application Protocol Version" },
	{ 0, NULL }
};

static const value_string ssl_identity_type_vals[] = {
	{ 0x01, "Certificate Serial Number" },
	{ 0x02, "Certificate Subject" },
	{ 0x03, "IP Address" },
	{ 0x04, "Port" },
	{ 0, NULL }
};

static const value_string ssl_status_vals[] = {
	{ 0x01, "Encrypted and Authenticated" },
	{ 0x02, "Encrypted" },
	{ 0x03, "Rejected Bad Certificate" },
	{ 0x04, "Rejected Revoked Certificate" },
	{ 0, NULL }
};

static const value_string ctrl_ie_ei_vals[] = {
	{ 0x00, "Identity Info" },
	{ 0x01, "Identity Type" },
	{ 0x02, "Identity Reason Code" },
	{ 0x03, "Additional Text" },
	{ 0x04, "SSL Status" },
	{ 0x05, "SSL Identity" },
	{ 0, NULL }
};

static const ie_def_t ctrl_ie_defs[] = {
	{ 0x00, "Identity Info", -1, dissect_ie_identity_info, NULL },
	{ 0x01, "Identity Type", 1, NULL, (void*)identity_type_vals },
	{ 0x02, "Identity Reason Code", 1, NULL, NULL },
	{ 0x03, "Additional Text", -1, NULL, NULL },
	{ 0x04, "SSL Status", 1, NULL, (void*)ssl_status_vals },
	{ 0x05, "SSL Identity", -1, dissect_ie_ssl_identity, NULL },
	{ 0, NULL, 0, NULL, NULL },
};

static GMemChunk *analysis_chunk = NULL;

static int analysis_count = IPA_LINK_MAX;	/* one for each conversation */

struct analysis {
	guint32 frame;
	guint32 heartbeat_frame;
	int heartbeats;
	nstime_t last_heartbeat_ts;
	nstime_t cumulative_heartbeat_roundtrip_ts;
	nstime_t peak_heartbeat_roundtrip_ts;
} analysis;

struct frame_analysis {
	guint32 heartbeat_frame;
	guint32 rtt;
	guint32 peak_rtt;
	guint32 avrg_rtt;
} frame_analysis;

static struct analysis *get_conversation_data(packet_info *pinfo)
{
	conversation_t *conv=NULL;
	struct analysis *a=NULL;

	/* Have we seen this conversation before? */
	if( (conv=find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0)) == NULL){
		/* No this is a new conversation. */
		conv=conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
	}

	/* check if we have any data for this conversation */
	a=conversation_get_proto_data(conv, proto_ipa);
	if(!a){
		/* No no such data yet. Allocate and init it */
		a=g_mem_chunk_alloc(analysis_chunk);
		if (a) {
			a->frame=pinfo->fd->num;
			a->heartbeat_frame=0;
			a->heartbeats=0;
			nstime_set_zero(&a->last_heartbeat_ts);
			nstime_set_zero(&a->cumulative_heartbeat_roundtrip_ts);
			nstime_set_zero(&a->peak_heartbeat_roundtrip_ts);

			conversation_add_proto_data(conv, proto_ipa, (void *)a);
		}
	}

	return a;
}

static void analyze_init(void)
{
	/*
	 * Now destroy the chunk from which the conversation table
	 * structures were allocated.
	 */
	if (analysis_chunk) {
		g_mem_chunk_destroy(analysis_chunk);
		analysis_chunk = NULL;
	}

	/* and create new chunks */

	analysis_chunk = g_mem_chunk_new("analysis_chunk",
			sizeof(struct analysis),
			analysis_count * sizeof(struct analysis),
			G_ALLOC_AND_FREE);
}

static struct frame_analysis *analyze_heartbeats(int msg_type, packet_info *pinfo)
{
	struct analysis *a;
	struct frame_analysis *fa;
	nstime_t roundtrip_ts, diff_ts;
	guint32 rtt, peak_rtt, avrg_rtt;

	/* Use existing packet info if available */
	fa = p_get_proto_data(pinfo->fd, proto_ipa);
	if (fa) {
		return fa;
	}

	a = get_conversation_data(pinfo);

	if (msg_type == IPA_HEARTBEAT) {
		if (a->frame <= pinfo->fd->num)	{ /* if already fully parsed, no need to continue */
			a->last_heartbeat_ts  = pinfo->fd->rel_ts;	/* ignore for now second heartbeat without Ack - FIXME */
			a->heartbeat_frame = pinfo->fd->num;	/* remember this frame for later */
			a->frame = pinfo->fd->num;	/* update frame pointer */
		}
		return NULL;
	}
	else if (msg_type == IPA_HEARTBEAT_ACK) {
		if (a->frame > pinfo->fd->num) /* if already fully parsed, no need to continue */
			return FALSE;		/* already parsed, but no info present */

		if (a->heartbeat_frame){	/* last received a hearbeat, not another ack */
			a->heartbeats++;
			nstime_delta(&roundtrip_ts, &pinfo->fd->rel_ts, &a->last_heartbeat_ts);
			nstime_add(&a->cumulative_heartbeat_roundtrip_ts, &roundtrip_ts);
			nstime_delta(&diff_ts, &roundtrip_ts, &a->peak_heartbeat_roundtrip_ts);

			if (diff_ts.secs>0 || (diff_ts.secs==0 && diff_ts.nsecs>0))
				a->peak_heartbeat_roundtrip_ts = roundtrip_ts;

			nstime_set_zero(&a->last_heartbeat_ts);				/* zero to indicate awaiting next heartbeat */

			avrg_rtt = (a->cumulative_heartbeat_roundtrip_ts.secs*1000+a->cumulative_heartbeat_roundtrip_ts.nsecs/1000000) / a->heartbeats;
			peak_rtt = (a->peak_heartbeat_roundtrip_ts.secs*1000+a->peak_heartbeat_roundtrip_ts.nsecs/1000000);
			rtt = (roundtrip_ts.secs*1000+roundtrip_ts.nsecs/1000000);

			a->frame = pinfo->fd->num;	/* update frame pointer */

			fa = se_alloc(sizeof(struct frame_analysis));
			if (fa) {
				fa->rtt = rtt;
				fa->peak_rtt = peak_rtt;
				fa->avrg_rtt = avrg_rtt;
				fa->heartbeat_frame = a->heartbeat_frame;
				p_add_proto_data(pinfo->fd, proto_ipa, fa);
			}
			return fa;
		}
		else {	/* second ack without heartbeat - ignore for now - FIXME */
			a->frame = pinfo->fd->num;	/* update frame pointer */
			return NULL;
		}
	}

	a->frame = pinfo->fd->num;	/* update frame pointer */
	return NULL;
}

#ifdef USE_HEURISTIC_SOIP
static gboolean
dissect_ipa_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	gint offset=0;
	guint16 ipa_len;
	guint8 stream;
	gboolean rc=FALSE;
	conversation_t *conversation;

	if (tvb_length(tvb)>=3) {
		ipa_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		stream = tvb_get_guint8(tvb, offset);
		offset++;

		if (stream>=0 && stream<RSL_STREAM_MAX && ipa_len==tvb_length_remaining(tvb,offset) && ipa_len<=273)	/* sanity check on length - the 273 matches the protocol use limits */
			rc=TRUE;
		else {
			switch (stream) {	/* only bother to check A-bis links as only these are really configurable ports */
			case CTRL_STREAM:
			case OML_STREAM:
				if (ipa_len<=tvb_length_remaining(tvb,offset) && ipa_len<=273)	/* sanity check on length - the 273 matches the protocol use limits */
					rc=TRUE;
				break;

			default:
				rc=FALSE;
			}
		}
	}

	if (rc==TRUE) {
		conversation = find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
		if (!conversation)
			conversation = conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
		conversation_set_dissector(conversation, ipa_handle);
		dissect_ipa(tvb, pinfo, tree);
	}

	return rc;
}
#endif

/* Code to actually dissect the packets - actually only a stub to allow the real
   routine (which this calls) to be called multiple times for packets which contain more than one message
*/
static int
dissect_ipa(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset=0, this_msg_len, desegment;

	if (check_col(pinfo->cinfo, COL_INFO))
		col_add_str(pinfo->cinfo, COL_INFO, "");	/* the COL_INFO is appended every time, so start blank */

	this_msg_len = tvb_length(tvb); /* for safety until we know better */

	while (this_msg_len > offset) {
		offset += dissect_ipa_transport(tvb, offset, pinfo, tree, &desegment);
		if (desegment)
			return offset;

		/* If packet contains a follow-on message, add a tab separator */
		if (tvb_length_remaining(tvb,offset)>0) {
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_str(pinfo->cinfo, COL_INFO, "\t");
		}
	}

	return offset;
}

static int
dissect_ipa_transport(tvbuff_t *old_tvb, int old_offset, packet_info *pinfo, proto_tree *main_tree, int *desegment)
{
	guint32 msg_len, this_msg_len;
	guint8 stream;
	int offset=0, isUdp=FALSE;
	tvbuff_t *tvb, *next_tvb;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	*desegment = 0;
	this_msg_len = tvb_length_remaining(old_tvb, old_offset); /* for safety until we know better */

	/* FIXME - kludge: assume packet is a complete message */
	msg_len = tvb_get_ntohs(old_tvb, old_offset);
	if (msg_len+IPA_UDP_HEADER_LEN == this_msg_len)	{	/* kludge to check for UDP messages - 1 octet length difference! */
		tvb = tvb_new_subset(old_tvb, old_offset, msg_len+IPA_UDP_HEADER_LEN, msg_len+IPA_UDP_HEADER_LEN);
		isUdp = TRUE;
	}
	else if (msg_len+IPA_HEADER_LEN > this_msg_len) {
		if (pinfo->can_desegment) { /* The message is split across segment boundaries */
            /* Tell the TCP dissector where the data for this
             * message starts in the data it handed us, and how many
             * more bytes we need, and return.
             */
            pinfo->desegment_offset = offset;
            pinfo->desegment_len = msg_len + IPA_HEADER_LEN - this_msg_len;
			*desegment = 1;
            return offset;
		}
		else
			tvb = tvb_new_subset(old_tvb, old_offset, msg_len+IPA_HEADER_LEN, msg_len+IPA_HEADER_LEN);
	}
	else
		tvb = tvb_new_subset(old_tvb, old_offset, msg_len+IPA_HEADER_LEN, msg_len+IPA_HEADER_LEN);

	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_ipa, tvb, offset, msg_len+IPA_HEADER_LEN, FALSE);

		/* add an item to the subtree, see section 1.6 for more information */
		tree = proto_item_add_subtree(ti, ett_ipa);
	}

	/* set up length of message or fragment in this packet appropriately */
	/*if (offset+msg_len+MSG_LEN_MIN<=this_msg_len)
		this_msg_len = offset + msg_len + MSG_LEN_MIN;
	  else leave this_msg_len set to full buffer size */

	if (tree) {
		if ((msg_len < 1) || (msg_len>MSG_LEN_MAX))
			proto_tree_add_text(tree, tvb, offset, 2,"Illegal Msg length %d (TCP segmentation ?): Skipping rest", msg_len);
		else
			proto_tree_add_uint_hidden(tree, hf_ipa_msg_len, tvb, offset, 2, msg_len);
	}
	offset += 2;

	/* sanity check the length before commiting to use it! */
	if ((msg_len < 1) || (msg_len>MSG_LEN_MAX)) {
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Illegal Message Length %d (TCP segmentation ?): Skipping rest", msg_len);
		return this_msg_len; /* bomb out here for safety */
	}

	stream = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_ipa_stream, tvb, offset, 1, stream);
	offset += 1;

	if (isUdp)
		offset += 1;	/* skip the extra UDP octet */

	/* construct a subset TVB for the payload to be disected with - assuming its in this packet */
	next_tvb = tvb_new_subset(tvb, offset, msg_len, msg_len);
	if (tvb_length_remaining(tvb,offset)>0) {
		offset += msg_len;

		if (((stream == OML_STREAM) || (stream == 0)) && (pinfo->match_port==TCP_PORT_IPA_BSCMGMT)) {	/* to cope with old BSC Mgmt protocol on the OML StreamId */
			call_dissector(bscmgmt_v0_handle,next_tvb,pinfo,main_tree);
		}
		else if (!dissector_try_port(subdissector_table, stream, next_tvb, pinfo, main_tree)) {

			if (stream == CTRL_STREAM) {
				dissect_ipa_ctrl(next_tvb,pinfo,tree);
			}
			else if ((pinfo->match_port==TCP_PORT_IPA_SPIRIT) && (stream == SPIRIT_STREAM)) {
				call_dissector(ipa_spirit_handle,next_tvb,pinfo,main_tree);
			}
			else if ((pinfo->match_port==TCP_PORT_IPA_SITE_IML) && (stream == IML_STREAM)) {
				call_dissector(data_handle,next_tvb,pinfo,main_tree);
			}
			else if ((pinfo->match_port==TCP_PORT_IPA_BTS_IML) && (stream == IML_STREAM)) {
				call_dissector(data_handle,next_tvb,pinfo,main_tree);
			}
			else if (stream == SCCP_STREAM) {
				call_dissector(sccp_handle,next_tvb,pinfo,main_tree);
			}
			else if (stream == BSC_MGMT_STREAM) {
				call_dissector(bscmgmt_v1_handle,next_tvb,pinfo,main_tree);
			}
			else if (stream == BSC_SUBAGENT_STREAM) {
				call_dissector(bscmgmt_v2_handle,next_tvb,pinfo,main_tree);
			}
			else if (stream < RSL_STREAM_MAX) {
				call_dissector(gsm0858_handle,next_tvb,pinfo,main_tree);
			}
			else {
				call_dissector(data_handle,next_tvb,pinfo,main_tree);
			}
		}
		if (tree)
			proto_item_set_len(ti,offset);
	}

	return offset;
}

/* Code to dissect the Control Channel packets */
static void
dissect_ipa_ctrl(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type;
	struct frame_analysis *fa=NULL;

	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "ip.access";

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "ipa");

	if (main_tree) {
		ti = proto_tree_add_text(main_tree, tvb, offset, 1, "CTRL" );
		tree = proto_item_add_subtree(ti, ett_ipa_ctrl);
	}

	msg_len = tvb_length_remaining(tvb,offset);

	msg_type = tvb_get_guint8(tvb, offset);

	if (tree)
		proto_tree_add_uint(tree, hf_ipa_ctrl_msg_type, tvb, offset, 1, msg_type);
	offset += 1;

	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_fstr(pinfo->cinfo, COL_INFO, "CTRL:%s(", rd_match_strval(msg_type,ctrl_msg_type_vals));

	if (main_tree)
		proto_item_set_text(ti, "CTRL: %s", rd_match_strval(msg_type,ctrl_msg_type_vals));

	/* create a new subset TVB for disecting the payload, and call the correct payload disector */
	next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
	if (msg_len > offset) {
		dissect_ipa_ctrl_ie(msg_type, ctrl_ie_defs, next_tvb, pinfo, tree);
	}

	if ((msg_type == IPA_HEARTBEAT) || (msg_type == IPA_HEARTBEAT_ACK)) {
		fa = analyze_heartbeats(msg_type, pinfo);
		if (fa && check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, "RTT=%dms,peak=%dms", fa->rtt, fa->peak_rtt);
	}

	if (tree) {
		if (fa) {
			ti = proto_tree_add_uint(tree, hf_ipa_ctrl_heartbeat_frame, tvb, 0, 0, fa->heartbeat_frame);
			PROTO_ITEM_SET_GENERATED(ti);
			ti = proto_tree_add_uint_format(tree, hf_ipa_ctrl_heartbeat_rtt, tvb, 0, 0, fa->rtt, "RTT=%d(ms)", fa->rtt);
			PROTO_ITEM_SET_GENERATED(ti);
			ti = proto_tree_add_uint_format(tree, hf_ipa_ctrl_heartbeat_peak_rtt, tvb, 0, 0, fa->peak_rtt, "Peak RTT=%d(ms)", fa->peak_rtt);
			PROTO_ITEM_SET_GENERATED(ti);
			ti = proto_tree_add_uint_format(tree, hf_ipa_ctrl_heartbeat_avrg_rtt, tvb, 0, 0, fa->avrg_rtt, "Average RTT=%d(ms)", fa->avrg_rtt);
			PROTO_ITEM_SET_GENERATED(ti);
		}
	}

	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_fstr(pinfo->cinfo, COL_INFO, ")");

	return;
}

/* Code to dissect the CTRL IEs */
static void
dissect_ipa_ctrl_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len, i;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		ie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_ipa_ctrl_ie_ei, tvb, offset, 1, ie_ei);
		offset += 1;

		for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==ie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				ie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				ie_len = p_ie_def->ie_len;
				if (tvb_length_remaining(tvb,offset)<ie_len) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len, tvb_length_remaining(tvb,offset));
					ie_len=tvb_length_remaining(tvb,offset);
				}

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)(p_ie_def->info)));
					for (i=2; i<ie_len+1; i++)
						proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_ctrl_ie);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				ie_len8 = tvb_get_guint8(tvb, offset);
				if (tvb_length_remaining(tvb,offset+1)<ie_len8) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, tvb_length_remaining(tvb,offset+1)+2, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len8, tvb_length_remaining(tvb,offset+1));
					ie_len8=tvb_length_remaining(tvb,offset+1);
				}

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ipa_ctrl_ie_len8, tvb, offset, 1, ie_len8);
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						if (ie_len8==0) {
							ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: <empty>", p_ie_def->ie_name);
						}
						else {
							ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+1),(value_string *)(p_ie_def->info)));
							for (i=3; i<ie_len8+2; i++)
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_ctrl_ie);
					}
				}
				offset += 1;
				ie_len = ie_len8;
			}
			else if (p_ie_def->ie_len == -2) {
				ie_len16 = tvb_get_ntohs(tvb, offset);
				if (tvb_length_remaining(tvb,offset+2)<ie_len16) {
					if (main_tree)
						proto_tree_add_text(main_tree, tvb, 0, tvb_length_remaining(tvb,offset+2)+3, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len16, tvb_length_remaining(tvb,offset+2));
					ie_len16=tvb_length_remaining(tvb,offset+2);
				}

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ipa_ctrl_ie_len16, tvb, offset, 2, ie_len16);
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						if (ie_len16==0) {
							ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: <empty>", p_ie_def->ie_name);
						}
						else {
							ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+2),(value_string *)(p_ie_def->info)));
							for (i=4; i<ie_len16+3; i++)
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ipa_ctrl_ie);
					}
				}
				offset += 2;
				ie_len = ie_len16;
			}
			else {			/* should never happen */
				ie_len=0; /* safety! */
			}

			if (ie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, main_tree, p_ie_def->info);
				else
					call_dissector(data_handle,new_tvb,pinfo,tree);
				offset+=ie_len;
			}

				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);
				dissect_ipa_ctrl_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree);
			}
		}
	}
	return;
}

static void dissect_ie_identity_info(int msg_type _U_, tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *tree, proto_tree *main_tree _U_, void *info _U_)
{
	int offset=0, msg_len, identity_type;
	guint8 *identity_value;

	msg_len = tvb_length_remaining(tvb,offset);

	identity_type = tvb_get_guint8(tvb, offset);
	identity_value = tvb_get_string(tvb, offset+1, msg_len-(offset+1));

	if (tree)
		proto_tree_add_text(tree, tvb, offset, msg_len, "%s=%s", rd_match_strval(identity_type,identity_type_vals),identity_value);
	offset += msg_len;

	g_free(identity_value);

	return;
}

static void dissect_ie_ssl_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len, ssl_identity_type;
	guint8 *ssl_identity_value;

	msg_len = tvb_length_remaining(tvb,offset);

	ssl_identity_type = tvb_get_guint8(tvb, offset);
	ssl_identity_value = tvb_get_string(tvb, offset+1, msg_len-(offset+1));

	if (tree)
		proto_tree_add_text(tree, tvb, offset, msg_len, "%s=%s", rd_match_strval(ssl_identity_type,ssl_identity_type_vals),ssl_identity_value);
	offset += msg_len;

	g_free(ssl_identity_value);

	return;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_ipa(void)
{
    module_t *ipa_module;

    /* Setup list of header fields */
	static hf_register_info hf[] = {
		{ &hf_ipa_msg_len,
			{ "Message Length", "ipa.msg_len",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Message Length", HFILL }
		},
		{ &hf_ipa_stream,
			{ "Stream Id", "ipa.stream",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Stream Id", HFILL }
		},
		{ &hf_ipa_ctrl_msg_type,
			{ "CTRL Msg Type", "ipa.ctrl.msg_type",
			FT_UINT8, BASE_HEX, VALS(ctrl_msg_type_vals), 0x0,
			"CTRL Message Type", HFILL }
		},
		{ &hf_ipa_ctrl_ie_ei,
			{ "CTRL IE", "ipa.ctrl.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ctrl_ie_ei_vals), 0x0,
			"CTRL IE Element Identifier", HFILL }
		},
		{ &hf_ipa_ctrl_ie_len8,
			{ "CTRL IE Len", "ipa.ctrl.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"CTRL IE Length", HFILL }
		},
		{ &hf_ipa_ctrl_ie_len16,
			{ "CTRL IE Len", "ipa.ctrl.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"CTRL IE Length", HFILL }
		},
		{ &hf_ipa_ctrl_heartbeat_rtt,
			{ "Heartbeat roundtrip time (ms)", "ipa.ctrl.heartbeat.rtt",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"The time for the Heartbeat Ack response to a Heartbeat (in ms)", HFILL }
		},
		{ &hf_ipa_ctrl_heartbeat_peak_rtt,
			{ "Heartbeat peak roundtrip time (ms)", "ipa.ctrl.heartbeat.peak_rtt",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"The peak time for the Heartbeat Ack response to a Heartbeat (in ms)", HFILL }
		},
		{ &hf_ipa_ctrl_heartbeat_avrg_rtt,
			{ "Heartbeat average roundtrip time (ms)", "ipa.ctrl.heartbeat.avrg_rtt",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"The average time for the Heartbeat Ack response to a Heartbeat (in ms)", HFILL }
		},
		{ &hf_ipa_ctrl_heartbeat_frame,
			{ "Heartbeat frame", "ipa.ctrl.heartbeat.frame",
			FT_FRAMENUM, BASE_NONE, NULL, 0x0,
			"The frame of the Heartbeat", HFILL }
		},
	};

    /* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_ipa,
		&ett_ipa_ctrl,
		&ett_ipa_ctrl_ie,
	};

	proto_ipa = proto_register_protocol("ip.access Signalling over IP", "ip.access", "ipa");
    register_dissector("ipa", dissect_ipa, proto_ipa);

    /* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_ipa, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	/* subdissector code */
    subdissector_table = register_dissector_table("ipa.stream", "ip.access Stream", FT_UINT8, BASE_DEC);

	/* Init rountines - for conversations */
	register_init_routine(analyze_init);

    /* Register preferences module */
    ipa_module = prefs_register_protocol(proto_ipa, proto_reg_handoff_ipa);

    /* Register preferences */
    prefs_register_bool_preference(ipa_module, "gprs_use_ipa",
         "Use ip.access Gb dissector",
         "Allows the use of the ip.access or the standard Gb dissector.",
         &global_gprs_use_ipa);

    prefs_register_uint_preference(ipa_module, "gprs.udp.start_port",
         "Gb UDP Start Port",
         "Defines the start of the UDP port range for Gb NS over UDP",
         10, &global_gprs_start_port);

    prefs_register_uint_preference(ipa_module, "gprs.udp.end_port",
         "Gb UDP End Port",
         "Defines the end of the UDP port range for Gb NS over UDP",
         10, &global_gprs_end_port);

    prefs_register_uint_preference(ipa_module, "oml.port",
         "OML Port",
         "Defines the TCP port used for the A-bis OML",
         10, &global_oml_port);

    prefs_register_uint_preference(ipa_module, "rsl.port",
         "RSL Port",
         "Defines the TCP port used for the A-bis RSL",
         10, &global_rsl_port);
}

/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.

   We use a Heuristic registration if the TCP port number isn't fixed
   but if there is a defined port number register that instead
*/
void
proto_reg_handoff_ipa(void)
{
    static int initialized = FALSE;
    static guint16 gprs_start_port;
    static guint16 gprs_end_port;
    static guint16 oml_port;
    static guint16 rsl_port;
    static gboolean gprs_use_ipa;
    dissector_handle_t use_handle;
    guint16 port;

    if (!initialized) {
        ipa_handle = find_dissector("ipa");

        data_handle = find_dissector("data");
        sccp_handle = find_dissector("sccp");
        gsm0858_handle = find_dissector("gsm0858");
        gsm0816_handle = find_dissector("gsm0816");
        gprs_ns_handle = find_dissector("gprs_ns");
	   	bscmgmt_v0_handle = find_dissector("bscmgmt_v0");
    	bscmgmt_v1_handle = find_dissector("bscmgmt_v1");
    	bscmgmt_v2_handle = find_dissector("bscmgmt_v2");
    	ipa_spirit_handle = find_dissector("ipa_spirit");

#ifdef TCP_PORT_IPA_2NDOML
    	dissector_add("tcp.port", TCP_PORT_IPA_2NDOML, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_BSCMGMT
    	dissector_add("tcp.port", TCP_PORT_IPA_BSCMGMT, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_HW_MON
    	dissector_add("tcp.port", TCP_PORT_IPA_HW_MON, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_FRIP
    	dissector_add("tcp.port", TCP_PORT_IPA_FRIP, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_GNAT
    	dissector_add("tcp.port", TCP_PORT_IPA_GNAT, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_A_IF
    	dissector_add("tcp.port", TCP_PORT_IPA_A_IF, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_SPIRIT
    	dissector_add("tcp.port", TCP_PORT_IPA_SPIRIT, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_SITE_IML
    	dissector_add("tcp.port", TCP_PORT_IPA_SITE_IML, ipa_handle);
#endif
#ifdef TCP_PORT_IPA_BTS_IML
    	dissector_add("tcp.port", TCP_PORT_IPA_BTS_IML, ipa_handle);
#endif
#ifdef UDP_PORT_IPA
	    dissector_add("udp.port", UDP_PORT_IPA, ipa_handle);
#endif

#ifdef USE_HEURISTIC_SOIP
    	heur_dissector_add("tcp", dissect_ipa_heur, proto_ipa);
#endif
        initialized = TRUE;
	}
	else {
        if (gprs_use_ipa)
            use_handle = gsm0816_handle;
        else
            use_handle = gprs_ns_handle;

        for (port=gprs_start_port; port<=gprs_end_port; port++)
            dissector_delete("udp.port", port, use_handle);
    	dissector_delete("tcp.port", oml_port, ipa_handle);
    	dissector_delete("tcp.port", rsl_port, ipa_handle);
	}

    if (global_gprs_start_port>=0 && global_gprs_start_port<=65535 && global_gprs_end_port>=0 && global_gprs_end_port<=65535) {
        gprs_start_port = global_gprs_start_port;
        gprs_end_port = global_gprs_end_port;
        gprs_use_ipa = global_gprs_use_ipa;
    }
    else {  /* a safe non-functional setting */
        gprs_start_port = gprs_end_port + 1;
        gprs_use_ipa = global_gprs_use_ipa;
    }
    oml_port = global_oml_port;
    rsl_port = global_rsl_port;

    if (gprs_use_ipa)
        use_handle = gsm0816_handle;
    else
        use_handle = gprs_ns_handle;

    for (port=gprs_start_port; port<=gprs_end_port; port++)
        dissector_add("udp.port", port, use_handle);

   	dissector_add("tcp.port", oml_port, ipa_handle);
   	dissector_add("tcp.port", rsl_port, ipa_handle);

	return;
}
