/* packet-oyster3g-domain.c
 * Routines for ip.access oyster3G XML message domains
 * Copyright 2007, ip.access Ltd <support@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-oyster3g-domain.c 1.6 2009/02/18 09:42:23GMT Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/emem.h>
#include <epan/packet.h>
#include <epan/expert.h>

#include "packet-oyster3g-domain.h"

/* function prototypes for this file */
static int dissect_oyster3g_domain_ie( tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *tree, proto_item *ti, const oyster3g_domain_t *p_domain, const oyster3g_message_t **p_p_message, const oyster3g_ie_t **p_p_ie);
static int dissect_oyster3g_domain_minimal_ie( tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *tree, proto_item *ti, const oyster3g_domain_t *p_domain, const oyster3g_message_t **p_p_message, const oyster3g_ie_t **p_p_ie);
static gboolean check_is_maximal(tvbuff_t *tvb, const oyster3g_domain_t *p_domain);
static int get_ie_len_minimal(tvbuff_t *tvb, int offset, int len_max, int *p_len_size);

/* static for the domain disection - how deep is the ett tree ? */
static int ett_depth;

/* the generic domain dissector */
void dissect_oyster3g_domain(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, const oyster3g_domain_t *p_domain)
{
    int offset=0, msg_len;
    tvbuff_t *next_tvb;
    const oyster3g_message_t  *p_message=NULL;
    const oyster3g_ie_t  *p_ie=NULL;
    proto_item *ti2=NULL;
    gboolean maximal;

    /* Set up structures needed to add the protocol subtree and manage it */
    proto_item *ti=NULL;
    proto_tree *tree=NULL;

    /* load the display labels  */
    pinfo->current_proto = p_domain->label;

    if (check_col(pinfo->cinfo, COL_PROTOCOL))
        col_add_str(pinfo->cinfo, COL_PROTOCOL, p_domain->label);

    msg_len = tvb_length_remaining(tvb,offset);
    if (main_tree) {
        /* create display subtree for the protocol */
        ti = proto_tree_add_item(main_tree, *(p_domain->proto), tvb, offset, msg_len, FALSE);

        /* add an item to the subtree */
        ett_depth = 0;
        tree = proto_item_add_subtree(ti, *(p_domain->ett[ett_depth++]));
    }

    /* could try to pull the first IE to get message type and use this in the label
     * Probably need this generally to pass a string back up one level as a summary
     * for display on all tree elements
     */
    if (tree)
        proto_item_set_text(ti, "%s", p_domain->label);

    maximal = check_is_maximal(tvb, p_domain);

    /* create a new subset TVB for disecting the message, and call the IE disector with undefine message and IE */
    while (offset<msg_len) {
        next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
        if (maximal == TRUE)
            offset += dissect_oyster3g_domain_ie(next_tvb, pinfo, main_tree, tree, ti, p_domain, &p_message, &p_ie);
        else
            offset += dissect_oyster3g_domain_minimal_ie(next_tvb, pinfo, main_tree, tree, ti, p_domain, &p_message, &p_ie);
        if (!p_message) { /* No such message - abort */
            if (tree) {
                ti2 = proto_tree_add_text(tree, tvb, offset, msg_len-offset, "Unknown Message");
                expert_add_info_format(pinfo, ti2, PI_UNDECODED, PI_WARN, "Unknown Message");
            }
            break;
		}
        else if (!p_ie->label) { /* end of the IE list - abort */
            if (tree) {
                ti2 = proto_tree_add_text(tree, tvb, offset, msg_len-offset, "Extra data after last defined IE");
                expert_add_info_format(pinfo, ti2, PI_UNDECODED, PI_WARN, "Extra data after last defined IE");
            }
            break;
		}
        p_ie++; /* move to the next IE for the next call */
    }

    return;
}

/* Code to dissect the maximally encoded IEs */
static int dissect_oyster3g_domain_ie( tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *tree, proto_item *main_ti, const oyster3g_domain_t *p_domain, const oyster3g_message_t **p_p_message, const oyster3g_ie_t **p_p_ie)
{
    int offset=0, msg_len=0, value_offset=0, consumed;
    gint32 ie_tag=-1, ie_len=-1, msg_tag=-1, display_ie_len;
    guint32 value, i;
    gint32 svalue;
    tvbuff_t *new_tvb=NULL;
    proto_item *ti=NULL;
    proto_tree *new_tree=NULL;
    const oyster3g_message_t *p_message=NULL;
    const oyster3g_syntax_t *p_syntax=NULL;
    const oyster3g_ie_t  *p_new_ie=NULL, *p_ie=NULL;
    int index=0;

    msg_len = tvb_length_remaining(tvb,offset);

    /* all IEs are a TLV, with T and L sized by domain constants */

    /* First read the T value */
    if (p_domain->tag_size==1)
        ie_tag = tvb_get_guint8(tvb, offset);
    else if (p_domain->tag_size==2)
        ie_tag = tvb_get_ntohs(tvb, offset);
    else if (p_domain->tag_size==4)
        ie_tag = tvb_get_ntohl(tvb, offset);
    else
        g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

    if (tree)
        proto_tree_add_uint_hidden(tree, *(p_domain->hf_ie_tag), tvb, offset, p_domain->tag_size, ie_tag);
    offset += p_domain->tag_size;

    /* Second read the L value */
    if (p_domain->len_size==1)
        ie_len = tvb_get_guint8(tvb, offset);
    else if (p_domain->len_size==2)
        ie_len = tvb_get_ntohs(tvb, offset);
    else if (p_domain->len_size==4)
        ie_len = tvb_get_ntohl(tvb, offset);
    else
        g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

    if (tree)
        proto_tree_add_uint_hidden(tree, *(p_domain->hf_ie_len), tvb, offset, p_domain->len_size, ie_len);
    offset += p_domain->len_size;
    value_offset = offset;

    if (offset+ie_len>msg_len) {
        if (tree) {
            ti = proto_tree_add_text(tree, tvb, offset-p_domain->len_size, p_domain->len_size, "IE Length exceeds size of enclosing message or IE");
            expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "IE Length exceeds size of enclosing message or IE");
		}
		ie_len = msg_len - offset; /* correct ie_len to max allowed for later use */
    }

    /* now move the p_ie pointer until the current tag is found.
       NOTE: p_ie may be pointing to the message IE list or a struct/set/array IE list */
    /* if message is undefined, we have yet to read the message Id - assume the first tag is the message ID */
    if (!(*p_p_message)) {
	    /* Read the message tag value from the first IE value */
		if (ie_len==1)
			msg_tag = tvb_get_guint8(tvb, offset);
	    else if (ie_len==2)
		    msg_tag = tvb_get_ntohs(tvb, offset);
	    else if (ie_len==4)
		    msg_tag = tvb_get_ntohl(tvb, offset);
		else {
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset, ie_len, "Illegal Message Id size");
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Illegal Message Id size");
            }
            offset += ie_len;
            return offset;
		}

        for (index=0,p_message=(p_domain->message)[0]; p_message && (p_message->tag!=-1); p_message=(p_domain->message)[++index]) {
            if (p_message->tag == msg_tag) {
                *p_p_message = p_message;
                p_ie = p_message->ie;   /* start at the first IE in the message */
                *p_p_ie = p_ie;
                if (tree)
                    proto_item_set_text(main_ti, "%s: %s", p_domain->display_name, p_message->display_name);
			    if (check_col(pinfo->cinfo, COL_INFO))
			        col_prepend_fence_fstr(pinfo->cinfo, COL_INFO,"%s, ", p_message->display_name);
                break;
            }
        }
		if (!p_message) { /* walked off the message array */
            *p_p_message = p_message;
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset, ie_len, "No such Message Tag %d found", msg_tag);
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "No such Message Tag %d found", msg_tag);
            }
            return 0;
		}
    }
    /* if the message is already known, find the IE in the message or struct based on the tag */
    else {
        p_message = *p_p_message;
        /* check the current tag against the next IE */
        for (p_ie=*p_p_ie; p_ie->syntax>=0; p_ie++) {
			if (!p_ie->label) {	/* end of IE list - abort */
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, 0, value_offset+ie_len, "Extra data after last defined IE");
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Extra data after last defined IE");
                }
                offset += ie_len;
                *p_p_ie = p_ie;
                return offset;
			}
            else if ((p_domain->syntax)[p_ie->syntax]->tag == ie_tag) { /* found the IE we are looking for */
                break;
            }
            else if (p_ie->mandatory) { /* missing mandatory IE */
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, 0, value_offset+ie_len, "Missing mandatory IE %s", (p_domain->syntax)[p_ie->syntax]->label);
                    expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Missing mandatory IE %s", (p_domain->syntax)[p_ie->syntax]->label);
                }
                /* Try carrying on to see if there are more IEs we do recognise */
            }
        }
        *p_p_ie = p_ie;

        if (p_ie->syntax<0) {   /* end of IE list - extra data found */
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, 0, value_offset+ie_len, "Unexpected extra IE");
                expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Unexpected extra IE");
            }
            offset += ie_len;
            return offset;
        }
    }

    /* We now have p_ie set to the current IE in the packet, so we can decode the value based on the IE syntax */

    /* First check if a custom dissector has been registered, and if so, use it */
    if ((p_domain->custom_dissector)[p_ie->syntax])
    {
        new_tvb = tvb_new_subset(tvb, offset, ie_len, ie_len);
		consumed = (p_domain->custom_dissector)[p_ie->syntax](new_tvb,pinfo,main_tree);
		if (consumed>0) {
			offset += consumed;
			if (offset<value_offset+ie_len) {
	            if (tree) {
	                proto_tree_add_text(tree, tvb, offset, value_offset+ie_len, "Extra data");
	            }
			}
			offset=value_offset+ie_len;

	        return offset;
		}
	}

    p_syntax = (p_domain->syntax)[p_ie->syntax];

    if ((p_syntax->type!=OYSTER3G_SYNTAX_SET)&&(p_syntax->type!=OYSTER3G_SYNTAX_ARRAY)&&(p_syntax->type!=OYSTER3G_SYNTAX_STRUCT)) {  /* if this is a simple field value just display the value directly */
        display_ie_len = ie_len;

        if ((p_syntax->type==OYSTER3G_SYNTAX_STRING)||(p_syntax->type==OYSTER3G_SYNTAX_UNI_STRING)||(p_syntax->type==OYSTER3G_SYNTAX_OCTET_ARRAY)) { /* except for these 3 variable length elements, the rest should be 1..4 octets - check lengths */
			/* check the string/octet array bounds - unless min==max==0 which indicates unlimited */
			if (((p_syntax->min>0) || (p_syntax->max>0))
			 && ((ie_len<p_syntax->min) || (ie_len>p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "String/Octet Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "String/Octet Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
				}
			}
		}
		else { /* integers, boolean and enums */
            if (p_syntax->type==OYSTER3G_SYNTAX_UINT8) {
            	display_ie_len = 1;
            	value = tvb_get_guint8(tvb,offset);
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT8) {
            	display_ie_len = 1;
            	svalue = (gint32)((gint8)tvb_get_guint8(tvb,offset));
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_UINT16) {
            	display_ie_len = 2;
            	value = tvb_get_ntohs(tvb,offset);
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT16) {
            	display_ie_len = 2;
            	svalue = (gint32)((gint8)tvb_get_ntohs(tvb,offset));
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_UINT32) {
            	display_ie_len = 4;
            	value = tvb_get_ntohl(tvb,offset);
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT32) {
            	display_ie_len = 4;
            	svalue = (gint32)tvb_get_ntohl(tvb,offset);
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_BOOLEAN) {
            	display_ie_len = 1;
			}
            else if (p_syntax->type==OYSTER3G_SYNTAX_ENUM) {
            	display_ie_len = 2;
			}

            if (display_ie_len!=ie_len) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, ie_len, "Illegal length for this Syntax type");
                    expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Illegal length for this Syntax type");
				}
            }
			/* check the integer bounds */
			if (((p_syntax->type==OYSTER3G_SYNTAX_SINT8) || (p_syntax->type==OYSTER3G_SYNTAX_SINT16) || (p_syntax->type==OYSTER3G_SYNTAX_SINT32))
				&& ((svalue<p_syntax->min) || (svalue>p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "Signed Integer value out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Signed Integer value out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
				}
			}
			else if (((p_syntax->type==OYSTER3G_SYNTAX_UINT8) || (p_syntax->type==OYSTER3G_SYNTAX_UINT16) || (p_syntax->type==OYSTER3G_SYNTAX_UINT32))
				&& ((value<(guint32)(p_syntax->min)) || (value>(guint32)(p_syntax->max)))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "Unsigned Integer value out of defined valid range (%u to %u)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Unsigned Integer value out of defined valid range (%u to %u)", p_syntax->min, p_syntax->max);
				}
			}
		}

        if (tree) {
            proto_tree_add_item(tree, *(p_syntax->hf), tvb, offset, display_ie_len, FALSE);
        }
        offset += ie_len;
    }
    else {  /* structs, sets/arrays */
        /* First create a new node for the complex value */
        if (tree) {
            ti = proto_tree_add_text(tree, tvb, 0, ie_len+offset, "%s", p_ie->label);
            new_tree = proto_item_add_subtree(ti, *(p_domain->ett[ett_depth++]));
        }

        if (p_syntax->type == OYSTER3G_SYNTAX_STRUCT) {  /* Struct value is a set of info elements */
            p_new_ie = p_syntax->ie;   /* start a new pointer at the front of the Struct IE list */
            /* Recurse to this disector to dissect the struct IEs - assuming there are any !!! */
            for(; (offset-value_offset)<ie_len; ) {
                /* use a new subset TVB for the recursion */
                new_tvb = tvb_new_subset(tvb, offset, ie_len+value_offset-offset, ie_len+value_offset-offset);
                offset += dissect_oyster3g_domain_ie(new_tvb, pinfo, main_tree, new_tree, ti, p_domain, p_p_message, &p_new_ie);
				if (p_new_ie->label)	/* careful not to walk off the end of the array, but keep going round the loop to dissect unexpected extra IEs */
					p_new_ie++;
            }
        }

        else if ((p_syntax->type == OYSTER3G_SYNTAX_SET)||(p_syntax->type == OYSTER3G_SYNTAX_ARRAY)) { /* set/array value is a set of elements */
            /* Recurse to this disector to dissect the set of IEs - assuming there are any !!! */
            for(i=0; (offset-value_offset)<ie_len; i++) {
                /* use a new subset TVB for the recursion */
                new_tvb = tvb_new_subset(tvb, offset, ie_len+value_offset-offset, ie_len+value_offset-offset);
                p_new_ie = p_syntax->ie;   /* always re-point to the single Set/Array element */
                offset += dissect_oyster3g_domain_ie(new_tvb, pinfo, main_tree, new_tree, ti, p_domain, p_p_message, &p_new_ie);
            }
			/* check the set/array bounds */
			if ((i<(guint32)(p_syntax->min)) || (i>(guint32)(p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, value_offset, ie_len, "Set/Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Set/Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
				}
			}
        }

        else {  /* something wrong... */
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset, ie_len, "Unknown syntax type %d", (p_domain->syntax)[p_ie->syntax]->type);
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Unknown syntax type %d", (p_domain->syntax)[p_ie->syntax]->type);
            }
            return offset;
        }

        /* restore ett depth if needed */
        if (tree) {
            ett_depth--;
        }
    }

    return offset;
}

/* Code to dissect the minimally encoded IEs */
static int dissect_oyster3g_domain_minimal_ie( tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *tree, proto_item *main_ti, const oyster3g_domain_t *p_domain, const oyster3g_message_t **p_p_message, const oyster3g_ie_t **p_p_ie)
{
    int offset=0, msg_len=0, value_offset=0, consumed;
    int datatype, len_size=0;
    gboolean is_msg_id=FALSE;
    gint32 ie_tag=-1, ie_len=-1, msg_tag=-1, display_ie_len;
    guint32 value, i;
    gint32 svalue;
    tvbuff_t *new_tvb=NULL;
    proto_item *ti=NULL;
    proto_tree *new_tree=NULL;
    const oyster3g_message_t *p_message=NULL;
    const oyster3g_syntax_t *p_syntax=NULL;
    const oyster3g_ie_t  *p_new_ie=NULL, *p_ie=NULL;
    int index=0;

    msg_len = tvb_length_remaining(tvb,offset);

    /* if message is undefined, we have yet to read the message Id - assume the first tag is the message ID */
    if (!(*p_p_message)) {
        is_msg_id = TRUE; /* save flag for later */
        
        /* The MessageId is the first IE in every message, but we do not yet know how big it is.
         * Look at the syntax of the first IE of the first message.
         */
        datatype = (p_domain->syntax)[((p_domain->message)[0]->ie)->syntax]->type;
        
        /* Read the message tag value from the first IE value */
        if ((datatype==OYSTER3G_SYNTAX_UINT8) || (datatype==OYSTER3G_SYNTAX_SINT8)) {
            msg_tag = tvb_get_guint8(tvb, offset);
            ie_len = 1;
        }
        else if ((datatype==OYSTER3G_SYNTAX_UINT16) || (datatype==OYSTER3G_SYNTAX_SINT16)) {
            msg_tag = tvb_get_ntohs(tvb, offset);
            ie_len = 2;
        }
        else if ((datatype==OYSTER3G_SYNTAX_UINT32) || (datatype==OYSTER3G_SYNTAX_SINT32)) {
            msg_tag = tvb_get_ntohl(tvb, offset);
            ie_len = 4;
        }
        else
            g_assert_not_reached(); /* something wrong with this code or the source code derived from the XML... */

        for (index=0,p_message=(p_domain->message)[0]; p_message && (p_message->tag!=-1); p_message=(p_domain->message)[++index]) {
            if (p_message->tag == msg_tag) {
                *p_p_message = p_message;
                p_ie = p_message->ie;   /* start at the first IE in the message */
                *p_p_ie = p_ie;
                if (tree)
                    proto_item_set_text(main_ti, "%s: %s", p_domain->display_name, p_message->display_name);
                if (check_col(pinfo->cinfo, COL_INFO))
                    col_prepend_fence_fstr(pinfo->cinfo, COL_INFO,"%s, ", p_message->display_name);
                break;
            }
        }
        if (!p_message) { /* walked off the message array */
            *p_p_message = p_message;
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset, ie_len, "No such Message Tag %d found", msg_tag);
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "No such Message Tag %d found", msg_tag);
            }
            return 0;
        }
    }

    else {
        /* if the message is already known, find the IE in the message or struct based on the message definition
         * if the IE is mandatory, and a tag is not needed, there will be no tag, so just get the IE from the message definition.
         * if we hit an optional IE, read the tag and keep checking
         */
        p_message = *p_p_message;
        /* check the current tag against the next IE */
        for (p_ie=*p_p_ie; p_ie->syntax>=0; p_ie++) {
            if (!p_ie->label) { /* end of IE list - abort */
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, 0, value_offset+msg_len, "Extra data after last defined IE");
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Extra data after last defined IE");
                }
                offset += msg_len;
                *p_p_ie = p_ie;
                return offset;
            }
            
            /* check if we need to read a tag value, and if so, get it */
            if ((ie_tag<0) && !(p_ie->mandatory)) {
                if (p_domain->tag_size==1)
                    ie_tag = tvb_get_guint8(tvb, offset);
                else if (p_domain->tag_size==2)
                    ie_tag = tvb_get_ntohs(tvb, offset);
                else if (p_domain->tag_size==4)
                    ie_tag = tvb_get_ntohl(tvb, offset);
                else
                    g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

                if (tree)
                    proto_tree_add_uint_hidden(tree, *(p_domain->hf_ie_tag), tvb, offset, p_domain->tag_size, ie_tag);
                offset += p_domain->tag_size;
            }
            
            /* we now have a tag if needed */
            if (p_ie->mandatory) { /* a mandatory IE next, so this must be it */
                if ((ie_tag>=0) && ((p_domain->syntax)[p_ie->syntax]->tag != ie_tag)) { /* if we have a tag value anyway, validate it */
                    if (tree) {
                        ti = proto_tree_add_text(tree, tvb, 0, value_offset+p_domain->tag_size, "Tag value does not match expected Mandatory IE tag");
                        expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Tag value does not match expected Mandatory IE tag");
                    }
                }
                break;
            }
            else if ((p_domain->syntax)[p_ie->syntax]->tag == ie_tag) { /* found the optional IE we are looking for */
                break;
            }
        }
        if (!p_ie->label) {
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset-(p_domain->tag_size), p_domain->tag_size, "End of message definition looking for IE Tag %d", ie_tag);
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "End of message definition looking for IE Tag %d", ie_tag);
            }
        return offset;
        }
        *p_p_ie = p_ie;
        p_syntax = (p_domain->syntax)[p_ie->syntax];
        
        /* does this IE need a length field ? If so, get it now */
        datatype = p_syntax->type;
        if ((datatype==OYSTER3G_SYNTAX_UINT8)||(datatype==OYSTER3G_SYNTAX_SINT8)||(datatype==OYSTER3G_SYNTAX_BOOLEAN))
            ie_len = 1;
        else if ((datatype==OYSTER3G_SYNTAX_UINT16)||(datatype==OYSTER3G_SYNTAX_SINT16)||(datatype==OYSTER3G_SYNTAX_ENUM))
            ie_len = 2;
        else if ((datatype==OYSTER3G_SYNTAX_UINT32)||(datatype==OYSTER3G_SYNTAX_SINT32))
            ie_len = 4;
        else {
            ie_len = get_ie_len_minimal(tvb, offset, p_domain->len_size, &len_size);

            if (!len_size) { /* variable length encoding messed up - abort */
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, 0, value_offset+p_domain->len_size, "Variable length field encoding corrupt");
                    expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Variable length field encoding corrupt");
                }
                return offset;
            }
            if (tree)
                proto_tree_add_uint_hidden(tree, *(p_domain->hf_ie_len), tvb, offset, len_size, ie_len);

            offset += len_size;
            value_offset = offset;
        }

        if (offset+ie_len>msg_len) {
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset-p_domain->len_size, p_domain->len_size, "IE Length exceeds size of enclosing message or IE");
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "IE Length exceeds size of enclosing message or IE");
            }
            ie_len = msg_len - offset; /* correct ie_len to max allowed for later use */
        }

        if (p_ie->syntax<0) {   /* end of IE list - extra data found */
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, 0, value_offset+ie_len, "Unexpected extra IE");
                expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Unexpected extra IE");
            }
            offset += ie_len;
            return offset;
        }
    }

    /* We now have p_ie set to the current IE in the packet, so we can decode the value based on the IE syntax */

    /* First check if a custom dissector has been registered, and if so, use it */
    if ((p_domain->custom_dissector)[p_ie->syntax])
    {
        new_tvb = tvb_new_subset(tvb, offset, ie_len, ie_len);
        consumed = (p_domain->custom_dissector)[p_ie->syntax](new_tvb,pinfo,main_tree);
        if (consumed>0) {
            offset += consumed;
            if (offset<value_offset+ie_len) {
                if (tree) {
                    proto_tree_add_text(tree, tvb, offset, value_offset+ie_len, "Extra data");
                }
            }
            offset=value_offset+ie_len;

            return offset;
        }
    }

    p_syntax = (p_domain->syntax)[p_ie->syntax];

    if ((p_syntax->type!=OYSTER3G_SYNTAX_SET)&&(p_syntax->type!=OYSTER3G_SYNTAX_ARRAY)&&(p_syntax->type!=OYSTER3G_SYNTAX_STRUCT)) {  /* if this is a simple field value just display the value directly */
        display_ie_len = ie_len;

        if ((p_syntax->type==OYSTER3G_SYNTAX_STRING)||(p_syntax->type==OYSTER3G_SYNTAX_UNI_STRING)||(p_syntax->type==OYSTER3G_SYNTAX_OCTET_ARRAY)) { /* except for these 3 variable length elements, the rest should be 1..4 octets - check lengths */
            /* check the string/octet array bounds - unless min==max==0 which indicates unlimited */
            if (((p_syntax->min>0) || (p_syntax->max>0))
             && ((ie_len<p_syntax->min) || (ie_len>p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "String/Octet Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "String/Octet Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                }
            }
        }
        else { /* integers, boolean and enums */
            if (p_syntax->type==OYSTER3G_SYNTAX_UINT8) {
                display_ie_len = 1;
                value = tvb_get_guint8(tvb,offset);
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT8) {
                display_ie_len = 1;
                svalue = (gint32)((gint8)tvb_get_guint8(tvb,offset));
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_UINT16) {
                display_ie_len = 2;
                value = tvb_get_ntohs(tvb,offset);
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT16) {
                display_ie_len = 2;
                svalue = (gint32)((gint8)tvb_get_ntohs(tvb,offset));
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_UINT32) {
                display_ie_len = 4;
                value = tvb_get_ntohl(tvb,offset);
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_SINT32) {
                display_ie_len = 4;
                svalue = (gint32)tvb_get_ntohl(tvb,offset);
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_BOOLEAN) {
                display_ie_len = 1;
            }
            else if (p_syntax->type==OYSTER3G_SYNTAX_ENUM) {
                display_ie_len = 2;
            }

            if (display_ie_len!=ie_len) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, ie_len, "Illegal length for this Syntax type");
                    expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Illegal length for this Syntax type");
                }
            }
            /* check the integer bounds */
            if (((p_syntax->type==OYSTER3G_SYNTAX_SINT8) || (p_syntax->type==OYSTER3G_SYNTAX_SINT16) || (p_syntax->type==OYSTER3G_SYNTAX_SINT32))
                && ((svalue<p_syntax->min) || (svalue>p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "Signed Integer value out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Signed Integer value out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                }
            }
            else if (((p_syntax->type==OYSTER3G_SYNTAX_UINT8) || (p_syntax->type==OYSTER3G_SYNTAX_UINT16) || (p_syntax->type==OYSTER3G_SYNTAX_UINT32))
                && ((value<(guint32)(p_syntax->min)) || (value>(guint32)(p_syntax->max)))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, offset, display_ie_len, "Unsigned Integer value out of defined valid range (%u to %u)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Unsigned Integer value out of defined valid range (%u to %u)", p_syntax->min, p_syntax->max);
                }
            }
        }

        if (tree) {
            proto_tree_add_item(tree, *(p_syntax->hf), tvb, offset, display_ie_len, FALSE);
        }
        offset += ie_len;
    }

    else {  /* structs, sets/arrays */
        /* First create a new node for the complex value */
        if (tree) {
            ti = proto_tree_add_text(tree, tvb, 0, ie_len+offset, "%s", p_ie->label);
            new_tree = proto_item_add_subtree(ti, *(p_domain->ett[ett_depth++]));
        }

        if (p_syntax->type == OYSTER3G_SYNTAX_STRUCT) {  /* Struct value is a set of info elements */
            p_new_ie = p_syntax->ie;   /* start a new pointer at the front of the Struct IE list */
            /* Recurse to this disector to dissect the struct IEs - assuming there are any !!! */
            for(; (offset-value_offset)<ie_len; ) {
                /* use a new subset TVB for the recursion */
                new_tvb = tvb_new_subset(tvb, offset, ie_len+value_offset-offset, ie_len+value_offset-offset);
                offset += dissect_oyster3g_domain_minimal_ie(new_tvb, pinfo, main_tree, new_tree, ti, p_domain, p_p_message, &p_new_ie);
                if (p_new_ie->label)    /* careful not to walk off the end of the array, but keep going round the loop to dissect unexpected extra IEs */
                    p_new_ie++;
            }
        }

        else if ((p_syntax->type == OYSTER3G_SYNTAX_SET)||(p_syntax->type == OYSTER3G_SYNTAX_ARRAY)) { /* set/array value is a set of elements */
            /* Recurse to this disector to dissect the set of IEs - assuming there are any !!! */
            for(i=0; (offset-value_offset)<ie_len; i++) {
                /* use a new subset TVB for the recursion */
                new_tvb = tvb_new_subset(tvb, offset, ie_len+value_offset-offset, ie_len+value_offset-offset);
                p_new_ie = p_syntax->ie;   /* always re-point to the single Set/Array element */
                offset += dissect_oyster3g_domain_minimal_ie(new_tvb, pinfo, main_tree, new_tree, ti, p_domain, p_p_message, &p_new_ie);
            }
            
            /* check the set/array bounds */
            if ((i<(guint32)(p_syntax->min)) || (i>(guint32)(p_syntax->max))) {
                if (tree) {
                    ti = proto_tree_add_text(tree, tvb, value_offset, ie_len, "Set/Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                    expert_add_info_format(pinfo, ti, PI_UNDECODED, PI_WARN, "Set/Array length out of defined valid range (%d to %d)", p_syntax->min, p_syntax->max);
                }
            }
        }

        else {  /* something wrong... */
            if (tree) {
                ti = proto_tree_add_text(tree, tvb, offset, ie_len, "Unknown syntax type %d", (p_domain->syntax)[p_ie->syntax]->type);
                expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Unknown syntax type %d", (p_domain->syntax)[p_ie->syntax]->type);
            }
            return offset;
        }

        /* restore ett depth if needed */
        if (tree) {
            ett_depth--;
        }
    }

    return offset;
}

static gboolean check_is_maximal(tvbuff_t *tvb, const oyster3g_domain_t *p_domain)
{
    int offset=0, msg_len=0, msg_id_len;
    int datatype, len_size=0, msg_tag;
    gint32 ie_tag=-1, ie_len=-1;
    const oyster3g_syntax_t *p_syntax;

    msg_len = tvb_length_remaining(tvb,offset);

    /* The MessageId is the first IE in every message, but we do not yet know how big it is.
     * Look at the syntax of the first IE of the first message.
     */
    p_syntax = (p_domain->syntax)[((p_domain->message)[0]->ie)->syntax];
    datatype = p_syntax->type;
    msg_tag = p_syntax->tag;
    
    if ((datatype==OYSTER3G_SYNTAX_UINT8) || (datatype==OYSTER3G_SYNTAX_SINT8)) {
        msg_id_len = 1;
    }
    else if ((datatype==OYSTER3G_SYNTAX_UINT16) || (datatype==OYSTER3G_SYNTAX_SINT16)) {
        msg_id_len = 2;
    }
    else if ((datatype==OYSTER3G_SYNTAX_UINT32) || (datatype==OYSTER3G_SYNTAX_SINT32)) {
        msg_id_len = 4;
    }
    else
        g_assert_not_reached(); /* something wrong with this code or the source code derived from the XML... */

    /* check whether the message is long enough to have a TLV for the Message type IE */
    if (msg_len < (int)(p_domain->tag_size + p_domain->len_size + msg_id_len))
        return FALSE;

    /* OK, it is long enough to read a TLV to validate the msg id */
    
    /* First read the T value */
    if (p_domain->tag_size==1)
        ie_tag = tvb_get_guint8(tvb, offset);
    else if (p_domain->tag_size==2)
        ie_tag = tvb_get_ntohs(tvb, offset);
    else if (p_domain->tag_size==4)
        ie_tag = tvb_get_ntohl(tvb, offset);
    else
        g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

    offset += p_domain->tag_size;

    /* This IE tag should be the Message Id tag */
    if (msg_tag != ie_tag)
        return FALSE;

    /* Second read the L value */
    if (p_domain->len_size==1)
        ie_len = tvb_get_guint8(tvb, offset);
    else if (p_domain->len_size==2)
        ie_len = tvb_get_ntohs(tvb, offset);
    else if (p_domain->len_size==4)
        ie_len = tvb_get_ntohl(tvb, offset);
    else
        g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

    offset += p_domain->len_size;

    if (ie_len != msg_id_len)
        return FALSE;

    offset += ie_len;

    /* The message Id TLV worked, so just parse for the whoe, message being a set of TLVs */

    while (offset < msg_len) {
        /* check whether the message is long enough to have a TLV for the IE */
        if (msg_len - offset < (int)(p_domain->tag_size + p_domain->len_size))
            return FALSE;

        /* First read the T value */
        if (p_domain->tag_size==1)
            ie_tag = tvb_get_guint8(tvb, offset);
        else if (p_domain->tag_size==2)
            ie_tag = tvb_get_ntohs(tvb, offset);
        else if (p_domain->tag_size==4)
            ie_tag = tvb_get_ntohl(tvb, offset);
        else
            g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

        offset += p_domain->tag_size;
    
        /* Second read the L value */
        if (p_domain->len_size==1)
            ie_len = tvb_get_guint8(tvb, offset);
        else if (p_domain->len_size==2)
            ie_len = tvb_get_ntohs(tvb, offset);
        else if (p_domain->len_size==4)
            ie_len = tvb_get_ntohl(tvb, offset);
        else
            g_assert_not_reached(); /* something wrong with the source code derived from the XML... */

        offset += p_domain->len_size;

        if (msg_len - offset < ie_len)
            return FALSE;
        
        offset += ie_len;
    }

    return TRUE;
}

static int get_ie_len_minimal(tvbuff_t *tvb, int offset, int len_max, int *p_len_size)
{
    int count;
    int len=0;
    guint8 octet;
    
    for (count=0; count<len_max; count++) {
        octet = tvb_get_guint8(tvb, offset++);
        len = (len<<7) + (octet & 0x7f); /* mask off the extension bit */
        if (!(octet & 0x80)) { /* extension bit cleared - no more */
            *p_len_size = count+1;
            break;
        }
    }
    if (octet & 0x80) { /* extension bit still set - must have hit size limit, so something wrong... */
        len = 0;        /* flag back to caller - error handling needed there */
        *p_len_size = 0;
    }
    
    return len;
}

void oyster3g_register_custom_dissectors(const oyster3g_domain_t *p_domain, custom_syntax_dissector_t *p_dissector)
{
    custom_syntax_dissector_t *p_d;
    const oyster3g_syntax_t *p_syntax;
	int	i;

    for (p_d=p_dissector; p_d->dissector; p_d++)
    {
		p_syntax=*(p_domain->syntax);

		for (i=0; p_syntax; i++,p_syntax=p_domain->syntax[i])
		{
			if (p_syntax->tag == p_d->syntax)
			{
				p_domain->custom_dissector[i] = p_d->dissector;
				break;
			}
		}
	}

    return;
}

void oyster3g_register_custom_hf(const oyster3g_domain_t *p_domain, hf_register_info *p_custom_hf, int custom_hf_len)
{
    hf_register_info *p_c_hf, *p_d_hf;
	int	i, j;

    for (p_c_hf=p_custom_hf,i=0; i<custom_hf_len; i++,p_c_hf++)
    {
		for (p_d_hf=p_domain->hf,j=0; j<p_domain->hf_len; j++,p_d_hf++)
		{
			if (!strcmp(p_c_hf->hfinfo.abbrev,p_d_hf->hfinfo.abbrev))  /* match the formal field name */
			{
				/* copy the custom hf fields to the set to be registered */
				p_d_hf->hfinfo = p_c_hf->hfinfo;
				break;
			}
		}
	}

    return;
}
