/* packet-oyster3g-domain.h
 * Header file for ip.access oyster3G message domain dissection
 * Copyright 2007, ip.access Ltd <support@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-oyster3g-domain.h 1.2 2008/02/26 21:31:34GMT Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef PACKET_OYSTER3G_DOMAIN_H
#define PACKET_OYSTER3G_DOMAIN_H

typedef enum {
    /* The first lot of Syntax types are the base types */
    OYSTER3G_SYNTAX_BOOLEAN,
    OYSTER3G_SYNTAX_ENUM,
    OYSTER3G_SYNTAX_SINT8,
    OYSTER3G_SYNTAX_SINT16,
    OYSTER3G_SYNTAX_SINT32,
    OYSTER3G_SYNTAX_UINT8,
    OYSTER3G_SYNTAX_UINT16,
    OYSTER3G_SYNTAX_UINT32,
    OYSTER3G_SYNTAX_STRING,
    OYSTER3G_SYNTAX_UNI_STRING,
    /* OYSTER3G_SYNTAX_VERSION_STRING, a normal string can be used by wireshark for this */
    OYSTER3G_SYNTAX_OCTET_ARRAY,
    /* The last lot of Syntax types are the complex types */
    OYSTER3G_SYNTAX_STRUCT,
    OYSTER3G_SYNTAX_ARRAY,
    OYSTER3G_SYNTAX_SET,
} oyster3g_syntax_type_t;

typedef struct oyster3g_ie_t   /* used for info elements and elements */
{
    gchar       *label;         /* optional - syntax label used if this is NULL */
    gint32      syntax;         /* the element index of the syntax in the syntax array - -1 used to indicate end of list */
    /* guint32     ordinal;     / * determines order of IEs - should always
                                   be defined here in order, so not needed */
    gboolean    mandatory;      /* include conditional and optional as false */
} oyster3g_ie_t;

typedef struct oyster3g_syntax_t
{
    gchar               *label;     /* short name for this Syntax - NULL to terminate list */
    gint32              tag;        /* the IE tag used for this Syntax - -1 to terminate list */
    gchar               *display;   /* the (long) Display name of this Syntax - NULL to terminate list */
    oyster3g_syntax_type_t   type;       /* the type of the value defined by this Syntax */
    const struct oyster3g_ie_t *ie;      /* only used for those syntax types which link to other elements:
                                       for sets/arrays it is a single element,
                                       for structs it is a list of IEs in presence/ordinal order */
    gint32              min;        /* the minimum value (int) or length (set, array, string) of the syntax */
    gint32              max;        /* the maximum value (int) or length (set, array, string) of the syntax */
    int                 *hf;        /* pointer to the wireshark display handle for the field - initialise to -1 */
} oyster3g_syntax_t;

typedef struct oyster3g_message_t
{
    gchar           *label;         /* NULL if end of list */
    gchar           *display_name;  /* NULL if end of list */
    gint32          tag;            /* -1 indicates the end of the list */
    const struct oyster3g_ie_t *ie;            /* a list of IE elements, in presence/ordinal order - NULL at the end of the list */
} oyster3g_message_t;

typedef int (*custom_dissector_t)(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);

typedef struct custom_syntax_dissector_t
{
	int                 syntax;
	custom_dissector_t  dissector;
} custom_syntax_dissector_t;

typedef struct oyster3g_domain_t
{
    /* the first part of this structure are constants loaded from the XML definitions */
    gchar                   *label;
    gchar                   *display_name;
    guint                   len_size;
    guint                   tag_size;
    gboolean                maximal;
    const struct oyster3g_message_t    **message;       /* a list of Message elements - terminated by a "Null" element */
    const struct oyster3g_syntax_t     **syntax;        /* a list of ALL Syntax elements - terminated by an "Null" element */

    /* elements below this line are handles and state values used in the dissection */
    /* these must be initialised during the registration and initialisation */

    custom_dissector_t     *custom_dissector; /* an array of custom syntax dissectors */
    int                     *proto;      /* the internal handle for this protocol */
    int                     ett_size;   /* size of the array of tree handles */
    gint                    **ett;      /* pointer to array of tree handles */
    int                     *hf_ie_tag;  /* standard filter for IE tag value */
    int                     *hf_ie_len;  /* standard filter for IE length value */
    hf_register_info		*hf;		 /* the complete hf array for this domain */
    int              	    hf_len;		 /* and its length */
} oyster3g_domain_t;

/* Function prototypes for all oyster3G dissectors */
void dissect_oyster3g_domain(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, const oyster3g_domain_t *p_domain);

/* Function prototype of the generic custom dissector registration */
void oyster3g_register_custom_dissectors(const oyster3g_domain_t *domain, custom_syntax_dissector_t *p_dissector);
void oyster3g_register_custom_hf(const oyster3g_domain_t *p_domain, hf_register_info *p_hf, int custom_hf_len);

#endif /* PACKET_OYSTER3G_DOMAIN_H */
