/* packet-rtp_mux.c
 * Routines for RTP Multiplex dissection
 * Copyright 2006, Neil Piercy <Neil.Piercy[at]ipaccess.com>
 *
 * $Id: epan/dissectors/packet-rtp_mux.c 1.5 2008/09/09 22:07:49BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <epan/packet.h>
#include <epan/proto.h>
#include <epan/conversation.h>

#include "prefs.h"
#include "packet-gsm.h"

static int dissect_mux_header(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
static int dissect_mux_samples(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
static void dissect_stream_id(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
static void dissect_ssrc(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);

/* Initialize the protocol and registered fields */
static int proto_rtp_mux = -1;
static int hf_header_len = -1;
static int hf_embedded_header_len = -1;
static int hf_header_type = -1;
static int hf_header_tag = -1;
static int hf_max_mux_pkt_size = -1;
static int hf_max_mux_pkt_interval = -1;
static int hf_stream_id = -1;
static int hf_stream_id_timeslot = -1;
static int hf_stream_id_channel_info = -1;
static int hf_stream_id_stream_create = -1;
static int hf_dest_ip_addr = -1;
static int hf_dest_port = -1;
static int hf_proxy_port = -1;
static int hf_sample_info_csd = -1;
static int hf_payload_type = -1;
static int hf_ssrc = -1;
static int hf_ssrc_site_id = -1;
static int hf_ssrc_trx_id = -1;
static int hf_ssrc_stream_type = -1;
static int hf_seq_no = -1;
static int hf_timestamp = -1;
static int hf_m_bit = -1;
static int hf_sample_len = -1;

/* Initialize the subtree pointers */
static int ett_rtp_mux = -1;
static int ett_rtp_mux_hdr = -1;
static int ett_rtp_mux_sample = -1;

/* Dissector handles for chaining */
static dissector_handle_t amr_handle=NULL;

/* The dynamic payload type which will be dissected as RTP-MUX */

static guint dynamic_payload_type = 0x60;
static guint temp_dynamic_payload_type = 0x60;

/* Standard lengths for fixed length samples */
#define GSM_FR_SAMPLE_LEN				33
#define GSM_EFR_SAMPLE_LEN				31
#define GSM_CSD_8_SAMPLE_LEN			20
#define GSM_CSD_16_SAMPLE_LEN			40

#define HEADER_TYPE_UNIPLEX				0x01
#define HEADER_TYPE_UNIPLEX_ACK			0x02
#define HEADER_TYPE_MULTIPLEX			0x03
#define HEADER_TYPE_MULTIPLEX_ACK		0x04
#define HEADER_TYPE_COMPRESSION			0x05
#define HEADER_TYPE_COMPRESSION_ACK		0x06


static const value_string sample_info_csd_vals[] = {
	{ 0, "Not CSD" },
	{ 1, "Uniplex Stream contains 8kbps intermediate rate CSD" },
	{ 2, "Uniplex Stream contains 16kbps intermediate rate CSD" },
	{ 3, "Uniplex Stream contains 32kbps intermediate rate CSD" },
	{ 4, "Uniplex Stream contains 64kbps intermediate rate CSD" },
	{ 0,	NULL }
};

static const value_string channel_info_vals[] = {
	{ 0, "Full Rate" },
	{ 1, "Half rate sub-channel 0" },
	{ 2, "Half rate sub-channel 1" },
	{ 0,	NULL }
};

static const value_string stream_type_vals[] = {
	{ 0, "Normal RTP" },
	{ 1, "Uniplex Stream" },
	{ 2, "RTP Multiplex Stream (BTS->BSC)" },
	{ 3, "SRTP Multiplex Stream (BTS->BSC)" },
	{ 4, "RTP Multiplex Stream (BSC->BTS)" },
	{ 5, "SRTP Multiplex Stream (BSC->BTS)" },
	{ 0,	NULL }
};

static const value_string header_type_vals[] = {
	{ HEADER_TYPE_UNIPLEX, "Uniplex Header" },
	{ HEADER_TYPE_UNIPLEX_ACK, "Uniplex Ack Header" },
	{ HEADER_TYPE_MULTIPLEX, "Multiplex Header" },
	{ HEADER_TYPE_MULTIPLEX_ACK, "Multiplex Ack Header" },
	{ HEADER_TYPE_COMPRESSION, "Compression Header" },
	{ HEADER_TYPE_COMPRESSION_ACK, "Compression Ack Header" },
	{ 0,	NULL }
};

static void
dissect_stream_id(tvbuff_t *tvb, int offset, packet_info *pinfo _U_, proto_tree *tree)
{
	guint8 stream_id;

	stream_id = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_uint_format(tree, hf_stream_id, tvb, offset, 1, stream_id, "StreamId: Timeslot=%d; ChannelInfo=%s; StreamCreate=%d", (stream_id&0xE0)>>5, rd_match_strval((stream_id&0x18)>>3,channel_info_vals), (stream_id&0x03)>>0 );
		proto_tree_add_uint_hidden(tree, hf_stream_id_timeslot, tvb, offset, 1, stream_id);
		proto_tree_add_uint_hidden(tree, hf_stream_id_channel_info, tvb, offset, 1, stream_id);
		proto_tree_add_uint_hidden(tree, hf_stream_id_stream_create, tvb, offset, 1, stream_id);
	}
}

static void
dissect_ssrc(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree)
{
	guint32 ssrc;

	ssrc = tvb_get_ntohl(tvb, offset);
	if (tree) {
		proto_tree_add_uint_format(tree, hf_ssrc, tvb, offset, 4, ssrc, "SSRC: SiteId=%d; TrxId=%d; StreamType=%s", ssrc>>16, (ssrc&0xf000)>>12, rd_match_strval((ssrc&0x0700)>>8,stream_type_vals));
	}
	dissect_stream_id(tvb, offset+3, pinfo, tree);
}

static int
dissect_mux_header(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *main_tree)
{
	guint8 len, header_type, embedded_header_len, header_tag, sample_info, payload_type;
	guint16 header_len, max_mux_pkt_size, max_mux_pkt_interval, port;
	guint32 ip_addr;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	header_len = tvb_get_ntohs(tvb,offset);
	if (main_tree)
		proto_tree_add_uint(main_tree, hf_header_len, tvb, offset, 1, header_len);
	offset+=2;

	for (len=0; len<header_len; len+=embedded_header_len+2) {
		header_type = tvb_get_guint8(tvb,offset);
		if (main_tree) {
			ti = proto_tree_add_text(main_tree, tvb, offset, -1, "Embedded Header" );
			tree = proto_item_add_subtree(ti, ett_rtp_mux_hdr);
			if (tree)
				proto_tree_add_uint(tree, hf_header_type, tvb, offset, 1, header_type);
		}
		offset++;

		embedded_header_len = tvb_get_guint8(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_embedded_header_len, tvb, offset, 1, embedded_header_len);
		offset++;
		if (ti)
			proto_item_set_len(ti, embedded_header_len+2);

		header_tag = tvb_get_guint8(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_header_tag, tvb, offset, 1, header_tag);
		offset++;

		switch (header_type) {
		case HEADER_TYPE_UNIPLEX:
			dissect_stream_id(tvb, offset, pinfo, tree);
			offset++;

			ip_addr = tvb_get_ipv4(tvb, offset);
			if (tree)
				proto_tree_add_ipv4(tree, hf_dest_ip_addr, tvb, offset, 4, ip_addr);
			offset += 4;

			port = tvb_get_ntohs(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_dest_port, tvb, offset, 2, port);
			offset += 2;

			port = tvb_get_ntohs(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_proxy_port, tvb, offset, 2, port);
			offset += 2;

			sample_info = tvb_get_guint8(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_sample_info_csd, tvb, offset, 1, sample_info);
			offset++;
			break;

		case HEADER_TYPE_UNIPLEX_ACK:
			dissect_stream_id(tvb, offset, pinfo, tree);
			offset++;
			break;

		case HEADER_TYPE_MULTIPLEX:
			max_mux_pkt_size = tvb_get_ntohs(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_max_mux_pkt_size, tvb, offset, 2, max_mux_pkt_size);
			offset += 2;
			max_mux_pkt_interval = tvb_get_ntohs(tvb, offset);
			if (tree)
				proto_tree_add_uint_format(tree, hf_max_mux_pkt_interval, tvb, offset, 2, max_mux_pkt_interval, "Maximum Multiplex RTP Time Interval: %dms", (guint32)max_mux_pkt_interval*20);
			offset += 2;
			break;

		case HEADER_TYPE_COMPRESSION:
			dissect_stream_id(tvb, offset, pinfo, tree);
			offset++;
			payload_type = tvb_get_guint8(tvb, offset);
			if (tree)
				proto_tree_add_uint(tree, hf_payload_type, tvb, offset, 1, payload_type);
			offset++;

			dissect_ssrc(tvb, offset, pinfo, tree);
			offset += 4;
			break;

		case HEADER_TYPE_COMPRESSION_ACK:
			dissect_stream_id(tvb, offset, pinfo, tree);
			offset++;
			break;

		default:
			/* only the standard tag */
			break;
		}
	}
	return offset;
}

static int
dissect_mux_samples(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *main_tree)
{
	guint16 msg_len, seq_no, len_m_e, len;
	guint32 timestamp, start;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	msg_len = tvb_length(tvb);

	for (; offset<msg_len; ) {
		start = offset;

		if (main_tree) {
			ti = proto_tree_add_text(main_tree, tvb, offset, -1, "Sample" );
			tree = proto_item_add_subtree(ti, ett_rtp_mux_hdr);
		}
		dissect_stream_id(tvb, offset, pinfo, tree);
		offset++;

		seq_no = tvb_get_ntohs(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_seq_no, tvb, offset, 2, seq_no);
		offset += 2;

		timestamp = tvb_get_ntohl(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_timestamp, tvb, offset, 4, timestamp);
		offset += 4;

		len_m_e = tvb_get_guint8(tvb,offset);
		if (tree)
			proto_tree_add_uint(tree, hf_m_bit, tvb, offset, 1, len_m_e);

		if (len_m_e & 0x40) {	/* test for extension bit */
			len = tvb_get_ntohs(tvb,offset) & 0x3fff;
			if (tree)
				proto_tree_add_uint(tree, hf_sample_len, tvb, offset, 2, len);
			offset += 2;
		}
		else {
			len = len_m_e & 0x3f;
			if (tree)
				proto_tree_add_uint(tree, hf_sample_len, tvb, offset, 1, len);
			offset++;
		}
		if (ti)
			proto_item_set_len(ti, len+offset-start);

		/* Cheat! In our restricted usage we should be able to work out the right payload dissector from the sample length! */
		switch (len) {
		case 0:
			break;
		case GSM_FR_SAMPLE_LEN:
			if (tree)
				proto_tree_add_text(tree, tvb, offset, len, "GSM FR Speech");
			break;
		case GSM_EFR_SAMPLE_LEN:
			if (tree)
				proto_tree_add_text(tree, tvb, offset, len, "GSM EFR Speech");
			break;
		case GSM_CSD_8_SAMPLE_LEN:
			if (tree)
				proto_tree_add_text(tree, tvb, offset, len, "GSM Circuit-Switched Data 8kb/s");
			break;
		case GSM_CSD_16_SAMPLE_LEN:
			if (tree)
				proto_tree_add_text(tree, tvb, offset, len, "GSM Circuit-Switched Data 16kb/s");
			break;
		default:	/* Assume AMR */
			next_tvb = tvb_new_subset(tvb, offset, len, len);
	        call_dissector(amr_handle, next_tvb, pinfo, tree);
		}

		offset += len;
	}

	return offset;
}

/* Code to actually dissect the packets */
static void
dissect_rtp_mux(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0, reported_length;
	guint8 byte;

/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti = NULL;
	proto_tree *rtp_mux_tree = NULL;

/* Make entries in Protocol column and Info column on summary display */
	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "RTP-MUX");
	if (tree) {
		ti = proto_tree_add_item(tree, proto_rtp_mux, tvb, 0, -1, FALSE);
		rtp_mux_tree = proto_item_add_subtree(ti, ett_rtp_mux);
	}

	/* first check if this is an empty payload - used for keepalives */
	reported_length = tvb_reported_length(tvb);
	if (reported_length==0) {
		if (tree)
			proto_tree_add_text(tree, tvb, offset, tvb_length(tvb), "Payload empty - Ethernet padding only");
		return;
	}

	/* first check for and, if found, dissect the optional Payload Header,
	   which is a set of Embedded Payload Headers */
	byte = tvb_get_guint8(tvb,offset);
	if (byte == 0xff) {
		offset++;
		offset = dissect_mux_header(tvb, offset, pinfo, rtp_mux_tree);
	}

	offset = dissect_mux_samples(tvb, offset, pinfo, rtp_mux_tree);

	return;
}

/* Register the protocol with Ethereal */
/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_rtp_mux(void)
{
	dissector_handle_t rtp_mux_handle;
	static int rtp_mux_prefs_initialized = FALSE;

	rtp_mux_handle = create_dissector_handle(dissect_rtp_mux, proto_rtp_mux);

	amr_handle = find_dissector("amr");

	if (!rtp_mux_prefs_initialized) {
		rtp_mux_prefs_initialized = TRUE;
	  }
	else {
		if ( dynamic_payload_type > 95 )
			dissector_delete("rtp.pt", dynamic_payload_type, rtp_mux_handle);
	}
	dynamic_payload_type = temp_dynamic_payload_type;

	if ( dynamic_payload_type > 95 ){
		dissector_add("rtp.pt", dynamic_payload_type, rtp_mux_handle);
	}
	dissector_add_string("rtp_dyn_payload_type","RTP-MUX", rtp_mux_handle);
}

/* This format is required because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_rtp_mux(void)
{
	module_t *rtp_mux_module;

	static hf_register_info hf[] = {
		{ &hf_header_len,
			{ "Header Length","rtp-mux.header.len",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Header Length", HFILL }
		},
		{ &hf_header_type,
			{ "Header Type",           "rtp-mux.header.type",
			FT_UINT8, BASE_DEC, VALS(header_type_vals), 0x0,
			"Header Type", HFILL }
		},
		{ &hf_embedded_header_len,
			{ "Embedded Header Length","rtp-mux.embedded_header.len",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Embedded Header Length", HFILL }
		},
		{ &hf_header_tag,
			{ "Header Tag",           "rtp-mux.header.tag",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"Header Tag", HFILL }
		},
		{ &hf_max_mux_pkt_size,
			{ "Maximum Mux Packet Size",      "rtp-mux.header.max_mux_pkt_size",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Maximum Multiplex Packet Size", HFILL }
		},
		{ &hf_max_mux_pkt_interval,
			{ "Maximum Mux Pkt Interval",   "rtp-mux.header.max_mux_pkt_interval",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Maximum Multiplex Packet Size", HFILL }
		},
		{ &hf_stream_id,
			{ "Stream Id",   "rtp-mux.stream_id",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"Stream Identifier", HFILL }
		},
		{ &hf_stream_id_timeslot,
			{ "Timeslot",   "rtp-mux.stream_id.timeslot",
			FT_UINT8, BASE_DEC, NULL, 0xE0,
			"Stream Identifier: Timeslot", HFILL }
		},
		{ &hf_stream_id_channel_info,
			{ "Channel Info",   "rtp-mux.stream_id.channel_info",
			FT_UINT8, BASE_HEX, VALS(channel_info_vals), 0x18,
			"Stream Identifier: Channel Info", HFILL }
		},
		{ &hf_stream_id_stream_create,
			{ "Stream Create",   "rtp-mux.stream_id.stream_create",
			FT_UINT8, BASE_DEC, NULL, 0x03,
			"Stream Identifier: Stream Create", HFILL }
		},
		{ &hf_dest_ip_addr,
			{ "Dest.IP Address",   "rtp-mux.header.dest_ip_addr",
			FT_IPv4, BASE_HEX, NULL, 0x0,
			"The Destination IP Address of the RTP stream", HFILL }
		},
		{ &hf_dest_port,
			{ "Dest. Port",   "rtp-mux.header.dest_port",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"The Destination Port on the RTP Stream", HFILL }
		},
		{ &hf_proxy_port,
			{ "Proxy Port",   "rtp-mux.header.proxy_port",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"The UDP Port on the A interface side of the BSC Proxy", HFILL }
		},
		{ &hf_sample_info_csd,
			{ "Sample Info (CSD)",   "rtp-mux.sample_info_csd",
			FT_UINT8, BASE_HEX, VALS(sample_info_csd_vals), 0x0,
			"The Sample Information CSD data rate", HFILL }
		},
		{ &hf_payload_type,
			{ "Payload Type",   "rtp-mux.payload_type",
			FT_UINT8, BASE_HEX, NULL, 0x0,
			"The RTP payload type of the Sample", HFILL }
		},
		{ &hf_ssrc,
			{ "SSRC",   "rtp-mux.ssrc",
			FT_UINT32, BASE_HEX, NULL, 0x0,
			"The SSRC of the Sample", HFILL }
		},
		{ &hf_ssrc_site_id,
			{ "SSRC(SiteId)",   "rtp-mux.ssrc.site_id",
			FT_UINT32, BASE_DEC, NULL, 0xffff0000,
			"The SSRC(SiteId)", HFILL }
		},
		{ &hf_ssrc_trx_id,
			{ "SSRC(TrxId)",   "rtp-mux.ssrc.trx_id",
			FT_UINT32, BASE_DEC, NULL, 0xf000,
			"The SSRC(TrxId)", HFILL }
		},
		{ &hf_ssrc_stream_type,
			{ "SSRC(StreamType)",   "rtp-mux.ssrc.stream_type",
			FT_UINT32, BASE_HEX, VALS(stream_type_vals), 0x0700,
			"The SSRC(StreamType)", HFILL }
		},
		{ &hf_seq_no,
			{ "Sequence No.",   "rtp-mux.seq_no",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"The RTP Header Sequence Number", HFILL }
		},
		{ &hf_timestamp,
			{ "Timestamp",   "rtp-mux.timestamp",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"The RTP Header Timestamp", HFILL }
		},
		{ &hf_m_bit,
			{ "M",   "rtp-mux.m_bit",
			FT_UINT8, BASE_DEC, NULL, 0x80,
			"The RTP Header M bit", HFILL }
		},
		{ &hf_sample_len,
			{ "Sample Length",   "rtp-mux.sample_len",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"RTP payload Sample Length", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_rtp_mux,
		&ett_rtp_mux_hdr,
		&ett_rtp_mux_sample,
	};

	/* Register the protocol name and description */
	proto_rtp_mux = proto_register_protocol("nanoGSM RTP Multiplex Data","RTP-MUX", "rtp-mux");

	/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_rtp_mux, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
	/* Register a configuration option for port */

	rtp_mux_module = prefs_register_protocol(proto_rtp_mux, proto_reg_handoff_rtp_mux);

	prefs_register_uint_preference(rtp_mux_module, "dynamic.payload.type",
								   "RTP-MUX dynamic payload type",
								   "The dynamic payload type which will be interpreted as RTP-MUX",
								   10,
								   &temp_dynamic_payload_type);
}
