/* packet-sccp-lite.c
 * Routines for SCCP-lite dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-sccp-lite.c 1.1 2006/07/13 18:50:53BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"
#include "packet-mtp3.h"

/* function prototypes for this file */
static int dissect_sccp_lite_transport(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);

/* fixed TCP port number */
#define TCP_PORT_SCCP_LITE	      5000

/* Constants used to validate message lengths and in the heuristic checker */
#define MSG_LEN_MIN		  3
#define MSG_LEN_MAX		254

/* Initialize the protocol and registered fields */
static int proto_sccp_lite = -1;
static int hf_sccp_lite_msg_len = -1;

static dissector_handle_t sccp_handle;
static dissector_handle_t ipa_handle;

/* Initialize the subtree pointers */
static gint ett_sccp_lite = -1;

/* Code to actually dissect the packets - actually only a stub to allow the real
   routine (which this calls) to recurse into itself for packets which contain more than one message
*/
static void
dissect_sccp_lite(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	guint8 msg_len;
	int offset=0, this_msg_len;

	if (check_col(pinfo->cinfo, COL_INFO))
		col_add_str(pinfo->cinfo, COL_INFO, "");	/* the COL_INFO is appended every time, so start blank */

	this_msg_len = tvb_length(tvb); /* for safety until we know better */

	/* FIXME - kludge: assume packet is a complete message or more */
	msg_len = tvb_get_guint8(tvb, offset);
	
	if (msg_len < 2) {	/* probably be the ip.access SCCP-over-IP format */
		/* load the display labels 	*/
		pinfo->current_proto = "SCCP-over-IP";
	
		call_dissector(ipa_handle, tvb, pinfo, tree);
	}
	else {
		/* load the display labels 	*/
		pinfo->current_proto = "SCCP-lite";
		
		while (this_msg_len > offset) {
			offset += dissect_sccp_lite_transport(tvb, offset, pinfo, tree);

			/* If packet contains a follow-on message, add a tab separator */
			if (tvb_length_remaining(tvb,offset)>0) {
				if (check_col(pinfo->cinfo, COL_INFO))
					col_append_str(pinfo->cinfo, COL_INFO, "\t");

			}
		}
	}
}

static int
dissect_sccp_lite_transport(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree)
{
	guint8 msg_len;
	tvbuff_t *new_tvb;

	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "sccp");

	msg_len = tvb_get_guint8(tvb, offset);
	offset += 1;

	/* use a new subset TVB for the decode of the SCCP message */
	new_tvb = tvb_new_subset(tvb, offset, msg_len, -1);

	call_dissector(sccp_handle,new_tvb,pinfo,tree);

	offset += msg_len;

	return msg_len + 1;
}

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_sccp_lite(void)
{		 

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_sccp_lite_msg_len,
			{ "Msg Len", "sccp_lite.msg_len",
			FT_UINT8, BASE_DEC, NULL, 0x0,	  
			"SCCP-lite Message Length", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_sccp_lite,
	};

	proto_sccp_lite = proto_register_protocol("SCCP-lite", "SCCP-lite", "sccp_lite");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_sccp_lite, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_sccp_lite(void)
{
	dissector_handle_t sccp_lite_handle;

	sccp_lite_handle = create_dissector_handle(dissect_sccp_lite, proto_sccp_lite);

    sccp_handle = find_dissector("sccp");
    ipa_handle = find_dissector("ipa");

#ifdef TCP_PORT_SCCP_LITE
	dissector_add("tcp.port", TCP_PORT_SCCP_LITE, sccp_lite_handle);
#endif
}
