/* packet-umts_macd.c
 *
 * Routines for UMTS MAC-D packet disassembly
 *
 * $Id: epan/dissectors/packet-umts_macd.c 1.7 2008/11/26 16:51:04GMT Neil P Piercy (npp) Exp  $
 *
 * Copyright (c) 2007 by ip.access Ltd <support@ipaccess.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1999 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gmodule.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/expert.h>

#include "packet-umts_macd.h"

/* Function prototypes */
void proto_register_macd (void);
void proto_reg_handoff_macd (void);
void dissect_dcch_macd (tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
void dissect_dtch_macd (tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);

/* Statics for handles etc */
static int proto_dtch_macd = -1;	/*stores protocol id that is returned during registration*/
static int proto_dcch_macd = -1;	/*stores protocol id that is returned during registration*/

static dissector_handle_t rlc_um_handle;
static dissector_handle_t rlc_am_handle;
static dissector_handle_t data_handle;

static int hf_macd_logical_channel = -1;

static hf_register_info hf [] =
{
	{
		&hf_macd_logical_channel,
		{
			"(C/T)Logical Channel",
			"macd.ct",
			FT_UINT8,
			BASE_DEC,
			NULL,
			0x0,
			"",
			HFILL
		}
	}
};

static int ett_macd = -1;
static gint *ett[] =
{
	&ett_macd,
};

static tvbuff_t *new_octet_aligned_subset(tvbuff_t *tvb, guint32 offset, guint32 length)
{
  tvbuff_t *sub_tvb = NULL;
  guint32 boffset = offset >> 3;
  unsigned int i, shift0, shift1;
  guint8 octet0, octet1, *buf;

  if (offset & 0x07) {  /* unaligned */
    shift1 = offset & 0x07;
    shift0 = 8 - shift1;
    buf = ep_alloc(length);
    octet0 = tvb_get_guint8(tvb, boffset);
    for (i=0; i<length; i++) {
      octet1 = octet0;
      octet0 = tvb_get_guint8(tvb, boffset + i + 1);
      buf[i] = (octet1 << shift1) | (octet0 >> shift0);
    }
    sub_tvb = tvb_new_real_data(buf, length, length);
    tvb_set_child_real_data_tvbuff(tvb, sub_tvb);
  } else {  /* aligned */
    sub_tvb = tvb_new_subset(tvb, boffset, length, length);
  }
  return sub_tvb;
}

static const guint16 bit_mask16[] = {
    0xffff,
    0x7fff,
    0x3fff,
    0x1fff,
    0x0fff,
    0x07ff,
    0x03ff,
    0x01ff
};

/* Fetch a number of bits to a new tvb right adjusted to the nearest number of bytes.
 * (add proceeding zeros).
 */
tvbuff_t *new_octet_aligned_subset_bits(tvbuff_t *tvb, guint32 offset, guint32 no_of_bits)
{
  tvbuff_t *sub_tvb = NULL;
  guint32 boffset = offset >> 3;
  unsigned int i, shift0, shift1;
  guint8 octet0, octet1, *buf;
  guint16	word;
  guint32	length;
  guint32	remainder;

  /* Calculate the size reqired */

  length = no_of_bits/8;
  remainder = no_of_bits % 8;
  if(remainder){
	  length++;
  }else{
	  /* Number of bits = even number of octets */
	return new_octet_aligned_subset(tvb, offset, length);
  }
  buf = ep_alloc(length);

  /* get the 'odd' bits */
  shift1 = offset & 0x07;
  word = tvb_get_ntohs(tvb,boffset) & bit_mask16[offset & 0x07];
  word = word >> (16-(shift1+remainder));
  buf[0] = word & 0x00ff;

  offset = offset + remainder;
  boffset = offset >> 3;
  if (length >1){
    shift1 = offset & 0x07;
    shift0 = 8 - shift1;
    octet0 = tvb_get_guint8(tvb, boffset);
    for (i=1; i<length; i++) {
      octet1 = octet0;
      octet0 = tvb_get_guint8(tvb, boffset + i);
      buf[i] = (octet1 << shift1) | (octet0 >> shift0);
    }
  }
  sub_tvb = tvb_new_real_data(buf, length, length);
  tvb_set_child_real_data_tvbuff(tvb, sub_tvb);

  return sub_tvb;
}

void
proto_register_macd (void)
{
	/* Preference registration - none yet
	module_t *macd_dtch_module,*macd_dcch_module;
    */

	proto_dcch_macd = proto_register_protocol("UMTS MACD DCCH over DCH", "MACD-DCCH-DCH", "macd_dcch_dch");
    register_dissector("macd_dcch_dch", dissect_dcch_macd, proto_dcch_macd);

	proto_dtch_macd = proto_register_protocol("UMTS MACD DTCH over DCH", "MACD-DTCH-DCH", "macd_dtch_dch");
    register_dissector("macd_dtch_dch", dissect_dtch_macd, proto_dtch_macd);

	proto_register_field_array (proto_dtch_macd, hf, array_length (hf));

	proto_register_subtree_array (ett, array_length (ett));

	/* Preference registration - none yet
	macd_dtch_module = prefs_register_protocol (proto_dtch_macd, proto_reg_handoff_macd);
	macd_dcch_module = prefs_register_protocol (proto_dcch_macd, proto_reg_handoff_macd);
	*/
}

void
proto_reg_handoff_macd (void)
{
	static int Initialized = FALSE;

	if (!Initialized) {
		rlc_am_handle = find_dissector("rlc_am");
		rlc_um_handle = find_dissector("rlc_um");
	    data_handle = find_dissector ("data");
        Initialized = TRUE;
	}
}

void dissect_dcch_macd (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    tvbuff_t *next_tvb;
    gint offset = 0;
    proto_item *ti = NULL;
    proto_tree *root_ptree=NULL,*macd_tree = NULL;
    guint8  macd_header,ct_field;

	if (!tvb_bytes_exist (tvb, 0,1 ))
		return;
    
	if (ptree) {
        root_ptree =  proto_tree_get_root(ptree);
		ti = proto_tree_add_item (root_ptree, proto_dcch_macd, tvb, 0,
				1, FALSE);
		macd_tree = proto_item_add_subtree (ti, ett_macd);
    }

    macd_header = tvb_get_guint8(tvb,offset);
    ct_field = (macd_header >> 4);

    /* In case of DCCH multiplexed over DCH only the (C/T) field is present
     * which is of 4 bits
     */
    proto_tree_add_uint (macd_tree, hf_macd_logical_channel, tvb, offset, 1, ct_field);

    /* the following produces a new TVB, omitting the first 4 (and last 4, which must be padding) bits */
    next_tvb = new_octet_aligned_subset_bits(tvb, 4, (tvb_length_remaining(tvb,offset)-1)*8);
    add_new_data_source(pinfo, next_tvb, "Shifted RLC PDU");
    if (ct_field == 0 ) /* Logical DCCH Channel 0 which shd be RLC-UM */
    {
        call_dissector (rlc_um_handle, next_tvb, pinfo, root_ptree);
    }
    else
    {
        call_dissector (rlc_am_handle, next_tvb,pinfo, root_ptree);
    }
    return;
}

void dissect_dtch_macd (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    tvbuff_t *next_tvb;
    gint offset = 0;
    proto_item *ti = NULL;
    proto_tree *root_ptree=NULL,*macd_tree = NULL;
    guint8  macd_header,ct_field;

	if (!tvb_bytes_exist (tvb, 0,1 ))
		return;
    
	if (ptree) {
        root_ptree =  proto_tree_get_root(ptree);
    }

    if (tvb_length(tvb)==(344/8)) { /* include MAC C/T header for multiplexing */
        if (ptree) {
		    ti = proto_tree_add_item (root_ptree, proto_dcch_macd, tvb, 0, 1, FALSE);
			macd_tree = proto_item_add_subtree (ti, ett_macd);
		}
		macd_header = tvb_get_guint8(tvb,offset);
		ct_field = (macd_header >> 4);

		/* In case of multiplexed over DCH only the (C/T) field is present
		 * which is of 4 bits
		 */
		proto_tree_add_uint (macd_tree, hf_macd_logical_channel, tvb, offset, 1, ct_field);

		/* the following produces a new TVB, omitting the first 4 (and last 4, which must be padding) bits */
		next_tvb = new_octet_aligned_subset_bits(tvb, 4, (tvb_length_remaining(tvb,offset)-1)*8);
		add_new_data_source(pinfo, next_tvb, "Shifted RLC PDU");
		if (ct_field == 0 ) /* Logical DCCH Channel 0 which should be RLC-UM */
		{
			call_dissector (rlc_um_handle, next_tvb, pinfo, root_ptree);
		}
		else
		{
			call_dissector (rlc_am_handle, next_tvb,pinfo, root_ptree);
	    }
	}
	else
		call_dissector (rlc_am_handle, tvb, pinfo, root_ptree);

    return;
}
