/* packet-umts_rlc.c
 *
 * Dissectors for UMTS RLC packet disassembly
 *
 * $Id: epan/dissectors/packet-umts_rlc.c 1.7 2008/10/23 23:08:21BST Neil P Piercy (npp) Exp  $
 *
 * Copyright (c) 2007 by ip.access Ltd <support@ipaccess.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1999 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gmodule.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/expert.h>

#include "packet-umts_rlc.h"

/* Function prototypes */
void proto_register_rlc (void);
void proto_reg_handoff_rlc (void);
void dissect_um_rlc (tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
void dissect_am_rlc (tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);

guint8 lsb_bitmask[] = { 0x0,0x1,0x3,0x7,0xf,0x1f,0x3f,0x7f,0xff };
guint8 msb_bitmask[] = { 0x0,0x80,0xC0,0xE0,0xF0,0xF8,0xFC,0xFE,0xFF };

/* Statics for handles etc */
static int proto_am_rlc = -1;    /*stores protocol id that is returned during registration*/
static int proto_um_rlc = -1;    /*stores protocol id that is returned during registration*/

static dissector_handle_t rlc_am_handle;
static dissector_handle_t rlc_um_handle;

/* Global preferences */
static gboolean global_dissect_am_payloads = FALSE;
static gboolean global_dissect_um_payloads = FALSE;

static const value_string rlc_sufi_proto_strs[] = {
    { NO_MORE_SUFI,"NO_MORE"},
    { WINDOW_SUFI,"WINDOW"},
    { ACK_SUFI,"ACK"},
    { LIST_SUFI,"LIST"},
    { BITMAP_SUFI,"BITMAP"},
    { RLIST_SUFI,"RLIST"},
    { MRW_SUFI,"MRW"},
    { MRW_ACK_SUFI,"MRW_ACK"},
    { 0, NULL }
};

static const value_string dc_strs[] = {
    { 0,"Control PDU"},
    { 1,"Data PDU"},
    { 0, NULL }
};

static const value_string pdu_type_strs[] = {
    { 0,"STATUS"},
    { 1,"RESET"},
    { 2,"RESET ACK"},
    { 0, NULL }
};

static const value_string p_strs[] = {
    { 0,"Status Report not Requested"},
    { 1,"Request a Status Report"},
    { 0, NULL }
};

static const value_string e_strs[] = {
    { 0,"No Length Indicator follows"},
    { 1,"Length Indicator follows"},
    { 0, NULL }
};

static const value_string he_strs[] = {
    { 0,"No Length Indicator follows"},
    { 1,"Length Indicator follows"},
    { 0, NULL }
};

static const value_string rlc_um_7bits_special_li_strs[] = {
    { 0,"Previous RLC PDU was exactly filled with last seg of an RLC SDU and no li in previous SDU to indicate end of RLC SDU"},
    { 124,"First data octet of RLC SDU"},
    { 125,"RESERVED"},
    { 126,"RESERVED"},
    { 127,"Rest of RLC PDU is padding"},
    { 0, NULL }
};

static const value_string rlc_am_7bits_special_li_strs[] = {
    { 0,"Previous RLC PDU was exactly filled with last seg of an RLC SDU and no li in previous SDU to indicate end of RLC SDU"},
    { 124,"RESERVED"},
    { 125,"RESERVED"},
    { 126,"Rest of RLC PDU includes a piggybacked STATUS PDU"},
    { 127,"Rest of RLC PDU is padding"},
    { 0, NULL }
};

static int hf_rlc_am_seq_no = -1;
static int hf_rlc_um_seq_no = -1;
static int hf_rlc_li = -1;
static int hf_rlc_li_tree = -1;
static int hf_rlc_status_pdu = -1;
static int hf_rlc_reset_pdu = -1;
static int hf_rlc_reset_ack_pdu = -1;
static int hf_rlc_am_data_pdu = -1;
static int hf_rlc_dc_field = -1;
static int hf_rlc_pdu_type = -1;
static int hf_rlc_polling_bit = -1;
static int hf_rlc_extn_bit = -1;
static int hf_rlc_he = -1;
static int hf_rlc_rsn = -1;
static int hf_rlc_lsn = -1;
static int hf_rlc_wsn = -1;
static int hf_rlc_hfni = -1;
static int hf_rlc_fsn = -1;
static int hf_bitmap_sufi = -1;
static int hf_ack_sufi = -1;
static int hf_list_sufi = -1;
static int hf_rlist_sufi = -1;
static int hf_window_sufi = -1;
static int hf_rlc_missing_seq_data = -1;
static int hf_rlc_user_data = -1;

static hf_register_info hf [] =
{
    {
        &hf_rlc_am_seq_no,
        {
            "Sequence No",
            "rlc_am.seq_no",
            FT_UINT16,
            BASE_DEC,
            NULL,
            0x7ff8,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_um_seq_no,
        {
            "Sequence No",
            "rlc_um.seq_no",
            FT_UINT8,
            BASE_DEC,
            NULL,
            0xfe,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_status_pdu,
        {
            "Status PDU",
            "rlc_am.status_pdu",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_reset_ack_pdu,
        {
            "Reset Ack PDU",
            "rlc_am.reset_ack_pdu",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_am_data_pdu,
        {
            "AM Data PDU",
            "rlc_am.am_data_pdu",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_reset_pdu,
        {
            "Reset PDU",
            "rlc_am.reset_pdu",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_li,
        {
            "Length Indicator",
            "rlc.li",
            FT_UINT8,
            BASE_DEC,
            NULL,
            0xfe,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_li_tree,
        {
            "Length Indicators",
            "rlc.li_tree",
            FT_NONE,
            BASE_NONE,
            NULL,
            0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_dc_field,
        {
            "D/C Field",
            "rlc.dc_field",
            FT_UINT16,
            BASE_DEC,
            VALS(dc_strs),
            0x8000,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_pdu_type,
        {
            "PDU Type",
            "rlc.pdu_type",
            FT_UINT8,
            BASE_HEX,
            VALS(pdu_type_strs),
            0x70,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_polling_bit,
        {
            "Polling Bit",
            "rlc.polling_bit",
            FT_UINT16,
            BASE_DEC,
            VALS(p_strs),
            0x0004,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_extn_bit,
        {
            "Extn Bit",
            "rlc.extn_bit",
            FT_UINT8,
            BASE_DEC,
            VALS(e_strs),
            0x01,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_he,
        {
            "Header Extension",
            "rlc.he",
            FT_UINT16,
            BASE_DEC,
            VALS(he_strs),
            0x0003,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_lsn,
        {
            "Last Sequence Number",
            "rlc.lsn",
            FT_UINT16,
            BASE_DEC,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_wsn,
        {
            "Window Size Number",
            "rlc.wsn",
            FT_UINT16,
            BASE_DEC,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_fsn,
        {
            "Bitmap First SN",
            "rlc.fsn",
            FT_UINT16,
            BASE_DEC,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_rsn,
        {
            "Reset Sequence Number",
            "rlc.rsn",
            FT_UINT8,
            BASE_DEC,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_hfni,
        {
            "Hyper Frame Number Indicator",
            "rlc.hfni",
            FT_UINT32,
            BASE_DEC,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_user_data,
        {
            "User data",
            "rlc.user_data",
            FT_BYTES,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlc_missing_seq_data,
        {
            "Missing Sequence No.s List",
            "rlc.missing_seq_data",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_bitmap_sufi,
        {
            "BITMAP SUFI",
            "rlc.bitmap_sufi",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_ack_sufi,
        {
            "ACK SUFI",
            "rlc.ack_sufi",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_window_sufi,
        {
            "WINDOW SUFI",
            "rlc.window_sufi",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_list_sufi,
        {
            "LIST SUFI",
            "rlc.list_sufi",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
    {
        &hf_rlist_sufi,
        {
            "RLIST SUFI",
            "rlc.rlist_sufi",
            FT_NONE,
            BASE_NONE,
            NULL,
            0x0,
            "",
            HFILL
        }
    },
};

static int ett_rlc = -1;
static int ett_bitmap = -1;
static int ett_li = -1;
static int ett_rlist = -1;
static int ett_ack = -1;
static int ett_list = -1;
static int ett_window = -1;
static int ett_missing_sn = -1;
static gint *ett[] =
{
    &ett_rlc,
    &ett_bitmap,
    &ett_li,
    &ett_rlist,
    &ett_ack,
    &ett_list,
    &ett_window,
    &ett_missing_sn,
};

static unsigned char is_special_li(unsigned char li)
{
    if ((li == 0) || (li == 124) || (li == 125)
            || (li == 126) || (li == 127))
    {
        return 1;
    }
    else
        return 0;
}

static dissector_handle_t data_handle;


void
proto_register_rlc (void)
{
    module_t *rlc_am_module,*rlc_um_module;

    proto_am_rlc = proto_register_protocol("RLC Acknowledged Mode", "RLC_AM", "rlc_am");
    register_dissector("rlc_am", dissect_am_rlc, proto_am_rlc);

    proto_um_rlc = proto_register_protocol("RLC UnAcknowledged Mode", "RLC_UM", "rlc_um");
    register_dissector("rlc_um", dissect_um_rlc, proto_um_rlc);

    proto_register_field_array (proto_um_rlc, hf, array_length (hf));

    proto_register_subtree_array (ett, array_length (ett));

    /* Preference registration */
    rlc_am_module = prefs_register_protocol (proto_am_rlc, proto_reg_handoff_rlc);
    rlc_um_module = prefs_register_protocol (proto_um_rlc, proto_reg_handoff_rlc);

    /* Register preferences */
    prefs_register_bool_preference(rlc_am_module, "dissect_payloads",
         "Dissect Payloads",
         "Allows the dissection of payloads, but fails if they are encrypted.",
         &global_dissect_am_payloads);

    /* Register preferences */
    prefs_register_bool_preference(rlc_um_module, "dissect_payloads",
         "Dissect Payloads",
         "Allows the dissection of payloads, but fails if they are encrypted.",
         &global_dissect_um_payloads);
}

void
proto_reg_handoff_rlc (void)
{
    static int Initialized = FALSE;

    if (!Initialized) {
        rlc_am_handle = find_dissector("rlc_am");
        rlc_um_handle = find_dissector("rlc_um");
        data_handle = find_dissector ("data");
        Initialized = TRUE;
    }
}

const char *rlc_convert_to_sufi_str(int sufi) {
    const char *str;

    str = match_strval(sufi, rlc_sufi_proto_strs);
    if (NULL == str)
        str = "Reserved";
    return str;
}

const char *rlc_convert_7bit_special_lis(int li,int type) {
    const char *str=NULL;

    if (type == RLC_UM_LI)
        str = match_strval(li, rlc_um_7bits_special_li_strs);
    if (type == RLC_AM_LI)
        str = match_strval(li, rlc_am_7bits_special_li_strs);
    if (NULL == str)
        str = "Unknown";
    return str;
}

#define LI_MAX 15 /* the maximum number of LIs which an RLC PDU can contain */
static guint8 li_arr[LI_MAX];
static guint8 li_counter;

void dissect_length_indicator(tvbuff_t *tvb,packet_info *pinfo, proto_tree *ptree,int *offset,int *data_pdu_len_in_bits,guint8 li_type)
{
    guint8  li, extn_bit, li_last=0, pdu_len;
    proto_item *ti=NULL, *error_ti=NULL;
    proto_tree *li_tree=NULL;

    pdu_len =  tvb_length_remaining(tvb,*offset);
    li_counter = 0;
    
    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_rlc_li_tree, tvb, *offset, -1, FALSE);
        li_tree = proto_item_add_subtree (ti, ett_li);
    }

    do
    {
        if (li_counter>=LI_MAX) {
            error_ti = proto_tree_add_text(li_tree, tvb, *offset, 1, "Maximum number of Length Indicators exceeded - Encrypted Payloads? Aborted.");
            expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Maximum number of Length Indicators exceeded - Encrypted Payloads? Aborted.");
            li_counter = 0; /* abort and dont let anything beyond try to use these LIs */
            break; /* something has gone wrong - abort before we run off the end of the array */
        }
        
        li = tvb_get_guint8(tvb,*offset);
        extn_bit = li & 0x1;
        li_arr[li_counter] = li>>1;
        if (is_special_li(li_arr[li_counter]) != 0)
        {
            if (li_tree) {
                proto_tree_add_uint_format_value (li_tree, hf_rlc_li, tvb, *offset, 1, li,"%s (Special Li)", rlc_convert_7bit_special_lis(li_arr[li_counter],li_type));
                proto_tree_add_item(li_tree, hf_rlc_extn_bit, tvb, *offset, 1, FALSE);
            }
            if (li_last>=124) {
                error_ti = proto_tree_add_text(li_tree, tvb, *offset, 1, "Illegal second special Length Indicator - Encrypted Payloads? Aborted.");
                expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Illegal second special Length Indicator - Encrypted Payloads? Aborted.");
                li_counter = 0; /* abort and dont let anything beyond try to use these LIs */
                break; /* something has gone wrong - abort before we run off the end of the array */                
            }
            else if ((li_arr[li_counter]==0)&&(li_last>0)) {
                error_ti = proto_tree_add_text(li_tree, tvb, *offset, 1, "Illegal special Length Indicator 0 not the first Length Indicator - Encrypted Payloads? Aborted.");
                expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Illegal special Length Indicator 0 not the first Length Indicator - Encrypted Payloads? Aborted.");
                li_counter = 0; /* abort and dont let anything beyond try to use these LIs */
                break; /* something has gone wrong - abort before we run off the end of the array */                
            }
        }
        else
        {
            (*data_pdu_len_in_bits) += li_arr[li_counter]<<3;
            
            if (li_tree) {
                proto_tree_add_item(li_tree, hf_rlc_li, tvb, *offset, 1, FALSE);
                proto_tree_add_item(li_tree, hf_rlc_extn_bit, tvb, *offset, 1, FALSE);
            }
            if ((li_arr[li_counter]<=li_last) || (li_counter+li_arr[li_counter]>=pdu_len)) {
                error_ti = proto_tree_add_text(li_tree, tvb, *offset, 1, "Inconsistent Length Indicator - Encrypted Payloads? Aborted.");
                expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Inconsistent Length Indicator - Encrypted Payloads? Aborted.");
                li_counter = 0; /* abort and dont let anything beyond try to use these LIs */
                break; /* something has gone wrong - abort before we run off the end of the array */                
            }
        }
        li_last = li_arr[li_counter];
        ++(*offset);
        ++li_counter;
    } while(extn_bit == 1);

    proto_item_set_len(ti,li_counter);

    return;
}

void dissect_um_rlc (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    tvbuff_t *next_tvb;
    gint offset = 0;
    proto_item *ti = NULL;
    proto_tree *rlc_um_tree = NULL;
    guint8     rlc_um_header;
    gint  no_of_data_bits;

    if (!tvb_bytes_exist (tvb, 0,RLC_MIN_UM_HEADER_SIZE ))
        return;

    if (check_col (pinfo->cinfo, COL_PROTOCOL))
        col_set_str (pinfo->cinfo, COL_PROTOCOL, "RLC-UM");

    if (ptree) {
        ti = proto_tree_add_item (ptree, proto_um_rlc, tvb, 0,
                tvb_length(tvb), FALSE);
        rlc_um_tree = proto_item_add_subtree (ti, ett_rlc);
    }

    rlc_um_header = tvb_get_guint8(tvb,offset);
    if (rlc_um_tree) {
        proto_tree_add_uint (rlc_um_tree, hf_rlc_um_seq_no, tvb, offset, 1, rlc_um_header);
        proto_tree_add_uint (rlc_um_tree, hf_rlc_extn_bit, tvb, offset, 1, rlc_um_header);
    }
    offset++;

    if (check_col(pinfo->cinfo, COL_INFO)) {
        col_prepend_fstr(pinfo->cinfo,COL_INFO,"SeqNo: %d ", (rlc_um_header >>1));
    }

    li_counter = 0;
    if (global_dissect_um_payloads && ((rlc_um_header & 0x1) == 1)) /* If extn bit states that the next field is a length indicator and e-bit */
    {
        dissect_length_indicator(tvb,pinfo,rlc_um_tree,&offset,&no_of_data_bits,RLC_UM_LI);
    }

    if (tvb_bytes_exist (tvb, offset, 1))
    {
        next_tvb = tvb_new_subset (tvb, offset, -1, -1);
        call_dissector (data_handle, next_tvb, pinfo, ptree);
    }

    return;
}

guint8  extract_4_bit_number (tvbuff_t *tvb,int * byte_offset, int * bit_offset)
{
    guint8  no,temp_no = tvb_get_guint8( tvb,*byte_offset);
    /** Assuming here that the *bit_offset would be either 4 or 0 **/
    no =  (temp_no >> (4 - *bit_offset) ) & 0xf;
    *bit_offset = 4 - *bit_offset;
    if (*bit_offset == 0)
    {
        (*byte_offset)++;
    }
    return no;
}

guint16 extract_12_bit_number(tvbuff_t *tvb,int * byte_offset, int * bit_offset)
{
    guint8  data_bits_to_copy;
    guint16  sn ,sn_temp;
    
    data_bits_to_copy = 8 - *bit_offset;
    sn_temp =  tvb_get_guint8(tvb,*byte_offset);
    sn = (sn_temp & lsb_bitmask [data_bits_to_copy]) << (12 - data_bits_to_copy) ;
    (*byte_offset)++;
    data_bits_to_copy = 12 - data_bits_to_copy;
    sn_temp =  tvb_get_guint8(tvb,*byte_offset);
    *bit_offset = 8 -data_bits_to_copy;
    sn |= (sn_temp & msb_bitmask [data_bits_to_copy] ) >> (*bit_offset);
    if (*bit_offset == 0)
    {
        (*byte_offset)++;
    }
    return sn;
}

void dissect_window_sufi (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
    proto_tree *window_tree=NULL;
    proto_item *ti=NULL;
    guint16  wsn ;
    int      start_byte_offset = *byte_offset;

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_window_sufi, tvb, *byte_offset,
                 2, FALSE);
        window_tree = proto_item_add_subtree (ti, ett_window);
    }

    wsn = extract_12_bit_number ( tvb,byte_offset,bit_offset);

    if (window_tree)
        proto_tree_add_uint (window_tree, hf_rlc_wsn, tvb, start_byte_offset, 2, wsn);

    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO,"WIN_SUFI WSN: %d ", wsn);
    }
}

void dissect_ack_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
    proto_tree *ack_tree=NULL;
    proto_item *ti=NULL;
    guint16  lsn ;
    int      start_byte_offset = *byte_offset;

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_ack_sufi, tvb, *byte_offset,
                 2, FALSE);
        ack_tree = proto_item_add_subtree (ti, ett_ack);
    }
    
    lsn = extract_12_bit_number (tvb,byte_offset, bit_offset);

    if (ack_tree)
        proto_tree_add_uint (ack_tree, hf_rlc_lsn, tvb, start_byte_offset, 2, lsn);

    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO,"ACK_SUFI LSN: %d ", lsn);
    }

    return;
}

void dissect_mrw_ack_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
}

void dissect_mrw_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
}

void dissect_rlist_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
    guint8  i,count,cw,no_of_cw ,err_burst_ind = 0;
    int missing_sn,temp_byte_offset  ,fsn ,start_byte_offset =  *byte_offset;
    proto_tree *rlist_tree=NULL, *missing_sn_tree=NULL;
    proto_item *ti=NULL;

    no_of_cw = extract_4_bit_number(tvb,byte_offset , bit_offset);
    temp_byte_offset = *byte_offset;
    fsn = extract_12_bit_number( tvb,byte_offset , bit_offset);

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_rlist_sufi, tvb, start_byte_offset,
                (no_of_cw/2+1) + 2 /*4 bits of no_of_cw and 12 bits of fsn*/, FALSE);
        rlist_tree = proto_item_add_subtree (ti, ett_rlist);
    }

    if (rlist_tree) {
        proto_tree_add_uint (rlist_tree, hf_rlc_fsn, tvb, temp_byte_offset, 2, fsn);
        ti = proto_tree_add_item (rlist_tree, hf_rlc_missing_seq_data, tvb, *byte_offset,
                no_of_cw , FALSE);
        missing_sn_tree = proto_item_add_subtree (ti, ett_missing_sn);
    }

    count = 0;
    missing_sn = 0;
    for (i = 0 ;i < no_of_cw ; i++)
    {
        temp_byte_offset = *byte_offset;
        cw = extract_4_bit_number(tvb,byte_offset , bit_offset) ;
        if (cw == 1)
        {
            err_burst_ind = 1; /** cw = 1 is error burst indicator flag**/
            missing_sn = 0;
            count = 0;
        }
        else
        {
            missing_sn |= (((cw >>1) & 0x7) << (count*3));
            if ((cw & 0x1) == 1) /*The missing_sn number terminates */
            {
                if (err_burst_ind == 1)
                {
                    if (missing_sn_tree)
                        proto_tree_add_text(missing_sn_tree,tvb,temp_byte_offset,count/2,"No. of subsequent consecutive erroreous AMD PDUs: %d ",missing_sn);
                }
                else
                {
                    if (missing_sn_tree)
                        proto_tree_add_text(missing_sn_tree,tvb,temp_byte_offset,count/2,"Distance between previous indicated erroreous AMD PDUs and upto (including) the next erroneous AMD PDU: %d ",missing_sn);
                }
                err_burst_ind = 0; /* Error burst Indicator**/
                missing_sn = 0;
                count = 0;
            }
        }

    }
    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO," RLIST_SUFI ");
    }

}

void dissect_bitmap_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
    guint8  i,j,length ;
    int temp_byte_offset  ,fsn ,start_byte_offset =  *byte_offset;
    proto_tree *bitmap_tree=NULL, *missing_sn_tree=NULL;
    proto_item *ti=NULL;

    length = extract_4_bit_number(tvb,byte_offset , bit_offset);
    temp_byte_offset = *byte_offset;
    fsn = extract_12_bit_number( tvb,byte_offset , bit_offset);

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_bitmap_sufi, tvb, start_byte_offset,
                length + 2 /*4 bits of length and 12 bits of fsn*/, FALSE);
        bitmap_tree = proto_item_add_subtree (ti, ett_bitmap);
    }

    if (bitmap_tree) {
        proto_tree_add_uint (bitmap_tree, hf_rlc_fsn, tvb, temp_byte_offset, 2, fsn);
        ti = proto_tree_add_item (bitmap_tree, hf_rlc_missing_seq_data, tvb, *byte_offset,
                length , FALSE);
        missing_sn_tree = proto_item_add_subtree (ti, ett_missing_sn);
    }

    for (i = 0 ;i <= length ; i++)
    {
        guint8 bitmap;

        temp_byte_offset = *byte_offset;
        bitmap = (extract_4_bit_number(tvb,byte_offset , bit_offset) << 4) & 0xf0;
        bitmap = (bitmap | (extract_4_bit_number(tvb,byte_offset , bit_offset) & 0x0f));
        for (j=0; j<=7; j++)
        {
            if (missing_sn_tree)
                proto_tree_add_text(missing_sn_tree,tvb,temp_byte_offset,1,"Seq No %d %s",(fsn +(i*8)+j)%4096,
                        (bitmap& (1<< (7-j))?"Received":"Not Received"));
        }
    }
    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO," BITMAP_SUFI ");
    }
}

void dissect_list_sufi  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int * byte_offset, int * bit_offset)
{
    guint8  i,li,length ;
    int missing_sn,temp_byte_offset,start_byte_offset =  *byte_offset;
    proto_tree *list_tree=NULL, *missing_sn_tree=NULL;
    proto_item *ti=NULL;

    length = extract_4_bit_number(tvb,byte_offset , bit_offset);

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_list_sufi, tvb, start_byte_offset,
                (length *2) +1 /*4 bits of length and 12 bits of fsn*/, FALSE);
        list_tree = proto_item_add_subtree (ti, ett_list);
    }

    if (list_tree) {
        ti = proto_tree_add_item (list_tree, hf_rlc_missing_seq_data, tvb, *byte_offset,
                length , FALSE);
        missing_sn_tree = proto_item_add_subtree (ti, ett_missing_sn);
    }

    for (i = 0 ;i < length ; i++)
    {
        temp_byte_offset = *byte_offset;
        missing_sn = extract_12_bit_number(tvb,byte_offset , bit_offset) ;
        li = extract_4_bit_number(tvb,byte_offset , bit_offset) ;
        if (missing_sn_tree)
            proto_tree_add_text(missing_sn_tree,tvb,temp_byte_offset,2,"Missing Seq No: %d and subsequent no. of PDUs missing: %d",missing_sn,li);
    }
    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO," LIST_SUFI ");
    }

}

void dissect_am_reset_pdu  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    proto_tree *rlc_reset_pdu_tree = NULL;
    proto_item *ti = NULL;
    int hfni, reset_header,offset = 0;

    if (ptree) {
        ti = proto_tree_add_item (ptree,hf_rlc_reset_pdu, tvb, 0,
            tvb_length(tvb),FALSE);
        rlc_reset_pdu_tree = proto_item_add_subtree (ti, ett_rlc);
    }
    reset_header = tvb_get_guint8(tvb,offset);

    if (rlc_reset_pdu_tree)
        proto_tree_add_uint(rlc_reset_pdu_tree, hf_rlc_rsn,tvb, offset,1,
                    (reset_header & 0x8) >>3);

    hfni = tvb_get_guint8(tvb,offset++) << 8;
    hfni = (hfni | tvb_get_guint8(tvb,offset++)) << 4;
    hfni = hfni | ((tvb_get_guint8(tvb,offset) & 0xf0) >> 4);

    if (rlc_reset_pdu_tree)
        proto_tree_add_uint(rlc_reset_pdu_tree, hf_rlc_hfni,tvb, 1,24,
                    hfni);

    return;
}

void dissect_am_reset_ack_pdu  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    proto_tree *rlc_reset_ack_pdu_tree = NULL;
    proto_item *ti = NULL;
    int hfni,reset_ack_header,offset = 0;

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_rlc_reset_ack_pdu, tvb, 0,
            tvb_length(tvb),FALSE);
        rlc_reset_ack_pdu_tree = proto_item_add_subtree (ti, ett_rlc);
    }

    reset_ack_header = tvb_get_guint8(tvb,offset);
    if (rlc_reset_ack_pdu_tree)
        proto_tree_add_uint(rlc_reset_ack_pdu_tree, hf_rlc_rsn,tvb, offset,1
                    ,(reset_ack_header & 0x8) >>3);

    hfni = tvb_get_guint8(tvb,offset++) << 8;
    hfni = (hfni | tvb_get_guint8(tvb,offset++)) << 4;
    hfni = hfni | ((tvb_get_guint8(tvb,offset) & 0xf0) >> 4);

    if (rlc_reset_ack_pdu_tree)
        proto_tree_add_uint(rlc_reset_ack_pdu_tree, hf_rlc_hfni,tvb, 1,24
                    ,hfni);

    return;
}

void dissect_am_status_pdu  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    proto_tree *rlc_status_pdu_tree = NULL;
    guint8 sufi_type = 0;
    proto_item *ti = NULL;
    int byte_offset = 0,bit_offset=0;

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_rlc_status_pdu, tvb, 0,
            tvb_length(tvb),FALSE);
        rlc_status_pdu_tree = proto_item_add_subtree (ti, ett_rlc);
    }

    sufi_type = tvb_get_guint8(tvb,byte_offset);
    sufi_type = sufi_type & 0xf;
    ++byte_offset;
    while (sufi_type != NO_MORE_SUFI)
    {
        if (rlc_status_pdu_tree)
            proto_tree_add_uint_format_value (rlc_status_pdu_tree, hf_rlc_pdu_type,tvb, byte_offset-1,
                    1,sufi_type," %s",rlc_convert_to_sufi_str(sufi_type));

        switch(sufi_type)
        {
            case WINDOW_SUFI:
                dissect_window_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case ACK_SUFI:
                dissect_ack_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case LIST_SUFI:
                dissect_list_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case BITMAP_SUFI:
                dissect_bitmap_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case RLIST_SUFI:
                dissect_rlist_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case MRW_SUFI:
                dissect_mrw_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
                break;
            case MRW_ACK_SUFI:
                dissect_mrw_ack_sufi(tvb,pinfo,rlc_status_pdu_tree,&byte_offset,&bit_offset);
        }
        sufi_type = extract_4_bit_number(tvb,&byte_offset,&bit_offset);
    }
    return;
}

void dissect_am_data_pdu (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    tvbuff_t *next_tvb;
    proto_item *ti = NULL,*error_ti = NULL;
    proto_tree *rlc_am_data_tree = NULL;
    int seq_no = 0, offset = 0, no_of_data_bits, size;
    guint16 hdr;
    guint8 li, li_last;

    if (ptree) {
        ti = proto_tree_add_item (ptree, hf_rlc_am_data_pdu, tvb, 0,
            tvb_length(tvb),FALSE);
        rlc_am_data_tree = proto_item_add_subtree (ti, ett_rlc);
    }

    hdr = tvb_get_ntohs(tvb, offset);
    seq_no = (hdr & 0x7ff8) >> 3;

    if (rlc_am_data_tree) {
        proto_tree_add_item(rlc_am_data_tree, hf_rlc_dc_field, tvb, offset, 2, FALSE);
        proto_tree_add_item(rlc_am_data_tree, hf_rlc_am_seq_no, tvb, offset, 2, FALSE);
        proto_tree_add_item(rlc_am_data_tree, hf_rlc_polling_bit, tvb, offset, 2, FALSE);
        proto_tree_add_item(rlc_am_data_tree, hf_rlc_he, tvb, offset, 2, FALSE);
    }
    offset += 2;
    
    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_prepend_fstr(pinfo->cinfo,COL_INFO,"Data PDU SeqNo: %d ", seq_no);
    }

    li_counter = 0;
    if (global_dissect_am_payloads && ((hdr & 0x3 ) == 1)) /* Next byte is a Li */
    {
        dissect_length_indicator(tvb, pinfo,rlc_am_data_tree,
                                    &offset,&no_of_data_bits,RLC_AM_LI);
    }

    if (tvb_bytes_exist (tvb, offset, 1))
    {
        if (li_counter==0) {
            next_tvb = tvb_new_subset (tvb, offset, -1, -1);
            call_dissector (data_handle, next_tvb, pinfo, ptree);
        }
        else {
            li_last = 0;
            for (li=0; li<li_counter; li++) {
                if (li_arr[li]==0){ /* indicates end of SDU at end of previous PDU only */
                }
                else if (li_arr[li]<124) {
                    if (tvb_length_remaining(tvb,offset)==0) {
                        error_ti = proto_tree_add_text(ptree, tvb, offset-1, 1, "Length indicated, but no data remains");
                        expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Length indicated, but no data remains");
                        break;
                    }
                    size = li_arr[li] - li_last;
                    if (size>tvb_length_remaining(tvb,offset)||size<0) {
                        size = tvb_length_remaining(tvb,offset);
                        error_ti = proto_tree_add_text(ptree, tvb, offset, size, "Length error - truncated");
                        expert_add_info_format(pinfo, error_ti, PI_MALFORMED, PI_ERROR, "Length error - truncated");
                    }
                    next_tvb = tvb_new_subset (tvb, offset, size, -1);
                    call_dissector (data_handle, next_tvb, pinfo, ptree);
                    offset += size;
                    li_last = li_arr[li];
                }
                else if (li_arr[li]==126) {
                    proto_tree_add_text(ptree, tvb, offset, -1, "Piggy-backed Status PDU");
                    next_tvb = tvb_new_subset (tvb, offset, -1, -1);
                    dissect_am_status_pdu(next_tvb, pinfo, ptree);
                    offset=tvb_length(tvb);
                }
                else if ((li_arr[li]==127)&&(tvb_length_remaining(tvb,offset))) {
                    proto_tree_add_text(ptree, tvb, offset, -1, "Padding");
                    next_tvb = tvb_new_subset (tvb, offset, -1, -1);
                    call_dissector (data_handle, next_tvb, pinfo, ptree);
                    offset=tvb_length(tvb);
                }
            }
            if (tvb_length_remaining(tvb,offset)>0) {
                next_tvb = tvb_new_subset (tvb, offset, -1, -1);
                call_dissector (data_handle, next_tvb, pinfo, ptree);
            }
        }
    }
}

void dissect_am_control_pdu  (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    guint8 pdu_type =0;

    pdu_type =  tvb_get_guint8(tvb,0);
    pdu_type = (pdu_type >>4) & 0x7;
    switch(pdu_type)
    {
        case STATUS_PDU:
            dissect_am_status_pdu(tvb,pinfo,ptree);
            break;
        case RESET_PDU:
            dissect_am_reset_pdu(tvb,pinfo,ptree);
            break;
        case RESET_ACK_PDU:
            dissect_am_reset_ack_pdu(tvb,pinfo,ptree);
    }
    return;
}

void dissect_am_rlc (tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree)
{
    guint8 dc_field,offset = 0;
    proto_tree *rlc_am_tree=NULL;
    proto_item *ti;

    dc_field = tvb_get_guint8(tvb,offset);

    if (check_col (pinfo->cinfo, COL_PROTOCOL))
        col_set_str (pinfo->cinfo, COL_PROTOCOL, "RLC-AM");

    if (ptree) {
        ti = proto_tree_add_item (ptree, proto_am_rlc, tvb, 0,
                tvb_length(tvb), FALSE);
        rlc_am_tree = proto_item_add_subtree (ti, ett_rlc);
    }

    if (((dc_field & 0x80 ) >>7) == 0)
        dissect_am_control_pdu(tvb,pinfo,rlc_am_tree);
    else
        dissect_am_data_pdu(tvb,pinfo,rlc_am_tree);

    return;
}

