/*
 * $Id: gsm_csd.c 1.2 2008/06/06 05:11:24BST Neil P Piercy (npp) Exp  $
 *
 * Copyright 2003 ip.access Ltd.  See the accompanying file "COPYRIGHT" for
 * details.  THERE IS ABSOLUTELY NO WARRANTY FOR THIS SOFTWARE.
 */

#include    <assert.h>
#include	<stdlib.h>
#include	<string.h>
#include	"gsm_csd.h"

/* from 04.21 table 1 */
#define E13_NULL		7
#define E13_9600_4800	6
#define E13_2400		3
#define E13_1200		2
#define E13_600			1

extern int guess_sub_chan_rate(unsigned char *c);

csd_state *gsm_csd_create()
{
	csd_state *s;

	s = (csd_state *)malloc(sizeof(struct csd_state));
	if (!s) return s;

	memset((char *)s, 0, sizeof(*s));
	s->sub_chan_rate = 0;				/* sub channel rate not yet determined */

	return s;
}

void gsm_csd_destroy(csd_state *S)
{
	if (S) free((char *)S);
}

/* this function receives the 160 octet RTP frames every 20ms, and returns the same! */
int gsm_csd_decode(csd_state *s _U_, const guint8 *c, unsigned char *target)
{
	/* save raw subchannel data to temp file - not found sync yet */
	memcpy(target, c, NUM_CSD_OCTETS);

	return NUM_CSD_OCTETS; /* number of octets in return buffer */
}

/* this function receives the 160 octet RTP frames every 20ms, and returns the raw subchannel data
 * at 8 or 16 kb/s - basically 20 or 40 octets of raw (unsynchronised) data */
int gsm_csd_decode_real(csd_state *s, unsigned char *c, unsigned char *target)
{
	if (!(s->sub_chan_rate))	/* maybe we should continue to check, but... */
		s->sub_chan_rate = guess_sub_chan_rate(c);

	if (!(s->sub_chan_rate)) {
		printf("Failed to guess CSD Sub Channel Rate!\n");
		return -1;
	}

	/* save raw subchannel data to temp file - not found sync yet */
	copy_sub_chan(s->sub_chan_rate, c, target);

	return 20*s->sub_chan_rate; /* number of octets in return buffer - 20 per sub channel bit used */
}

#if 0
/* This function is passed the raw subchannel octet at a time
 * it returns the number of bytes written to the output - basically 0 or a frame's worth
 * (but beware 1.2kb/s and 600bps - less than a full output octet per V.110 frame
		 - take several frames before putting any output */
int gsm_csd_decode_stream(csd_state *s, char in, char *out)
{
	int i, data_rate, bit_reps;
	char * buf;

/* Find sync - treat as a bit stream until we find sync!!!
	- is it 14.4 format (16+ zero bits followed by a 1),
	- is it V.110 format (8+ zero bits followed by 9 octets each starting with a 1 bit */

	if (s->frame_shift == -1) {		/* data passed in byte at a time until sync found - then N frames */
		find_sync(s, in);		/* finds first 1 bit after sync of zeros */
		if (s->frame_shift == 0) {	/* sync found and already byte aligned - we have the first data octet */
			return s->frame_len - 1;
		}
		else if (s->frame_shift == 0) {
			return s->frame_len;	/* sync found but not enough data to continue */
		}
		else
			return 0;				/* no sync yet */
	}

	/* to get here we have found sync and been passed another byte */

	if (store_data_bits(s, in)==0) /* returns non-zero when it has a full frame */
		return 0;

	/* by the time we get here, the data in the buffer is correctly octet aligned,
	and contains a full data frame - 320 bits for 14400, 80 bits for the v.110 frames */

	/* we know from the sync if its 14400 data, so go ahead and decode it now */
	if (s->frame_type == FRAME_14K4) {
		return decode_data_14400(s, out, p_out_len);
	}

	/* otherwise its in V.110 frames - find the data rate within the frame
		- but by now 9600 and 4800 look the same - its basically how many bit_reps there are */
	return decode_data_v110(in, out, p_out_len);
}

void find_sync(csd_state *s, unsigned char in)
{
	int j;

	for (j=0; j<7; j++, in>>=1) {
		if (in & 0x01) {					/* found a 1 */
			if (s->frame_zeros>15) {		/* its a 14k4 sync */
				s->frame_type = FRAME_14K4;
				s->frame_len = FRAME_14K4_LENGTH;
				s->frame_shift = j;
				s->frame_bits = 0;			/* valid data bits */
				s->frame_bit[s->frame_bits++] = in & 0x01;		/* save data bit */
				break;
			}
			else if (s->frame_zeros>7) {	/* its a V110 sync */
				s->frame_type = FRAME_V110;
				s->frame_len = FRAME_V110_LENGTH;
				s->frame_shift = j;
				s->frame_bits = 0;			/* valid data bits */
				s->frame_bit[s->frame_bits++] = in & 0x01;		/* save data bit */
				break;
			}
			else if (s->frame_type == FRAME_NULL) {	/* no sync yet - reset the zero count and keep looking */
				s->frame_zeros = 0;
			}
			else {							/* store remaining data bits */
				s->frame_bit[s->frame_bits++] = in & 0x01;		/* save data bit */
			}
		}
		else if (s->frame_type == FRAME_NULL) {	/* no sync yet - found a zero */
			s->frame_zeros++;
		}
		else {							/* store remaining data bits */
			s->frame_bit[s->frame_bits++] = in & 0x01;		/* save data bit */
		}
	}

	return;
}

int store_data_bits(csd_state *s, unsigned char in)
{
	int j, done=0;

	for (j=0; j<7; j++, in>>=1) {
		s->frame_bit[s->frame_bits++] = in & 0x01;		/* save data bit */
	}

	if (s->frame_bits >= s_frame_len)
		return 1;
	else
		return 0;
}

int decode_data_bytes_v110(char *in, char *out)
{
	char temp[8];
	int i, j, bit_reps, bits;

	if (in[0] != 0)
		printf("Wrong V.110 sync word %d\n", *in);

	if (in[5] == 0xff)
		printf("Null frame!!! \n");

	/* FIXME - extract bit reps from the E1-3 bits in octet 5 (see 04.21), and set bits accordingly */

	for (j=0,i=1; i<10; i++) {
		if (i!= 5)		/* skip the middle line - it contains no data */
			temp[j++]  = (in[i] & 0x7e) << 1;	/* first get the D1..D6 etc as the first 6 bits in temp */
	}

	if (bit_reps == 1) {
		out[0] = ((temp[0] & 0x3f) >> 0) | ((temp[1] & 0x03) << 6);
		out[1] = ((temp[1] & 0x3c) >> 2) | ((temp[2] & 0x0f) << 4);
		out[2] = ((temp[2] & 0x30) >> 4) | ((temp[3] & 0x3f) << 2);
		out[3] = ((temp[4] & 0x3f) >> 0) | ((temp[5] & 0x03) << 6);
		out[4] = ((temp[5] & 0x3c) >> 2) | ((temp[6] & 0x0f) << 4);
		out[5] = ((temp[6] & 0x30) >> 4) | ((temp[7] & 0x3f) << 2);
	}
	else {
		printf("Cant handle rate which have bit repetitions yet\n");
	}

	return bits;	/* the number of bits actually extracted to the output */
}

int decode_data_v110(csd_state *s)
{
	/* decode the bit buffer of 80 bits, starting with the first 1 bit of the frame */
	char *in = s->bit;
	int i, j, k, bit_reps;

	/* check the leading 1s */
	for (j=0; j<72; j+=8) {
		if (in[j] != 1)
			printf("Wrong V.110 sync 1 bit position %d\n", j+8);
	}
	/* check the next set of 8 0s */
	for (j=72; j<80; j++) {
		if (in[j] != 0)
			printf("Wrong V.110 sync 0 bit position %d\n", j-72);
	}

	/* extract the frame rate bits E1..3 */
	e13 = (in[41] << 2) | (in[42] << 1) | in[43]
	switch (e13) {
	case E13_9600_4800:
		bit_reps = 1;
		break

	case E13_2400:
		bit_reps = 2;
		break

	case E13_1200:
		bit_reps = 4;
		break

	case E13_600:
		bit_reps = 8;
		break

	case E13_NULL:
		printf("Null frame!!! \n");
		break;

	default:
		printf("Unknown frame rate bits E1..E3 %d\n", e13);
		break;
	}

	/* extract data bits */
	k = 0;					/* initialise the bit repetition counter */
	for (i=0; i<9; i++) {
		if (i!= 4) {		/* skip the middle line - it contains no data */
			for (j=1; j<7; j++) {	/* skip sync 1 bits and status bits */
				if (++k==bit_reps) {
					s->data[s->data_len++] = in[(i<<3)+j];
					k = 0;	/* reset the bit repetition counter */
				}
			}
		}
	}

	return;
}

int decode_data_14400(data_rate, sub_chan, target)
{

}

#endif

unsigned char get_byte(int sub_chan_rate, unsigned char *c)
{
	int byte=0;

	switch(sub_chan_rate)
	{
		case 8:		/* 64kb/s - just copy */
			byte = *c++;
			break;

		case 4:		/* 32kb/s - 4 bits per octet */
			byte  = ((*c++) & 0x0f)<<4;
			byte |= ((*c++) & 0x0f)<<0;
			break;

		case 2:		/* 16kb/s - 2 bits per octet */
			byte  = ((*c++) & 0x03)<<6;
			byte |= ((*c++) & 0x03)<<4;
			byte |= ((*c++) & 0x03)<<2;
			byte |= ((*c++) & 0x03)<<0;
			break;

		case 1:		/* 8kb/s - 1 bit per octet */
			byte  = ((*c++) & 0x01)<<7;
			byte |= ((*c++) & 0x01)<<6;
			byte |= ((*c++) & 0x01)<<5;
			byte |= ((*c++) & 0x01)<<4;
			byte |= ((*c++) & 0x01)<<3;
			byte |= ((*c++) & 0x01)<<2;
			byte |= ((*c++) & 0x01)<<1;
			byte |= ((*c++) & 0x01)<<0;
			break;

		default:
			break;

	}
	return byte;
}

void copy_sub_chan(int sub_chan_rate, unsigned char *c, unsigned char *sub_chan)
{
	int i, bytes_per_octet, sub_chan_octets;

	bytes_per_octet = 8/sub_chan_rate;
	sub_chan_octets = NUM_CSD_OCTETS/bytes_per_octet;

	for (i=0; i<sub_chan_octets; i++) {
		*sub_chan++ = get_byte(sub_chan_rate, c);
		c+=bytes_per_octet;
	}
	return;
}

unsigned char mask[9]={0, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff};

int find_sync(int sub_chan_rate, unsigned char *c, int len)
{
	int v, i, zeros=0, sync=-1;

	for (i=0; i<len; i++) {
		v = (*c++) & mask[sub_chan_rate];
		if (v==0)
			zeros+=sub_chan_rate;
		else
			zeros = 0;
		if (zeros >= 8) {
			sync = i;
			break;
		}
	}

	/* should really check for the following ones at the start of every octet here too */

	return sync;
}

int check_sync(int sub_chan_rate, unsigned char *c, int start)
{
	int v, i, zeros=0, sync=-1;

	for (i=start; i<start+8/sub_chan_rate; i++) {
		v = (*c++) & mask[sub_chan_rate];
		if (v==0)
			zeros+=sub_chan_rate;
		else
			zeros = 0;
		if (zeros >= 8) {
			sync = i;
			break;
		}
	}

	/* should really check for the following ones at the start of every octet here too */

	return sync;
}

int guess_sub_chan_rate(unsigned char *c)
{
	int i, sub_chan_rate=0;
	unsigned char b=0xff;

	for (i=0; i<NUM_CSD_OCTETS; i++)
		b = b & *c++;				/* unused bits are 1s, data is mix of 0s and 1s - b will acquire 0s in the data bits */

	for (i=0; i<=8; i++) {
		if ((mask[i]|b) == 0xff) {	/* when the mask 1s make up for the data 0s, we have the right rate */
			sub_chan_rate = i;
			break;
		}
	}

	return sub_chan_rate;
}

