/* tap_gsmstat.c
 * tap_gsmstat   npp 2003
 *
 * $Id: gtk/tap_gsmstat.c 1.4 2008/06/06 05:11:29BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* This module provides GSM statistics to ethereal.
 * It is only used by ethereal and not tethereal
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <epan/epan_dissect.h>
#include <epan/filesystem.h>

#include "util.h"
#include <epan/tap.h>
#include "register.h"

/* in /gtk ... */
#include <gtk/gtk.h>
#include "gtkglobals.h"

#include <epan/stat_cmd_args.h>
#include "dlg_utils.h"
#include "gui_utils.h"
#include "alert_box.h"
#include "simple_dialog.h"
#include <../stat_menu.h>
#include "gui_stat_menu.h"
#include "main.h"
#include "progress_dlg.h"
#include "compat_macros.h"
#include "../color.h"

#include <glib.h>
#include <epan/packet_info.h>
#include <epan/packet.h>
#include "menu.h"
#include <epan/dissectors/packet-gsm.h>
#include "../globals.h"

/* used to keep track of the statistics for an entire ethereal tap instance
 - only one for GSM, but... */
typedef struct _gsmstat_t {
	char *filter;
	GtkWidget *win;
	GtkWidget *dlabel;
	gchar sdlabel[60];
	GtkWidget *table;
	GtkWidget *wcount[GSMSTAT_MAX][2];
	GtkWidget *label[GSMSTAT_MAX];
	gchar scount[GSMSTAT_MAX][2][32];
	GtkWidget *utable;
	GtkWidget *uwcount[GSMSTAT_CHAN_TYPE_MAX][4];
	GtkWidget *ulabel[GSMSTAT_CHAN_TYPE_MAX];
	gchar uscount[GSMSTAT_CHAN_TYPE_MAX][4][32];
	gsmstat_common_t gsc;

} gsmstat_t;

void gtk_gsmstat_cb(GtkWidget *w _U_, gpointer d _U_);

/* This callback is invoked whenever the tap system has seen a packet
 * we might be interested in.
 * The function is to be used to only update internal state information
 * in the *tapdata structure, and if there were state changes which requires
 * the window to be redrawn, return 1 and (*draw) will be called sometime
 * later.
 *
 * This function should be as lightweight as possible since it executes together
 * with the normal ethereal dissectors. Try to push as much processing as
 * possible into (*draw) instead since that function executes asynchronously
 * and does not affect the main threads performance.
 *
 * If it is possible, try to do all "filtering" explicitely as we do below in
 * this example since you will get MUCH better performance than applying
 * a similar display-filter in the register call.
 *
 * The third parameter is tap dependant. Since we register this one to the "gsm"
 * tap the third parameters type is gsm_call_info_value.
 *
 *
 * The filtering we do is just to check the gsm_call_info_value struct that
 * we were called for the proper program and version. We didnt apply a filter
 * when we registered so we will be called for ALL gsm packets and not just
 * the ones we are collecting stats for.
 *
 *
 * function returns :
 *  0: no updates, no need to call (*draw) later
 * !0: state has changed, call (*draw) sometime later
 */
int
gsmstat_packet(void *prs, packet_info *pinfo _U_, epan_dissect_t *edt _U_, const void *pri)
{
	gsmstat_t		*rs=prs;

	/* all packet handling is done with the common routine */
	return gsmstat_common_packet(&(rs->gsc), pinfo, pri);
}


/* This callback is never used by tethereal but it is here for completeness.
 * When registering below, we could just have left this function as NULL.
 *
 * When used by ethereal, this function will be called whenever we would need
 * to reset all state. Such as when ethereal opens a new file, when it
 * starts a new capture, when it rescans the packetlist after some prefs have
 * changed etc.
 * So if your aplication has some state it needs to clean up in those
 * situations, here is a good place to put that code.
 */
void
gsmstat_reset(void *prs)
{
	gsmstat_t		*rs=prs;

	/* all reset handling is done with the common routine */
	gsmstat_common_reset(&(rs->gsc));

	return;
}


static void
gsmstat_draw(gsmstat_t *rs)
{
	int i;
	gsmstat_common_t *gsc=&(rs->gsc);
	float r, d;

	d = rs->gsc.secs + (float)(rs->gsc.secs)/1000000;
	sprintf(rs->sdlabel,"Duration: %d.%03ds", gsc->secs, gsc->usecs/1000);
	gtk_label_set_text(GTK_LABEL(rs->dlabel), rs->sdlabel);

	for (i=0; i<gsmstat_counts_max(rs->gsc.intface); i++)
	{
		sprintf(rs->scount[i][0],"%d", rs->gsc.count[gsmstat_counts(rs->gsc.intface,i)]);
		gtk_label_set_text(GTK_LABEL(rs->wcount[i][0]), rs->scount[i][0]);
		if (rs->gsc.secs > 0) {
			r = (float)(gsc->count[gsmstat_counts(gsc->intface,i)]) / d;
			if (gsc->count[gsmstat_counts(gsc->intface,i)]==0)
				sprintf(rs->scount[i][1],"-" );
			else if (r>1)
				sprintf(rs->scount[i][1],"%8.3f/s", r );
			else
				sprintf(rs->scount[i][1],"%8.3f/hr", 3600*r);
			gtk_label_set_text(GTK_LABEL(rs->wcount[i][1]), rs->scount[i][1]);
		}
	}

	for (i=0; i<gsmstat_usage_max(rs->gsc.intface); i++)
	{
		sprintf(rs->uscount[i][0],"%8.3fE", (gsc->usage[gsmstat_usage(gsc->intface,i)].chan_secs + (float)(gsc->usage[gsmstat_usage(gsc->intface,i)].chan_usecs)/1000000) / d);
		gtk_label_set_text(GTK_LABEL(rs->uwcount[i][0]), rs->uscount[i][0]);

		sprintf(rs->uscount[i][1],"Peak %d", gsc->usage[gsmstat_usage(gsc->intface,i)].peak);
		gtk_label_set_text(GTK_LABEL(rs->uwcount[i][1]), rs->uscount[i][1]);

		sprintf(rs->uscount[i][2],"for %d.%03ds", gsc->usage[gsmstat_usage(gsc->intface,i)].peak_secs, gsc->usage[gsmstat_usage(gsc->intface,i)].peak_usecs/1000);
		gtk_label_set_text(GTK_LABEL(rs->uwcount[i][2]), rs->uscount[i][2]);

		sprintf(rs->uscount[i][3],"on %d occurances", gsc->usage[gsmstat_usage(gsc->intface,i)].peak_count);
		gtk_label_set_text(GTK_LABEL(rs->uwcount[i][3]), rs->uscount[i][3]);
	}

}

/* since the gtk2 implementation of tap is multithreaded we must protect
 * remove_tap_listener() from modifying the list while draw_tap_listener()
 * is running.  the other protected block is in main.c
 *
 * there should not be any other critical regions in gtk2
 */
void protect_thread_critical_region(void);
void unprotect_thread_critical_region(void);
static void
win_destroy_cb(GtkWindow *win _U_, gpointer data)
{
	gsmstat_t *rs=(gsmstat_t *)data;

	protect_thread_critical_region();
	remove_tap_listener(rs);
	unprotect_thread_critical_region();

	g_free(rs);
}

/* When called, this function will create a new instance of gtk2-gsmstat.
 */
static void
gtk_gsmstat_init(const char *optarg,void* userdata _U_)
{
	gsmstat_t *rs;
	guint32 intface;
	char title_string[60];
	char filter_string[256];
	GtkWidget *vbox;
	GtkWidget *stat_label;
	GtkWidget *filter_label;
	GtkWidget *tmp;
	int pos, i;
	char *filter=NULL;

	pos=0;
	i=sscanf(optarg,"gsm,%d,%n",&intface, &pos);
	if(i==1 || i==2){
		if(pos){
			filter=(char *)optarg+pos;
		} else {
			filter=NULL;
		}
	} else {
		fprintf(stderr, "ethereal: invalid \"-z gsm,interface[,<filter>]\" argument\n");
		exit(1);
	}

	rs=g_malloc(sizeof(gsmstat_t));

	rs->win=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	rs->gsc.intface=intface;
	sprintf(title_string,"%s",gsmstat_intface_name(intface));
	gtk_window_set_title(GTK_WINDOW(rs->win), title_string);
	SIGNAL_CONNECT(rs->win, "destroy", win_destroy_cb, rs);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(rs->win), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	gtk_widget_show(vbox);

	stat_label=gtk_label_new(title_string);
	gtk_box_pack_start(GTK_BOX(vbox), stat_label, FALSE, FALSE, 0);
	gtk_widget_show(stat_label);

	g_snprintf(filter_string,255,"Filter:%s",filter?filter:"");
	filter_label=gtk_label_new(filter_string);
	gtk_box_pack_start(GTK_BOX(vbox), filter_label, FALSE, FALSE, 0);
	gtk_widget_show(filter_label);

	sprintf(rs->sdlabel,"Duration: 0.0s");
	rs->dlabel=gtk_label_new(rs->sdlabel);
	gtk_box_pack_start(GTK_BOX(vbox), rs->dlabel, FALSE, FALSE, 0);
	gtk_widget_show(rs->dlabel);

	rs->table=gtk_table_new(gsmstat_counts_max(intface), 3, TRUE);
	gtk_container_add(GTK_CONTAINER(vbox), rs->table);

	tmp=gtk_label_new("Counts");
	gtk_table_attach_defaults(GTK_TABLE(rs->table), tmp, 1,2,0,1);
	gtk_label_set_justify(GTK_LABEL(tmp), GTK_JUSTIFY_RIGHT);
	gtk_widget_show(tmp);

	tmp=gtk_label_new("Rate");
	gtk_table_attach_defaults(GTK_TABLE(rs->table), tmp, 2,3,0,1);
	gtk_label_set_justify(GTK_LABEL(tmp), GTK_JUSTIFY_RIGHT);
	gtk_widget_show(tmp);

	for (i=0; i<gsmstat_counts_max(intface); i++)
	{
		rs->label[i]=gtk_label_new(gsmstat_name(gsmstat_counts(intface,i)));
		gtk_label_set_justify(GTK_LABEL(rs->label[i]), GTK_JUSTIFY_LEFT);
		gtk_table_attach_defaults(GTK_TABLE(rs->table), rs->label[i], 0,1,i+1,i+2);
		gtk_widget_show(rs->label[i]);

		rs->wcount[i][0]=gtk_label_new("0");
		gtk_table_attach_defaults(GTK_TABLE(rs->table), rs->wcount[i][0], 1,2,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->wcount[i][0]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->wcount[i][0]);

		rs->wcount[i][1]=gtk_label_new("0");
		gtk_table_attach_defaults(GTK_TABLE(rs->table), rs->wcount[i][1], 2,3,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->wcount[i][1]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->wcount[i][1]);
	}

	rs->utable=gtk_table_new(gsmstat_usage_max(intface), 5, TRUE);
	gtk_container_add(GTK_CONTAINER(vbox), rs->utable);

	for (i=0; i<gsmstat_usage_max(intface); i++)
	{
		rs->ulabel[i]=gtk_label_new(gsmstat_usage_name(gsmstat_usage(intface,i)));
		gtk_label_set_justify(GTK_LABEL(rs->ulabel[i]), GTK_JUSTIFY_LEFT);
		gtk_table_attach_defaults(GTK_TABLE(rs->utable), rs->ulabel[i], 0,1,i+1,i+2);
		gtk_widget_show(rs->ulabel[i]);

		rs->uwcount[i][0]=gtk_label_new("0E");
		gtk_table_attach_defaults(GTK_TABLE(rs->utable), rs->uwcount[i][0], 1,2,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->uwcount[i][0]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->uwcount[i][0]);

		rs->uwcount[i][1]=gtk_label_new("Peak 0");
		gtk_table_attach_defaults(GTK_TABLE(rs->utable), rs->uwcount[i][1], 2,3,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->uwcount[i][1]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->uwcount[i][1]);

		rs->uwcount[i][2]=gtk_label_new("for 0.0s");
		gtk_table_attach_defaults(GTK_TABLE(rs->utable), rs->uwcount[i][2], 3,4,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->uwcount[i][2]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->uwcount[i][2]);

		rs->uwcount[i][3]=gtk_label_new("on 0 occurrances");
		gtk_table_attach_defaults(GTK_TABLE(rs->utable), rs->uwcount[i][3], 4,5,i+1,i+2);
		gtk_label_set_justify(GTK_LABEL(rs->uwcount[i][3]), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(rs->uwcount[i][3]);
	}

	gsmstat_reset(rs);

	gtk_widget_show(rs->table);

	if(register_tap_listener("gsm", rs, filter, (void*)gsmstat_reset, (void*)gsmstat_packet, (void*)gsmstat_draw)){
		char str[256];
		/* error, we failed to attach to the tap. clean up */
		g_snprintf(str,255,"Could not attach to tap using filter:%s",filter?filter:"");
		simple_dialog(ESD_TYPE_WARN, ESD_BTN_OK, str);
		g_free(rs);
		return;
	}

	gtk_widget_show_all(rs->win);
	cf_redissect_packets(&cfile);
}

static GtkWidget *dlg=NULL, *dlg_box;
static GtkWidget *filter_box;
static GtkWidget *filter_label, *filter_entry;
static GtkWidget *a_button;
static GtkWidget *abis_button;


static void
gsmstat_start_button_clicked(GtkWidget *item _U_, gpointer data _U_, int intface)
{
	char *filter;
	char str[256];

	filter=(char *)gtk_entry_get_text(GTK_ENTRY(filter_entry));
	if(filter[0]==0){
		sprintf(str, "gsm,%d", intface);
		filter="";
	} else {
		sprintf(str, "gsm,%d,%s", intface, filter);
	}
	gtk_gsmstat_init(str, NULL);
}

static void
gsmstat_a_button_clicked(GtkWidget *item _U_, gpointer data _U_)
{
	gsmstat_start_button_clicked(item,data,GSMSTAT_INTERFACE_A);
}

static void
gsmstat_abis_button_clicked(GtkWidget *item _U_, gpointer data _U_)
{
	gsmstat_start_button_clicked(item,data,GSMSTAT_INTERFACE_A_BIS);
}

static void
dlg_destroy_cb(void)
{
	dlg=NULL;
}

void
gtk_gsmstat_cb(GtkWidget *w _U_, gpointer d _U_)
{
	/* if the window is already open, bring it to front */
	if(dlg){
		gdk_window_raise(dlg->window);
		return;
	}

	dlg=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(dlg), "GSM Statistics");
	SIGNAL_CONNECT(dlg, "destroy", dlg_destroy_cb, NULL);
	dlg_box=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(dlg), dlg_box);
	gtk_widget_show(dlg_box);

	/* filter box */
	filter_box=gtk_hbox_new(FALSE, 10);
	/* Filter label */
	gtk_container_set_border_width(GTK_CONTAINER(filter_box), 10);
	filter_label=gtk_label_new("Filter:");
	gtk_box_pack_start(GTK_BOX(filter_box), filter_label, FALSE, FALSE, 0);
	gtk_widget_show(filter_label);

	filter_entry=gtk_entry_new_with_max_length(250);
	gtk_box_pack_start(GTK_BOX(filter_box), filter_entry, FALSE, FALSE, 0);
	gtk_widget_show(filter_entry);

	gtk_box_pack_start(GTK_BOX(dlg_box), filter_box, TRUE, TRUE, 0);
	gtk_widget_show(filter_box);

	/* the A button */
	a_button=gtk_button_new_with_label("A Interface");
        SIGNAL_CONNECT_OBJECT(a_button, "clicked",
                              gsmstat_a_button_clicked, NULL);
	gtk_box_pack_start(GTK_BOX(dlg_box), a_button, TRUE, TRUE, 0);
	gtk_widget_show(a_button);

	/* the A-bis button */
	abis_button=gtk_button_new_with_label("A-bis Interface");
        SIGNAL_CONNECT_OBJECT(abis_button, "clicked",
                              gsmstat_abis_button_clicked, NULL);
	gtk_box_pack_start(GTK_BOX(dlg_box), abis_button, TRUE, TRUE, 0);
	gtk_widget_show(abis_button);

	gtk_widget_show_all(dlg);
}

void
register_tap_listener_gtkgsmstat(void)
{
	register_stat_cmd_arg("ipa-gsm,stat,", gtk_gsmstat_init, NULL);

	register_stat_menu_item("IPA-GSM", REGISTER_STAT_GROUP_TELEPHONY, gtk_gsmstat_cb, NULL, NULL, NULL);
}
