/* tap-gsmstat.c
 * gsmstat   2002 Neil Piercy
 *
 * $Id: tap-gsmstat.c 1.3 2008/06/06 05:11:36BST Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* This module provides GSM statistics to tethereal.
 * It is only used by tethereal and not ethereal
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <glib.h>

#include <string.h>
#include "epan/packet_info.h"
#include <epan/packet.h>
#include <epan/tap.h>
#include <epan/stat_cmd_args.h>
#include "epan/dissectors/packet-gsm.h"
#include "register.h"

/* used to keep track of the statistics for an entire tethereal tap instance
 - only one for GSM, but... */
typedef struct _gsmstat_t {
	char				*filter;
	gsmstat_common_t	gsc;

} gsmstat_t;

/* This callback is invoked whenever the tap system has seen a packet
 * we might be interested in.
 * The function is to be used to only update internal state information
 * in the *tapdata structure, and if there were state changes which requires
 * the window to be redrawn, return 1 and (*draw) will be called sometime
 * later.
 *
 * This function should be as lightweight as possible since it executes together
 * with the normal ethereal dissectors. Try to push as much processing as
 * possible into (*draw) instead since that function executes asynchronously
 * and does not affect the main threads performance.
 *
 * If it is possible, try to do all "filtering" explicitely as we do below in
 * this example since you will get MUCH better performance than applying
 * a similar display-filter in the register call.
 *
 * The third parameter is tap dependant. Since we register this one to the "gsm"
 * tap the third parameters type is gsm_call_info_value.
 *
 *
 * The filtering we do is just to check the gsm_call_info_value struct that
 * we were called for the proper program and version. We didnt apply a filter
 * when we registered so we will be called for ALL gsm packets and not just
 * the ones we are collecting stats for.
 *
 *
 * function returns :
 *  0: no updates, no need to call (*draw) later
 * !0: state has changed, call (*draw) sometime later
 */
int
gsmstat_packet(void *prs, packet_info *pinfo _U_, epan_dissect_t *edt _U_, const void *pri)
{
	gsmstat_t		*rs=prs;

	/* all packet handling is done with the common routine */
	return gsmstat_common_packet(&(rs->gsc), pinfo, pri);
}


/* This callback is never used by tethereal but it is here for completeness.
 * When registering below, we could just have left this function as NULL.
 *
 * When used by ethereal, this function will be called whenever we would need
 * to reset all state. Such as when ethereal opens a new file, when it
 * starts a new capture, when it rescans the packetlist after some prefs have
 * changed etc.
 * So if your aplication has some state it needs to clean up in those
 * situations, here is a good place to put that code.
 */
void
gsmstat_reset(void *prs)
{
	gsmstat_t		*rs=prs;

	/* all reset handling is done with the common routine */
	gsmstat_common_reset(&(rs->gsc));

	return;
}

/* This callback is used when tethereal wants us to draw/update our
 * data to the output device. Since this is tethereal only output is
 * stdout.
 * Tethereal will only call this callback once, which is when tethereal has
 * finished reading all packets and exists.
 * If used with ethereal this may be called any time, perhaps once every 3
 * seconds or so.
 * This function may even be called in parallell with (*reset) or (*draw)
 * so make sure there are no races. The data in the gsmstat_t can thus change
 * beneath us. Beware.
 */
static void
gsmstat_draw(void *prs)
{
	gsmstat_t *rs=prs;
	gsmstat_common_t *gsc=&(rs->gsc);

	int i;
	float r, d;

	d = rs->gsc.secs + (float)(rs->gsc.secs)/1000000;
	printf("\n");
	printf("===================================================================\n");
	printf("GSM Statistics:\n");
	printf("Filter: %s\n",rs->filter?rs->filter:"");
	printf("   Duration: %8.3fs\n", d);
	for (i=0; i<gsmstat_counts_max(gsc->intface); i++) {
		r = (float)(gsc->count[gsmstat_counts(gsc->intface,i)]) / d;
		printf("%13s %8d", gsmstat_name(gsmstat_counts(gsc->intface,i)), gsc->count[gsmstat_counts(gsc->intface,i)]);
		if (gsc->count[gsmstat_counts(gsc->intface,i)]==0)
			printf("\n");
		else if (r>1)
			printf(" %8.3f/s\n", r);
		else
			printf(" %8.3f/hr\n", 3600*r);
	}

	for (i=0; i<gsmstat_usage_max(gsc->intface); i++) {
		printf( "%13s %8.3fE Peak %d for %d.%03ds over %d occurrances\n", gsmstat_usage_name(gsmstat_usage(gsc->intface,i)),
			(float)(gsc->usage[gsmstat_usage(gsc->intface,i)].chan_secs + (gsc->usage[gsmstat_usage(gsc->intface,i)].chan_usecs/1000000)) / d,
			gsc->usage[gsmstat_usage(gsc->intface,i)].peak,
			gsc->usage[gsmstat_usage(gsc->intface,i)].peak_secs,
			gsc->usage[gsmstat_usage(gsc->intface,i)].peak_usecs/1000,
			gsc->usage[gsmstat_usage(gsc->intface,i)].peak_count
			);
	}
	printf("===================================================================\n");
}

/* When called, this function will create a new instance of gsmstat.
 * Arg determines what we want to collect statistics for.
 * This function is called from tethereal when it parses the -Z gsm, arguments
 * and it creates a new instance to store statistics in and registers this
 * new instance for the gsm tap.
 */
static void
gsmstat_init(const char *optarg,void* userdata _U_)
{
	gsmstat_t *rs;
	int i, intface;
	int pos=0;
	char *filter=NULL;

	i=sscanf(optarg,"gsm,%d,%n",&intface, &pos);
	if(i==1 || i==2){
		if(pos){
			filter=(char *)optarg+pos;
		} else {
			filter=NULL;
		}
	} else {
		fprintf(stderr, "tethereal: invalid \"-z gsm,interface[,<filter>]\" argument\n");
		exit(1);
	}

	rs=g_malloc(sizeof(gsmstat_t));
	gsmstat_reset(rs);
	rs->gsc.intface = intface;

	if(filter){
		rs->filter=g_malloc(strlen(filter)+1);
		strcpy(rs->filter, filter);
	} else {
		rs->filter=NULL;
	}

/* It is possible to create a filter and attach it to the callbacks. Then the
 * callbacks would only be invoked if the filter matched.
 * Evaluating filters is expensive and if we can avoid it and not use them
 * we gain performance.
 * In this case we do the filtering for protocol and version inside the
 * callback itself but use whatever filter the user provided.
 * (Perhaps the user only want the stats for nis+ traffic for certain objects?)
 *
 */

	if(register_tap_listener("gsm", rs, filter, gsmstat_reset, gsmstat_packet, gsmstat_draw)){
		/* error, we failed to attach to the tap. clean up */
		g_free(rs->filter);
		g_free(rs);

		fprintf(stderr,"tethereal: gsmstat_init() failed to attach to tap.\n");
		exit(1);
	}
}


void
register_tap_listener_gsmstat(void)
{
	register_stat_cmd_arg("gsm", gsmstat_init, NULL);
}
